/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../form/bean/server_response.dart';
import '../config/config.dart';
import '../config/defaults.dart';
import '../error/illegal_argument_exception.dart';
import '../error/unsupported_operation_exception.dart';
import '../structure/i18n/message_array.dart';
import '../structure/i18n/option.dart';
import '../structure/model/script.dart';
import '../structure/ui/method/method.dart';
import '../ui/discrete/form_controller.dart';
import '../ui/discrete/renderer/ui_renderer.dart';
import '../ui/discrete/view_state.dart';
import '../util/bean_utils.dart';
import '../util/form_utils.dart';
import '../util/log/logger.dart';
import '../util/script/script_evaluator.dart';
import '../util/utils.dart';
import '../validation/converter/c_boolean.dart';
import '../validation/converter/c_string.dart';
import 'i_action.dart';
import 'i_array_provider.dart';
import 'i_constraint.dart';
import 'i_get_value.dart';
import 'i_on_value_change.dart';
import 'i_page_handler.dart';
import 'i_phase.dart';
import 'i_rmi.dart';
import 'i_table_action.dart';

class ContractUtils {
  static const UnsupportedOperationException UNSUPPORTED_OPERATION_EXCEPTION =
      UnsupportedOperationException.msg(
          "Custom code is deprecated, please use script.");

  const ContractUtils._();

  // IAction listener - Start

  static bool doAct(String id, ViewState view, UIRenderer renderer) {
    if (!Config.CUSTOM_CODE) {
      throw UNSUPPORTED_OPERATION_EXCEPTION;
    }

    return BeanUtils.newInstance<IAction>(id)
        .doAct(view, renderer.getController());
  }

  // IAction listener - End

  // IArrayProvider listener - Start

  static List<Option> load(String id, MessageArray array, UIRenderer renderer) {
    if (!Config.CUSTOM_CODE) {
      throw UNSUPPORTED_OPERATION_EXCEPTION;
    }

    return BeanUtils.newInstance<IArrayProvider>(id).load(array, renderer);
  }

  // IArrayProvider listener - End

  // IConstraint listener - Start

  static String doCheck(String id, ViewState view, Map<String, Object> args) {
    if (!Config.CUSTOM_CODE) {
      throw UNSUPPORTED_OPERATION_EXCEPTION;
    }

    return BeanUtils.newInstance<IConstraint>(id).doCheck(view, args);
  }

  // IConstraint listener - End

  // IGetValue listener - Start

  static Object getValue(String id, UIRenderer renderer) {
    if (!Config.CUSTOM_CODE) {
      throw UNSUPPORTED_OPERATION_EXCEPTION;
    }

    try {
      return BeanUtils.newInstance<IGetValue>(id)
          .getValue(renderer.getController());
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return null;
    }
  }

  // IGetValue listener - End

  // IOnValueChange listener - Start

  static void onValueChange(Object value, ViewState view, UIRenderer renderer) {
    try {
      final String id = view.getOnValueChange();

      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (Script.notInline(script)) {
          // 1. Attempt to run the script if it really is
          ScriptEvaluator.runScript_(script, view, null, renderer);
        } else {
          if (!Config.CUSTOM_CODE) {
            throw UNSUPPORTED_OPERATION_EXCEPTION;
          }

          // 2. If it's not a script, it has to be a custom class
          IOnValueChange listener = BeanUtils.newInstance(id);
          FormController controller = renderer.getController();

          if (value is bool) {
            listener.onValueChange_(value, view, controller);
          } else {
            listener.onValueChange(CString.convert(value), view, controller);
          }
        }
      }
    } on UnsupportedOperationException catch (e) {
      throw e;
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }
  }

  // IOnValueChange listener - End

  // IPageHandler listener - Start

  static IPageHandler getHandler(String id) {
    try {
      return BeanUtils.newInstance(id);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      throw IllegalArgumentException("Invalid page handler: $id");
    }
  }

  // IPageHandler listener - End

  // IPhase listener - Start

  static void onRender(String id, UIRenderer renderer) {
    try {
      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (Script.notInline(script)) {
          // 1. Attempt to run the script if it really is
          ScriptEvaluator.runScript_(script, null, null, renderer);
        } else {
          if (!Config.CUSTOM_CODE) {
            throw UNSUPPORTED_OPERATION_EXCEPTION;
          }

          // 2. If it's not a script, it has to be a custom class
          BeanUtils.newInstance<IPhase>(id)
              .onRender(renderer.isRendering(), renderer);
        }
      }
    } on UnsupportedOperationException catch (e) {
      throw e;
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }
  }

  static String checkData(String id, UIRenderer renderer) {
    try {
      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (Script.notInline(script)) {
          // 1. Attempt to run the script if it really is
          return CString.convert(
              ScriptEvaluator.runScript_(script, null, null, renderer));
        }

        if (!Config.CUSTOM_CODE) {
          throw UNSUPPORTED_OPERATION_EXCEPTION;
        }

        // 2. If it's not a script, it has to be a custom class
        return BeanUtils.newInstance<IPhase>(id)
            .checkData(renderer.getData(), renderer);
      }
    } on UnsupportedOperationException catch (e) {
      throw e;
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }

    return null;
  }

  // IPhase listener - End

  // IRmi listener - Start

  static IRmi getIRmi(String id) {
    try {
      return Config.CUSTOM_CODE ? BeanUtils.newInstance(id) : null;
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      throw IllegalArgumentException(e);
    }
  }

  static bool isRmiReady(String id, ViewState vSrc, UIRenderer renderer) {
    try {
      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (script != null) {
          return Utils.toTrue(CBoolean.convertNoErr(
              ScriptEvaluator.runScript_(script, vSrc, null, renderer)));
        }
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }

    return true;
  }

  static Map<String, Object> onRmiPrepare(String id, ViewState vSrc,
      Map<String, Object> data, bool all, Method method, UIRenderer renderer) {
    try {
      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (script != null) {
          Map<String, Object> args = {
            "data": data,
            "all": all,
            "method": method
          };
          return ScriptEvaluator.runScript_(script, vSrc, args, renderer)
              as Map<String, Object>;
        }
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }

    return null;
  }

  static bool onRmiPost(String id, ViewState vSrc, bool success, Method method,
      ServerResponse response, UIRenderer renderer) {
    try {
      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (script != null) {
          Map<String, Object> args = {
            "success": success,
            "method": method,
            "response": response
          };
          return Utils.toTrue(CBoolean.convertNoErr(
              ScriptEvaluator.runScript_(script, vSrc, args, renderer)));
        }
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }

    return success;
  }

  static bool onRmiError(String id, ViewState vSrc, Exception e, Method method,
      UIRenderer renderer) {
    try {
      if (!Utils.isBlank(id)) {
        Script script = FormUtils.getScript(id, renderer);

        if (script != null) {
          Map<String, Object> args = {"error": e, "method": method};
          return Utils.toTrue(CBoolean.convertNoErr(
              ScriptEvaluator.runScript_(script, vSrc, args, renderer)));
        }
      }
    } catch (e1, stackTrace) {
      Logger.error(e1, stackTrace: stackTrace);
    }

    return true;
  }

  // IRmi listener - End

  // ITableAction listener - Start

  static bool doTableAct(
      String id, List<String> row, ViewState vTable, UIRenderer renderer) {
    if (!Config.CUSTOM_CODE) {
      throw UNSUPPORTED_OPERATION_EXCEPTION;
    }

    return BeanUtils.newInstance<ITableAction>(id)
        .doAct(row, vTable, renderer.getController());
  }

  // ITableAction listener - End

  // DataConsumer - Start

  static Object prepareData(Object data, ViewState vSrc, UIRenderer renderer) {
    Script script =
        FormUtils.getScript("${vSrc.getId()}.save.prepare", renderer);

    if (Script.notInline(script)) {
      Map<String, Object> args = {Defaults.VALUE: data};
      return ScriptEvaluator.eval__(script, vSrc, args, renderer);
    }

    return data;
  }

  static void postPersist(bool success, ServerResponse response, ViewState vSrc,
      UIRenderer renderer) {
    Script script = FormUtils.getScript("${vSrc.getId()}.save.post", renderer);

    if (Script.notInline(script)) {
      Map<String, Object> args = {"success": success, "response": response};
      ScriptEvaluator.eval__(script, vSrc, args, renderer);
    }
  }

// DataConsumer - End

}
