/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../form/structure/t_map.dart';
import '../enums/action.dart';
import '../enums/mode.dart';
import '../enums/permission.dart';
import '../enums/view_type.dart';
import '../error/illegal_argument_exception.dart';
import '../structure/ui/view.dart';
import '../util/enum_utils.dart';
import '../util/utils.dart';

class Permissions {
  // Reserved permissions are being used from Permission enum.
  final TMap<bool> permissions = TMap();

  Permissions(
      bool canCreate,
      bool canView,
      bool canUpdate,
      bool canAccept,
      bool canReject,
      bool canRevise,
      bool canPrint,
      bool canChangePin,
      bool canAdd,
      bool canRemoteValidate,
      bool canOfflineLogin)
      : this.full(
            canCreate,
            canView,
            canUpdate,
            canAccept,
            canReject,
            canRevise,
            canPrint,
            canChangePin,
            canAdd,
            canRemoteValidate,
            canOfflineLogin,
            null);

  Permissions.list(List<String> permissions)
      : this.full(false, false, false, false, false, false, false, false, false,
            false, false, permissions);

  Permissions.full(
      bool canCreate,
      bool canView,
      bool canUpdate,
      bool canAccept,
      bool canReject,
      bool canRevise,
      bool canPrint,
      bool canChangePin,
      bool canAdd,
      bool canRemoteValidate,
      bool canOfflineLogin,
      List<String> permissions) {
    assign(Permission.CREATE, canCreate);
    assign(Permission.VIEW, canView);
    assign(Permission.UPDATE, canUpdate);
    assign(Permission.ACCEPT, canAccept);
    assign(Permission.REJECT, canReject);
    assign(Permission.REVISE, canRevise);
    assign(Permission.PRINT, canPrint);
    assign(Permission.CHANGEPIN, canChangePin);
    assign(Permission.ADD, canAdd);
    assign(Permission.REMOTE_VALIDATE, canRemoteValidate);
    assign(Permission.OFFLINE_LOGIN, canOfflineLogin);

    if (permissions != null) {
      for (String permission in permissions) {
        this.permissions.put(permission, true);
      }
    }
  }

  static Permissions all() {
    return Permissions(
        true, true, true, true, true, true, true, true, true, true, true);
  }

  static Permissions none() {
    return Permissions(false, false, false, false, false, false, false, false,
        false, false, false);
  }

  bool permitted(Mode mode) {
    if (Mode.PREVIEW == mode) {
      return check(Permission.VIEW);
    }

    return check(Permission.toPermission(mode.toString(), false));
  }

  bool permitted_(String sAction) {
    bool permitted = true;
    Action action = Action.toAction(sAction, false);

    if (action != null) {
      Permission permission;

      if (EnumUtils.match_(action, [Action.EDIT, Action.UPDATE])) {
        permission = Permission.UPDATE;
      } else {
        permission = Permission.toPermission(action.toString(), false);
      }

      permitted = check(permission);
    }

    return permitted;
  }

  bool permitted__(View view) {
    bool permitted = true;

    if (EnumUtils.match_(view.getType(), [ViewType.BUTTON, ViewType.SUBMIT])) {
      final String permission = view.getPermission();

      if (!Utils.isBlank(permission)) {
        permitted = check_(permission) && permitted_(view.getOnClick());
      }
    }

    return permitted;
  }

  void assign(Permission permission, bool permitted) {
    assign_(permission.toString(), permitted);
  }

  void assign_(String permission, bool permitted) {
    if (Utils.isBlank(permission)) {
      throw IllegalArgumentException("Permission can't be blank.");
    }

    permissions.put(permission, permitted);
  }

  bool check(Permission permission) {
    return permission == null || check_(permission.toString());
  }

  bool check_(String permission) {
    // TODO LOG or THROW if permission is missing?
    return Utils.toFalse(permissions[permission]);
  }
}
