/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:collection' show LinkedHashSet;

import '../../../form/structure/ordered_map.dart';
import '../../../form/structure/t_map.dart';
import '../../config/defaults.dart';
import '../../error/illegal_argument_exception.dart';
import '../../error/value_required_exception.dart';
import '../../ui/discrete/renderer/messages.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../ui/discrete/util/key_value_pair.dart';
import '../../ui/discrete/util/language/languages.dart';
import '../../util/form_utils.dart';
import '../../util/map_utils.dart';
import '../../util/utils.dart';
import '../root_tag.dart';
import 'locale.dart';
import 'message.dart';
import 'message_array.dart';
import 'option.dart';

class I18n extends RootTag {
  // @Attribute(required = true)
  String defaultLocale;

  // @ElementList(required = false)
  List<Locale> locales;

  // @Transient
  TMap<Locale> localeMap;

  // @ElementList(required = false)
  List<Message> messages;

  // @Transient
  TMap<Message> messageMap;

  // @ElementList(required = false)
  List<MessageArray> messageArrays;

  // @Transient
  TMap<MessageArray> messageArrayMap;

  I18n() {
    // default constructor required for de/serialization
  }

  I18n.detailed(int id, int version, String defaultLocale, List<Locale> locales,
      List<Message> messages, List<MessageArray> messageArrays)
      : super.id(id, version) {
    this.defaultLocale = defaultLocale;
    this.locales = locales;
    this.messages = messages;
    this.messageArrays = messageArrays;
  }

  I18n.json(Map json) : super.json(json) {
    this.defaultLocale = json['defaultLocale'];

    if (Utils.isBlank(this.defaultLocale)) {
      throw ValueRequiredException("I18n", "defaultLocale");
    }

    if (json.containsKey('locales')) {
      this.locales = [];

      for (Map map in json['locales']) {
        this.locales.add(Locale.json(map));
      }
    }

    if (json.containsKey('messages')) {
      this.messages = [];

      for (Map map in json['messages']) {
        this.messages.add(Message.json(map));
      }
    }

    if (json.containsKey('messageArrays')) {
      this.messageArrays = [];

      for (Map map in json['messageArrays']) {
        this.messageArrays.add(MessageArray.json(map));
      }
    }
  }

  @override
  void toMap() {
    if (Utils.isBlank(locales)) {
      this.locales = [Locale.id(Languages.DEFAULT.id)];
    }

    this.localeMap = MapUtils.toMap(locales, false);
    _validate(); // don't call this before localeMap's population
    this.messageMap = MapUtils.toMap(messages, false);
    this.messageArrayMap = MapUtils.toMap(messageArrays, false);
  }

  @override
  void toList() {
    this.locales = MapUtils.toList(localeMap, true);
    this.messages = MapUtils.toList(messageMap, true);
    this.messageArrays = MapUtils.toList(messageArrayMap, true);
  }

  void _validate() {
    if (!localeMap.containsKey(defaultLocale)) {
      throw IllegalArgumentException(
          "Default locale isn't defined as a supported locale!");
    }

    final int localesCount = localeMap.length;

    if (messages != null) {
      for (Message msg in messages) {
        final String id = msg.getId();
        final String refMsgId = msg.getRefMsgId();

        if (Utils.equalsIgnoreCase(id, refMsgId)) {
          throw IllegalArgumentException(
              "Message '$id' can't refer to itself.");
        }

        Set<String> msgLocales = MapUtils.keys(msg.getValue());
        final int msgLocalesCount = Utils.getSize(msgLocales);

        if (Utils.isBlank(refMsgId) && msgLocalesCount != localesCount) {
          throw IllegalArgumentException(
              "Message '$id' should have $localesCount locale(s).");
        }

        if (msgLocalesCount > 0) {
          for (String locale in msgLocales) {
            if (!localeMap.containsKey(locale)) {
              throw IllegalArgumentException(
                  "Message '$id' has unknown locale '$locale'");
            }
          }
        }
      }
    }
  }

  String getText(String msgId, String language, UIRenderer renderer) {
    return getText_(msgId, toLocale(language), null, renderer);
  }

  String getText_(
      String msgId, Locale locale, Messages msgs, UIRenderer renderer) {
    String text;

    if (!Utils.isBlank(msgId) && locale != null && !Utils.isBlank(messageMap)) {
      Message msg = messageMap[msgId];

      if (msg != null) {
        text = msg.get(locale,
            renderer); // translation might be overridden even if refMsgId is declared
        final String refMsgId = msg.getRefMsgId();

        if (text == null && !Utils.isBlank(refMsgId)) {
          // TODO Reference message shouldn't be in the visited messages.
          text = msgs == null
              ? getText_(refMsgId, locale, null, renderer)
              : msgs.get(refMsgId, null, null);
        }
      }
    }

    return text;
  }

  MessageArray getMsgArray(String id) {
    return Utils.get_(id, messageArrayMap);
  }

  MessageArray getMsgArray_(String parentId, String optionKey) {
    if (!Utils.isBlank(parentId) &&
        !Utils.isBlank(optionKey) &&
        !Utils.isBlank(messageArrayMap)) {
      for (MessageArray array in messageArrayMap.values) {
        if (Utils.equalsIgnoreCase(parentId, array.getParentId()) &&
            Utils.equalsIgnoreCase(optionKey, array.getOptionKey())) {
          return array;
        }
      }
    }

    return null;
  }

  List<String> getArrayKeys(String arrayId) {
    MessageArray msgArray = getMsgArray(arrayId);
    OrderedMap<Option> options =
        msgArray == null ? null : msgArray.getAllOptions();
    return Utils.toArray(MapUtils.keys(options));
  }

  OrderedMap<Option> getArrayOptions(String id, UIRenderer renderer) {
    return getArrayOptions_(getMsgArray(id), renderer);
  }

  OrderedMap<Option> getArrayOptions_(
      MessageArray msgArray, UIRenderer renderer) {
    return msgArray == null ? null : msgArray.getOptions_(renderer);
  }

  Option getArrayOption(String id, String arrayId) {
    MessageArray msgArray = getMsgArray(arrayId);
    return msgArray == null ? null : msgArray.getOption(id);
  }

  String getArrayDefaultOption(String id) {
    MessageArray msgArray = getMsgArray(id);
    return msgArray == null ? null : msgArray.getDefaultOption();
  }

  bool hasArrayBlankItem(String id) {
    MessageArray msgArray = getMsgArray(id);
    return msgArray == null ? Defaults.BLANK_ITEM : msgArray.hasBlankItem();
  }

  String getArrayBlankMsg(String id) {
    MessageArray msgArray = getMsgArray(id);
    return msgArray == null ? null : msgArray.getBlankMsg();
  }

  String getDefaultLocale() {
    return defaultLocale;
  }

  Locale toLocale(String language) {
    if (Utils.isBlank(localeMap)) {
      return null; // configuration error
    }

    Locale locale = localeMap[language];

    // if locale for the given language can't be found, use the configured default locale.
    if (locale == null) {
      locale = localeMap[defaultLocale];

      // if default locale isn't one of the supported locale, just use the first locale from the map.
      if (locale == null) {
        locale = localeMap.values.first;
      }
    }

    return locale;
  }

  Set<KeyValuePair> toText(MessageArray msgArray, UIRenderer renderer) {
    OrderedMap<Option> options = getArrayOptions_(msgArray, renderer);

    if (Utils.isBlank(options)) {
      return null;
    }

    final bool toProperCase = msgArray.toProperCase_();
    Locale locale = FormUtils.getLocale_(renderer);
    Messages msgs = renderer.getMsgs();
    Set<KeyValuePair> array = LinkedHashSet();

    for (Option option in options.values) {
      final String msgId = option.getMsgId();
      final String msg =
          Utils.toDefault(getText_(msgId, locale, msgs, renderer), msgId);
      array.add(KeyValuePair(
          option.getId(), toProperCase ? Utils.toProperCase(msg) : msg));
    }

    return array;
  }

  TMap<Locale> getLocales() {
    return localeMap;
  }

  List<Locale> getLocaleList() {
    return locales;
  }

  List<Message> getMsgs() {
    return messages;
  }

  List<MessageArray> getMsgArrays() {
    return messageArrays;
  }
}
