/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../form/structure/ordered_map.dart';
import '../../../form/structure/t_map.dart';
import '../../config/defaults.dart';
import '../../contract/contract_utils.dart';
import '../../enums/code.dart';
import '../../enums/sort_method.dart';
import '../../error/illegal_argument_exception.dart';
import '../../ui/discrete/renderer/features.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../util/form_utils.dart';
import '../../util/log/logger.dart';
import '../../util/map_utils.dart';
import '../../util/utils.dart';
import '../model/model.dart';
import '../model/script.dart';
import '../required_id_tag.dart';
import 'option.dart';

class MessageArray extends RequiredIdTag {
  // @Attribute(required = false)
  String parentId;

  // @Attribute(required = false)
  String optionKey;

  // @Attribute(required = false)
  String defaultOption;

  // @Attribute(required = false)
  bool blankItem;

  // @Attribute(required = false)
  String blankMsg;

  // @Attribute(required = false)
  bool toProperCase;

  // @Attribute(required = false)
  String source;

  // @Attribute(required = false)
  bool cache;

  // @Attribute(required = false)
  bool onDemand;

  // @Attribute(required = false)
  String delimiter;

  // @Attribute(required = false)
  SortMethod sortBy;

  // @Element(name = Defaults.OPTIONS, required = false)
  String rawOptions;

  // @Element(name = Defaults.MESSAGES, required = false)
  String rawMsgs;

  // @ElementList(required = false, inline = true)
  List<Option> options;

  // @Transient
  OrderedMap<Option> optionsMap = OrderedMap();

  // @Transient
  bool changeable = false;

  MessageArray() {
    // default constructor required for de/serialization
  }

  MessageArray.json(Map json) : super.json(json) {
    this.parentId = json['parentId'];
    this.optionKey = json['optionKey'];
    this.defaultOption = json['defaultOption'];
    this.blankItem = json['blankItem'];
    this.blankMsg = json['blankMsg'];
    this.toProperCase = json['toProperCase'];
    this.source = json['source'];
    this.cache = json['cache'];
    this.onDemand = json['onDemand'];
    this.delimiter = json['delimiter'];
    this.sortBy = SortMethod.toMethod(json['sortBy'], true);
    this.rawOptions = json['rawOptions'];
    this.rawMsgs = json['rawMsgs'];

    if (json.containsKey('options')) {
      this.options = [];

      for (Map map in json['options']) {
        this.options.add(Option.json(map));
      }
    }
  }

  @override
  void toMap() {
    final String id = getId();

    if (!Utils.isBlank(options) && !Utils.isBlank(rawOptions)) {
      throw IllegalArgumentException(
          "Provide either options or rawOptions for message-array '$id'");
    }

    rawToOptions();

    final bool notExternal = !isExternal();

    if (notExternal && Utils.isBlank(options)) {
      throw IllegalArgumentException(
          "Provide either source or options for message-array '$id'");
    }

    if (!Utils.isBlank(parentId) && isCacheable()) {
      throw IllegalArgumentException(
          "Can't cache dependent message-array '$id'");
    }

    if (notExternal && isOnDemand()) {
      throw IllegalArgumentException(
          "On-demand is only for external message-array '$id'");
    }

    if (notExternal) {
      options = FormUtils.sortOptions(options, sortBy);
      setOptions(MapUtils.toMap(options, true));
    }

    rawOptions = rawMsgs = null;
  }

  @override
  void toList() {
    this.options = MapUtils.toList(optionsMap, true);
  }

  void rawToOptions() {
    if (Utils.isBlank(options) && !Utils.isBlank(rawOptions)) {
      final String delimiter = getDelimiter();
      List<String> csvOptions =
          Utils.splitByDelimiters(rawOptions, delimiter, false);

      if (!Utils.isBlank(csvOptions)) {
        options = [];
        List<String> csvMsgs =
            Utils.splitByDelimiters(rawMsgs, delimiter, false);
        final int msgsCount = Utils.getSize(csvMsgs);

        for (int index = 0; index < csvOptions.length; index++) {
          final String msgId = index < msgsCount ? csvMsgs[index] : null;
          options.add(Option.id(csvOptions[index], Utils.toNull(msgId)));
        }
      }
    }
  }

  String getParentId() {
    return parentId;
  }

  String getOptionKey() {
    return optionKey;
  }

  String getDefaultOption() {
    return defaultOption;
  }

  bool hasBlankItem() {
    return Utils.toDefault(blankItem, Defaults.BLANK_ITEM);
  }

  String getBlankMsg() {
    return blankMsg;
  }

  bool toProperCase_() {
    return Utils.toFalse(toProperCase);
  }

  bool isChangeable() {
    return changeable;
  }

  bool isCacheable() {
    return Utils.toFalse(cache);
  }

  bool isOnDemand() {
    return Utils.toFalse(onDemand);
  }

  String getDelimiter() {
    return Utils.toDefault(delimiter, ",");
  }

  bool isExternal() {
    return !Utils.isBlank(source);
  }

  Option getOption(String optionId) {
    return getAllOptions()[optionId];
  }

  List<Option> getOptions() {
    return options;
  }

  OrderedMap<Option> getAllOptions() {
    return optionsMap;
  }

  OrderedMap<Option> getOptions_(UIRenderer renderer) {
    if (isExternal()) {
      loadOptions(renderer);
    }

    if (changeable) {
      OrderedMap<Option> filtered = OrderedMap();

      for (Option option in optionsMap.values) {
        if (option.isIncluded(renderer)) {
          filtered.put(option.getId(), option);
        }
      }

      return filtered;
    }

    return optionsMap;
  }

  void setOptions(Map<String, Option> optionsMap) {
    this.changeable = false;

    // auto-detect if this message-array can grow/shrink with options
    if (Utils.isBlank(rawOptions) &&
        this.optionsMap != optionsMap &&
        !Utils.isBlank(optionsMap)) {
      for (Option option in optionsMap.values) {
        if (!Utils.isBlank(option.getInclusionIf())) {
          this.changeable = true;
          break;
        }
      }
    }

    this.optionsMap = optionsMap as OrderedMap<Option>;
  }

  Future<void> loadOptions(UIRenderer renderer) async {
    // No need to fetch the list from external source if:
    // 1) it's already loaded and cached.
    // 2) the form is being constructed and this message-array is configured to be loaded on-demand only.
    // 3) the form's locale is being changed.
    if ((!isCacheable() || Utils.isBlank(optionsMap)) &&
        (!isOnDemand() || !renderer.isRendering()) &&
        !renderer.isLocaleChanging()) {
      List<Option> array;

      try {
        if (Code.CUSTOM.starts(source)) {
          array = ContractUtils.load(Utils.skipFirst(source), this, renderer);
        } else if (Code.RMI.starts(source)) {
          array = await Features.getRMIAction(renderer)
              .invokeFetchList(Utils.skipFirst(source));
        } else {
          throw IllegalArgumentException(
              "Unrecognized external source: $source");
        }
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
      } finally {
        // TODO On error, we should not cache the default options? Rather, try to re-invoke the external source?
        // if we get 'null' from external source, we can use default options listed in the XML
        List<Option> finalOptions =
            FormUtils.sortOptions(array == null ? options : array, sortBy);
        setOptions(MapUtils.toMap_(finalOptions, true, false));
      }
    }
  }

  Set<Script> getConditions(TMap<Script> scripts) {
    Set<Script> conditions = {};

    if (changeable) {
      for (Option option in optionsMap.values) {
        Script condition = Model.getScript(option.getInclusionIf(), scripts);

        if (condition != null) {
          conditions.add(condition);
        }
      }
    }

    return conditions;
  }

  @override
  String toString() =>
      "MessageArray [id=${getId()}, parentId=$parentId, optionKey=$optionKey, source=$source, options=$options]";
}
