/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../config/config.dart';
import '../../config/defaults.dart';
import '../../enums/code.dart';
import '../../enums/constraint_type.dart';
import '../../error/illegal_argument_exception.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../util/form_utils.dart';
import '../../util/script/condition_evaluator.dart';
import '../../util/structure/unique_no_blank_string_map.dart';
import '../../util/utils.dart';
import '../../validation/converter/c_boolean.dart';
import '../args_tag.dart';
import 'model.dart';

class Constraint extends ArgsTag {
  // @Attribute(required = false)
  String msg;

  // @Attribute(required = false)
  String enabled;

  // @Attribute(required = false)
  bool ignoreIfDisabled; // skip this constraint if the component is disabled?

  // @Attribute(required = false)
  String immediate;

  // @Transient
  bool canImmediate;

  Constraint() {
    // default constructor required for de/serialization
  }

  Constraint.full(String id, String msg, String enabled, bool ignoreIfDisabled,
      String immediate, UniqueNoBlankStringMap args)
      : super.full(id, args) {
    this.msg = msg;
    this.enabled = enabled;
    this.ignoreIfDisabled = ignoreIfDisabled;
    this.immediate = immediate;
  }

  Constraint.constraint(Constraint constraint)
      : this.full(constraint.getId(), constraint.msg, constraint.enabled,
            constraint.ignoreIfDisabled, constraint.immediate, constraint.args);

  Constraint.json(Map json) : super.json(json) {
    this.msg = json['msg'];
    this.enabled = json['enabled'];
    this.ignoreIfDisabled = json['ignoreIfDisabled'];
    this.immediate = json['immediate'];
  }

  @override
  void copyAttributes(Constraint constraint) {
    super.copyAttributes(constraint);

    if (!Utils.isBlank(constraint.msg)) {
      this.msg = constraint.msg;
    }

    if (!Utils.isBlank(constraint.enabled)) {
      this.enabled = constraint.enabled;
    }

    if (constraint.ignoreIfDisabled != null) {
      this.ignoreIfDisabled = constraint.ignoreIfDisabled;
    }

    if (constraint.immediate != null) {
      this.immediate = constraint.immediate;
    }
  }

  ConstraintType getType() {
    return ConstraintType.toType(getId(), false);
  }

  String getIdUnprefixed() {
    return Utils.skipFirst(getId());
  }

  bool isCustom() {
    return Code.CUSTOM.starts(getId());
  }

  bool isGlobal() {
    return Code.GLOBAL.starts(getId());
  }

  @override
  String resolveId() {
    final String id = getId();

    if (isCustom() || isGlobal()) {
      return id;
    }

    ConstraintType type = getType();

    if (type == null) {
      throw IllegalArgumentException(
          "Constraint '$id' isn't supported or wrongly configured!");
    }

    return type.toString();
  }

  String getMsg() {
    return msg == null ? msg : msg.trim();
  }

  String getEnabledIf(UIRenderer renderer) {
    String enabled = this.enabled;

    if (Utils.isBlank(enabled) && isGlobal()) {
      enabled = Model.getConstraint_(
              getIdUnprefixed(), FormUtils.getConstraints(renderer))
          .enabled;
    }

    return CBoolean.parseNoErr(enabled) == null ? enabled : null;
  }

  bool isDisabled(UIRenderer renderer) {
    return !ConditionEvaluator.eval_(enabled, Defaults.ENABLED, renderer);
  }

  bool ignoreIfDisabled_() {
    return Utils.toTrue(ignoreIfDisabled);
  }

  bool canImmediate_(UIRenderer renderer) {
    if (canImmediate == null) {
      String immediate = this.immediate;

      if (Utils.isBlank(immediate) && isGlobal()) {
        immediate = Model.getConstraint_(
                getIdUnprefixed(), FormUtils.getConstraints(renderer))
            .immediate;
      }

      if (Utils.isBlank(immediate)) {
        canImmediate = Config.IMMEDIATE;
      } else {
        canImmediate = !CBoolean.NO == CBoolean.parseNoErr(immediate);
      }
    }

    return canImmediate;
  }

  bool isImmediate(UIRenderer renderer) {
    return ConditionEvaluator.eval_(immediate, Config.IMMEDIATE, renderer);
  }

  @override
  String toString() =>
      "Constraint [id=${getId()}, msg=$msg, enabled=$enabled, args=$args]";
}
