/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../form/structure/ordered_map.dart';
import '../../../form/structure/t_map.dart';
import '../../config/defaults.dart';
import '../../enums/constraint_type.dart';
import '../../enums/field_type.dart';
import '../../ui/discrete/form_controller.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../util/enum_utils.dart';
import '../../util/form_utils.dart';
import '../../util/map_utils.dart';
import '../../util/script/condition_evaluator.dart';
import '../../util/structure/string_set.dart';
import '../../util/utils.dart';
import '../../validation/constraint/c_length.dart' as CLength;
import '../../validation/constraint/pattern/c_alpha_numeric.dart'
    as CAlphaNumeric;
import '../../validation/converter.dart';
import '../../validation/converter/c_boolean.dart';
import '../../validation/converter/c_number.dart';
import '../../validation/converter/c_string.dart';
import '../required_id_tag.dart';
import 'constraint.dart';
import 'global_constraint.dart';
import 'model.dart';

class Field extends RequiredIdTag {
  // @Attribute(required = false)
  FieldType type;

  // @Attribute(required = false)
  String inclusion;

  // @Attribute(required = false)
  String value;

  // Display this string if value is null in preview/update mode
  // @Attribute(required = false)
  String previewValue;

  // @Attribute(required = false)
  String array;

  // @Attribute(required = false)
  String parentOf;

  // @Attribute(required = false)
  bool hidden;

  // @Attribute(required = false)
  String persist;

  // @Attribute(required = false)
  String update;

  // @Attribute(required = false)
  String distribution;

  // @Transient
  List<String> distributionList;

  // @Attribute(required = false)
  String delimiters;

  // @Attribute(required = false)
  String replicate;

  // @Transient
  List<String> replicateList;

  // @Attribute(required = false)
  bool enDigitsOnly;

  // @Attribute(required = false)
  bool excludeOnFetch;

  // TODO Can we remove Model.extras?
  // @Attribute(required = false)
  bool extra;

  // @Attribute(required = false)
  int luhn;

  // if a field can be rejected in the 'reject' form mode. If yes, tick-box is attached to the input view.
  // @Attribute(required = false)
  bool rejectable;

  // @Attribute(required = false)
  String conversionError;

  // @Attribute(required = false)
  String invalidItemMsg;

  // @ElementList(required = false, inline = true)
  List<Constraint> constraints;

  // @Transient
  OrderedMap<Constraint> constraintMap;

  // @Transient
  StringSet views;

  // @Transient
  bool hasImmediateConstraints;

  Field() {
    // default constructor required for de/serialization
  }

  Field.json(Map json) : super.json(json) {
    this.type = FieldType.toType(json['type'], true);
    this.inclusion = json['inclusion'];
    this.value = json['value'];
    this.previewValue = json['previewValue'];
    this.array = json['array'];
    this.parentOf = json['parentOf'];
    this.hidden = json['hidden'];
    this.persist = json['persist'];
    this.update = json['update'];
    this.distribution = json['distribution'];
    this.delimiters = json['delimiters'];
    this.replicate = json['replicate'];
    this.enDigitsOnly = json['enDigitsOnly'];
    this.excludeOnFetch = json['excludeOnFetch'];
    this.extra = json['extra'];
    this.luhn = json['luhn'];
    this.rejectable = json['rejectable'];
    this.conversionError = json['conversionError'];
    this.invalidItemMsg = json['invalidItemMsg'];

    if (json.containsKey('constraints')) {
      this.constraints = [];

      for (Map map in json['constraints']) {
        this.constraints.add(Constraint.json(map));
      }
    }
  }

  @override
  void toMap() {
    this.constraintMap = MapUtils.toMap(constraints, true);
  }

  @override
  void toList() {
    this.constraints = MapUtils.toList(constraintMap, true);
  }

  int getMaxLength(UIRenderer renderer) {
    Constraint cLength = get(CLength.CLength.TYPE, renderer);

    if (cLength == null) {
      return null;
    }

    return CNumber.convertNoErr(cLength.getArg___(CLength.Args.MAX, renderer));
  }

  bool isNumberType(UIRenderer renderer) {
    Constraint cAlphaNumeric = get(CAlphaNumeric.CAlphaNumeric.TYPE, renderer);

    if (cAlphaNumeric == null) {
      return EnumUtils.match_(type, [
        FieldType.BYTE,
        FieldType.SHORT,
        FieldType.NUMBER,
        FieldType.LONG,
        FieldType.FLOAT,
        FieldType.DECIMAL
      ]);
    }

    String mode = CString.convert(
        cAlphaNumeric.getArg___(CAlphaNumeric.Args.MODE, renderer));
    return CAlphaNumeric.Mode.NUMERIC == CAlphaNumeric.Mode.toMode(mode, false);
  }

  Constraint get(ConstraintType type, UIRenderer renderer) {
    if (Utils.anyNull([type, renderer])) {
      return null;
    }

    Constraint cRequired = Utils.get_(type.toString(), constraintMap);

    if (cRequired == null) {
      // find global constraint
      TMap<GlobalConstraint> globalConstraints =
          FormUtils.getConstraints(renderer);

      for (Constraint constraint in constraintMap.values) {
        if (constraint.isGlobal()) {
          constraint = Model.getConstraint(constraint, globalConstraints);

          if (type == constraint.getType() &&
              !constraint.isDisabled(renderer)) {
            return constraint;
          }
        }
      }
    }

    return cRequired == null || cRequired.isDisabled(renderer)
        ? null
        : cRequired;
  }

  FieldType getType() {
    return type;
  }

  String getInclusion() {
    return inclusion;
  }

  String getValue() {
    return value;
  }

  Object getConvertedValue(UIRenderer renderer) {
    final Object value = FormUtils.resolve_(getValue(), renderer);
    return type == null ? value : Converter.convertNoErr(type, value);
  }

  String getPreviewValue() {
    return previewValue;
  }

  String getArray() {
    return array;
  }

  String getParentOf() {
    return parentOf;
  }

  bool isHidden() {
    return Utils.toFalse(hidden);
  }

  bool canPersist(bool defaultValue, UIRenderer renderer) {
    return ConditionEvaluator.eval_(persist, defaultValue, renderer);
  }

  bool canUpdate(UIRenderer renderer) {
    return ConditionEvaluator.eval_(update, CBoolean.YES, renderer);
  }

  String getDelimiters() {
    return Utils.toDefault(delimiters, ";,");
  }

  List<String> getDistribution(UIRenderer renderer) {
    return getDistribution_(renderer.getController());
  }

  List<String> getDistribution_(FormController controller) {
    // don't cache if script is used (even if script is cached)
    final bool hasScript = FormUtils.hasScript(distribution);

    // don't use isBlank here
    if (hasScript || distributionList == null) {
      distributionList =
          Utils.split(FormUtils.resolveAsString(distribution, controller), ";");
      distributionList = distributionList == null ? [] : distributionList;
    }

    return distributionList;
  }

  List<String> getReplicate(UIRenderer renderer) {
    return getReplicate_(renderer.getController());
  }

  List<String> getReplicate_(FormController controller) {
    // don't cache if script is used (even if script is cached)
    final bool hasScript = FormUtils.hasScript(replicate);

    // don't use isBlank here
    if (hasScript || replicateList == null) {
      replicateList =
          Utils.split(FormUtils.resolveAsString(replicate, controller), ";");
      replicateList = replicateList == null ? [] : replicateList;
    }

    return replicateList;
  }

  bool getEnDigitsOnly() {
    return Utils.toFalse(enDigitsOnly);
  }

  bool isExcludeOnFetch() {
    return Utils.toFalse(excludeOnFetch);
  }

  bool isExtra() {
    return Utils.toDefault(extra, Defaults.EXTRA);
  }

  int getLuhn() {
    return luhn;
  }

  bool isRejectable() {
    return rejectable;
  }

  String getConversionError() {
    return conversionError;
  }

  String getInvalidItemMsg() {
    return invalidItemMsg;
  }

  bool hasImmediateConstraints_(UIRenderer renderer) {
    if (hasImmediateConstraints == null) {
      hasImmediateConstraints = CBoolean.NO;

      for (Constraint constraint in getConstraints().values) {
        if (constraint.canImmediate_(renderer)) {
          hasImmediateConstraints = CBoolean.YES;
          break;
        }
      }
    }

    return hasImmediateConstraints;
  }

  OrderedMap<Constraint> getConstraints() {
    return constraintMap;
  }

  StringSet getViews() {
    return views;
  }

  void addView(String viewId) {
    if (views == null) {
      views = StringSet();
    }

    views.add(viewId);
  }

  @override
  String toString() => "Field [id=${getId()}, type=$type, value=$value]";
}
