/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../form/structure/ordered_map.dart';
import '../../../form/structure/t_map.dart';
import '../../config/config.dart';
import '../../config/defaults.dart';
import '../../enums/code.dart';
import '../../error/illegal_argument_exception.dart';
import '../../util/map_utils.dart';
import '../../util/utils.dart';
import '../root_tag.dart';
import 'constraint.dart';
import 'extras.dart';
import 'field.dart';
import 'global_constraint.dart';
import 'script.dart';

class Model extends RootTag {
  // @Attribute(required = false)
  bool immediate;

  // @ElementList(required = false)
  List<Field> fields;

  // @Transient
  OrderedMap<Field> fieldMap;

  // @Element(required = false)
  Extras extras;

  // @ElementList(required = false)
  List<Script> scripts;

  // @Transient
  TMap<Script> scriptMap;

  // @ElementList(required = false)
  List<GlobalConstraint> constraints;

  // @Transient
  TMap<GlobalConstraint> constraintMap;

  Model() {
    // default constructor required for de/serialization
  }

  Model.full(int id, int version, List<Field> fields) : super.id(id, version) {
    this.fields = fields;
  }

  Model.json(Map json) : super.json(json) {
    this.immediate = json['immediate'];

    if (json.containsKey(Defaults.FIELDS)) {
      this.fields = [];

      for (Map map in json[Defaults.FIELDS]) {
        this.fields.add(Field.json(map));
      }
    }

    this.extras =
        json.containsKey('extras') ? Extras.json(json['extras']) : null;

    if (json.containsKey('scripts')) {
      this.scripts = [];

      for (Map map in json['scripts']) {
        this.scripts.add(Script.json(map));
      }
    }

    if (json.containsKey('constraints')) {
      this.constraints = [];

      for (Map map in json['constraints']) {
        this.constraints.add(GlobalConstraint.json(map));
      }
    }
  }

  @override
  void toMap() {
    this.fieldMap = MapUtils.toMap(fields, true);
    this.scriptMap = MapUtils.toMap(scripts, false);
    this.constraintMap = MapUtils.toMap(constraints, false);

    builtInScripts();
  }

  @override
  void toList() {
    this.fields = MapUtils.toList(fieldMap, true);
    this.scripts = MapUtils.toList(scriptMap, true);
    this.constraints = MapUtils.toList(constraintMap, true);
  }

  bool isImmediate() {
    return Utils.toDefault(immediate, Config.IMMEDIATE);
  }

  OrderedMap<Field> getFieldMap() {
    return fieldMap;
  }

  Field getField(String id) {
    return Utils.get_(id, fieldMap);
  }

  Extras getExtras() {
    return extras;
  }

  bool hasScript(String id) {
    return Utils.got(id, getScriptMap());
  }

  static Script getScript(String id, TMap<Script> scripts) {
    if (Utils.isBlank(id)) {
      return null;
    }

    Script script = Utils.get_(Code.NOT.remove(id), scripts);
    return script == null
        ? Script.inline_(id)
        : script; // if it's not a script id, it's used as inline script.
  }

  static Constraint getConstraint(
      Constraint reference, TMap<GlobalConstraint> constraints) {
    Constraint constraint = Constraint.constraint(
        getConstraint_(reference.getIdUnprefixed(), constraints));
    constraint.copyAttributes(reference);
    return constraint;
  }

  static GlobalConstraint getConstraint_(
      String id, TMap<GlobalConstraint> constraints) {
    if (Utils.isBlank(id)) {
      return null;
    }

    GlobalConstraint constraint = Utils.get_(id, constraints);

    if (constraint == null) {
      throw IllegalArgumentException("Missing global constraint: $id");
    }

    if (constraint.getType() == null) {
      throw IllegalArgumentException(
          "Global constraint '$id' of type '${constraint.getId()}' isn't supported or wrongly configured!");
    }

    if (constraint.isGlobal()) {
      throw IllegalArgumentException(
          "Global constraint '$id' can't refer to another global constraint '${constraint.getIdUnprefixed()}'");
    }

    return constraint;
  }

  TMap<Script> getScriptMap() {
    return scriptMap;
  }

  TMap<GlobalConstraint> getConstraintMap() {
    return constraintMap;
  }

  // TODO Move all such scripts to .properties bundle?
  void builtInScripts() {
    addScript("isAndroid", true, "platform('Android')");
    addScript("isIOS", true, "platform('IOS')");
    addScript("isFx", true, "platform('JavaFX')");
    addScript("isWeb", true, "platform('JSF')");
    addScript("isVirtual", true, "platform('Virtual')");

    addScript("isCreate", true, "mode('Create')");
    addScript("isPreview", true, "mode('Preview')");
    addScript("isUpdate", true, "mode('Update')");
    addScript(
        "isCreateOrUpdate", true, "runC('_isCreate') or runC('_isUpdate')");
    addScript(
        "isPreviewOrUpdate", true, "runC('_isPreview') or runC('_isUpdate')");

    addScript("isProd", true, "resolve('?PRODUCTION_MODE')");

    // scripts that shouldn't be cached
    addScript("locale", false, "resolve('?I18N_LOCALE')");
    addScript("isFormValid", false, "resolve('?FORM_IS_VALID')");
  }

  void addScript(String id, bool cache, String text) {
    final String _id = Code.UNDERSCORE.prefix(id);
    scriptMap.put(_id, Script.full(_id, text, cache));
  }

  List<Script> getBuiltInScripts() {
    List<Script> scripts = [];

    for (Script script in scriptMap.values) {
      if (Code.UNDERSCORE.starts(script.getId())) {
        // if it's built-in
        scripts.add(script);
      }
    }

    return scripts;
  }
}
