/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../form/structure/string_object_map.dart';
import '../../error/value_required_exception.dart';
import '../../ui/discrete/action/data/data_comparator.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../ui/discrete/view_state.dart';
import '../../util/form_utils.dart';
import '../../util/log/logger.dart';
import '../../util/script/script_evaluator.dart';
import '../../util/structure/string_set.dart';
import '../../util/utils.dart';
import '../required_id_tag.dart';

class Script extends RequiredIdTag {
  // @Text(required = true)
  String text;

  // @Transient
  String preparedText; // cache prepared script so we improve performance

  // @Attribute(required = false)
  bool cache;

  // @Attribute(required = false)
  bool onClient;

  // @Attribute(required = false)
  String views;

  // @Transient
  StringSet viewsSet;

  // @Transient
  StringObjectMap lastViewValues;

  // @Transient
  Object result;

  // @Transient
  bool cached = false; // result is already cached?

  Script() {
    // default constructor required for de/serialization
  }

  Script.full(String id, String text, bool cache) : super.id(id) {
    this.text = text;
    this.cache = cache;
  }

  Script.json(Map json) : super.json(json) {
    this.text = json['text'];

    if (Utils.isBlank(this.text)) {
      throw ValueRequiredException("Script", "text");
    }

    this.cache = json['cache'];
    this.onClient = json['onClient'];
    this.views = json['views'];
  }

  static Script inline_(String text) {
    return Script.full(null, text, null);
  }

  String getText() {
    if (preparedText == null) {
      preparedText = ScriptEvaluator.prepareText(text);

      Logger.debug('Provided script: $text');
      Logger.debug(
          'Prepared script: ${text == preparedText ? 'Same' : preparedText}');
    }

    return preparedText;
  }

  bool inline() {
    return Utils.isBlank(getId());
  }

  static bool notInline(Script script) {
    return script != null && !script.inline();
  }

  bool cache_(ViewState vSrc, UIRenderer renderer) {
    if (cache == null) {
      final bool viewsUndefined = Utils.isBlank(getViews(vSrc, renderer));
      cache = Utils.toDefault(cache, viewsUndefined);

      if (cache && !viewsUndefined) {
        Logger.warn(
            "Script '${getId()}' having views defined shouldn't be cached!");
      }
    }

    return cache;
  }

  bool onClient_() {
    return Utils.toFalse(onClient);
  }

  StringSet getViews(ViewState vSrc, UIRenderer renderer) {
    if ((viewsSet == null && !onClient_()) || hasSrcVar(text)) {
      viewsSet = ScriptEvaluator.findViews(text, vSrc, renderer);
      Utils.addAll(viewsSet, [Utils.splitByComma(views)]);
    }

    return viewsSet;
  }

  Object getResult() {
    return this.result;
  }

  void setResult(Object result, ViewState vSrc, UIRenderer renderer) {
    bool cache = cache_(vSrc, renderer);

    if (cache) {
      this.result = result;
      this.cached = true;
      this.lastViewValues = inline() ? null : getViewValues(vSrc, renderer);
    }
  }

  bool cached_(ViewState vSrc, UIRenderer renderer) {
    if (!cached || inline()) {
      return false;
    }

    if (cache_(vSrc, renderer)) {
      return true;
    }

    return !DataComparator.isChanged(
        lastViewValues, getViewValues(vSrc, renderer));
  }

  StringObjectMap getViewValues(ViewState vSrc, UIRenderer renderer) {
    StringObjectMap values;
    StringSet views = getViews(vSrc, renderer);

    if (!Utils.isBlank(views)) {
      values = StringObjectMap();

      for (String viewId in views) {
        values.put(viewId, FormUtils.getValue_(viewId, renderer));
      }
    }

    return values;
  }

  static bool hasSrcVar(String script) {
    return Utils.contains_(script, false, [ScriptEvaluator.SRC]);
  }

  @override
  String toString() =>
      "Script [${(inline() ? "" : "id=${getId()}, ")}text=$text]";
}
