/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/bean/server_response.dart';
import '../../../config/defaults.dart';
import '../../../contract/contract_utils.dart';
import '../../../contract/i_rmi.dart';
import '../../../error/value_required_exception.dart';
import '../../../ui/discrete/action/handler/rmi/rmi_cycle.dart';
import '../../../ui/discrete/renderer/ui_renderer.dart';
import '../../../ui/discrete/view_state.dart';
import '../../../util/form_utils.dart';
import '../../../util/script/condition_evaluator.dart';
import '../../../util/utils.dart';
import '../../../validation/converter/c_boolean.dart';
import '../../required_id_tag.dart';
import 'data.dart';
import 'response.dart';

class Method extends RequiredIdTag {
  // @Attribute(required = true)
  String service;

  // @Attribute(required = true)
  String name;

  // @Attribute(required = false)
  int timeout; // in milliseconds

  // @Attribute(required = false)
  String enabled;

  // @Attribute(required = false)
  String disabledMsg;

  // @Attribute(required = false)
  bool toggleIfDisabled;

  // @Attribute(required = false)
  bool standardRmi; // coerces to use specific method in EAI

  // @Attribute(required = false)
  String networkRequired;

  // @Attribute(required = false)
  String networkRequiredMsg;

  // @Attribute(required = false)
  String networkOptionalMsg;

  // @Attribute(required = false)
  String impl;

  // @Attribute(required = false)
  String isReady; // script to invoke

  // @Attribute(required = false)
  String onPrepare; // script to invoke, should return Map<String, Object>

  // @Attribute(required = false)
  String onPost; // script to invoke

  // @Attribute(required = false)
  String onError; // script to invoke

  // @Attribute(required = false)
  String namespace;

  // @Attribute(required = false)
  int maxAttempts;

  // @Attribute(required = false)
  String blockedMsg;

  // @Attribute(required = false)
  bool toggleIfBlocked;

  // @Attribute(required = false)
  bool detectChange;

  // @Attribute(required = false)
  String noChangeMsg;

  // @Attribute(required = false)
  bool toggleIfNoChange;

  // @Attribute(required = false)
  bool soap; // Use SOAP or CMI? Only used by JSF

  // @Attribute(required = false)
  String endpoint;

  // @Element(required = true)
  Data data;

  // @Element(required = true)
  Response response;

  Method() {
    // default constructor required for de/serialization
  }

  Method.json(Map json) : super.json(json) {
    this.service = json['service'];

    if (Utils.isBlank(this.service)) {
      throw ValueRequiredException("Method", "service");
    }

    this.name = json['name'];

    if (Utils.isBlank(this.name)) {
      throw ValueRequiredException("Method", "name");
    }

    this.timeout = json['timeout'];
    this.enabled = json['enabled'];
    this.disabledMsg = json['disabledMsg'];
    this.toggleIfDisabled = json['toggleIfDisabled'];
    this.standardRmi = json['standardRmi'];
    this.networkRequired = json['networkRequired'];
    this.networkRequiredMsg = json['networkRequiredMsg'];
    this.networkOptionalMsg = json['networkOptionalMsg'];
    this.impl = json['impl'];
    this.isReady = json['isReady'];
    this.onPrepare = json['onPrepare'];
    this.onPost = json['onPost'];
    this.onError = json['onError'];
    this.namespace = json['namespace'];
    this.maxAttempts = json['maxAttempts'];
    this.blockedMsg = json['blockedMsg'];
    this.toggleIfBlocked = json['toggleIfBlocked'];
    this.detectChange = json['detectChange'];
    this.noChangeMsg = json['noChangeMsg'];
    this.toggleIfNoChange = json['toggleIfNoChange'];
    this.soap = json['soap'];
    this.endpoint = json['endpoint'];
    this.data = json.containsKey('data') ? Data.json(json['data']) : null;

    if (this.data == null) {
      throw ValueRequiredException("Method", "data");
    }

    this.response =
        json.containsKey('response') ? Response.json(json['response']) : null;

    if (this.response == null) {
      throw ValueRequiredException("Method", "response");
    }
  }

  @override
  void toMap() {
    response.validate(getId());
    data.toMap();
  }

  @override
  void toList() {
    data.toList();
  }

  String getService() {
    return service;
  }

  String getName() {
    return name;
  }

  int getTimeout() {
    return timeout;
  }

  bool isEnabled(UIRenderer renderer) {
    return ConditionEvaluator.eval_(enabled, Defaults.ENABLED, renderer);
  }

  String getDisabledMsg() {
    return disabledMsg;
  }

  bool isToggleIfDisabled() {
    return Utils.toTrue(toggleIfDisabled);
  }

  bool isStandardRmi() {
    return Utils.toFalse(standardRmi);
  }

  bool isNetworkRequired(UIRenderer renderer) {
    return ConditionEvaluator.eval_(networkRequired, CBoolean.NO, renderer);
  }

  String getNetworkRequiredMsg() {
    return Utils.toDefault(networkRequiredMsg, "network.required.unavailable");
  }

  String getNetworkOptionalMsg() {
    return Utils.toDefault(networkOptionalMsg, "network.optional.unavailable");
  }

  IRmi getImpl() {
    return ContractUtils.getIRmi(impl);
  }

  bool isReady_(IRmi iRmi, ViewState vSrc, UIRenderer renderer) {
    return iRmi == null
        ? ContractUtils.isRmiReady(isReady, vSrc, renderer)
        : iRmi.isReady(renderer.getController());
  }

  void onPrepare_(bool all, Map<String, Object> data, IRmi iRmi, ViewState vSrc,
      UIRenderer renderer) {
    if (iRmi == null) {
      Map<String, Object> scriptData = ContractUtils.onRmiPrepare(
          onPrepare, vSrc, data, all, this, renderer);

      if (scriptData != null) {
        data.addAll(scriptData);
      }
    } else {
      iRmi.prepare(data, all, this, renderer.getController());
    }
  }

  bool onPost_(bool success, ServerResponse response, IRmi iRmi, ViewState vSrc,
      RMICycle cycle, UIRenderer renderer) {
    return iRmi == null
        ? ContractUtils.onRmiPost(
            onPost, vSrc, success, this, response, renderer)
        : iRmi.doPost(success, this, response, cycle, cycle.getController());
  }

  bool onError_(Exception e, IRmi iRmi, ViewState vSrc, UIRenderer renderer) {
    if (iRmi == null) {
      return ContractUtils.onRmiError(onError, vSrc, e, this, renderer);
    }

    iRmi.onError(e, this, vSrc, renderer.getController());
    return false;
  }

  String getNamespace() {
    return namespace;
  }

  int getMaxAttempts() {
    return Utils.toDefault(maxAttempts, Defaults.MISSING); // -1 means infinity
  }

  String getBlockedMsg() {
    return blockedMsg;
  }

  bool isToggleIfBlocked() {
    return Utils.toFalse(toggleIfBlocked);
  }

  bool detectChange_() {
    return Utils.toFalse(detectChange);
  }

  String getNoChangeMsg() {
    return noChangeMsg;
  }

  bool isToggleIfNoChange() {
    return Utils.toTrue(toggleIfNoChange);
  }

  bool isSoap(UIRenderer renderer) {
    return soap == null ? FormUtils.isMobile(renderer) : soap;
  }

  String getEndpoint_(String platformEndpoint) {
    return Utils.toDefault(getEndpoint(), platformEndpoint);
  }

  String getEndpoint() {
    return endpoint;
  }

  Data getData() {
    return data;
  }

  Response getResponse() {
    return response;
  }

  @override
  String toString() => "Method [id=${getId()}, name=$name]";
}
