/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../form/structure/t_map.dart';
import '../../config/config.dart';
import '../../enums/mode.dart';
import '../../enums/orientation.dart';
import '../../enums/scope.dart';
import '../../error/illegal_argument_exception.dart';
import '../../error/value_required_exception.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../util/form_utils.dart';
import '../../util/map_utils.dart';
import '../../util/structure/string_set.dart';
import '../../util/structure/unique_no_blank_string_map.dart';
import '../../util/utils.dart';
import '../root_tag.dart';
import 'method/method.dart';
import 'page.dart';
import 'popup.dart';
import 'view.dart';

class UI extends RootTag {
  // @Attribute(required = false)
  String title;

  // @Attribute(required = false)
  String subTitle;

  // @Attribute(required = false)
  String screenOrientation; // Screen orientation just for Android

  // @Attribute(required = false)
  Orientation orientation; // Root layout's orientation

  // @Attribute(required = false)
  bool encode;

  // @Attribute(required = false)
  String focus;

  // @Attribute(required = false)
  String baseFontColor;

  // @Attribute(required = false)
  String baseFontSize;

  // @Attribute(required = false)
  bool forceRenderInvalidView;

  // @Attribute(required = false)
  bool showHiddenFields;

  // @Attribute(required = false)
  String dataConsumer;

  // @Attribute(required = false)
  String dataProvider;

  // @Attribute(required = false)
  bool debug;

  // @Attribute(required = false)
  String preRender;

  // @Attribute(required = false)
  String postRender;

  // @Attribute(required = false)
  String dataCheck;

  // @Attribute(required = false)
  String modes;

  // @ElementList(required = true)
  List<View> layout;

  // @ElementMap(entry = "group", key = Defaults.ID, attribute = true, inline = false, keyType = String,
  //         valueType = String, required = false)
  UniqueNoBlankStringMap groups;

  // @ElementList(required = false)
  List<Popup> popups;

  // @Transient
  TMap<Popup> popupMap;

  // @ElementList(required = false)
  List<Page> pages;

  // @Transient
  TMap<Page> pageMap;

  // @ElementList(required = false)
  List<Method> methods;

  // @Transient
  TMap<Method> methodMap;

  UI() {
    // default constructor required for de/serialization
  }

  UI.full(
      int id,
      int version,
      String title,
      Orientation orientation,
      bool encode,
      String focus,
      List<View> layout,
      List<Popup> popups,
      List<Method> methods)
      : super.id(id, version) {
    this.title = title;
    this.orientation = orientation;
    this.encode = encode;
    this.focus = focus;
    this.layout = layout;
    this.popups = popups;
    this.methods = methods;
  }

  UI.json(Map json) : super.json(json) {
    this.title = json['title'];
    this.subTitle = json['subTitle'];
    this.screenOrientation = json['screenOrientation'];
    this.orientation = Orientation.toOrientation(json['orientation'], true);
    this.encode = json['encode'];
    this.focus = json['focus'];
    this.baseFontColor = json['baseFontColor'];
    this.baseFontSize = json['baseFontSize'];
    this.forceRenderInvalidView = json['forceRenderInvalidView'];
    this.showHiddenFields = json['showHiddenFields'];
    this.dataConsumer = json['dataConsumer'];
    this.dataProvider = json['dataProvider'];
    this.debug = json['debug'];
    this.preRender = json['preRender'];
    this.postRender = json['postRender'];
    this.dataCheck = json['dataCheck'];
    this.modes = json['modes'];

    if (json.containsKey('layout')) {
      this.layout = [];

      for (Map map in json['layout']) {
        this.layout.add(View.json(map));
      }
    } else {
      throw ValueRequiredException("UI", "layout");
    }

    if (json.containsKey('groups')) {
      this.groups = UniqueNoBlankStringMap();

      for (String key in (json['groups'] as Map).keys) {
        this.groups[key] = json['groups'][key];
      }
    }

    if (json.containsKey('popups')) {
      this.popups = [];

      for (Map map in json['popups']) {
        this.popups.add(Popup.json(map));
      }
    }

    if (json.containsKey('pages')) {
      this.pages = [];

      for (Map map in json['pages']) {
        this.pages.add(Page.json(map));
      }
    }

    if (json.containsKey('methods')) {
      this.methods = [];

      for (Map map in json['methods']) {
        this.methods.add(Method.json(map));
      }
    }
  }

  @override
  void toMap() {
    if (!Utils.isBlank(layout)) {
      final bool allowSpace = layout[0].allowSpace(); // just call it once

      for (View view in layout) {
        view.validate(allowSpace);
        view.generateId();
      }
    }

    this.popupMap = MapUtils.toMap(popups, false);
    this.pageMap = MapUtils.toMap(pages, false);
    this.methodMap = MapUtils.toMap(methods, false);
  }

  @override
  void toList() {
    this.popups = MapUtils.toList(popupMap, true);
    this.pages = MapUtils.toList(pageMap, true);
    this.methods = MapUtils.toList(methodMap, true);
  }

  String getTitle() {
    return Utils.toDefault(title, "form.title");
  }

  String getSubTitle() {
    return subTitle; // don't want default 'form.subtitle' as it's merely optional
  }

  Orientation getOrientation() {
    return Utils.toDefault(orientation, Orientation.VERTICAL);
  }

  String getScreenOrientation() {
    return screenOrientation;
  }

  bool getEncode() {
    return Utils.toDefault(encode, Config.ENCODE_TEXT);
  }

  List<View> getLayout() {
    return layout;
  }

  bool hasGroup(String groupId) {
    return Utils.got(groupId, groups);
  }

  StringSet getGroupViews(String groupId) {
    return Utils.splitByComma(Utils.get_(groupId, groups));
  }

  Popup getPopup(String id) {
    return Utils.get_(id, popupMap);
  }

  Page getPage(String id) {
    return Utils.get_(id, pageMap);
  }

  Method getMethod(String id) {
    return Utils.get_(id, methodMap);
  }

  String getFocus() {
    return focus;
  }

  String getBaseFontColor() {
    return baseFontColor;
  }

  String getBaseFontSize() {
    return baseFontSize;
  }

  bool isForceRenderInvalidView(UIRenderer renderer) {
    return forceRenderInvalidView == null
        ? FormUtils.isMobile(renderer)
        : forceRenderInvalidView;
  }

  bool isShowHiddenFields() {
    return Utils.toFalse(showHiddenFields);
  }

  String getDataConsumer() {
    return dataConsumer;
  }

  String getDataProvider() {
    return dataProvider;
  }

  bool isDebug() {
    return Utils.toDefault(debug, !Config.PRODUCTION_MODE);
  }

  String getPreRender() {
    return preRender;
  }

  String getPostRender() {
    return postRender;
  }

  String getDataCheck() {
    return dataCheck;
  }

  bool supportsMode(Mode mode) {
    StringSet modes = mode == null ? null : Utils.splitByComma(this.modes);

    if (Utils.isBlank(modes) ||
        !modes.contains(Scope.NONE.id()) && modes.contains(mode.toString())) {
      return true;
    }

    throw IllegalArgumentException("Mode not supported: $mode");
  }
}
