/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../config/config.dart';
import '../../config/defaults.dart';
import '../../enums/alignment.dart';
import '../../enums/orientation.dart';
import '../../enums/view_type.dart';
import '../../error/illegal_argument_exception.dart';
import '../../error/value_required_exception.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../util/form_utils.dart';
import '../../util/utils.dart';
import '../optional_id_tag.dart';

class View extends OptionalIdTag {
  // @Attribute(required = true)
  String type;

  // @Transient
  ViewType resolvedType;

  // @Attribute(required = false)
  String field;

  // @Attribute(required = false)
  String parentOf;

  // @Attribute(required = false)
  String label;

  // @Attribute(required = false)
  String hint;

  // @Attribute(required = false)
  String title;

  // @Attribute(required = false)
  String subTitle;

  // Orientation of the view if its a container
  // @Attribute(required = false)
  Orientation orientation;

  // Inner alignment of the view
  // @Attribute(required = false)
  Alignment align;

  // Outer alignment of the view (Android only)
  // @Attribute(required = false)
  Alignment layoutAlign;

  // @Attribute(required = false)
  String rendered;

  // @Attribute(required = false)
  String enabled;

  // @Attribute(required = false)
  bool manualInput;

  // @Attribute(required = false)
  bool autoComplete;

  // @Attribute(required = false)
  String modes;

  // @Attribute(required = false)
  String render;

  // @Attribute(required = false)
  String enable;

  // @Attribute(required = false)
  String validate_;

  // @Attribute(required = false)
  bool immediate;

  // @Attribute(required = false)
  String onClick;

  // @Attribute(required = false)
  String params;

  // @Attribute(required = false)
  String popup;

  // @Attribute(required = false)
  String permission;

  // allow camera/fingerprint to capture from device?
  // @Attribute(required = false)
  String allowCapture;

  // allow camera/fingerprint file to be uploaded?
  // @Attribute(required = false)
  String allowUpload;

  // allow camera/fingerprint to rotate captured image?
  // @Attribute(required = false)
  String allowRotate;

  // Width of the view (value depends on platform)
  // @Attribute(required = false)
  String width;

  // Height of the view (value depends on platform)
  // @Attribute(required = false)
  String height;

  // Width of the view's layout (value depends on platform, Android only)
  // @Attribute(required = false)
  String layoutWidth;

  // Height of the view's layout (value depends on platform, Android only)
  // @Attribute(required = false)
  String layoutHeight;

  // Weight of the view (strictly for Android only)
  // @Attribute(required = false)
  String weight;

  // WeightSum of the view (strictly for Android only)
  // @Attribute(required = false)
  String weightSum;

  // number of rows for the data-table
  // @Attribute(required = false)
  int rows;

  // number of columns for the GridLayout/Choices
  // @Attribute(required = false)
  String columns;

  // text color of label
  // @Attribute(required = false)
  String fontColor;

  // text size of all labels, in input-components as well
  // @Attribute(required = false)
  String fontSize;

  // text style of all labels, in input-components as well
  // @Attribute(required = false)
  String fontStyle; // should be one of the enum in FontStyle

  // @Attribute(required = false)
  String style;

  // @Attribute(required = false)
  String focus;

  // @Attribute(required = false)
  String clear;

  // @Attribute(required = false)
  String set; // TODO Shouldn't have duplicate assignments

  // @Attribute(required = false)
  String reset;

  // @Attribute(required = false)
  String forId;

  // if a table column can import data from CSV.
  // @Attribute(required = false)
  bool importable;

  // if a table column should export to PDF, XML etc.
  // @Attribute(required = false)
  bool exportable;

  // @Attribute(required = false)
  String onValueChange;

  // Date to show when calendar pops up for DatePicker
  // @Attribute(required = false)
  String selectionDate;

  // Pattern for Date/Time Picker
  // @Attribute(required = false)
  String pattern;

  // @Attribute(required = false)
  String onSuccess; // TODO Support onFailure?

  // @Attribute(required = false)
  bool toastError;

  // Keyboard class type for InputText (Android only)
  // @Attribute(required = false)
  String keyboard;

  // @Attribute(required = false)
  bool twoLinesMrz;

  // @Attribute(required = false)
  String mask;

  // @Attribute(required = false)
  // FIXME Space is parsed to null because of StringTransform (JSON is fine)
  String placeholder;

  // @Attribute(required = false)
  bool useMaskedValue;

  // @Attribute(required = false)
  String link;

  // @Attribute(required = false)
  String previewEdit;

  // @Attribute(required = false)
  String icon;

  // @ElementList(required = false, entry = Defaults.VIEW, inline = true)
  List<View> layout;

  View() {
    // default constructor required for de/serialization
  }

  // constructor for all mandatory attributes
  View.id(String id, ViewType type)
      : this.resolvedType = type,
        super.id(id);

  View.modes(String id, ViewType type, bool rendered, this.modes)
      : super.id(id) {
    this.resolvedType = type;
    this.rendered = rendered.toString();
  }

  View.columns(String id, ViewType type, this.orientation, this.columns)
      : super.id(id) {
    this.resolvedType = type;
  }

  View.onClick(String id, ViewType type, this.label, this.onClick)
      : super.id(id) {
    this.resolvedType = type;
  }

  View.layout(String id, ViewType type, this.title, this.layout)
      : super.id(id) {
    this.resolvedType = type;
  }

  View.json(Map json) : super.json(json) {
    this.type = json['type'];

    if (Utils.isBlank(this.type)) {
      throw ValueRequiredException("View", "type");
    }

    this.field = json['field'];
    this.parentOf = json['parentOf'];
    this.label = json['label'];
    this.hint = json['hint'];
    this.title = json['title'];
    this.subTitle = json['subTitle'];
    this.orientation = Orientation.toOrientation(json['orientation'], true);
    this.align = Alignment.toAlignment(json['align'], true);
    this.layoutAlign = Alignment.toAlignment(json['layoutAlign'], true);
    this.rendered = json['rendered'];
    this.enabled = json['enabled'];
    this.manualInput = json['manualInput'];
    this.autoComplete = json['autoComplete'];
    this.modes = json['modes'];
    this.render = json['render'];
    this.enable = json['enable'];
    this.validate_ = json['validate'];
    this.immediate = json['immediate'];
    this.onClick = json['onClick'];
    this.params = json['params'];
    this.popup = json['popup'];
    this.permission = json['permission'];
    this.allowCapture = json['allowCapture'];
    this.allowUpload = json['allowUpload'];
    this.allowRotate = json['allowRotate'];
    this.width = json['width'];
    this.height = json['height'];
    this.layoutWidth = json['layoutWidth'];
    this.layoutHeight = json['layoutHeight'];
    this.weight = json['weight'];
    this.weightSum = json['weightSum'];
    this.rows = json['rows'];
    this.columns = json['columns'];
    this.fontColor = json['fontColor'];
    this.fontSize = json['fontSize'];
    this.fontStyle = json['fontStyle'];
    this.style = json['style'];
    this.focus = json['focus'];
    this.clear = json['clear'];
    this.set = json['set'];
    this.reset = json['reset'];
    this.forId = json['forId'];
    this.importable = json['importable'];
    this.exportable = json['exportable'];
    this.onValueChange = json['onValueChange'];
    this.selectionDate = json['selectionDate'];
    this.pattern = json['pattern'];
    this.onSuccess = json['onSuccess'];
    this.toastError = json['toastError'];
    this.keyboard = json['keyboard'];
    this.twoLinesMrz = json['twoLinesMrz'];
    this.mask = json['mask'];
    this.placeholder = json['placeholder'];
    this.useMaskedValue = json['useMaskedValue'];
    this.link = json['link'];
    this.previewEdit = json['previewEdit'];
    this.icon = json['icon'];

    if (json.containsKey('layout')) {
      this.layout = [];

      for (Map map in json['layout']) {
        this.layout.add(View.json(map));
      }
    }
  }

  void generateId() {
    if (Utils.isBlank(getId())) {
      setId("_" + Utils.getShortUUID());
    }

    if (!Utils.isBlank(layout)) {
      for (View view in layout) {
        view.generateId();
      }
    }
  }

  void resolveType(UIRenderer renderer) {
    if (resolvedType == null) {
      final String resolvedValue =
          FormUtils.resolveAsString_(this.type, renderer);
      this.resolvedType = ViewType.toViewType(resolvedValue, false);

      if (resolvedType == null) {
        throw IllegalArgumentException(
            "Couldn't resolve type '$resolvedValue' for view '${getId()}'. Configured: ${this.type}");
      }
    }
  }

  ViewType getType() {
    return resolvedType;
  }

  String getField() {
    return Utils.toDefault(field, getId());
  }

  String getParentOf() {
    return parentOf;
  }

  String getLabel(UIRenderer renderer) {
    final String label = getMsgId(this.label, "label", renderer);
    // default label for 'link' component is it's hyper-link
    return Utils.isBlank(label) && ViewType.LINK == getType()
        ? getLink(renderer)
        : label;
  }

  String getHint(UIRenderer renderer) {
    return getMsgId(hint, "hint", renderer);
  }

  String getTitle(UIRenderer renderer) {
    return getMsgId(title, "title", renderer);
  }

  String getSubTitle(UIRenderer renderer) {
    return getMsgId(subTitle, "subtitle", renderer);
  }

  Orientation getOrientation() {
    return Utils.toDefault(
        orientation, Orientation.getDefault(resolvedType.isContainer()));
  }

  Alignment getAlign() {
    return align;
  }

  Alignment getLayoutAlign() {
    return layoutAlign;
  }

  String getRendered() {
    return rendered;
  }

  String getEnabled() {
    return enabled;
  }

  bool allowManualInput() {
    return Utils.toDefault(manualInput,
        !Config.PRODUCTION_MODE && ViewType.SELECTONE != resolvedType);
  }

  bool allowAutoComplete() {
    return Utils.toDefault(autoComplete, !Config.PRODUCTION_MODE);
  }

  String getModes() {
    return modes;
  }

  String getRender() {
    return render;
  }

  String getEnable() {
    return enable;
  }

  String getValidate() {
    return validate_;
  }

  bool isImmediate() {
    return immediate;
  }

  String getOnClick() {
    return onClick;
  }

  String getParams() {
    return params;
  }

  String getPopup() {
    return popup;
  }

  String getPermission() {
    return permission;
  }

  String getAllowCapture() {
    return allowCapture;
  }

  String getAllowUpload() {
    return Utils.toDefault(
        allowUpload,
        (!Config.PRODUCTION_MODE || ViewType.CAMERA == resolvedType)
            .toString());
  }

  String getAllowRotate() {
    return allowRotate;
  }

  String getWidth() {
    return width;
  }

  String getHeight() {
    return height;
  }

  String getLayoutWidth() {
    return layoutWidth;
  }

  String getLayoutHeight() {
    return layoutHeight;
  }

  String getWeight() {
    return weight;
  }

  String getWeightSum() {
    return weightSum;
  }

  int getRows() {
    return rows;
  }

  String getColumns() {
    return columns;
  }

  String getFontColor() {
    return fontColor;
  }

  String getFontSize() {
    return fontSize;
  }

  String getFontStyle() {
    return fontStyle;
  }

  String getStyle() {
    return style;
  }

  String getFocus() {
    return focus;
  }

  String getClear() {
    return clear;
  }

  String getSet() {
    return set;
  }

  String getReset() {
    return reset;
  }

  String getForId() {
    return forId;
  }

  bool isImportable() {
    return Utils.toFalse(importable);
  }

  bool isExportable() {
    return Utils.toTrue(exportable);
  }

  String getOnValueChange() {
    return onValueChange;
  }

  String getSelectionDate() {
    return selectionDate;
  }

  String getPattern() {
    return pattern;
  }

  String getOnSuccess() {
    return onSuccess;
  }

  bool isToastError() {
    return toastError;
  }

  String getKeyboard() {
    return keyboard;
  }

  bool getTwoLinesMrz() {
    return twoLinesMrz;
  }

  String getMask() {
    return mask;
  }

  String getPlaceholder() {
    return placeholder;
  }

  bool isUseMaskedValue() {
    return Utils.toDefault(useMaskedValue, Defaults.USE_MASKED_VALUE);
  }

  String getLink(UIRenderer renderer) {
    return FormUtils.resolveAsString_(link, renderer);
  }

  String getPreviewEdit() {
    return previewEdit;
  }

  String getIcon() {
    return icon;
  }

  String getMsgId(String label, String suffix, UIRenderer renderer) {
    if (Utils.isBlank(label)) {
      label = Utils.suffix(
          getId(), ".$suffix"); // default is the id with the suffix
      label = Utils.hasMsg(label, renderer)
          ? label
          : null; // only if translation is available in i18n
    }

    return label;
  }

  List<View> getLayout() {
    return layout;
  }

  @override
  String toString() => "View [id=${getId()}, type=$resolvedType]";
}
