/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../form/bean/register/crm_form_data.dart';
import '../../../../../form/structure/string_map.dart';
import '../../../../../form/structure/string_object_map.dart';
import '../../../../error/file_not_found_exception.dart';
import '../../../../util/cipher/algorithm.dart';
import '../../../../util/cipher/cypher.dart';
import '../../../../util/io_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/stop_watch.dart';
import '../../../../util/utils.dart';
import '../../../../validation/converter/c_binary.dart';

/// Reads the content of all the files in the data map of CrmFormData. If any problem occurs in this process, the reason
/// is recorded in a missing reasons map of CrmFormData.
///
/// This should only be used by a platform that keeps the binary field type value as a file path using byte[] data type.
class DataAnalyzer {
  DataAnalyzer() {
    // nothing to do here
  }

  void perform(CrmFormData data) {
    try {
      perform_(data, false, null, true);
    } catch (e) {
      // should never occur
    }
  }

  void perform_(CrmFormData data, bool throwError, List<String> onlyIds,
      bool includeBinary) {
    StringObjectMap dataMap = data.getMap();

    if (Utils.isBlank(dataMap)) {
      return;
    }

    StopWatch stopWatch = StopWatch.start();
    StringMap missingReasons = data.getMissingReasons();
    missingReasons = missingReasons == null ? StringMap() : missingReasons;

    for (String id in dataMap.keys) {
      if (onlyIds != null && !onlyIds.contains(id)) {
        continue;
      }

      final Object value = dataMap[id];

      if (includeBinary && CBinary.isBinary(value)) {
        List<int> content = value as List<int>; // value is actually a file path

        if (!Utils.startsWith(content, ['/'])) {
          continue; // some fields might already been converted from path to real data during check status
        }

        try {
          content = readFile(
              Utils.toString_(content)); // content is actually a file path
        } catch (e, stackTrace) {
          content = null;
          Logger.error("Error reading file content");

          if (throwError) {
            throw e;
          }

          final String reason = e.toString();
          Logger.error(reason, stackTrace: stackTrace);
          missingReasons.put(id, reason);
        } finally {
          dataMap.put(id, content);
        }
      }
    }

    data.setMissingReasons(missingReasons);
    StopWatch.stop(stopWatch);
  }

  // TODO For FilePicker, we don't need encryption.
  List<int> readFile(String filePath) {
    if (!IOUtils.fileExists_(filePath)) {
      throw FileNotFoundException.msg("The file '$filePath' is missing.");
    }

    List<int> content = IOUtils.readFile(filePath);

    // We don't want to catch decryption error because sometimes byte[] will be encrypted sometimes not.
    return Cypher(Algorithm.DES128).decryptNoErr_(content);
  }
}
