/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'package:collection/collection.dart' show ListEquality;

import '../../../../../form/structure/string_object_map.dart';
import '../../../../enums/mode.dart';
import '../../../../error/illegal_argument_exception.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/map_utils.dart';
import '../../../../util/structure/string_set.dart';
import '../../../../util/utils.dart';
import '../../../../validation/converter/c_binary.dart';
import '../../../../validation/converter/c_boolean.dart';
import '../../../../validation/error_marker.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';

class DataComparator {
  final UIRenderer renderer;

  DataComparator(this.renderer);

  bool detectChange(StringObjectMap oldData, StringObjectMap newData) {
    if (!renderer.checkMode([Mode.UPDATE, Mode.REVISE])) {
      throw IllegalArgumentException(
          "DataComparator only supports update and revise modes.");
    }

    if (oldData == newData) {
      return false;
    }

    final bool oldDataBlank = Utils.isBlank(oldData);
    final bool newDataBlank = Utils.isBlank(newData);

    if (oldDataBlank && newDataBlank) {
      return false;
    }

    if (oldDataBlank && !newDataBlank) {
      return !MapUtils.allNullValued(newData);
    }

    if (!oldDataBlank && newDataBlank) {
      return !MapUtils.allNullValued(oldData);
    }

    addMissingKeys(oldData, newData);
    addMissingKeys(newData, oldData);

    if (oldData.length != newData.length) {
      return true;
    }

    // TODO Make it configurable?
    return renderer.checkMode([Mode.UPDATE])
        ? checkForAny(oldData, newData)
        : checkForAll(oldData, newData);
  }

  bool checkForAny(StringObjectMap oldData, StringObjectMap newData) {
    for (String viewId in oldData.keys) {
      ViewState view = FormUtils.getView_(viewId, renderer);

      if (CBoolean.YES == isValueChanged(view, oldData, newData)) {
        return true;
      }
    }

    return false;
  }

  bool checkForAll(StringObjectMap oldData, StringObjectMap newData) {
    bool changed = true;

    for (String viewId in oldData.keys) {
      ViewState view = FormUtils.getView_(viewId, renderer);

      if (CBoolean.NO == isValueChanged(view, oldData, newData)) {
        ErrorMarker.mark(view, renderer, null, "action.value.not.revised");
        changed = false;
      } else if (view != null) {
        ErrorMarker.clear(view); // clears the previous error message
      }
    }

    return changed;
  }

  /// Compares the view's value in old and new data map. If the value is changed, it returns null and false
  /// otherwise. If the view has to be ignored, it returns null.
  bool isValueChanged(
      ViewState view, StringObjectMap oldData, StringObjectMap newData) {
    if (view == null) {
      return null;
    }

    if ((renderer.checkMode([Mode.UPDATE]) && !view.canUpdate(renderer)) ||
        (renderer.checkMode([Mode.REVISE]) && !view.isRejected(renderer))) {
      return null;
    }

    final String fieldId = view.getFieldId();
    Object oldValue = oldData[fieldId];
    Object newValue = newData[fieldId];

    final bool isOldValueString = oldValue is String;

    oldValue = isOldValueString
        ? Utils.trimOrToBlank(oldValue as String, true)
        : oldValue;
    newValue =
        newValue is String ? Utils.trimOrToBlank(newValue, true) : newValue;

    if (oldValue == newValue) {
      // check for reference and null value
      return false;
    }

    if ((oldValue == null && newValue != null) ||
        (oldValue != null && newValue == null)) {
      return true;
    }

    if (oldValue.runtimeType != newValue.runtimeType) {
      return true;
    }

    if (CBinary.isBinary(oldValue)) {
      if (!ListEquality().equals(oldValue, newValue as List<int>)) {
        return true;
      }
    } else if (oldValue is List<Object>) {
      oldValue.sort();
      (newValue as List<Object>).sort();

      if (!ListEquality().equals(oldValue, newValue as List<int>)) {
        return true;
      }
    } else if (view.getType().canArray()) {
      if (isOldValueString) {
        StringSet oldSet = Utils.splitByComma(oldValue as String);
        StringSet newSet = Utils.splitByComma(newValue as String);

        if (!Utils.isBlank(oldSet) && !Utils.isBlank(newSet)) {
          if (oldSet != newSet) {
            return true;
          }
        }
      }
    } else if (isOldValueString) {
      if (!Utils.equalsIgnoreCase(oldValue as String, newValue as String)) {
        return true;
      }
    } else if (oldValue != newValue) {
      return true;
    }

    return false;
  }

  static void addMissingKeys(
      Map<String, Object> src, Map<String, Object> target) {
    for (String key in src.keys) {
      if (!Utils.got(key, target)) {
        target[key] = null;
      }
    }
  }

  static bool isChanged(Object oldValue, Object newValue) {
    final bool isOldValueString = oldValue is String;

    oldValue = isOldValueString
        ? Utils.trimOrToBlank(oldValue as String, true)
        : oldValue;
    newValue =
        newValue is String ? Utils.trimOrToBlank(newValue, true) : newValue;

    if (oldValue == newValue) {
      // check for reference and null value
      return false;
    }

    if ((oldValue == null && newValue != null) ||
        (oldValue != null && newValue == null)) {
      return true;
    }

    if (oldValue.runtimeType != newValue.runtimeType) {
      return true;
    }

    if (CBinary.isBinary(oldValue)) {
      if (!ListEquality().equals(oldValue, newValue as List<int>)) {
        return true;
      }
    } else if (oldValue is List<Object>) {
      oldValue.sort();
      (newValue as List<Object>).sort();

      if (!ListEquality().equals(oldValue, newValue as List<Object>)) {
        return true;
      }
    } else if (isOldValueString) {
      if (!Utils.equalsIgnoreCase(oldValue as String, newValue as String)) {
        return true;
      }
    } else if (oldValue != newValue) {
      return true;
    }

    return false;
  }

  static bool isChanged_(
      Map<String, Object> oldData, Map<String, Object> newData) {
    if (oldData == newData) {
      return false;
    }

    final bool oldDataBlank = Utils.isBlank(oldData);
    final bool newDataBlank = Utils.isBlank(newData);

    if (oldDataBlank && newDataBlank) {
      return false;
    }

    if (oldDataBlank && !newDataBlank) {
      return !MapUtils.allNullValued(newData);
    }

    if (!oldDataBlank && newDataBlank) {
      return !MapUtils.allNullValued(oldData);
    }

    addMissingKeys(oldData, newData);
    addMissingKeys(newData, oldData);

    if (oldData.length != newData.length) {
      return true;
    }

    return checkForChange(oldData, newData);
  }

  static bool checkForChange(
      Map<String, Object> oldData, Map<String, Object> newData) {
    for (String key in oldData.keys) {
      if (isChanged(oldData[key], newData[key])) {
        return true;
      }
    }

    return false;
  }
}
