/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../form/structure/string_object_map.dart';
import '../../../../enums/code.dart';
import '../../../../error/no_changes_exception.dart';
import '../../../../generator/reader/json_reader.dart';
import '../../../../structure/model/constraint.dart';
import '../../../../structure/model/extras.dart';
import '../../../../structure/model/field.dart';
import '../../../../util/date_utils.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/io_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/numbers.dart';
import '../../../../util/stop_watch.dart';
import '../../../../util/structure/string_set.dart';
import '../../../../util/structure/unique_no_blank_string_map.dart';
import '../../../../util/utils.dart';
import '../../../../validation/constraint/c_mandatory.dart';
import '../../../../validation/constraint/pattern/c_alpha_numeric.dart'
    as CAlphaNumeric;
import '../../../../validation/converter/c_binary.dart';
import '../../../../validation/converter/c_string.dart';
import '../../renderer/features.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';

/// Every field's value is included despite it is:<br>
///
/// <br>
/// 1) null or not null (UPDATE mode).<br>
/// 2) changed or unchanged (UPDATE mode).<br>
/// 3) updatable or not (UPDATE mode).<br>
/// <br>
///
/// A field is included only if it can be persisted (persist = "true").
class DataPreparer {
  final UIRenderer renderer;

  DataPreparer(this.renderer);

  StringObjectMap prepare(
      bool detectChanges, bool ensureMandatoryField, bool readPath) {
    StopWatch stopWatch = StopWatch.start();

    StringObjectMap oldData = renderer.getData();
    StringObjectMap newData = createDataMap(ensureMandatoryField, readPath);

    // check if the data has any/all changes
    if (detectChanges &&
        !Features.getDataComparator(renderer).detectChange(oldData, newData)) {
      throw NoChangesException();
    }

    addSysData(oldData, newData);
    addExtras(newData);

    StopWatch.stop(stopWatch);

    return newData;
  }

  StringObjectMap createDataMap(bool ensureMandatoryField, bool readPath) {
    StringSet inputViews = renderer.getInputViews();
    StringObjectMap data = StringObjectMap();

    for (String viewId in inputViews) {
      // loop only on IInput views
      ViewState view = FormUtils.getView_(viewId, renderer);

      if (view.canPersist(renderer) &&
          checkActiveViews(view, ensureMandatoryField, data)) {
        // TODO Send only UPDATED/REVISED data in their modes?
        final String fieldId = getFieldId(view);
        final Object value = getValue(view, readPath);

        // if this view is mandatory, it's value shouldn't be null
        if (ensureMandatoryField &&
            view.isMandatory(renderer) &&
            CMandatory().isValid(value) != null) {
          throw Exception("No value for mandatory field '$fieldId'.");
        }

        data.put(fieldId, value);
      }
    }

    return data;
  }

  bool checkActiveViews(
      ViewState view, bool ensureMandatoryField, StringObjectMap data) {
    Field field = view.getField();

    if (field != null && field.isHidden()) {
      return true;
    }

    StringSet activeViews = view.getActiveViews();
    final String fieldId = view.getFieldId();
    final int activeViewsCount = Utils.getSize(activeViews);

    if (activeViewsCount > 1) {
      // ideally this should never happen
      throw Exception("Multiple active views found for field '$fieldId'.");
    }

    final bool noActiveView = activeViewsCount == 0;

    if (Utils.isBlank(view.getInclusion())) {
      // auto-detect 'inclusion'
      if (noActiveView) {
        data.put(getFieldId(view), null);
        return false;
      }
    } else if (view.isIncluded(renderer)) {
      // field is included
      if (noActiveView) {
        if (ensureMandatoryField && view.isMandatory(renderer)) {
          throw Exception(
              "No active view found for mandatory field '$fieldId'.");
        }

        data.put(getFieldId(view), null);
        return false;
      }
    } else {
      // field is excluded
      if (noActiveView) {
        return false; // ignore this view/field
      }

      throw Exception(
          "Field '$fieldId' is excluded but it has an active view: $activeViews");
    }

    return activeViews
        .contains(view.getId()); // we shouldn't use 'fieldId' here
  }

  Object getValue(ViewState view, bool readPath) {
    Object value = view.getValue();

    // convert every value to String except byte[].
    if (CBinary.isBinary(value)) {
      List<int> content = value as List<int>;
      // if value is actually a file path (like in Android), we should read it.
      value = readPath
          ? IOUtils.readFileNoErr(String.fromCharCodes(content))
          : content;
    } else if (value != null) {
      value = toString__(value, view.getField(), renderer);
    }

    return value;
  }

  String getFieldId(ViewState view) {
    final String fieldId = view.getFieldId();
    // if configured to be saved as extra, it should be prefixed.
    return view.isExtra() ? Code.EXTRA.prefix(fieldId) : fieldId;
  }

  void addSysData(StringObjectMap oldData, StringObjectMap newData) {
    if (!Utils.isBlank(oldData)) {
      for (String dataId in oldData.keys) {
        if (Code.SYS.starts(dataId)) {
          newData.put(dataId, oldData[dataId]);
        }
      }
    }
  }

  void addExtras(StringObjectMap data) {
    Extras extras = FormUtils.getModel(renderer).getExtras();
    UniqueNoBlankStringMap fields = extras == null ? null : extras.getFields();

    if (!Utils.isBlank(fields)) {
      for (String fieldId in fields.keys) {
        if (Utils.got(fieldId, data)) {
          Logger.warn("Ignoring duplicated extra field '$fieldId'.");
        } else {
          final Object value = FormUtils.resolve_(fields[fieldId], renderer);

          if (value != null) {
            // resolved value can be null.
            data.put(Code.EXTRAS.prefix(fieldId), toString_(value));
          }
        }
      }
    }
  }

  static String toString_(Object value) {
    return toString__(value, null, null);
  }

  static String toString__(Object value, Field field, UIRenderer renderer) {
    if (value is DateTime) {
      return DateUtils.formatDefault(value);
    } else if (value is List) {
      // table view
      return JsonReader().toJson(value);
    } else if (value is List<Object>) {
      // multi-valued views
      return Utils.toString_(value); // TODO Use JSON?
    }

    return toEnDigits(Utils.toString_O(value), field, renderer);
  }

  /// Convert non-English digits to English if field:<br>
  /// <br>
  ///
  /// 1) is configured to have English digits only.<br>
  /// 2) can only accept numbers.
  static String toEnDigits(String text, Field field, UIRenderer renderer) {
    if (Utils.isBlank(text) || field == null) {
      return text;
    }

    if (field.getEnDigitsOnly()) {
      return Numbers.toDigitsCharByChar(text);
    }

    Constraint cNumeric = field.get(CAlphaNumeric.CAlphaNumeric.TYPE, renderer);

    if (cNumeric != null) {
      String mode = CString.convert(
          cNumeric.getArg___(CAlphaNumeric.Args.MODE, renderer));

      if (Utils.equalsIgnoreCase(CAlphaNumeric.Mode.NUMERIC.toString(), mode)) {
        return Numbers.toEnDigits(text);
      }
    }

    return text;
  }
}
