/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:math' show Random;
import 'dart:typed_data' show Uint8List;

import '../../../../../form/structure/string_object_map.dart';
import '../../../../config/config.dart';
import '../../../../contract/contract_utils.dart';
import '../../../../enums/code.dart';
import '../../../../enums/mode.dart';
import '../../../../enums/platform.dart';
import '../../../../enums/value_type.dart';
import '../../../../error/illegal_argument_exception.dart';
import '../../../../structure/model/script.dart';
import '../../../../util/captcha_utils.dart';
import '../../../../util/date_utils.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/io_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/network.dart';
import '../../../../util/script/script_evaluator.dart';
import '../../../../util/stop_watch.dart';
import '../../../../util/utils.dart';
import '../../form_controller.dart';
import '../../renderer/features.dart';
import '../../renderer/ui_renderer.dart';

abstract class ValueResolver {
  static const String MOCK_FOLDER = "mock/";

  final UIRenderer renderer;
  final StringObjectMap extras;

  ValueResolver(FormController controller)
      : this.renderer = controller.getRenderer(),
        this.extras = controller.getExtras() == null
            ? StringObjectMap()
            : controller.getExtras();

  Object resolve(String value) {
    StopWatch stopWatch = StopWatch.start();
    Object resolved = value;

    if (Utils.isBlank(value)) {
      resolved = null;
    } else if (Code.SKIP_START.starts(value) && Code.SKIP_END.ends(value)) {
      resolved = Utils.skipFirstAndLast(value); // ignore value resolution
    } else {
      final String unprefixed = Utils.skipFirst(value.trim());

      if (Code.RESOLVE.starts(value)) {
        resolved = getValue(ValueType.toType(unprefixed, false));
      } else if (Code.PERMISSION.starts(value)) {
        resolved = renderer.getParams().getPermissions().check_(unprefixed);
      } else if (Code.EXTRA.starts(value)) {
        resolved = extras[unprefixed];
      } else if (Code.CONFIG.starts(value)) {
        resolved = getConfig(unprefixed);
      } else if (Code.FIELD_VALUE.starts(value)) {
        resolved = FormUtils.getFieldValue(unprefixed, renderer);
      } else if (Code.VIEW_VALUE.starts(value)) {
        resolved = FormUtils.getValue_(unprefixed, renderer);
      } else if (Code.VIEW.starts(value)) {
        resolved = FormUtils.getView_(unprefixed, renderer);
      } else if (Code.GET_VALUE.starts(value)) {
        resolved = ContractUtils.getValue(unprefixed, renderer);
      } else if (FormUtils.hasScript(value)) {
        resolved = evalScript(unprefixed);
      }

      // TODO Get object from a method in some controller/class?
    }

    StopWatch.stop_(value, stopWatch);

    return resolved;
  }

  Object getValue(ValueType type) {
    try {
      Object value;

      if (type == null || ValueType.NULL == type) {
        // let the value be null
      } else if (ValueType.PLATFORM == type) {
        return getPlatform();
      } else if (ValueType.PIN == type) {
        value = getPin();
      } else if (ValueType.IPv4 == type) {
        value = getIP(true);
      } else if (ValueType.IPv6 == type) {
        value = getIP(false);
      } else if (ValueType.MAC == type) {
        value = getMacAddress();
      } else if (ValueType.CAPTCHA == type) {
        value = getCaptcha();
      } else if (ValueType.TODAY == type) {
        value = getToday();
      } else if (ValueType.NOW == type) {
        value = getNow();
      } else if (ValueType.TIMESTAMP == type) {
        value = getTimestamp();
      } else if (ValueType.DEALER_CODE == type) {
        value = getDealerCode();
      } else if (ValueType.TRX_ID == type) {
        value = getTrxId();
      } else if (ValueType.FORM_TRX_ID == type) {
        value = getFormTrxId();
      } else if (ValueType.MODE == type) {
        value = getMode();
      } else if (ValueType.PRODUCTION_MODE == type) {
        value = Config.PRODUCTION_MODE;
      } else if (ValueType.BUILTIN_SCRIPTS == type) {
        value = getBuiltInScripts();
      } else if (ValueType.MODEL_ID == type) {
        value = getModelId();
      } else if (ValueType.MODEL_VERSION == type) {
        value = getModelVersion();
      } else if (ValueType.UI_ID == type) {
        value = getUiId();
      } else if (ValueType.UI_VERSION == type) {
        value = getUiVersion();
      } else if (ValueType.I18N_ID == type) {
        value = getI18nId();
      } else if (ValueType.I18N_VERSION == type) {
        value = getI18nVersion();
      } else if (ValueType.I18N_LOCALE == type) {
        value = getI18nLocale();
      } else if (ValueType.FORM_IS_VALID == type) {
        value = Features.getValidator(renderer).runAll(false);
      } else if (ValueType.FORM_DATA == type) {
        value = getFormData();
      } else if (type.isMock()) {
        value = getMockValue(type);
      } else {
        value = getPlatformValue(type);
      }

      return value;
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return null;
    }
  }

  Platform getPlatform();

  int getPin() {
    return 100000 + Random().nextInt(900000);
  }

  String getIP(bool useIPv4) {
    return Network.getHostAddress(useIPv4);
  }

  String getMacAddress() {
    return Network.getMacAddress();
  }

  String getCaptcha() {
    return CaptchaUtils.generate();
  }

  DateTime getToday() {
    return DateUtils.today();
  }

  DateTime getNow() {
    return DateUtils.now();
  }

  int getTimestamp() {
    return DateUtils.epoch();
  }

  String getDealerCode() {
    return extras.getAsString(ValueType.DEALER_CODE.toString());
  }

  static String getTrxId() {
    return Utils.getUUID();
  }

  String getFormTrxId() {
    return getTrxId();
  }

  Mode getMode() {
    return FormUtils.getMode(renderer);
  }

  List<Script> getBuiltInScripts() {
    return FormUtils.getModel(renderer).getBuiltInScripts();
  }

  int getModelId() {
    return FormUtils.getModel(renderer).getId();
  }

  int getModelVersion() {
    return FormUtils.getModel(renderer).getVersion();
  }

  int getUiId() {
    return FormUtils.getUI_(renderer).getId();
  }

  int getUiVersion() {
    return FormUtils.getUI_(renderer).getVersion();
  }

  int getI18nId() {
    return FormUtils.getI18n(renderer).getId();
  }

  int getI18nVersion() {
    return FormUtils.getI18n(renderer).getVersion();
  }

  String getI18nLocale() {
    return FormUtils.getLocale_(renderer).getId();
  }

  StringObjectMap getFormData() {
    // TODO It's an heavy operation, we don't want to call again.
    return Features.getDataPreparer(renderer)
        .prepare(false, false, FormUtils.isMobile_(this));
  }

  List<int> getMockValue(ValueType type) {
    return readResource(getMockFileName(type));
  }

  Uint8List readResource(String name) {
    return IOUtils.readResource(MOCK_FOLDER + name);
  }

  String getMockFileName(ValueType type) {
    if (ValueType.MOCK_PIC == type) {
      return "_flowers.png";
    } else if (ValueType.MOCK_FACE == type) {
      return "_face.jpg";
    } else if (ValueType.MOCK_DOC == type) {
      return "_document.png";
    } else if (ValueType.MOCK_SIGN == type) {
      return "_signature.jpg";
    } else if (ValueType.MOCK_FINGER == type) {
      return "_finger.wsq";
    }

    return null; // should never be null
  }

  Object getPlatformValue(ValueType type) => null;

  String getConfig(String name) {
    return null;
  }

  Object evalScript(String scriptId) {
    Script script = FormUtils.getScript(scriptId, renderer);

    if (script.onClient_()) {
      throw IllegalArgumentException(
          "Can't run client-side script to get a value: ${script.getId()}");
    }

    return ScriptEvaluator.eval_(script, null, renderer);
  }

  UIRenderer getRenderer() {
    return this.renderer;
  }
}
