/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../form/structure/string_object_map.dart';
import '../../../../contract/contract_utils.dart';
import '../../../../enums/action.dart';
import '../../../../enums/code.dart';
import '../../../../enums/mode.dart';
import '../../../../error/illegal_argument_exception.dart';
import '../../../../error/unsupported_operation_exception.dart';
import '../../../../structure/model/script.dart';
import '../../../../util/enum_utils.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/script/script_evaluator.dart';
import '../../../../util/stop_watch.dart';
import '../../../../util/utils.dart';
import '../../../../validation/converter/c_boolean.dart';
import '../../component/table/action_column.dart';
import '../../component/table/table.dart';
import '../../renderer/features.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';
import 'form_actions.dart';

// TODO Should be just one instance.
class ActionHandler {
  final String action;
  final bool immediate; // ignore logic before the real action invocation?
  final UIRenderer renderer;

  const ActionHandler._(this.action, this.immediate, this.renderer);

  static ActionHandler getInstance(
      String action, bool immediate, UIRenderer renderer) {
    return ActionHandler._(getAction(action, renderer), immediate, renderer);
  }

  static String getAction(String action, UIRenderer renderer) {
    if (Utils.isBlank(action)) {
      throw IllegalArgumentException("Action can't be blank!");
    }

    action = Utils.trim(action);

    if (Code.RESOLVE.starts(action)) {
      Object resolvedAction =
          Features.getResolver(renderer).evalScript(Utils.skipFirst(action));

      if (resolvedAction is String) {
        action = resolvedAction;

        if (Utils.isBlank(action)) {
          throw IllegalArgumentException("Resolved action can't be blank!");
        }
      } else {
        throw IllegalArgumentException(
            "$action doesn't resolve to String value.");
      }
    }

    return action;
  }

  Future<void> performAction(ViewState view) async {
    StopWatch stopWatch = StopWatch.start();

    try {
      if (Code.CUSTOM.starts(action)) {
        customAction(view);
      } else if (FormUtils.hasScript(action)) {
        runScript(view);
      } else if (Code.RMI.starts(action)) {
        if (immediate) {
          // so to put an end to possible infinite RMI invocations
          throw UnsupportedOperationException.msg(
              "Can't invoke RMI in immediate mode.");
        }

        await Features.getRMIAction(renderer)
            .invoke(Utils.skipFirst(action), view);
      } else if (await overriddenAction(action, view)) {
        // the action was overridden in the 'protected' method
      } else {
        await predefinedAction(view);
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    } finally {
      StopWatch.stop_(action, stopWatch);
    }
  }

  void customAction(ViewState view) {
    FormActions formActions = Features.getFormActions(renderer);

    if (immediate || formActions.validateRegion(view)) {
      final String clazz = Utils.skipFirst(action);
      bool success = false;

      if (view.getComponent() is ActionColumn) {
        ViewState vTable = view.getContainer();
        final List<String> selectedRow =
            (vTable.getComponent() as Table).getSelectedRow();
        success =
            ContractUtils.doTableAct(clazz, selectedRow, vTable, renderer);
      } else {
        renderer.closePopup();
        success = ContractUtils.doAct(clazz, view, renderer);
      }

      if (success) {
        formActions.toggleViews(view);
      }
    }
  }

  void runScript(ViewState view) {
    FormActions formActions = Features.getFormActions(renderer);

    if (immediate || formActions.validateRegion(view)) {
      renderer.closePopup();

      Script script = FormUtils.getScript(Utils.skipFirst(action), renderer);
      bool result = false;

      try {
        if (script.onClient_()) {
          result = ScriptEvaluator.runOnClient(script, view, null, renderer);
        } else {
          result = CBoolean.convertNoErr(
              ScriptEvaluator.eval__(script, view, null, renderer));
        }
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
      }

      if (result == null || CBoolean.YES == result) {
        // will be null if script return non-boolean or nothing
        formActions.toggleViews(view);
      }
    }
  }

  Future<void> predefinedAction(ViewState view) async {
    final bool controllerNotNull = renderer.getController() != null;

    if (controllerNotNull) {
      renderer.getController().log(action);
    }

    if (EnumUtils.match(action, [Action.NOTHING])) {
      return; // nothing to do here
    } else if (EnumUtils.match(action, [Action.RENDER, Action.ENABLE])) {
      Features.getFormActions(renderer).doRenderAndEnable(view);
    } else if (EnumUtils.match(action, [Action.CLEAR])) {
      Features.getDataActions(renderer).doClearAndReset(view);
    } else if (EnumUtils.match(action, [Action.RESET])) {
      // keeps the reset form to same initial mode
      await Features.getFormActions(renderer).doReset(null, view);
    } else if (EnumUtils.match(action, [Action.RECREATE])) {
      Features.getFormActions(renderer).doRecreate(view);
    } else if (EnumUtils.match(action, [Action.NAVIGATE])) {
      Features.getFormActions(renderer).doNavigate(view);
    } else if (EnumUtils.match(action, [Action.CLOSE])) {
      Features.getFormActions(renderer).doClose(view.isImmediate());
    } else if (EnumUtils.match(action, [Action.DESTROY])) {
      if (controllerNotNull) {
        renderer.getController().destroy(false);
      }
    } else if (EnumUtils.match(action, [Action.VALIDATE])) {
      Features.getDataActions(renderer).doValidate(view);
    } else if (EnumUtils.match(action, [Action.TABLE_SAVE])) {
      Features.getTableActions(renderer).doAdd(view);
    } else if (EnumUtils.match(action, [Action.TABLE_CLEAR])) {
      Features.getTableActions(renderer).doClear(view);
    } else if (EnumUtils.match(action, [Action.TABLE_CANCEL])) {
      Features.getTableActions(renderer).doCancel(view);
    } else if (EnumUtils.match(action, [Action.CREATE])) {
      Features.getFormActions(renderer).changeMode(Mode.CREATE, view, false);
    } else if (EnumUtils.match(action, [Action.PRE_SAVE_VIEW])) {
      Features.getFormActions(renderer)
          .changeMode(Mode.PRE_SAVE_VIEW, view, true);
    } else if (EnumUtils.match(action, [Action.PREVIEW])) {
      Features.getFormActions(renderer).changeMode(Mode.PREVIEW, view, false);
    } else if (EnumUtils.match(action, [Action.POST_SAVE_VIEW])) {
      Features.getFormActions(renderer)
          .changeMode(Mode.POST_SAVE_VIEW, view, false);
    } else if (EnumUtils.match(action, [Action.EDIT])) {
      Features.getFormActions(renderer).changeMode(Mode.UPDATE, view, false);
    } else if (EnumUtils.match(action, [Action.REJECTION])) {
      await Features.getFormActions(renderer).doReset(Mode.REJECT, view);
    } else if (EnumUtils.match(action, [Action.REVISION])) {
      await Features.getFormActions(renderer).doReset(Mode.REVISE, view);
    } else if (EnumUtils.match(action, [Action.ACCEPT])) {
      await Features.getDataActions(renderer).doAcceptReject(view, true);
    } else if (EnumUtils.match(action, [Action.REJECT])) {
      await Features.getDataActions(renderer).doAcceptReject(view, false);
    } else if (EnumUtils.match(action, [Action.FETCH])) {
      await Features.getDataActions(renderer).doFetch(view);
    } else if (EnumUtils.match(action, [Action.CLEAR_LIST])) {
      Features.getDataActions(renderer).doFetchOrClearList(view, false);
    } else if (EnumUtils.match(action, [Action.FETCH_LIST])) {
      Features.getDataActions(renderer).doFetchOrClearList(view, true);
    } else if (EnumUtils.match(action, [Action.SAVE])) {
      await Features.getDataActions(renderer).doPersist(view, false);
    } else if (EnumUtils.match(action, [Action.UPDATE, Action.REVISE])) {
      await Features.getDataActions(renderer).doPersist(view, true);
    } else if (EnumUtils.match(action, [Action.PRINT])) {
      Features.getDataActions(renderer).doPrint(view);
    } else if (EnumUtils.match(action, [Action.SHOWPOPUP])) {
      Features.getFormActions(renderer).showPopup(view);
    } else if (EnumUtils.match(action, [Action.CLOSEPOPUP])) {
      renderer.closePopup();
    } else if (EnumUtils.match(action, [Action.REJECT_ALL])) {
      Features.getFormActions(renderer).doRejectAllOrNone(true);
    } else if (EnumUtils.match(action, [Action.REJECT_NONE])) {
      Features.getFormActions(renderer).doRejectAllOrNone(false);
    } else if (EnumUtils.match(action, [Action.SMS])) {
      Features.getCommActions(renderer).doSendSms(view);
    } else {
      throw IllegalArgumentException("Unsupported action: $action");
    }
  }

  /// Override this method to custom implement the predefined actions. In case if a certain action is redefined, make
  /// sure to return true.
  Future<bool> overriddenAction(String action, ViewState view) async {
    return false;
  }

  UIRenderer getRenderer() {
    return renderer;
  }

  static Future<void> doPostAction(ViewState view, bool canFetch, bool update,
      StringObjectMap data, UIRenderer renderer) async {
    if (view != null) {
      await (SubActionHandler.getInstance(
              view.getOnSuccess(), true, renderer, canFetch, update, data))
          .performAction(view);
    }
  }
}

class SubActionHandler extends ActionHandler {
  final bool canFetch;
  final bool update;
  final StringObjectMap data;

  const SubActionHandler._(String action, bool immediate, UIRenderer renderer,
      this.canFetch, this.update, this.data)
      : super._(action, immediate, renderer);

  static SubActionHandler getInstance(String action, bool immediate,
      UIRenderer renderer, bool canFetch, bool update, StringObjectMap data) {
    return SubActionHandler._(ActionHandler.getAction(action, renderer),
        immediate, renderer, canFetch, update, data);
  }

  @override
  Future<bool> overriddenAction(String action, ViewState view) async {
    if (EnumUtils.match(action, [Action.FETCH])) {
      if (!canFetch) {
        throw UnsupportedOperationException();
      }

      renderer.setData(data); // update the persisted/fetched data

      Features.getFormActions(renderer)
          .changeMode(update ? Mode.PREVIEW : Mode.POST_SAVE_VIEW, view, false);

      return true;
    }

    return false;
  }
}
