/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../form/bean/register/response_message.dart';
import '../../../../../form/bean/server_response.dart';
import '../../../../../form/structure/string_object_map.dart';
import '../../../../config/defaults.dart';
import '../../../../contract/contract_utils.dart';
import '../../../../enums/mode.dart';
import '../../../../error/illegal_argument_exception.dart';
import '../../../../error/no_changes_exception.dart';
import '../../../../structure/model/field.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/map_utils.dart';
import '../../../../util/structure/string_set.dart';
import '../../../../util/structure/unique_string_map.dart';
import '../../../../util/utils.dart';
import '../../component/alert.dart';
import '../../component/behavior/i_array.dart';
import '../../data/data_handler.dart';
import '../../renderer/array_value_handler.dart';
import '../../renderer/features.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';
import '../data/accept_reject_action.dart';
import '../data/data_consumer.dart';
import '../data/data_provider.dart';
import '../data/reject_reason.dart';
import '../toggle_action.dart';
import 'action_handler.dart';
import 'form_actions.dart';

class DataActions {
  final UIRenderer renderer;

  const DataActions(this.renderer);

  void doClearAndReset(ViewState vSrc) {
    // we should remove duplicated views among these attributes,
    // else there is much more work to be done resulting in the
    // triggering of more onValueChange listeners
    StringSet viewsToClear =
        vSrc.getClear_(renderer); // views that should nullify their values
    StringSet viewsToReset =
        vSrc.getReset_(renderer); // views that should reset their values
    Set<String> viewsToSet = MapUtils.keys(vSrc.getSet());

    Utils.removeAll(viewsToClear, [viewsToReset, viewsToSet]);
    Utils.removeAll(viewsToReset, [viewsToSet]);

    doClear(viewsToClear);
    doReset(viewsToReset);
    doSet(vSrc);
  }

  void doClear(StringSet views) {
    if (!Utils.isBlank(views)) {
      for (String viewId in views) {
        ViewState view = FormUtils.getView_(viewId, renderer);

        if (view != null && view.setValue(null)) {
          view.clearErrMsg();
        }
      }
    }
  }

  void doReset(StringSet views) {
    if (!Utils.isBlank(views)) {
      StringObjectMap data = renderer.getData();
      ArrayValueHandler handler = Features.getArrayValueHandler(renderer);

      for (String viewId in views) {
        ViewState view = FormUtils.getView_(viewId, renderer);

        if (view != null && view.canInput()) {
          final Object value = handler.getValue(view, true, data);

          if (view.setValue(value)) {
            view.clearErrMsg();
          }
        }
      }
    }
  }

  void doSet(ViewState vSrc) {
    UniqueStringMap pairs = vSrc.getSet();

    if (!Utils.isBlank(pairs)) {
      for (String viewId in pairs.keys) {
        final Object val = FormUtils.resolve_(pairs[viewId], renderer);

        try {
          ViewState view = FormUtils.getView__(viewId, true, true, renderer);

          if (view.setValue(val)) {
            view.clearErrMsg();
          }
        } catch (e, stackTrace) {
          Logger.error(e, stackTrace: stackTrace);
        }
      }
    }
  }

  void doValidate(ViewState vSrc) {
    if (Features.getValidator(renderer)
        .run_(vSrc.getValidate_(renderer), false)) {
      renderer.closePopup();
      Features.getFormActions(renderer).toggleViews(vSrc);
    }
  }

  // TODO Support custom methods on error and on failure?
  Future<bool> doPersist(ViewState vSrc, bool update) async {
    doClearAndReset(vSrc);

    Alert alert = Features.getAlert_(renderer);

    if (!vSrc.isImmediate() && !Features.getValidator(renderer).runAll(true)) {
      ViewState vPopup = renderer.getPopupView();

      if (vPopup == null) {
        alert.error(renderer.get(Defaults.MSG_INVALID_DATA));
      } else if (!vPopup.hasError()) {
        renderer.closePopup();
        alert.error(renderer.get(Defaults.MSG_INVALID_DATA));
      }

      return false;
    }

    bool success = false;

    try {
      DataConsumer consumer = Features.getDataConsumer(renderer);
      final String readyMsgId = consumer.checkIfReady();

      if (!Utils.isBlank(readyMsgId)) {
        alert.error(renderer.get(readyMsgId));
        return false;
      }

      StringObjectMap data =
          Features.getDataPreparer(renderer).prepare(update, true, false);
      Object preparedData = ContractUtils.prepareData(
          consumer.prepare(data, update), vSrc, renderer);

      if (preparedData == null) {
        renderer.closePopup();
        return false; // if prepare script returned null, we have nothing to persist
      }

      ServerResponse response = update
          ? await consumer.update(preparedData)
          : await consumer.consume(preparedData);
      success = ServerResponse.isSuccessful(response);

      renderer.closePopup();

      ContractUtils.postPersist(success, response, vSrc, renderer);

      final String action = update ? "update" : "save";

      if (success) {
        renderer.setData(data);
        renderer.setChangedOnServer(true);

        alert.info(renderer.get("action.$action.successful"));
        updateData(response);
        await ActionHandler.doPostAction(vSrc, true, update, data, renderer);
      } else {
        alert.error(getErrMsg(action, response));

        if (ServerResponse.isAccepted(response)) {
          await ActionHandler.doPostAction(vSrc, true, update, data, renderer);
        }
      }
    } on NoChangesException catch (e) {
      Logger.error(e);
      alert.error(renderer.get(renderer.checkMode([Mode.REVISE])
          ? "action.revise.all"
          : "action.no.change"));
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      alert.error(Utils.getStackTrace(stackTrace));
    }

    return success;
  }

  String getErrMsg(String action, ServerResponse response) {
    String msg;

    if (response != null) {
      final int responseCode = response.getResponseCode();
      ResponseMessage customMsg = response.getCustomMessage();

      msg = customMsg == null
          ? msg
          : customMsg.getMessage(FormUtils.getLocale_(renderer).getId());
      msg = Utils.isBlank(msg) ? response.getResponseDesc() : msg;
      msg = Utils.isBlank(msg)
          ? renderer.get("action.$action.$responseCode.failure")
          : msg;
      msg = Utils.isBlank(msg)
          ? renderer.get("validation.$responseCode.invalid")
          : msg;
    }

    return Utils.isBlank(msg) ? renderer.get("action.unknown.reason") : msg;
  }

  Future<void> doAcceptReject(ViewState vSrc, bool accept) async {
    Alert alert = Features.getAlert_(renderer);
    RejectReason reason;
    StringSet rejectedFieldIds;
    String errMsgId;
    final bool isRejectMode = renderer.checkMode([Mode.REJECT]);

    if (!accept) {
      reason = getReason(vSrc, alert); // TODO Reason can be optional

      // TODO In projects, we don't want to show the error message.
      if (reason == null || reason.getReason() == null) {
        errMsgId = "action.reject.reason.missing";
      } else if (isRejectMode) {
        rejectedFieldIds = getRejectedFieldIds();

        if (Utils.isBlank(rejectedFieldIds)) {
          errMsgId = "action.reject.none.selected";
        }
      }
    }

    if (!Utils.isBlank(errMsgId)) {
      alert.error(renderer.get(errMsgId));
      return;
    }

    renderer.closePopup();

    try {
      AcceptRejectAction action = Features.getAcceptRejectAction(renderer);
      ServerResponse response =
          accept ? action.accept() : action.reject(reason, rejectedFieldIds);

      if (ServerResponse.isSuccessful(response)) {
        renderer.setChangedOnServer(true);
        vSrc.setEnabled(false);
        ToggleAction().toggleEnable(vSrc, renderer);
        alert.info(getAcceptRejectMsg(accept, true));
        updateData(response);

        if (isRejectMode) {
          await renderer.getController().onReset(Mode.PREVIEW);
        }
      } else {
        alert.error(response.getResponseDesc());
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      alert.error("${getAcceptRejectMsg(accept, false)} [$e]");
    }
  }

  void updateData(ServerResponse response) {
    final Map<String, Object> resData = response.getData();

    if (!Utils.isBlank(resData)) {
      StringObjectMap data = StringObjectMap.fromMap(resData);

      renderer.handleData(data);
      renderer.getData().addAll(data); // we don't expect getData() to be null
    }
  }

  String getAcceptRejectMsg(bool accept, bool success) {
    return renderer.get(
        "action.${accept ? Defaults.ACCEPT : 'reject'}.${success ? 'success' : 'failure'}");
  }

  RejectReason getReason(ViewState vSrc, Alert alert) {
    StringSet params = Utils.splitByComma(vSrc.getParams());

    if (Utils.isBlank(params)) {
      return null;
    }

    Iterator<String> iterator = params.iterator;
    // Both parameters should be provided from the UI
    iterator.moveNext();
    String reasonViewId = Utils.skipFirst(iterator.current);
    iterator.moveNext();
    String commentViewId = Utils.skipFirst(iterator.current);

    ViewState vReason = FormUtils.getView_(reasonViewId, renderer);
    Object reason;
    String comment;

    if (vReason == null) {
      alert.error(renderer.get(Defaults.MSG_REJECT_NO_REASON));
    } else {
      if (vReason.canInput()) {
        if (Features.getValidator(renderer)
            .run([reasonViewId, commentViewId])) {
          reason = vReason.getValue();
          ViewState vComment = FormUtils.getView_(commentViewId, renderer);

          if (vComment != null && vComment.canInput()) {
            comment = vComment.getValue() as String;
          }
        }
      } else {
        alert.error(renderer.get(Defaults.MSG_REJECT_NO_REASON));
      }
    }

    return RejectReason(reason, comment);
  }

  StringSet getRejectedFieldIds() {
    StringSet rejectedFieldIds = StringSet();

    for (ViewState view in renderer.getAllViewStates()) {
      if (view.canInput() &&
          view.isLabelledComponent() &&
          view.getLabelledComponent().isRejected()) {
        rejectedFieldIds.add(view.getFieldId());
      }
    }

    return rejectedFieldIds;
  }

  Future<void> doFetch(ViewState vSrc) async {
    FormActions formActions = Features.getFormActions(renderer);

    if (!formActions.validateRegion(vSrc)) {
      return;
    }

    renderer.closePopup();

    try {
      DataProvider provider = Features.getDataProvider(renderer);
      StringObjectMap data =
          Features.getDataPreparer(renderer).prepare(false, true, false);
      data = await provider.fetch_(
          null, data); // id is null as it's fetched from 'data'
      StringSet vOverridden = StringSet();
      StringSet viewsToReset = Utils.splitByComma(vSrc.getParams());

      // Reset the value of all views that are not loaded by data provider
      if (!Utils.isBlank(viewsToReset)) {
        if (data == null) {
          data = StringObjectMap();
        }

        for (String viewId in viewsToReset) {
          Object value = data[viewId]; // fetched value

          if (!DataHandler.isLoaded(value)) {
            // if it didn't fetch.
            Field field = FormUtils.getField__(viewId, renderer);

            if (field != null) {
              final String fieldValue =
                  field.getValue(); // TODO Use ValueResolver?

              if (!Utils.isBlank(fieldValue)) {
                vOverridden.add(viewId);
              }

              value = fieldValue; // default field value
            }

            data.put(viewId, value); // override the value if not loaded.
          }
        }
      }

      Features.getDataHandler(renderer).handle(data, vOverridden);

      formActions.toggleViews(vSrc);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }
  }

  void doFetchOrClearList(ViewState vSrc, bool fetch) {
    StringSet targetViewIds = Utils.splitByComma(vSrc.getForId());

    if (!Utils.isBlank(targetViewIds)) {
      FormActions formActions = Features.getFormActions(renderer);

      if (formActions.validateRegion(vSrc) &&
          fetchOrClearLists(fetch, targetViewIds)) {
        formActions.toggleViews(vSrc);
      }
    }
  }

  bool fetchOrClearLists(bool fetch, StringSet viewIds) {
    ArrayValueHandler arrayValueHandler =
        Features.getArrayValueHandler(renderer);
    bool success = true;

    for (String viewId in viewIds) {
      ViewState vTarget = FormUtils.getView_(viewId, renderer);

      if (vTarget == null || !vTarget.canArray()) {
        throw IllegalArgumentException(
            "Missing view or it can't IArray: $viewId");
      }

      IArray iArray = vTarget.getInputComponent() as IArray;

      if (fetch) {
        arrayValueHandler.setArrayAndValue(
            vTarget, iArray, null, renderer.getAllViews());
      } else {
        arrayValueHandler.clearArrayAndValue(vTarget, iArray);
      }

      success = success && (fetch ? iArray.haveItems() : !iArray.haveItems());
    }

    return success;
  }

  void doPrint(ViewState vSrc) {
    Alert alert = Features.getAlert_(renderer);

    if (!vSrc.isImmediate() && !Features.getValidator(renderer).runAll(true)) {
      alert.error(renderer.get(Defaults.MSG_INVALID_DATA));
      return;
    }

    renderer.closePopup();

    final String failure = renderer.get("action.unable.print");

    try {
      ServerResponse response =
          Features.getDataPrinter(renderer).print(getPrintingLanguage(vSrc));

      if (ServerResponse.isSuccessful(response)) {
        alert.info(renderer.get("action.print.success"));
      } else {
        // Set error description in case not null to be matched with the requirement.
        alert.error(Utils.toDefault(response.getResponseDesc(), failure));
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      alert.error("$failure [$e]");
    }
  }

  String getPrintingLanguage(ViewState vSrc) {
    String language = FormUtils.resolveAsString_(vSrc.getParams(), renderer);
    Logger.debug(
        "Printing language: ${Utils.toDefault(language, Defaults.UNKNOWN)}");
    return language;
  }
}
