/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../config/defaults.dart';
import '../../../../enums/mode.dart';
import '../../../../enums/scope.dart';
import '../../../../error/illegal_argument_exception.dart';
import '../../../../structure/ui/page.dart';
import '../../../../structure/ui/popup.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/structure/string_set.dart';
import '../../../../util/utils.dart';
import '../../form_controller.dart';
import '../../renderer/features.dart';
import '../../renderer/params/form_params.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';
import '../toggle_action.dart';

class FormActions {
  final UIRenderer renderer;

  const FormActions(this.renderer);

  void changeMode(Mode mode, ViewState view, bool validate) {
    if (!validate || validateRegion(view)) {
      renderer.closePopup();

      try {
        if (renderer.changeMode(mode)) {
          toggleViews(view);
          renderer.updateTitle(view.getTitle(), view.getSubTitle());
        }
      } on IllegalArgumentException catch (e, stacktrace) {
        // for unsupported mode error
        Features.getAlert_(renderer).error(Utils.getStackTrace(stacktrace));
      }
    }
  }

  void doRenderAndEnable(ViewState view) {
    if (validateRegion(view)) {
      renderer.closePopup();
      toggleViews(view);
    }
  }

  void toggleViews(ViewState view) {
    if (view != null) {
      ToggleAction toggleAction = ToggleAction();
      toggleAction.toggleRender_(view, renderer);
      toggleAction.toggleEnable(view, renderer);

      Features.getDataActions(renderer).doClearAndReset(view);
    }
  }

  bool validateRegion(ViewState view) {
    if (view == null || view.isImmediate()) {
      return true;
    }

    StringSet views = view.getValidate_(renderer);
    // if there is no view mentioned, we should filter.
    final bool noViews = Utils.isBlank(views);
    return Features.getValidator(renderer)
        .run_(noViews ? [Scope.REGION.id()] : views, noViews);
  }

  Future<void> doReset(Mode mode, ViewState view) async {
    renderer.closePopup();
    renderer.setChangedOnServer(false);
    renderer.reset();
    await renderer.getController().onReset(mode);

    toggleViews(view);

    // TODO Set focus again?
  }

  void doRecreate(ViewState view) {
    // Don't move it down, parsing should be done prior to anything
    FormParams params = FormParams.getInstance(view.getParams(), renderer);
    // get this before calling cleanUp() or else it will be null
    FormController controller = renderer.getController();

    renderer.closePopup();
    renderer.cleanUp();

    controller.destroy(true);
    controller.onRecreate(params);
  }

  void doNavigate(ViewState view) {
    final String pageId = view.getParams();

    if (!Utils.isBlank(pageId)) {
      if (Features.getValidator(renderer)
          .run_(view.getValidate_(renderer), false)) {
        navigateToPage(pageId);
      } else {
        Features.getAlert_(renderer)
            .error(renderer.get(Defaults.MSG_INVALID_DATA));
      }
    }
  }

  void navigateToPage(String pageId) {
    Page page = FormUtils.getUI_(renderer).getPage(pageId);

    if (page == null) {
      throw IllegalArgumentException("Page '$pageId' not found!");
    }

    renderer.getController().navigate(page);
  }

  void doClose(bool immediate) {
    renderer.getController().onClose(immediate);
  }

  void showPopup(ViewState view) {
    final String popupId = view.getPopupId();

    if (!Utils.isBlank(popupId)) {
      if (validateRegion(view)) {
        showPopup_(popupId);
        toggleViews(view);
      } else {
        Features.getAlert_(renderer)
            .error(renderer.get(Defaults.MSG_INVALID_DATA));
      }
    }
  }

  void showPopup_(String popupId) {
    Popup popup = FormUtils.getUI_(renderer).getPopup(popupId);

    if (popup == null) {
      throw IllegalArgumentException("Popup '$popupId' not found!");
    }

    renderer.createPopup(popup);
  }

  void doRejectAllOrNone(bool all) {
    if (renderer.checkMode([Mode.REJECT])) {
      for (ViewState view in renderer.getAllViewStates()) {
        if (view.canInput() &&
            view.isLabelledComponent() &&
            view.isRejectable(renderer)) {
          view.getLabelledComponent().setRejected(all);
        }
      }
    } else {
      Features.getAlert_(renderer)
          .error(renderer.get("action.reject.wrong.mode"));
    }
  }
}
