/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../../form/bean/i_reflect.dart';
import '../../../../../../form/structure/t_map.dart';
import '../../../../../enums/code.dart';
import '../../../../../error/illegal_argument_exception.dart';
import '../../../../../structure/ui/method/complex_type.dart';
import '../../../../../structure/ui/method/data.dart';
import '../../../../../structure/ui/method/property.dart';
import '../../../../../util/bean_utils.dart';
import '../../../../../util/form_utils.dart';
import '../../../../../util/io_utils.dart';
import '../../../../../util/stop_watch.dart';
import '../../../../../util/utils.dart';
import '../../../../../validation/constraint/c_mandatory.dart';
import '../../../../../validation/converter/c_binary.dart';
import '../../data/data_preparer.dart';
import '../../data/value_resolver.dart';

/// Responsible to prepare data as a map for the RMI method.
class MethodData {
  final ValueResolver resolver;

  const MethodData(this.resolver);

  /// Prepares the configured data for the method to be invoked. The result is a map with the property name with it's
  /// evaluated value.
  Map<String, Object> prepare(Data data, bool all) {
    StopWatch stopWatch = StopWatch.start();
    validateBean(data.getBean());

    TMap<Property> properties = data.getProperties();
    Map<String, Object> preparedData = {};

    for (Property property in properties.values) {
      if (all || !property.isEverChanging()) {
        final Object value = getValue(property.getValue(), data, all);

        if (!skipValue(value)) {
          // skip this value?
          final String id = property.getId();

          // halt the RMI cycle if a mandatory property is detected to be null
          if (all && property.isMandatory() && !CMandatory.check(value)) {
            throw IllegalArgumentException("$id can't be null.");
          }

          preparedData[id] = value;
        }
      }
    }

    StopWatch.stop_(data, stopWatch);
    return preparedData;
  }

  /// Validates if the provided class exists and is Serializable. If valid, returns a new instance of given class.
  Object validateBean(String clazz) {
    Object bean = BeanUtils.newInstance(clazz);

    if (bean == null || bean is IReflect) {
      return bean;
    }

    throw IllegalArgumentException("$clazz isn't serializable.");
  }

  /// Evaluates the property's value.
  Object getValue(String propertyValue, Data data, bool all) {
    if (Utils.isBlank(propertyValue)) {
      return null;
    }

    final String sVal = propertyValue.trim();

    // if value is of complex type, instantiate a new bean and process that.
    if (Code.COMPLEX_TYPE.starts(sVal)) {
      return newComplexType(Utils.skipFirst(sVal), data, all);
    }

    final Object val = resolver.resolve(sVal);
    // prepare the value (like DataPreparer do to persist) when it comes from field or view if it isn't byte[]
    return !CBinary.isBinary(val) &&
            (Code.FIELD_VALUE.starts(sVal) || Code.VIEW_VALUE.starts(sVal))
        ? DataPreparer.toString_(val)
        : val;
  }

  /// Instantiate and process the configured complex type.
  Object newComplexType(String id, Data data, bool all) {
    ComplexType complexType = data.getComplexType(id);

    if (complexType == null) {
      throw IllegalArgumentException("Missing complex type: $id");
    }

    Object bean = validateBean(complexType.getBean());
    final bool isMobile = FormUtils.isMobile_(resolver);
    final bool isMap = BeanUtils.isMap(bean);
    final bool isList = BeanUtils.isList(bean);

    for (Property property in complexType.getProperties().values) {
      if (!all && property.isEverChanging()) {
        continue;
      }

      // 1. Evaluate the configured value. (recursive)
      Object value = getValue(property.getValue(), data, all);

      if (skipValue(value)) {
        continue; // skip this value?
      }

      if (value is List<int> && isMobile) {
        value = IOUtils.readFileNoErr(String.fromCharCodes(value));
      }

      final String propertyId = property.getId();

      if (isMap) {
        // 2. put the evaluated value in the map.
        (bean as Map)[propertyId] = value;
      } else if (isList) {
        if (value != null) {
          // 2. put the evaluated value in the list.
          (bean as List).add(value);
        }
      } else {
        // 2. Get the field (throws error if it doesn't exist)
        String field = BeanUtils.getField(propertyId, bean);
        // 3. set the evaluated value in the field's attribute.
        BeanUtils.setFieldValue(value, field, bean);
      }
    }

    return bean;
  }

  bool skipValue(Object value) {
    return value is String && Code.SKIP_PROPERTY.match(value);
  }
}
