/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:convert' show utf8;
import 'dart:io'
    show
        HttpClient,
        HttpClientRequest,
        HttpClientResponse,
        HttpHeaders,
        SecurityContext;

import '../../../../../config/defaults.dart';
import '../../../../../util/date_utils.dart';
import '../../../../../util/log/logger.dart';
import '../../../../../util/utils.dart';
import 'transport.dart';

class MposHttpTransportSE extends Transport {
  static const String REQ_METHOD_GET = "GET";
  static const String REQ_METHOD_POST = "POST";
  static const String GZIP = "gzip";

  final String method;

  MposHttpTransportSE(String url, int timeout, {this.method})
      : super(url, timeout);

  @override
  Future<String> call(String envelope) async {
    return await call_(envelope, null);
  }

  @override
  Future<String> call_(String envelope, Map<String, String> headers) async {
    String response;

    try {
      HttpClientResponse clientResponse = await call__(envelope, headers);
      response = await clientResponse.transform(utf8.decoder).join();
      return response;
    } finally {
      if (debug) {
        Logger.debug("transport.requestDump = $envelope");
        Logger.debug("transport.responseDump = $response");
      }
    }
  }

  @override
  Future<HttpClientResponse> call__(
      String envelope, Map<String, String> headers) async {
    HttpClientRequest connection = await initConnection(envelope, headers);
    connection.write(envelope);

    // sends request to server and closes the connection
    // TODO Test if the communication is in GZIP for HTTP
    // TODO Test if the communication is encrypted and in GZIP for HTTPs
    HttpClientResponse clientResponse = await connection.close();
    // TODO Check if response has soap body then parse it otherwise throw exception
    // TODO (MT server responds with 500, 503, 505, all of which needs to be parsed and returned to caller)
    // final int statusCode = clientResponse.statusCode;

    // if (statusCode != BaseServerResponseCode.SUCCESS.getCode()) {
    //   throw HttpResponseException(statusCode);
    // }

    return clientResponse;
  }

  Future<HttpClientRequest> initConnection(
      String envelope, Map<String, String> headers) async {
    HttpClientRequest connection = await getServiceConnection();
    connection.headers
      ..set(HttpHeaders.userAgentHeader, "mPOS-Flutter")
      ..set(HttpHeaders.contentTypeHeader, Defaults.MIME_XML_UTF8)
      ..set(HttpHeaders.acceptEncodingHeader, GZIP)
      ..set(HttpHeaders.contentLengthHeader,
          envelope == null ? Defaults.MISSING : utf8.encode(envelope).length);

    _addNoCacheHeaders(connection.headers);

    // Pass the headers provided by the user along with the call
    if (headers != null) {
      for (String key in headers.keys) {
        connection.headers.set(key, headers[key]);
      }
    }

    return connection;
  }

  @override
  Future<HttpClientRequest> getServiceConnection() async {
    if (conn == null) {
      conn = await createHttpClient()
          .openUrl(Utils.toDefault(method, REQ_METHOD_POST), uri);
      conn.persistentConnection = false;
      setHttpConnectionClose(conn);
    }

    return conn;
  }

  HttpClient createHttpClient() {
    HttpClient httpClient = HttpClient(context: getSecurityContext());
    httpClient
      ..autoUncompress = true
      ..findProxy = null;

    if (timeout != null && timeout > 0) {
      httpClient.connectionTimeout = Duration(milliseconds: timeout);
    }

    return httpClient;
  }

  SecurityContext getSecurityContext() => null;

  void setHttpConnectionClose(HttpClientRequest conn) {
    conn.headers.set(HttpHeaders.connectionHeader, "Close");
  }

  void _addNoCacheHeaders(HttpHeaders headers) {
    headers
      ..set(HttpHeaders.expiresHeader, DateUtils.millisToDate(0))
      ..set(HttpHeaders.lastModifiedHeader, DateUtils.now().toString())
      ..set(HttpHeaders.cacheControlHeader,
          "no-store, no-cache, must-revalidate, private, max-age=0, max-stale=0, post-check=0, pre-check=0")
      ..set(HttpHeaders.pragmaHeader, "no-cache");
  }
}
