/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:io' show HttpClient, SecurityContext, Platform, X509Certificate;
import 'dart:typed_data' show Uint8List;

import '../../../../../util/utils.dart';
import 'mpos_http_transport_se.dart';

/// Discussion with Remi (18th Nov '21):
///
/// Case 1: backend fully trusted: just use below code (for Java)
///   URL url = new URL("https://wikipedia.org");
///   URLConnection urlConnection = url.openConnection();
///   InputStream in = urlConnection.getInputStream();
///   copyInputStreamToOutputStream(in, System.out);
///
/// Case 2: not fully trusted
/// 	-> option 1: ignore the warning and continue => WE DO NOT WANT. Because not secure
/// 	-> option 2: get backend certificate key / password and load into client. Use ONLY that as "trusted" => We do that
///
/// References:
///   https://developer.android.com/training/articles/security-ssl.html
///   https://www.codeproject.com/Articles/826045/Android-security-Implementation-of-Self-signed-SSL
///
/// Sample certificate.pem is stored in 'test/resources' folder.
///
/// More details are in SSLScenarios.txt stored in 'test/resources' folder.
class MposHttpsTransportSE extends MposHttpTransportSE {
  static const String delimiter = "-----END CERTIFICATE-----";

  static SecurityContext _context; // cache context once it's already setup

  final Uint8List keyIs;
  final String keyPass; // TODO IF this is not required, remove?

  MposHttpsTransportSE(String url, int timeout, this.keyIs, this.keyPass,
      {String method})
      : super(url, timeout, method: method);

  @override
  HttpClient createHttpClient() => super.createHttpClient()
    ..badCertificateCallback = _badCertificateCallback;

  @override
  SecurityContext getSecurityContext() {
    if (keyIs != null && _context == null) {
      _context = SecurityContext.defaultContext;
      // The key is NOT required.
      // The key is ONLY required if we use PKCS format that requires the server private key to be generated, and thus has to have the server private key to be opened.
      // There is no reason to bundle the server private key with the app, and the java app is not doing that, so we don't need to pass the key.
      _context.setTrustedCertificatesBytes(keyIs);
    }

    return _context;
  }

  // TODO-May How to reject expired certificate? Can we decode pem content and get certificate data?
  bool _badCertificateCallback(X509Certificate cert, String host, int port) {
    if (keyIs == null) {
      return false;
    }
    List<String> pemArray = Utils.toUtf8(keyIs).split(delimiter);
    String certificate = "";

    // TODO-Tanveer Move DeviceUtils.getPlatform to Utils class
    if (Platform.isAndroid) {
      if (pemArray.length > 0) {
        certificate = pemArray[0];
      }
    } else {
      if (pemArray.length > 1) {
        certificate = pemArray[1].replaceFirst("\n", "");
      }
    }

    return cert.pem == [certificate, "$delimiter\n"].join("");
  }
}
