/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../../form/bean/server_response.dart';
import '../../../../../contract/i_rmi.dart';
import '../../../../../error/illegal_argument_exception.dart';
import '../../../../../structure/i18n/option.dart';
import '../../../../../structure/ui/method/method.dart';
import '../../../../../util/form_utils.dart';
import '../../../../../util/log/logger.dart';
import '../../../../../util/stop_watch.dart';
import '../../../../../util/utils.dart';
import '../../../method_state.dart';
import '../../../renderer/features.dart';
import '../../../renderer/ui_renderer.dart';
import '../../../task/sync_task.dart';
import '../../../view_state.dart';
import 'array_cycle.dart';
import 'rmi_cycle.dart';

class RMIAction {
  final UIRenderer renderer;

  const RMIAction(this.renderer);

  Future<Object> invoke(String methodId, ViewState vSrc) async {
    renderer.closePopup(); // in any case

    // ViewState can be null when we don't have it to invoke RMI
    // if not immediate, validate the configured views before invocation.
    return Features.getFormActions(renderer).validateRegion(vSrc)
        ? await invoke_(methodId, vSrc, false)
        : null;
  }

  Future<List<Option>> invokeFetchList(String methodId) async {
    return await invoke_(methodId, null, true) as List<Option>;
  }

  Future<Object> invoke_(String methodId, ViewState vSrc, bool array) async {
    Method method =
        FormUtils.getUI_(renderer).getMethod(methodId); // locate method

    if (method == null) {
      throw IllegalArgumentException("Method missing: $methodId");
    }

    IRmi iRmi = method.getImpl();
    MethodState methodState = renderer.getMethodState(method);
    return await invoke__(
        vSrc, methodState, iRmi, getCycle(vSrc, array, methodState, iRmi));
  }

  Future<Object> invoke__(ViewState vSrc, MethodState methodState, IRmi iRmi,
      RMICycle cycle) async {
    StopWatch stopWatch = StopWatch.start();
    Method method = methodState.getMethod();

    if (method.isEnabled(renderer)) {
      try {
        if (!methodState.isBlocked() && method.isReady_(iRmi, vSrc, renderer)) {
          if (await checkNetwork(method, iRmi, vSrc, cycle)) {
            // don't run in BG thread as the action listener will already be
            await SyncTask(cycle).start();
          } else {
            methodState.incrementAttempt();
          }
        }
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
        cycle.onError(e, iRmi);
      } finally {
        if (methodState.isBlocked()) {
          onDiscontinue(vSrc, method, true);
        }
      }
    } else {
      onDiscontinue(vSrc, method, false);
    }

    StopWatch.stop_(method, stopWatch);
    return cycle.getResult();
  }

  RMICycle getCycle(
      ViewState vSrc, bool array, MethodState methodState, IRmi iRmi) {
    return array
        ? ArrayCycle(methodState, iRmi, renderer)
        : RMICycle(methodState, iRmi, vSrc, renderer);
  }

  Future<bool> checkNetwork(
      Method method, IRmi iRmi, ViewState view, RMICycle cycle) async {
    final bool available =
        await cycle.getController().getRMI(method).isNetworkAvailable();

    if (!available) {
      final bool required = method.isNetworkRequired(renderer);
      final String errKey = required
          ? method.getNetworkRequiredMsg()
          : method.getNetworkOptionalMsg();
      final String errMsg = renderer.get(errKey);

      if (!Utils.isBlank(errMsg) && errKey != errMsg) {
        Features.getAlert_(renderer).popup(errMsg);
      }

      if (required) {
        // nothing to do here
      } else if (method.onPost_(
          false, ServerResponse(), iRmi, view, cycle, renderer)) {
        Features.getFormActions(renderer).toggleViews(view);
      }
    }

    return available;
  }

  void onDiscontinue(ViewState vSrc, Method method, bool blocked) {
    if (blocked ? method.isToggleIfBlocked() : method.isToggleIfDisabled()) {
      Features.getFormActions(renderer).toggleViews(vSrc);
    }

    Features.getAlert_(renderer).error(renderer
        .get(blocked ? method.getBlockedMsg() : method.getDisabledMsg()));
  }
}
