/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../../form/bean/register/crm_form_data.dart';
import '../../../../../../form/bean/register/response_message.dart';
import '../../../../../../form/bean/server_response.dart';
import '../../../../../../form/structure/string_object_map.dart';
import '../../../../../contract/i_rmi.dart';
import '../../../../../error/no_changes_exception.dart';
import '../../../../../structure/ui/method/method.dart';
import '../../../../../structure/ui/method/response.dart';
import '../../../../../structure/ui/method/response_tag.dart';
import '../../../../../util/form_utils.dart';
import '../../../../../util/log/logger.dart';
import '../../../../../util/utils.dart';
import '../../../component/alert.dart';
import '../../../form_controller.dart';
import '../../../method_state.dart';
import '../../../renderer/features.dart';
import '../../../renderer/ui_renderer.dart';
import '../../../task/bg_cycle.dart';
import '../../../view_state.dart';
import '../../data/data_comparator.dart';
import '../../rmi/error/server_unavailable_exception.dart';
import '../action_handler.dart';
import 'base_irmi.dart';
import 'method_data.dart';
import 'soap_rmi.dart';

class RMICycle implements BgCycle {
  final MethodState methodState;
  final Method method;
  final IRmi iRmi;
  final ViewState vSrc;
  final UIRenderer renderer;

  Object result;

  RMICycle(this.methodState, this.iRmi, this.vSrc, this.renderer)
      : this.method = methodState.getMethod();

  @override
  void preRun() {
    // nothing to do here
  }

  @override
  Future<Object> run() async {
    try {
      if (method.detectChange_()) {
        // doesn't include all properties
        final Map<String, Object> data = prepareData(false);

        if (!DataComparator.isChanged(
            methodState.getLastPreparedData(), data)) {
          throw NoChangesException();
        }

        methodState.setLastPreparedData(data);
      }

      final Map<String, Object> data = prepareData(true);
      return await getController().getRMI(method).invoke(
          method.getService(),
          methodState.getName(),
          method.getNamespace(),
          method.getTimeout(),
          method.getData().getBean(),
          data);
    } catch (e) {
      return e;
    }
  }

  @override
  Future<void> postRun(Object response) async {
    if (response is ServerUnavailableException) {
      response = ServerResponse.response(1503, response.toString());
    }

    this.result = response;

    if (response is NoChangesException) {
      onNoChange();
    } else if (response is Exception) {
      onError(response, iRmi);
    } else {
      try {
        await handleResponse(response as ServerResponse);
      } catch (e) {
        onError(e, iRmi);
      }
    }
  }

  Object getResult() {
    return this.result;
  }

  /// Prepares the data map to be passed to Web Service/CMI.
  Map<String, Object> prepareData(bool all) {
    Map<String, Object> data = MethodData(Features.getResolver(renderer))
        .prepare(method.getData(), all);

    if (method.isStandardRmi()) {
      final String methodName = method.getName();
      data[SoapRMI.PROP_REQUEST_OPERATION_NAME] = methodName;

      CrmFormData formData = data[SoapRMI.CRM_FORM_DATA] as CrmFormData;

      if (formData != null) {
        formData.getMap().put("RMI_TASK_NAME", methodName);
      }
    }

    method.onPrepare_(all, data, iRmi, vSrc, renderer);
    return data;
  }

  /// 1) Figure out if the remote method invocation was a success.<br>
  /// 2) Call the controller's method, if configured.<br>
  /// 3) On a successful flag return, render/enable configured views.<br>
  /// 4) Alert the success/failure message, if configured.
  Future<void> handleResponse(ServerResponse serverResponse) async {
    FormController controller = getController();

    if (ServerResponse.isSecurityTokenExpired(serverResponse)) {
      controller.alertSecurityExpired();
      return;
    }

    Response response = method.getResponse();
    final bool noResponse = serverResponse == null;
    bool success = noResponse ||
        Utils.got(
            getResponseCode(serverResponse), response.getSuccess().getCodes());

    if (!noResponse) {
      gotResponse(success, serverResponse, response);
    }

    success =
        method.onPost_(success, serverResponse, iRmi, vSrc, this, renderer);

    if (success) {
      methodState.resetAttempts();

      Map<String, Object> data = serverResponse.getData();

      if (!Utils.isBlank(data)) {
        Features.getDataHandler(renderer)
            .handle(StringObjectMap.fromMap(data), null);
      }

      Features.getFormActions(renderer).toggleViews(vSrc);
      await ActionHandler.doPostAction(vSrc, false, false, null, renderer);
    } else {
      methodState.incrementAttempt();
      // TODO Toggle other configured views for failed case?
    }
  }

  /// Shows the popup if configured. If not, alerts the configured message for the server response.
  void gotResponse(
      bool success, ServerResponse serverResponse, Response response) {
    ResponseTag rTag = success
        ? response.getSuccess()
        : response.getFailure(); // Unmatched code leads to failure
    final String popupId = rTag.getPopup();

    if (Utils.isBlank(popupId)) {
      // If popup is undefined, show alert.
      ResponseMessage custom = serverResponse.getCustomMessage();
      String msg = custom == null
          ? ""
          : custom.getMessage(FormUtils.getLocale_(renderer).getId());

      if (Utils.isBlank(msg)) {
        final String msgId = getMsgId(getResponseCode(serverResponse), rTag);

        if (!Utils.isBlank(msgId)) {
          // if no msg is provided, we don't want to alert
          msg = renderer.get(msgId, [serverResponse.getResponseDesc()]);
        }
      }

      Alert alert = Features.getAlert_(renderer);

      if (success) {
        alert.info(msg);
      } else {
        alert.error(msg);
      }
    } else {
      Features.getFormActions(renderer).showPopup_(popupId);
    }
  }

  /// If there was no change in the current request data compared to the previous one.
  void onNoChange() {
    if (method.isToggleIfNoChange()) {
      Features.getFormActions(renderer).toggleViews(vSrc);
    }

    Features.getAlert_(renderer).info(method.getNoChangeMsg());
  }

  /// If there is any error occurred during RMI, either we alert it or we invoke the configured method to handle it.
  void onError(Exception e, IRmi iRmi) {
    methodState.incrementAttempt();

    Logger.error(e);
    String errMsg;

    try {
      if (method.onError_(e, iRmi, vSrc, renderer)) {
        errMsg = renderer.get(BaseIRMI.isDbErr(e)
            ? BaseIRMI.DB_UNAVAILABLE
            : BaseIRMI.UNAVAILABLE);
      }
    } catch (e1, stackTrace) {
      errMsg = e1.toString(); // This one should be pure technical error
      Logger.error(e1, stackTrace: stackTrace);
    }

    // When found error, just show popup with error message
    Features.getAlert_(renderer).error(errMsg);
  }

  String getResponseCode(ServerResponse serverResponse) {
    return serverResponse.getResponseCode().toString();
  }

  String getMsgId(String code, ResponseTag responseTag) {
    String msgId = responseTag.getCodes()[code];
    return Utils.isBlank(msgId)
        ? responseTag.getMsg()
        : msgId; // Use default message id if not provided.
  }

  FormController getController() {
    return renderer.getController();
  }
}
