/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:io' show IOException;
import 'dart:typed_data' show Uint8List;

import '../../../../../../form/bean/register/crm_form_data.dart';
import '../../../../../../form/bean/server_response.dart';
import '../../../../../config/config.dart';
import '../../../../../enums/code.dart';
import '../../../../../util/log/logger.dart';
import '../../../../../util/utils.dart';
import '../../data/rmi.dart';
import '../../rmi/error/remote_service_exception.dart';
import '../../rmi/error/server_unavailable_exception.dart';
import '../../rmi/soap/soap_builder.dart';
import '../../rmi/soap/soap_deserializer.dart';
import 'base_irmi.dart';
import 'transport.dart';

abstract class SoapRMI implements RMI {
  static const String PROP_RESPONSE_CODE = "responseCode";
  static const String PROP_RESPONSE_DESC = "responseDesc";

  static const String CRM_FORM_DATA = "crmFormData";
  static const String MPOS_INFO_DATA = "mPosInfoData";
  static const String PROP_REQUEST_SECURITY_TOKEN = "securityToken";
  static const String PROP_REQUEST_INSTANCE_ID = "instanceId";
  static const String PROP_REQUEST_OFFLINE_LOGIN_FLAG = "offlineLoginFlag";
  static const String PROP_REQUEST_OPERATION_NAME = "operationName";

  final String endpoint;
  final String certificate;
  final String certificateKey;
  final int defaultTimeout;
  final SoapBuilder soapBuilder;

  SoapRMI(
      this.endpoint, this.certificate, this.certificateKey, this.defaultTimeout)
      : this.soapBuilder = SoapBuilder();

  @override
  Future<ServerResponse> invoke(String service, String method, String namespace,
      int timeout, String clazz, Map<String, Object> data) async {
    try {
      String soap =
          await invokeWS__(service, method, namespace, timeout, data, false);
      return SoapDeserializer.deserialize(soap);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      throw e;
    }
  }

  Future<ServerResponse> invoke_(String service, String method,
      String namespace, Map<String, Object> data) async {
    return await invoke(service, method, namespace, null, null, data);
  }

  Future<String> invokeWS(String service, String method, String namespace,
      Map<String, Object> data) async {
    return await invokeWS_(service, method, namespace, null, data);
  }

  Future<String> invokeWS_(String service, String method, String namespace,
      int timeout, Map<String, Object> data) async {
    return await invokeWS__(service, method, namespace, timeout, data, true);
  }

  Future<String> invokeWS__(String service, String method, String namespace,
      int timeout, Map<String, Object> data, bool checkNetwork) async {
    if (checkNetwork && !await isNetworkAvailable()) {
      return null;
    }

    int lTimeout = Utils.toDefault(timeout, defaultTimeout);
    namespace = Utils.toDefault(namespace, Config.NAMESPACE);
    String soapAction = getSoapAction(namespace, method);

    final String soap = soapBuilder.build(method, namespace, data);
    return await callWS(
        soap, "$endpoint$service?wsdl", soapAction, lTimeout, data);
  }

  Future<String> callWS(String soap, String methodEndpoint, String action,
      int timeout, Map<String, Object> data) async {
    final bool secured = Code.HTTPS.starts(methodEndpoint);
    Uint8List cert = secured ? await getCertificate(certificate) : null;

    try {
      final String envelope = createEnvelope(soap, data);
      Transport transport = Transport.init(
          secured, methodEndpoint, timeout, cert, certificateKey);
      return await transport.call(envelope);
      // FIXME
// } catch (SSLProtocolException | SSLHandshakeException e) {
// logError(methodEndpoint, action, e);
// throw SSLProtocolException(e.toString());
// } catch (SSLException e) {
// logError(methodEndpoint, action, e);
// throw SSLException(e.toString());
    } on IOException catch (e) {
      logError(methodEndpoint, action, e);
      throw ServerUnavailableException(getMsg(BaseIRMI.UNAVAILABLE));
    } catch (e) {
      throw RemoteServiceException(e.toString());
    }
  }

  void logError(String methodEndpoint, String action, Exception e) {
    Logger.error("Can't connect to host : ${e.toString()}");
    Logger.error("endpoint=$methodEndpoint, action=$action");
    Logger.error(e);
  }

  String createEnvelope(String soap, Map<String, Object> data) {
    StringBuffer envelope = StringBuffer('''
      <v:Envelope
        xmlns:i="http://www.w3.org/2001/XMLSchema-instance"
        xmlns:d="http://www.w3.org/2001/XMLSchema"
        xmlns:c="http://schemas.xmlsoap.org/soap/encoding/"
        xmlns:v="http://schemas.xmlsoap.org/soap/envelope/">
    ''');
    addHeaders(data, envelope);
    envelope.write('<v:Body>$soap</v:Body>');
    envelope.write('</v:Envelope>');
    return envelope.toString();
  }

  void addHeaders(Map<String, Object> data, StringBuffer envelope) {
    envelope.write('<v:Header>');
    // TODO Should namespace here be hard-coded?
    envelope.write('<n0:info xmlns:n0="http://mobile-technologies.com">');

    addAuthElement(PROP_REQUEST_SECURITY_TOKEN, data, envelope);
    addAuthElement(PROP_REQUEST_INSTANCE_ID, data, envelope);
    addAuthElement_(
        PROP_REQUEST_OFFLINE_LOGIN_FLAG,
        (data[CRM_FORM_DATA] as CrmFormData)
            ?.getmPosInfoData()
            ?.isOfflineLogin(),
        envelope);
    addAuthElement(PROP_REQUEST_OPERATION_NAME, data, envelope);

    envelope.write('</n0:info>');
    envelope.write('</v:Header>');
  }

  void addAuthElement(
      String property, Map<String, Object> data, StringBuffer header) {
    addAuthElement_(property, data[property], header);
  }

  void addAuthElement_(String property, Object value, StringBuffer header) {
    if (value == null) {
      Logger.warn("Auth property $property is missing from SoapObject.");
    } else {
      header.write('<$property>$value</$property>');
    }
  }

  static String getSoapAction(String namespace, String method) {
    return '"$namespace$method"';
  }

  String getActionNamespace(String action) {
    try {
      List<String> slashSplit = action.replaceAll('"', '').split('/');
      return '${slashSplit[0]}/${slashSplit[1]}/${slashSplit[2]}';
    } catch (e) {
      return '';
    }
  }

  int getResponseCode(Map soap) {
    return int.parse(soap[PROP_RESPONSE_CODE]);
  }

  String getResponseDescription(Map soap) {
    return soap[PROP_RESPONSE_DESC];
  }

  Future<Uint8List> getCertificate(String certificate);

  String getMsg(String msgId, [Object args]);
}
