/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../../form/structure/ordered_map.dart';
import '../../../../config/defaults.dart';
import '../../../../enums/code.dart';
import '../../../../enums/field_type.dart';
import '../../../../generator/reader/json_reader.dart';
import '../../../../structure/model/field.dart';
import '../../../../util/base64.dart';
import '../../../../util/date_utils.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/numbers.dart';
import '../../../../util/utils.dart';
import '../../../../validation/converter/c_binary.dart';
import '../../../../validation/converter/c_boolean.dart';
import '../../component/base/component.dart';
import '../../component/table/table.dart';
import '../../renderer/features.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';

class TableActions {
  final UIRenderer renderer;

  const TableActions(this.renderer);

  /// Adds a new row or updates an existing row in the table.
  bool doAdd(ViewState vButton) {
    List<String> params = Utils.splitByCommaAsList(vButton.getParams());

    if (Utils.isBlank(params)) {
      return false;
    }

    final int valuesCount = params.length - 1;
    List<String> row = []; // row data that will be added/updated in the table
    Table
        table; // Table view to which the provided view's data will be pushed to
    List<String> referencedViews = [];

    for (int index = 0; index <= valuesCount; index++) {
      final String value = params[index];

      if (index == 0) {
        // first parameter should be the id of an existing table view
        table = getTableIfValid(value);

        if (table == null) {
          Logger.warn("Table '$value' is missing or it's invalid.");
          return false;
        }
      } else {
        row[index - 1] = toString_(FormUtils.resolve_(value, renderer));

        if (Code.VIEW_VALUE.starts(value)) {
          referencedViews.add(Code.VIEW_VALUE.remove(value));
        }
      }
    }

    if (Utils.isBlankDeep(row) ||
        !Features.getValidator(renderer)
            .run__(referencedViews, false, true, true)) {
      return false;
    }

    final int selectedRowIndex = table.getSelectedRowIndex();
    final int rowIndex = getRowIndex(row, getRows(table));

    if (!isRowMissing(rowIndex) && rowIndex != selectedRowIndex) {
      table.setErrMsg(renderer.get("table.row.exists", [rowIndex + 1]));
      return false;
    }

    saveRow(row, isRowMissing(selectedRowIndex), table);

    for (String viewId in referencedViews) {
      clearView(viewId); // If value is pushed to table, clear the source view
    }

    FormUtils.fireOnValueChange_(false, table);

    return true;
  }

  /// Adds a new row in the table during the CSV import process.
  bool doAddForImport(
      Table table, List<String> rowData, List<ViewState> vSrcs) {
    if (Utils.isBlankDeep(rowData)) {
      return false;
    }

    final int rowIndex = getRowIndex(rowData, getRows(table));

    if (!isRowMissing(rowIndex)) {
      return false;
    }

    int index = Defaults.MISSING;
    bool isValid = true;

    for (ViewState vSrc in vSrcs) {
      if (vSrc != null) {
        isValid = Features.getValidator(renderer)
                .runWithoutMarking(vSrc, rowData[++index]) &&
            isValid;
      }
    }

    if (isValid) {
      saveRow(rowData, true, table);
    }

    return isValid;
  }

  /// Remove all rows in the table.
  void doClear(ViewState vButton) {
    Table table = getTable(FormUtils.getView_(vButton.getParams(), renderer));

    if (table != null) {
      doClear_(table);
      FormUtils.fireOnValueChange_(false, table);
    }
  }

  void doClear_(Table table) {
    List<List<String>> rows = getRows(table);

    if (!Utils.isBlank(rows)) {
      Utils.clear(rows);
      table.onClear();
    }

    clearSelectedRow(table);
    table.setErrMsg(null);
  }

  /// Clears the source views and brings back table to addition mode.
  void doCancel(ViewState vButton) {
    ViewState vTable = FormUtils.getView_(vButton.getParams(), renderer);
    Table table = getTable(vTable);

    if (table != null) {
      clearSelectedRow(table); // go back to normal CRUD mode
      table.setEnabled(true);
      table.setErrMsg(null);

      // iterates on all the columns and clear the values of their mapped views.
      for (ViewState view in vTable.getChildren()) {
        clearView(view.getParams());
      }
    }
  }

  /// Find the ViewState of the given viewId and then it's Table component.
  ///
  /// If table is going to have a new row, check if it already has reached allowed maximum.
  Table getTableIfValid(String viewId) {
    ViewState vTable = FormUtils.getView_(viewId, renderer);
    Table table = getTable(vTable);

    if (table != null && isRowMissing(table.getSelectedRowIndex())) {
      // new row
      if (!table.isEnabled()) {
        // can't perform add operation on disabled table
        table.setErrMsg(renderer.get("table.disabled"));
        return null;
      }

      if (maxReached(table, vTable.getField())) {
        table.setErrMsg(renderer.get(Defaults.MSG_TABLE_MAX_REACHED));
        return null;
      }
    }

    return table;
  }

  /// Checks if Table has reached the allowed maximum.
  bool maxReached(Table table, Field field) {
    final int max = FormUtils.getMaxAllowed(field, renderer);
    return max != null && Utils.getSize(getRows(table)) >= max;
  }

  String toString_(Object val) {
    if (val is DateTime) {
      val = DateUtils.formatDefault(val); // TODO Use pattern from XML
    } else if (CBinary.isBinary(val)) {
      val = Base64.encode_(val);
    }

    return Utils.toString_O(val);
  }

  void clearView(String viewId) {
    ViewState view = FormUtils.getView_(viewId, renderer);

    if (view != null && view.isRendered() && view.isEnabled()) {
      // no need to clear view's value is hidden/disabled
      view.setValue(
          FieldType.BOOLEAN == view.getFieldType() ? CBoolean.NO : null);
    }
  }

  /// Saves or updates the given row data in the table.
  void saveRow(List<String> row, bool newRow, Table table) {
    List<List<String>> rows =
        getRows(table); // TODO Verify the row length as per the columns count?

    if (newRow) {
      if (rows == null) {
        rows = [];
        table.setVal(rows);
      }

      rows.add(row);
      table.onRowAdd();
    } else {
      // update mode
      if (rows != null) {
        rows[table.getSelectedRowIndex()] = row;
        table.onRowUpdate();
      }

      table.setEnabled(true);
      clearSelectedRow(table);
    }

    table.setErrMsg(null); // clears the message
  }

  static List<List<String>> getRows(Table table) {
    return table.getVal() as List<List<String>>;
  }

  /// Verifies if the candidate new row already exists in the table.
  ///
  /// TODO If there is enabled tickColumn, this can be misleading. E.g. {false, 1, 2} != {true, 1, 2}
  int getRowIndex(List<String> newRow, List<List<String>> rows) {
    if (Utils.isBlank(rows)) {
      return Defaults.MISSING;
    }

    // here we assume that provided array length is same as existing rows.
    final int size = newRow.length; // # of columns
    int rowIndex = Defaults.MISSING;

    for (List<String> row in rows) {
      rowIndex++;
      bool identical = true;

      for (int index = 0; index < size; index++) {
        final String x = Numbers.toDigitsCharByChar(newRow[index].trim());
        final String y = Numbers.toDigitsCharByChar(row[index].trim());

        if (!Utils.equalsIgnoreCase(x, y)) {
          identical = false;
          break;
        }
      }

      if (identical) {
        return rowIndex;
      }
    }

    return Defaults.MISSING;
  }

  void clearSelectedRow(Table table) {
    table.setSelectedRowIndex(Defaults.MISSING); // no row is selected.
  }

  /// Assigns the selected row data to mapped views and keeps the selected row index in the table for future.
  static void editRow(String tableId, OrderedMap<ViewState> allViews) {
    ViewState vTable = allViews[tableId];
    Table table = getTable(vTable);

    if (table != null) {
      table.markSelectedRowIndex();

      List<String> selectedRow = table.getSelectedRow();
      int counter = Defaults.MISSING;

      // iterates on all the columns and assign their values to mapped views.
      for (ViewState view in vTable.getChildren()) {
        ViewState vSrc = allViews[view.getParams()];

        if (vSrc != null) {
          vSrc.setValue(selectedRow[++counter]);
        }
      }

      table.setEnabled(
          false); // disable the table so that further actions can't happen
      table.setErrMsg(null);
    }
  }

  /// Remove the selected row from the table.
  static bool removeRow(String tableId, OrderedMap<ViewState> allViews) {
    return removeRow_(getTable(allViews[tableId]));
  }

  /// Remove the selected row from the table.
  static bool removeRow_(Table table) {
    if (table == null) {
      return false;
    }

    table.setErrMsg(null);
    final bool removed = getRows(table).remove(table.getSelectedRow());

    if (removed) {
      FormUtils.fireOnValueChange_(false, table);
    }

    return removed;
  }

  static List<List<String>> toList(String json) {
    // FIXME fromJson won't convert json to List
    List list = JsonReader().fromJson(json, List);
    List<List<String>> items;

    if (!Utils.isBlank(list)) {
      items = [];

      for (Object entry in list) {
        items.add(Utils.toArray(entry as List));
      }
    }

    return items;
  }

  /// Extract the Table component from the it's ViewState.
  static Table getTable(ViewState view) {
    Component component = view == null ? null : view.getComponent();
    return component is Table ? component : null;
  }

  bool isRowMissing(int rowIndex) {
    return rowIndex == Defaults.MISSING;
  }
}
