/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../config/defaults.dart';
import '../../../../util/form_utils.dart';
import '../../../../util/log/logger.dart';
import '../../../../util/utils.dart';
import '../../component/alert.dart';
import '../../component/table/table.dart';
import '../../form_controller.dart';
import '../../renderer/features.dart';
import '../../renderer/ui_renderer.dart';
import '../../view_state.dart';
import 'table_actions.dart';

class TableImporter {
  final Table table;
  final UIRenderer renderer;
  final ViewState view;

  TableImporter(this.table, FormController controller)
      : this.renderer = controller.getRenderer(),
        this.view = FormUtils.getView(table.getIdentity(), controller);

  void onFileUpload(List<int> data) {
    try {
      doImport(Utils.toUtf8(data));
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    }
  }

  void doImport(String content) {
    List<List<String>> csv =
        Utils.isBlank(content) ? null : parseContent(content);
    Alert alert = Features.getAlert_(renderer);

    if (Utils.isBlank(csv)) {
      alert.error(renderer.get("table.import.nothing"));
    } else {
      TableActions tableActions = Features.getTableActions(renderer);
      tableActions.doClear_(table); // TODO Make it configurable
      List<int> faultyRows = addRows(csv, tableActions);

      FormUtils.fireOnValueChange_(true, table);

      final String faultyLines =
          Utils.isBlank(faultyRows) ? "-" : faultyRows.toString();
      // TODO Enhance formatting
      alert.info(renderer.get("table.import.success",
          [table.getRowCount(), csv.length, faultyLines]));
    }
  }

  List<int> addRows(List<List<String>> csv, TableActions tableActions) {
    int max = FormUtils.getMaxAllowed(view.getField(), renderer);
    int rowIndex = 0, importCount = 0;
    List<int> faultyRows = [];

    List<ViewState> vSrcs = getSrcViews();

    for (List<String> row in csv) {
      if (max != null && importCount >= max) {
        table.setErrMsg(renderer.get(Defaults.MSG_TABLE_MAX_REACHED));
        break;
      }

      rowIndex++;

      if (tableActions.doAddForImport(table, row, vSrcs)) {
        importCount++;
      } else {
        faultyRows.add(rowIndex);
      }
    }

    return faultyRows;
  }

  List<ViewState> getSrcViews() {
    List<ViewState> vColumns = view.getChildren();
    List<ViewState> vSrcs = [];
    int index = Defaults.MISSING;

    for (ViewState vColumn in vColumns) {
      vSrcs[++index] = FormUtils.getView_(vColumn.getParams(), renderer);
    }

    return vSrcs;
  }

  List<List<String>> parseContent(String content) {
    List<List<String>> csv = [];
    final int columnCount = table.getColumnCount();

    for (String row in content.split(Defaults.EOL)) {
      List<String> columns = Utils.split(row, ",");
      columns =
          Utils.isBlank(columns) ? Utils.blankArray(columnCount) : columns;

      // if columns in the row are not what's expected, we create new empty array of correct size and copy
      // provided columns into it.
      if (columns.length != columnCount) {
        List<String> original = columns;
        columns = Utils.blankArray(columnCount);
        int lastIndex = original.length;
        lastIndex = lastIndex > columnCount ? columnCount : lastIndex;

        for (int index = 0; index < lastIndex; index++) {
          columns[index] = original[index];
        }
      }

      csv.add(columns);
    }

    return csv;
  }

  String getViewId() {
    return view.getId();
  }
}
