/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../structure/i18n/i18n.dart';
import '../../../structure/i18n/message_array.dart';
import '../../../util/form_utils.dart';
import '../../../util/stop_watch.dart';
import '../../../util/utils.dart';
import '../../../validation/converter/c_string.dart';
import '../component/behavior/i_array.dart';
import '../renderer/features.dart';
import '../renderer/ui_renderer.dart';
import '../util/key_value_pair.dart';
import '../view_state.dart';
import 'action_base.dart';
import 'handler/action_handler.dart';
import 'relative_render_action.dart';

class ReloadAction extends ActionBase {
  const ReloadAction(ViewState view, ActionHandler handler)
      : super(view, handler);

  void reload(Object optionKey) {
    if (handler != null) {
      StopWatch stopWatch = StopWatch.start();

      if (view.canArray()) {
        reload_(view.getParentOf(), view.getArray(), optionKey);
        RelativeRenderAction(view, handler).render(optionKey);

        UIRenderer renderer = getRenderer();
        Features.getReplicator(renderer).perform(view);
        Features.getDataActions(renderer).doClearAndReset(view);
      }

      Set<ViewState> dependents = view.getDependents();

      if (!Utils.isBlank(dependents)) {
        reloadArray(dependents);
      }

      StopWatch.stop_(optionKey, stopWatch);
    }
  }

  void reloadArray(Set<ViewState> views) {
    UIRenderer renderer = getRenderer();
    I18n i18n = FormUtils.getI18n(renderer);

    for (ViewState view in views) {
      if (view.canArray()) {
        MessageArray array = i18n.getMsgArray(view.getArray());

        if (array != null && array.isChangeable()) {
          final Object value = view.getValue(); // keep the preset value

          IArray iArray = view.getInputComponent() as IArray;
          iArray.clearArray();

          Set<KeyValuePair> entries = i18n.toText(array, renderer);

          if (!Utils.isBlank(entries)) {
            // 'null' sustains the 'blank message'
            iArray.setArray(entries, array.hasBlankItem(), null);
          }

          // onValueChange listener will itself invoke the hierarchical reloading of linked arrays
          view.setValue(iArray.isValidItem(value) ? value : null);
        }
      }
    }
  }

  void reload_(String parentOf, String arrayId, Object optionKey) {
    // Get the view linked to the view on which the change is detected.
    ViewState vRelated = FormUtils.getView_(parentOf, getRenderer());
    reload__(vRelated, arrayId, optionKey); // Reload the related view's array
  }

  void reload__(ViewState view, String parentArrayId, Object optionKey) {
    if (view != null) {
      view.setValue(null); // Nullify the component's value

      // Load the array for the component.
      String arrayId = resetArray(view, parentArrayId, optionKey);

      view.setArray(arrayId);

      // Reload connected view's in the hierarchy.
      reload_(view.getParentOf(), arrayId, null);

      UIRenderer renderer = getRenderer();
      Features.getReplicator(renderer).perform(view);
      Features.getDataActions(renderer).doClearAndReset(view);
    }
  }

  String resetArray(ViewState view, String parentArrayId, Object optionId) {
    IArray iArray = view.getComponent() as IArray;
    iArray.clearArray();

    UIRenderer renderer = getRenderer();
    I18n i18n = FormUtils.getI18n(renderer);
    final String definedArrayId = view.getDefinedArray();
    MessageArray array = Utils.isBlank(definedArrayId)
        ? i18n.getMsgArray_(parentArrayId, CString.convert(optionId))
        : i18n.getMsgArray(definedArrayId);

    if (array != null) {
      Set<KeyValuePair> entries = i18n.toText(array, renderer);

      if (!Utils.isBlank(entries)) {
        iArray.setArray(
            entries, array.hasBlankItem(), renderer.get(array.getBlankMsg()));
      }
    }

    return array == null ? null : array.getId();
  }
}
