/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/structure/ordered_map.dart';
import '../../../config/defaults.dart';
import '../../../enums/code.dart';
import '../../../enums/scope.dart';
import '../../../util/stop_watch.dart';
import '../../../util/structure/string_set.dart';
import '../../../util/utils.dart';
import '../renderer/ui_renderer.dart';
import '../view_state.dart';
import 'handler/action_handler.dart';

class ToggleAction {
  const ToggleAction();

  // when prefix symbol isn't provided as it is dependent on the component's boolean flag like in the case of TickBox.
  void toggleRender(ViewState view, bool show, ActionHandler handler) {
    StopWatch stopWatch = StopWatch.start();
    UIRenderer renderer = handler.getRenderer();
    StringSet viewIds = getViewIds(view, true, renderer);

    if (!Utils.isBlank(viewIds)) {
      StringSet idsWithSign = StringSet();
      Code code = show ? Code.INCLUDE : Code.EXCLUDE;

      for (String viewId in viewIds) {
        idsWithSign.add(code.prefix(viewId));
      }

      try {
        view.setRender(idsWithSign);
        handler.performAction(view);
      } finally {
        view.setRender(viewIds);
      }
    }

    StopWatch.stop_(view, stopWatch);
  }

  bool toggleRender_(ViewState view, UIRenderer renderer) {
    return toggle(view, true, renderer);
  }

  bool toggleEnable(ViewState view, UIRenderer renderer) {
    return toggle(view, false, renderer);
  }

  bool toggle(ViewState view, bool renderAttr, UIRenderer renderer) {
    StringSet viewIds = getViewIds(view, renderAttr, renderer);

    if (Utils.isBlank(viewIds)) {
      return false;
    }

    alterAttribute(
        viewIds, renderAttr, renderer.getAllViews(), renderer.getInputViews());
    return true;
  }

  void alterAttribute(StringSet viewIds, bool renderAttr,
      OrderedMap<ViewState> allViews, StringSet inputViews) {
    for (String viewId in viewIds) {
      viewId = viewId.trim();

      // TODO Use ValueResolver here for viewId?

      if (viewId.length < 2) {
        continue; // minimum length should be 2, one for prefix and one for viewId.
      }

      // initial character has to be '@' or '+' or '-'.
      if (Code.EFFECT.starts(viewId)) {
        if (Scope.ALLTOPVIEWS.check(viewId)) {
          setTopViewsToggled(allViews, true, renderAttr);
        } else if (Scope.NOTOPVIEWS.check(viewId)) {
          setTopViewsToggled(allViews, false, renderAttr);
        } else if (Scope.ALLINPUTVIEWS.check(viewId)) {
          setInputViewsToggled(inputViews, allViews, true, renderAttr);
        } else if (Scope.NOINPUTVIEWS.check(viewId)) {
          setInputViewsToggled(inputViews, allViews, false, renderAttr);
        } else if (Scope.ALL.check(viewId)) {
          setAllToggled(allViews, true, renderAttr);
        } else if (Scope.isNone(viewId)) {
          setAllToggled(allViews, false, renderAttr);
        }
      } else {
        setAttribute(allViews[Utils.skipFirst(viewId)],
            Code.INCLUDE.starts(viewId), renderAttr);
      }
    }
  }

  void setAllToggled(
      OrderedMap<ViewState> views, bool toggled, bool renderAttr) {
    for (ViewState view in views.values) {
      setAttribute(view, toggled, renderAttr);
    }
  }

  void setTopViewsToggled(
      OrderedMap<ViewState> views, bool toggled, bool renderAttr) {
    List<ViewState> topViews = views[Defaults.ID_ROOT].getChildren();

    if (!Utils.isBlank(topViews)) {
      for (ViewState topView in topViews) {
        setAttribute(topView, toggled, renderAttr);
      }
    }
  }

  void setInputViewsToggled(StringSet inputViews, OrderedMap<ViewState> views,
      bool toggled, bool renderAttr) {
    for (String inputViewId in inputViews) {
      setAttribute(views[inputViewId], toggled, renderAttr);
    }
  }

  StringSet getViewIds(ViewState view, bool render, UIRenderer renderer) {
    return render ? view.getRender_(renderer) : view.getEnable_(renderer);
  }

  void setAttribute(ViewState view, bool toggled, bool renderAttr) {
    if (view != null) {
      if (renderAttr) {
        view.setRendered(toggled);
      } else {
        view.setEnabled(toggled);
      }
    }
  }
}
