/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:math' show min;

import '../../../../util/form_utils.dart';
import '../../../../util/utils.dart';
import '../../../../validation/converter/c_string.dart';
import '../../component/behavior/i_input.dart';
import '../../component/behavior/i_value.dart';
import '../on_value_change_listener.dart';

class CInputText {
  static const String MASK_DIGIT = '9'; // numeric -> [0-9]
  static const String MASK_CHAR = 'a'; // alphabetical -> [A-Za-z]
  static const String MASK_DIGIT_OR_CHAR = '*'; // alphanumeric -> [A-Za-z0-9]
  static const String MASK_PLACEHOLDER = '_';

  static const String WHAT_MASK_CHAR = '#';
  static const String WHAT_MASK_NO_CHAR = '-';

  const CInputText._();

  static void setVal(Object val, IInput input, IValue<String> iValue,
      OnValueChangeListener onValueChange) {
    setVal_(val, input, iValue, false, onValueChange);
  }

  static void setVal_(Object val, IInput input, IValue<String> iValue,
      bool alwaysSetValue, OnValueChangeListener onValueChange) {
    final String oldVal = Utils.toBlank(input.getVal() as String);
    final String newVal = Utils.toBlank(CString.convert(val));
    final bool changed = oldVal != newVal;

    if (alwaysSetValue || changed) {
      iValue.setValue(newVal);
    }

    if (changed) {
      FormUtils.fireOnValueChange__(newVal, false, onValueChange);
    }
  }

  static String removeMask(String text, String mask) {
    if (Utils.anyBlank([text, mask])) {
      return text;
    }

    final int length = min(text.length, mask.length);
    StringBuffer plainText = StringBuffer();

    for (int index = 0; index < length; index++) {
      final String maskChar = mask[index];
      final String plainChar = text[index];
      Position position = Position(maskChar);

      if (position.isDigitOrChar()) {
        if (position.isCorrect(plainChar)) {
          plainText.write(plainChar);
        } else {
          break; // this will accept what's valid so far
        }
      } else if (maskChar != plainChar) {
        return ""; // mask should exactly map
      }
    }

    return plainText.toString();
  }

  static String dateToMask(String pattern) {
    // all these characters are constants in DateFormatSymbols class
    return pattern
        .replaceAll('G', MASK_CHAR)
        .replaceAll('y', MASK_DIGIT)
        .replaceAll('M', MASK_DIGIT_OR_CHAR)
        .replaceAll('d', MASK_DIGIT)
        .replaceAll('k', MASK_DIGIT)
        .replaceAll('H', MASK_DIGIT)
        .replaceAll('m', MASK_DIGIT)
        .replaceAll('s', MASK_DIGIT)
        .replaceAll('S', MASK_DIGIT)
        .replaceAll('E', MASK_DIGIT)
        .replaceAll('D', MASK_DIGIT)
        .replaceAll('F', MASK_DIGIT)
        .replaceAll('w', MASK_DIGIT)
        .replaceAll('W', MASK_DIGIT)
        .replaceAll('a', MASK_CHAR)
        .replaceAll('h', MASK_DIGIT)
        .replaceAll('K', MASK_DIGIT)
        .replaceAll('z', MASK_CHAR)
        .replaceAll('Z', MASK_CHAR)
        .replaceAll('Y', MASK_DIGIT)
        .replaceAll('u', MASK_DIGIT)
        .replaceAll('X', MASK_CHAR);
  }
}

class Position {
  final String mask;
  final String whatMask;
  final String placeholder;

  // Use this only if you don't care about whatMask and placeholder.
  const Position(String mask) : this.full(mask, ' ', ' ');

  const Position.full(this.mask, this.whatMask, this.placeholder);

  bool isPlainCharacter() {
    return whatMask == CInputText.WHAT_MASK_CHAR;
  }

  bool isCorrect(String c) {
    switch (mask) {
      case CInputText.MASK_DIGIT:
        return Utils.isDigit(c);
      case CInputText.MASK_CHAR:
        return Utils.isLetter(c);
      case CInputText.MASK_DIGIT_OR_CHAR:
        return Utils.isLetter(c) || Utils.isDigit(c);
      default:
        break;
    }

    return false;
  }

  bool isDigitOrChar() {
    return Utils.match(mask, [
      CInputText.MASK_DIGIT,
      CInputText.MASK_CHAR,
      CInputText.MASK_DIGIT_OR_CHAR
    ]);
  }
}
