/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../contract/contract_utils.dart';
import '../../../structure/model/field.dart';
import '../../../util/log/logger.dart';
import '../../../util/stop_watch.dart';
import '../../../util/utils.dart';
import '../../../util/value/luhn.dart';
import '../../../util/value/value_utils.dart';
import '../action/data/data_comparator.dart';
import '../action/handler/action_handler.dart';
import '../action/relative_render_action.dart';
import '../action/reload_action.dart';
import '../component/behavior/i_input.dart';
import '../component/input_text.dart';
import '../renderer/features.dart';
import '../renderer/ui_renderer.dart';
import '../view_state.dart';

class OnValueChangeListener {
  final IInput iInput;
  final ViewState view;
  final ActionHandler handler;

  bool doLuhn = false;
  bool doReplicate = false;
  bool doOnValueChange = false;
  RelativeRenderAction relativeRenderAction;
  ReloadAction reloadAction;

  Object currentValue;

  OnValueChangeListener(this.iInput, this.view, this.handler);

  static OnValueChangeListener init(
      IInput iInput, ViewState view, ActionHandler handler) {
    OnValueChangeListener listener =
        OnValueChangeListener(iInput, view, handler);
    final bool required = listener.isRequired();

    if (required) {
      final Object val = iInput.getVal();
      listener.currentValue = val;

      // setVal() is called before the onValueChange is attached. So, for the initialization we should invoke the
      // onValueChange if some value has been set
      if (!Utils.isBlank(val)) {
        listener.perform_(val, false);
      }
    }

    return required ? listener : null;
  }

  bool isRequired() {
    Field field = view.getField();
    final bool fieldDeclared = field != null;

    doLuhn = fieldDeclared && field.getLuhn() != null && iInput is InputText;

    final bool handlerProvided = handler != null;

    if (handlerProvided) {
      UIRenderer renderer = getRenderer();

      if (fieldDeclared) {
        List<String> views = ValueUtils.getViews(
            view.getId(), field.getReplicate(renderer), renderer);

        doReplicate = !Utils.isBlank(views);

        if (doReplicate) {
          Features.getReplicator(renderer).perform(view);
        }
      }

      doOnValueChange = !Utils.isBlank(view.getOnValueChange());

      if (!doOnValueChange && fieldDeclared) {
        doOnValueChange = field.hasImmediateConstraints_(renderer);
      }
    }

    final bool gotDependents = !Utils.isBlank(view.getDependents());

    if (gotDependents) {
      relativeRenderAction = RelativeRenderAction(view, handler);
    }

    if (gotDependents || view.canArray()) {
      reloadAction = ReloadAction(view, handler);
    }

    return doLuhn ||
        doReplicate ||
        doOnValueChange ||
        relativeRenderAction != null ||
        reloadAction != null ||
        !Utils.isBlank(view.getClear()) ||
        !Utils.isBlank(view.getReset());
  }

  void perform() {
    perform_(iInput.getVal(), true); // pass the new value
  }

  void perform_(Object newValue, bool ensureChange) {
    StopWatch stopWatch = StopWatch.start();

    try {
      if (!ensureChange || DataComparator.isChanged(currentValue, newValue)) {
        // ensure value change if it has to
        UIRenderer renderer = getRenderer();

        // we should ignore the call when value is reset because of the locale change
        if (renderer != null && !renderer.isLocaleChanging()) {
          performOnChange(newValue, renderer);
        }
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    } finally {
      StopWatch.stop_(newValue, stopWatch);
    }
  }

  void performOnChange(Object newValue, UIRenderer renderer) {
    currentValue = newValue;

    if (doLuhn) {
      newValue = Luhn.perform(view);
      iInput.setVal(newValue);
    }

    if (doReplicate) {
      Features.getReplicator(renderer).perform(view);
    }

    if (relativeRenderAction != null) {
      relativeRenderAction.render(newValue);
    }

    if (reloadAction != null) {
      reloadAction.reload(newValue);
    }

    Features.getDataActions(renderer).doClearAndReset(view);

    if (doOnValueChange &&
        Features.getValidator(renderer).run___(view, newValue, true, true)) {
      ContractUtils.onValueChange(newValue, view, renderer);
    }
  }

  UIRenderer getRenderer() {
    return handler == null ? null : handler.getRenderer();
  }
}
