/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/structure/ordered_map.dart';
import '../../../../form/structure/string_object_map.dart';
import '../../../structure/model/field.dart';
import '../../../util/form_utils.dart';
import '../../../util/map_utils.dart';
import '../../../util/stop_watch.dart';
import '../../../util/structure/string_set.dart';
import '../../../util/utils.dart';
import '../../../util/value/distributor.dart';
import '../../../util/value/replicator.dart';
import '../../../util/value/value_utils.dart';
import '../renderer/features.dart';
import '../renderer/ui_renderer.dart';
import '../view_state.dart';

class DataHandler {
  final UIRenderer renderer;

  const DataHandler(this.renderer);

  Set<String> handle(StringObjectMap data, StringSet vOverridden) {
    StopWatch stopWatch = StopWatch.start();
    Set<String> views = assignValues(data);
    views = getAffectedViews(views);
    refreshViews(views, vOverridden);
    StopWatch.stop(stopWatch);
    return views;
  }

  Set<String> assignValues(StringObjectMap data) {
    data = replicateToViews(data);
    Map<String, Object> orderedData = fixOrder(data);
    ValueUtils.setValues(orderedData, false, renderer);
    return MapUtils.keys(orderedData);
  }

  /// Single field can have multiple views and the data value in the data map is mapped using field id.
  StringObjectMap replicateToViews(StringObjectMap data) {
    StringObjectMap replicatedData = StringObjectMap();

    for (String fieldId in data.keys) {
      Field field = FormUtils.getField__(fieldId, renderer);
      StringSet views =
          field == null ? Utils.asSet([fieldId]) : field.getViews();

      if (!Utils.isBlank(views)) {
        final Object value = data[fieldId];

        for (String viewId in views) {
          replicatedData.put(viewId, value);
        }
      }
    }

    Utils.clear([data]);

    return replicatedData;
  }

  /// Arrange all views in the map as per their definition order in the UI.
  Map<String, Object> fixOrder(StringObjectMap data) {
    // We should use linked map here, if we don't we will loose the order that is being fixed.
    Map<String, Object> map = OrderedMap<Object>();

    if (!Utils.isBlank(data)) {
      Replicator replicator = Features.getReplicator(renderer);
      Distributor distributor = Features.getDistributor_(renderer);

      for (String viewId in renderer.getAllViews().keys) {
        if (Utils.got(viewId, data)) {
          ViewState view = FormUtils.getView_(viewId, renderer);

          if (view.canInput()) {
            map[viewId] =
                getValue(viewId, data[viewId], data, replicator, distributor);
          }
        }
      }

      Utils.clear([data]);
    }

    return map;
  }

  Object getValue(String viewId, Object inputVal, Map<String, Object> data,
      Replicator replicator, Distributor distributor) {
    ViewState view = FormUtils.getView_(viewId, renderer);
    Field field = FormUtils.getField(view);

    if (field == null) {
      return inputVal;
    }

    Object value = distribute(inputVal, viewId, data, distributor);
    value = getMappedValue(value, view);
    replicator.perform_(value, viewId, field.getReplicate(renderer), data);

    return value;
  }

  Object getMappedValue(Object val, ViewState view) {
    return val is String
        ? Features.getArrayValueHandler(renderer).getMappedValue(val, view)
        : val;
  }

  // TODO Should work for all data types.
  Object distribute(Object val, String viewId, Map<String, Object> data,
      Distributor distributor) {
    return val is String ? distributor.perform_(val, viewId, data) : val;
  }

  /// Get all the views which are affected due to the value modification of the provided views.
  Set<String> getAffectedViews(Iterable<String> modifiedViews) {
    Set<String> views = {};

    if (!Utils.isBlank(modifiedViews)) {
      for (String viewId in modifiedViews) {
        views.add(viewId);

        Set<ViewState> dependents =
            FormUtils.getView_(viewId, renderer).getDependents();

        if (!Utils.isBlank(dependents)) {
          for (ViewState dependent in dependents) {
            views.add(dependent.getId());
          }
        }
      }

      // TODO Include callers using scripts which are affected?
      // TODO Include callers using scripts with (in/ex)clusion?
    }

    return views;
  }

  /// Refresh enabled, inclusion and render attributes of provided views.
  void refreshViews(Iterable<String> views, StringSet vOverridden) {
    if (Utils.isBlank(views)) {
      return;
    }

    for (String viewId in views) {
      // reset excludeOnFetch
      FormUtils.getView_(viewId, renderer).setExcludedOnFetch(false);
    }

    for (String viewId in arrangeInInputOrder(views)) {
      ViewState view = FormUtils.getView_(viewId, renderer);

      // 1. Refresh view's enabled attribute.
      view.refresh(view.getEnabledIf(), false, renderer);

      // 2. Refresh Render
      view.refresh(view.getRenderedIf(), true, renderer);

      // 3. Refresh Inclusion
      Field field = view.getField();

      if (field != null &&
          field.isExcludeOnFetch() &&
          !vOverridden.contains(viewId) &&
          isLoaded(view.getValue())) {
        view.setExcludedOnFetch(true);

        // TODO Refresh all views depending on this view's inclusion.
      }
    }
  }

  /// Arrange the collection of views in following order:
  ///
  /// Input Components -> Labels -> Non-input Components
  ///
  /// TODO It's a temporary fix. refreshViews() should be recursive.
  List<String> arrangeInInputOrder(List<String> views) {
    List<String> inputComponents = [];
    List<String> labels = [];
    List<String> nonInputComponents = [];

    for (String viewId in views) {
      ViewState view = FormUtils.getView_(viewId, renderer);

      if (view.canInput()) {
        if (view.isLabel()) {
          labels.add(viewId);
        } else {
          inputComponents.add(viewId);
        }
      } else {
        nonInputComponents.add(viewId);
      }
    }

    // push all labels and non-input components at the end of this list.
    Utils.addAll(inputComponents, [labels, nonInputComponents]);
    return inputComponents;
  }

  /// Arrange the collection of views in following order:
  ///
  /// Input Components -> Labels -> Non-input Components
  ///
  /// TODO It's a temporary fix.
  static List<ViewState> arrangeInInputOrder_(Set<ViewState> views) {
    List<ViewState> inputComponents = [];
    List<ViewState> labels = [];
    List<ViewState> nonInputComponents = [];

    for (ViewState view in views) {
      if (view.canInput()) {
        if (view.isLabel()) {
          labels.add(view);
        } else {
          inputComponents.add(view);
        }
      } else {
        nonInputComponents.add(view);
      }
    }

    // push labels and non-input components at the end of the list
    Utils.addAll(inputComponents, [labels, nonInputComponents]);
    return inputComponents;
  }

  /// Checks if provided value does have a value.
  static bool isLoaded(Object val) {
    bool loaded = val != null;

    if (loaded) {
      loaded = val is List<int>
          ? !Utils.isBlank(val)
          : !Utils.isBlank(val.toString());
    }

    return loaded;
  }
}
