/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/structure/ordered_map.dart';
import '../../../../form/structure/string_object_map.dart';
import '../../../enums/mode.dart';
import '../../../structure/i18n/i18n.dart';
import '../../../structure/i18n/locale.dart';
import '../../../structure/i18n/message_array.dart';
import '../../../structure/i18n/option.dart';
import '../../../structure/model/field.dart';
import '../../../util/form_utils.dart';
import '../../../util/stop_watch.dart';
import '../../../util/utils.dart';
import '../../../validation/converter/c_string.dart';
import '../../../validation/error_marker.dart';
import '../component/behavior/i_array.dart';
import '../component/behavior/i_input.dart';
import '../component/input_text.dart';
import '../component/label.dart';
import '../util/key_value_pair.dart';
import '../view_state.dart';
import 'features.dart';
import 'messages.dart';
import 'ui_renderer.dart';

class ArrayValueHandler {
  final UIRenderer renderer;

  const ArrayValueHandler(this.renderer);

  // Handle array and value attributes of all the input components and the repercussions that follow
  void handle(StringObjectMap data, OrderedMap<ViewState> views) {
    StopWatch stopWatch = StopWatch.start();

    for (ViewState view in views.values) {
      IInput iInput = view.getInputComponent();

      if (iInput != null) {
        if (iInput is Label) {
          setMandatory(view, views);
        }

        if (view.canArray()) {
          setArrayAndValue(view, iInput as IArray, data, views);
        } else {
          view.setValue(toPreviewValue(getValue(view, false, data), view));
        }
      }
    }

    StopWatch.stop(stopWatch);
  }

  void setMandatory(ViewState view, OrderedMap<ViewState> allViews) {
    ViewState vFor = allViews[view.getForId()];

    if (vFor != null && vFor.isLabel()) {
      (view.getComponent() as Label).setMandatory(vFor.isMandatory(renderer));
    }
  }

  void clearArrayAndValue(ViewState view, IArray iArray) {
    iArray.clearArray();

    view.clearErrMsg();
    view.setValue(null);
  }

  // TODO Test this functionality for SingleChoice and MultipleChoice?
  void setArrayAndValue(ViewState view, IArray iArray, StringObjectMap data,
      OrderedMap<ViewState> views) {
    final String arrayId = view.getArray();

    if (Utils.isBlank(arrayId)) {
      return;
    }

    String value = CString.convert(getValue(view, false, data));
    value = CString.convert(toPreviewValue(value, view));
    bool arrayed = setArray(arrayId, iArray);

    view.clearErrMsg();

    if (!iArray.isValidItem(value)) {
      if (view.canInputAndErr()) {
        ErrorMarker.mark(
            view, renderer, null, view.getInvalidItemMsg(), [value]);
      }

      value = null;
    }

    if (arrayed) {
      // if array is blank, we don't need to set the value and attach listener
      view.setValue(value);
      // if the value is set, we need to reload IArray component which this view is a parent of.
      ViewState vChild = views[view.getParentOf()];
      setArrayAndValue_(vChild, arrayId, value, data, views);

      if (vChild != null) {
        // set onValueChange listener for the component
        Features.getEventHandler(renderer).addOnChangeValue(view);
      }
    }
  }

  void setArrayAndValue_(ViewState view, String parentArrayId,
      String parentValue, StringObjectMap data, OrderedMap<ViewState> views) {
    if (view == null || !view.canInput() || !view.canArray()) {
      return;
    }

    String arrayId = view.getArray();

    if (!Utils.isBlank(arrayId)) {
      return; // Array should be blank for child view's field.
    }

    ViewState vChild = views[view.getParentOf()];
    I18n i18n = FormUtils.getI18n(renderer);
    MessageArray array = i18n.getMsgArray_(parentArrayId, parentValue);

    if (array != null) {
      arrayId = array.getId(); // newly assigned id
      Set<KeyValuePair> entries = i18n.toText(array, renderer);

      if (!Utils.isBlank(entries)) {
        String value = CString.convert(getValue(view, false, data));
        value = CString.convert(toPreviewValue(value, view));
        IArray iArray = view.getInputComponent() as IArray;
        iArray.setArray(
            entries, array.hasBlankItem(), renderer.get(array.getBlankMsg()));

        bool valued = iArray.isValidItem(value) ? view.setValue(value) : false;

        view.clearErrMsg();

        String selectedValue =
            !valued || value == null ? null : value.toString();
        setArrayAndValue_(vChild, arrayId, selectedValue, data, views);
      }
    }

    view.setArray(arrayId);

    if (vChild != null) {
      // set onValueChange listener for the component
      Features.getEventHandler(renderer).addOnChangeValue(view);
    }
  }

  bool setArray(String arrayId, IArray iArray) {
    iArray.clearArray();

    I18n i18n = FormUtils.getI18n(renderer);
    Set<KeyValuePair> array = i18n.toText(i18n.getMsgArray(arrayId), renderer);

    if (Utils.isBlank(array)) {
      return false;
    }

    final bool blankItem = i18n.hasArrayBlankItem(arrayId);
    String blankMsg =
        blankItem ? renderer.get(i18n.getArrayBlankMsg(arrayId)) : null;
    iArray.setArray(array, blankItem, blankMsg);
    return true;
  }

  Object getValue(ViewState view, bool reset, StringObjectMap data) {
    Object value = getInitialValue(view, reset, data);

    if (value is String) {
      try {
        value = FormUtils.resolve_(value as String, renderer);
      } catch (e) {
        // nothing to do here
      }

      if (value is String) {
        value = getMappedValue(value as String, view);
      }
    }

    return value;
  }

  /// Try to use the loaded value first if its provided. If not, use the value set in the model for the view if we are
  /// resetting the view or we are in create mode.
  Object getInitialValue(ViewState view, bool reset, StringObjectMap data) {
    Object value = reset ? null : view.getValue();
    Field field = view.getField();

    if (reset || renderer.checkMode([Mode.CREATE])) {
      if (field != null) {
        value = Utils.toDefault(field.getValue(), value);
      }
    } else if (!Utils.isBlank(data)) {
      final String fieldId = view.getFieldId();

      if (Utils.got(fieldId, data)) {
        final Object valueInDataMap = data[fieldId];

        // if view is label and it's value is not in the data map, we should use the configured label from XML.
        if (!view.isLabelledComponent() && view.isLabel()) {
          value = Utils.toDefault(CString.convert(valueInDataMap), value);
        } else {
          value = valueInDataMap;
        }
      } else {
        // if the field's value isn't listed in the map, use the one in field declaration
        if (field != null) {
          value = Utils.toDefault(field.getValue(), value);
        }
      }
    }

    return value;
  }

  Object toPreviewValue(Object value, ViewState view) {
    String previewValue = value == null || value is String
        ? getPreviewValue(value as String, view)
        : null;
    return Utils.toDefault(previewValue, value);
  }

  String getPreviewValue(String value, ViewState view) {
    return Utils.isBlank(value) &&
            renderer.checkMode([Mode.PREVIEW, Mode.UPDATE])
        ? view.getPreviewValue()
        : null;
  }

  String getMappedValue(String value, ViewState view) {
    return getMappedValue_(value, view, false);
  }

  // gets the dynamically mapped value for a Label or InputText or container.
  String getMappedValue_(String value, ViewState view, bool forLabel) {
    String mappedValue = value;

    if (view.isContainer()) {
      mappedValue = getMappedValue__(value, view.getInitialTitle());
    } else {
      IInput iInput = view.getInputComponent();

      if (forLabel || iInput is Label || iInput is InputText) {
        mappedValue = getMappedValue__(
            value, Utils.toDefault(view.getArray(), view.getInitialLabel()));
      }
    }

    return mappedValue;
  }

  String getMappedValue__(String value, String arrayId) {
    String mappedValue = value;
    I18n i18n = FormUtils.getI18n(renderer);
    OrderedMap<Option> options = i18n.getArrayOptions(arrayId, renderer);

    if (!Utils.isBlank(options)) {
      Option option = options[value];
      final String msgId = option == null ? null : option.getMsg();
      Locale locale = FormUtils.getLocale_(renderer);
      Messages msgs = renderer.getMsgs();

      if (Utils.isBlank(msgId)) {
        String sOption = i18n.getArrayDefaultOption(arrayId);
        String text = Utils.toDefault(
            i18n.getText_(sOption, locale, msgs, renderer), sOption);
        mappedValue = Utils.toDefault(text, value);
      } else {
        mappedValue = Utils.toDefault(
            i18n.getText_(msgId, locale, msgs, renderer), msgId);
      }
    }

    return mappedValue;
  }
}
