/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/structure/string_object_map.dart';
import '../../../config/defaults.dart';
import '../../../enums/alignment.dart';
import '../../../enums/font_style.dart';
import '../../../enums/mode.dart';
import '../../../enums/orientation.dart';
import '../../../enums/table_action.dart';
import '../../../enums/view_type.dart';
import '../../../error/illegal_argument_exception.dart';
import '../../../structure/model/constraint.dart';
import '../../../structure/model/field.dart';
import '../../../structure/ui/ui.dart';
import '../../../structure/ui/view.dart';
import '../../../util/bean/range/int_range.dart';
import '../../../util/form_utils.dart';
import '../../../util/numbers.dart';
import '../../../util/stop_watch.dart';
import '../../../util/utils.dart';
import '../../../validation/constraint/c_file.dart' as CFile;
import '../../../validation/constraint/image/c_image.dart' as CImage;
import '../../../validation/converter/c_boolean.dart';
import '../../../validation/converter/c_float.dart';
import '../../../validation/converter/c_number.dart';
import '../../../validation/converter/c_string.dart';
import '../common/component/c_date_time_picker.dart';
import '../component/alert.dart';
import '../component/barcode.dart';
import '../component/base/component.dart';
import '../component/base/input_component.dart';
import '../component/behavior/i_align.dart';
import '../component/behavior/i_array.dart';
import '../component/behavior/i_capture.dart';
import '../component/behavior/i_column.dart';
import '../component/behavior/i_enable.dart';
import '../component/behavior/i_err.dart';
import '../component/behavior/i_export.dart';
import '../component/behavior/i_font.dart';
import '../component/behavior/i_hint.dart';
import '../component/behavior/i_icon.dart';
import '../component/behavior/i_import.dart';
import '../component/behavior/i_keyboard.dart';
import '../component/behavior/i_label.dart';
import '../component/behavior/i_layout_align.dart';
import '../component/behavior/i_layout_size.dart';
import '../component/behavior/i_length.dart';
import '../component/behavior/i_link.dart';
import '../component/behavior/i_locale.dart';
import '../component/behavior/i_mask.dart';
import '../component/behavior/i_orient.dart';
import '../component/behavior/i_pattern.dart';
import '../component/behavior/i_row.dart';
import '../component/behavior/i_select_date.dart';
import '../component/behavior/i_select_file.dart';
import '../component/behavior/i_sub_title.dart';
import '../component/behavior/i_title.dart';
import '../component/behavior/i_weight.dart';
import '../component/behavior/i_weight_sum.dart';
import '../component/button/button.dart';
import '../component/button/link.dart';
import '../component/button/submit.dart';
import '../component/captcha.dart';
import '../component/card.dart';
import '../component/composition/camera.dart';
import '../component/composition/date_picker.dart';
import '../component/composition/date_time_picker.dart';
import '../component/composition/fingerprint.dart';
import '../component/composition/labelled_component.dart';
import '../component/composition/signature.dart';
import '../component/composition/time_picker.dart';
import '../component/doc_viewer.dart';
import '../component/file_picker.dart';
import '../component/group/multiple_choice.dart';
import '../component/group/single_choice.dart';
import '../component/image.dart';
import '../component/input_area.dart';
import '../component/input_text.dart';
import '../component/label.dart';
import '../component/layout/grid_layout.dart';
import '../component/layout/linear_layout.dart';
import '../component/layout/popup.dart';
import '../component/layout/tab.dart';
import '../component/layout/tab_panel.dart';
import '../component/message.dart';
import '../component/ocr.dart';
import '../component/passport.dart';
import '../component/password.dart';
import '../component/radio_button.dart';
import '../component/scanner.dart';
import '../component/select_many.dart';
import '../component/select_one.dart';
import '../component/spacer.dart';
import '../component/table/action_column.dart';
import '../component/table/column.dart';
import '../component/table/fp_column.dart';
import '../component/table/id_column.dart';
import '../component/table/image_column.dart';
import '../component/table/radio_column.dart';
import '../component/table/table.dart';
import '../component/table/tick_column.dart';
import '../component/tick_box.dart';
import '../component/virtual.dart';
import '../form_controller.dart';
import '../util/focus_utils.dart';
import '../view_state.dart' as ViewState;
import 'features.dart';
import 'ui_renderer.dart';

abstract class ComponentFactory<T extends FormController> {
  final T controller;

  ComponentFactory(this.controller) {
    if (controller == null) {
      throw IllegalArgumentException("Controller can't be null!");
    }
  }

  ViewState.ViewState createRoot(
      String id, Orientation orientation, String focus) {
    ViewState.ViewState vRoot = ViewState.ViewState.popup(id, false, focus);
    View view = View.columns(id, vRoot.getType(), orientation, "1");
    vRoot.setComponent(
        create(vRoot, view, false)); // forPopup doesn't matter here
    return vRoot;
  }

  Component create(ViewState.ViewState viewState, View view, bool forPopup) {
    StopWatch stopWatch = StopWatch.start();

    Component component = newComponent(viewState, view, forPopup);

    if (component == null) {
      throw IllegalArgumentException("Couldn't create component for $view");
    }

    // one by one, apply basic attributes
    component.setIdentity(viewState.getId());
    component.setRendered(viewState.isRendered());

    setRows(view, component);
    setColumns(view, component, forPopup);
    setOrientation(view, component);
    setAlignment(view, component);
    setLayoutAlignment(view, component);
    setSize(view.getWidth(), view.getHeight(), component);
    setLayoutSize(view, component);
    setWeight(view, component);
    setWeightSum(view, component);
    setLabel(viewState, component);
    setTitle(viewState, component);
    setSubTitle(viewState, component);
    setFont(view, component);
    setStyle(view.getStyle(), component);
    setEnabled(viewState, component);
    setPattern(view, component);
    setSelectionDate(view, component);
    setImportable(view, component);
    setExportable(view, component);
    setKeyboard(view, component);
    setMask(view, component);
    setLink(view, component);
    setIcon(view, component);

    StopWatch.stop_(view, stopWatch);

    return component;
  }

  Component newComponent(
      ViewState.ViewState viewState, View view, bool forPopup) {
    ViewType type = viewState.getType();

    if (ViewType.LINEARLAYOUT == type) {
      return linearLayout(view.getOrientation());
    } else if (ViewType.GRIDLAYOUT == type) {
      return gridLayout();
    } else if (ViewType.TABPANEL == type) {
      return tabPanel();
    } else if (ViewType.TAB == type) {
      return tab();
    } else if (ViewType.TABLE == type) {
      String action = FormUtils.resolveAsString_(
          viewState.getParams(), controller.getRenderer());
      return table(Utils.toDefault(
          TableAction.toAction(action, false), TableAction.EDIT_REMOVE));
    } else if (ViewType.COLUMN == type) {
      return column();
    } else if (ViewType.IDCOLUMN == type) {
      return idColumn();
    } else if (ViewType.TICKCOLUMN == type) {
      return tickColumn();
    } else if (ViewType.RADIOCOLUMN == type) {
      return radioColumn();
    } else if (ViewType.IMAGECOLUMN == type) {
      return imageColumn();
    } else if (ViewType.FPCOLUMN == type) {
      return fpColumn();
    } else if (ViewType.ACTIONCOLUMN == type) {
      return actionColumn();
    } else if (ViewType.SPACER == type) {
      return spacer();
    } else if (ViewType.IMAGE == type) {
      return image();
    } else if (ViewType.DOCVIEWER == type) {
      return docViewer(FormUtils.resolveAsString(view.getParams(), controller));
    } else if (ViewType.LINK == type) {
      return link(FormUtils.resolveAsString(view.getParams(), controller));
    } else if (ViewType.BUTTON == type) {
      return button();
    } else if (ViewType.SUBMIT == type) {
      return submit();
    }

    InputComponent component = initInputComponent(type, view);

    if (component == null) {
      // for some reason, unable to create the component
      // This should never happen because there can't be an enum view type that can be parsed and mapped to
      // ViewType. Also, all the configured enum types should be handled above.
      throw IllegalArgumentException("$type view type isn't supported!");
    }

    // if component is a label and it can persist, we should wrap it as a labelled component.
    if (component is Label &&
        !viewState.canPersist_(false, controller.getRenderer())) {
      return component;
    }

    return ViewType.VIRTUAL == type
        ? component
        : labelled(viewState, component, forPopup);
  }

  InputComponent initInputComponent(ViewType type, View view) {
    if (ViewType.LABEL == type) {
      return label();
    } else if (ViewType.INPUTTEXT == type) {
      return inputText(view.allowAutoComplete());
    } else if (ViewType.INPUTAREA == type) {
      return inputArea();
    } else if (ViewType.PASSWORD == type) {
      return password();
    } else if (ViewType.VIRTUAL == type) {
      return virtual();
    } else if (ViewType.BARCODE == type) {
      return barcode(view.allowManualInput());
    } else if (ViewType.CARD == type) {
      return card(view.allowManualInput());
    } else if (ViewType.TICKBOX == type) {
      return tickBox();
    } else if (ViewType.SINGLECHOICE == type) {
      return singleChoice();
    } else if (ViewType.MULTIPLECHOICE == type) {
      return multipleChoice();
    } else if (ViewType.SELECTONE == type) {
      return selectOne(view.allowManualInput());
    } else if (ViewType.SELECTMANY == type) {
      return selectMany(view.allowManualInput());
    } else if (ViewType.CAPTCHA == type) {
      return captcha();
    } else if (ViewType.DATEPICKER == type) {
      return datePicker(view.allowManualInput());
    } else if (ViewType.TIMEPICKER == type) {
      return timePicker(view.allowManualInput());
    } else if (ViewType.DATETIMEPICKER == type) {
      return dateTimePicker(view.allowManualInput());
    } else if (ViewType.FILEPICKER == type) {
      return filePicker();
    } else if (ViewType.SCANNER == type) {
      return scanner();
    } else if (ViewType.CAMERA == type) {
      return camera(
          isAllowCapture(view), isAllowUpload(view), isAllowRotate(view));
    } else if (ViewType.FINGERPRINT == type) {
      return fingerprint(isAllowCapture(view), isAllowUpload(view));
    } else if (ViewType.SIGNATURE == type) {
      return signature(isAllowCapture(view), isAllowUpload(view),
          isAllowRotate(view), view.getParams());
    } else if (ViewType.OCR == type) {
      return ocr(view.getTwoLinesMrz(), isAllowCapture(view),
          isAllowUpload(view), isAllowRotate(view));
    } else if (ViewType.PASSPORT == type) {
      final bool allowCapture = isAllowCapture(view);
      return passport(allowCapture == null ? null : !allowCapture,
          isAllowUpload(view), isAllowRotate(view));
    }

    return null;
  }

  Component labelled(
      ViewState.ViewState view, InputComponent component, bool forPopup) {
    UIRenderer renderer = controller.getRenderer();

    // TODO If component can have it's own label, we shouldn't create extra label (like Tickbox)
    Label label = Utils.isBlank(view.getLabel()) ? null : this.label();
    // TODO If an external label is linked to input component and it can IErr then errMsg shouldn't be created.
    Message errMsg = label is IErr || component is IErr
        ? null
        : createMessage(view, renderer);

    // we don't need tick box holder outside reject mode or in popup.
    final bool isRejectMode = renderer.checkMode([Mode.REJECT]) && !forPopup;

    // If there is no label and error-message to render, we don't need LabelledComponent.
    // But if the form is in reject mode, we still need LabelledComponent so that we can have the TickBox.
    if (!isRejectMode && Utils.allNull([label, errMsg])) {
      return component;
    }

    LabelledComponent labelled = labelledComponent();

    if (isRejectMode) {
      // We put a blank label if view can't be rejected.
      labelled.setTickboxHolder(
          view.isRejectable(renderer) ? tickBox() : this.label());
    }

    labelled.setLblText(label);
    labelled.setInputComponent(component);
    labelled.setMsgErr(errMsg);
    labelled.setToastError(view.toastError);
    return labelled;
  }

  // Don't create Message to display errors if:
  //
  // 1. Error messages have to be toasted on the screen, or
  // 2. Input-component can show the error within itself
  Message createMessage(ViewState.ViewState view, UIRenderer renderer) {
    Message errMsg = view.toastError || view.canErr() ? null : message();
    // But we need it in revise mode, in any case.
    errMsg = errMsg == null && renderer.checkMode([Mode.REVISE])
        ? message()
        : errMsg;
    // And we still need it if view's container is a grid, merely to render consistent layout.
    return errMsg == null &&
            ViewType.GRIDLAYOUT == view.getContainer().getType()
        ? message()
        : errMsg;
  }

  void reflectConstraints(ViewState.ViewState view) {
    final bool mandatory = view.isMandatory(controller.getRenderer());
    ViewState.ViewState lblView = view.getLblView();

    if (lblView != null) {
      // if label is externalized, this will be true
      Label lblComponent = lblView.getInputComponent() as Label;
      lblComponent.setMandatory(mandatory);
      setLabel(lblView, lblComponent);
    }

    Component component = view.getComponent();
    InputComponent input;

    if (view.isLabelledComponent()) {
      LabelledComponent labelled = view.getLabelledComponent();
      labelled.setMandatory(mandatory);
      setLabel(view, component);
      input = labelled.getInputComponent();
    }

    setHint(view, component);

    Field field = view.getField();

    if (field != null) {
      setLengthAndNumeric(field, input == null ? component : input);
      setImageAttrs(field, component);
      setFileAttrs(field, component);
    }
  }

  void setRows(View view, Component component) {
    if (component is IRow) {
      int vRows = view.getRows();
      int rows =
          ViewType.TABLE == view.getType() ? vRows : Utils.toDefault(vRows, 2);

      if (rows != null && rows > 0) {
        (component as IRow).setRows(rows);
      }
    }
  }

  void setColumns(View view, Component component, bool forPopup) {
    if (component is IColumn) {
      int columns = CNumber.convertNoErr(
          FormUtils.resolve(view.getColumns(), controller));

      if (columns != null && (columns < Defaults.MISSING || columns == 0)) {
        throw IllegalArgumentException("Columns can be -1 or greater than 0!");
      }

      final bool isGrid = ViewType.GRIDLAYOUT == view.getType();
      columns =
          Utils.toDefault(columns, isGrid ? 3 : Defaults.INPUT_GROUP_COLUMNS);

      if (isGrid &&
          !forPopup &&
          controller.getRenderer().checkMode([Mode.REJECT])) {
        columns++; // in reject mode and outside popup, we need extra column in the grid for tick box.
      }

      (component as IColumn).setColumns(columns);
    }
  }

  void setOrientation(View view, Component component) {
    if (component is IOrient) {
      (component as IOrient).setOrientation(view.getOrientation());
    }
  }

  void setAlignment(View view, Component component) {
    if (component is IAlign) {
      Alignment alignment = view.getAlign();

      if (alignment != null) {
        (component as IAlign).setAlignment(alignment);
      }
    }
  }

  void setLayoutAlignment(View view, Component component) {
    if (component is ILayoutAlign) {
      Alignment alignment = view.getLayoutAlign();

      if (alignment != null) {
        (component as ILayoutAlign).setLayoutAlignment(alignment);
      }
    }
  }

  List<String> getSize(String sWidth, String sHeight) {
    String width = FormUtils.resolveAsString(sWidth, controller);
    String height = FormUtils.resolveAsString(sHeight, controller);
    return Utils.allBlank([width, height]) ? null : [width, height];
  }

  void setSize(String sWidth, String sHeight, Component component) {
    List<String> size = getSize(sWidth, sHeight);

    if (!Utils.isBlank(size)) {
      String width = size[0];

      if (!Utils.isBlank(width)) {
        component.setWidth(width);
      }

      String height = size[1];

      if (!Utils.isBlank(height)) {
        component.setHeight(height);
      }
    }
  }

  void setLayoutSize(View view, Component component) {
    if (component is ILayoutSize) {
      List<String> size =
          getSize(view.getLayoutWidth(), view.getLayoutHeight());

      if (!Utils.isBlank(size)) {
        ILayoutSize iSize = component as ILayoutSize;
        String width = size[0];

        if (!Utils.isBlank(width)) {
          iSize.setLayoutWidth(width);
        }

        String height = size[1];

        if (!Utils.isBlank(height)) {
          iSize.setLayoutHeight(height);
        }
      }
    }
  }

  void setWeight(View view, Component component) {
    if (component is IWeight) {
      double weight = CFloat.parseNoErr(view.getWeight());

      if (weight != null) {
        (component as IWeight).setWeight(weight);
      }
    }
  }

  void setWeightSum(View view, Component component) {
    if (component is IWeightSum) {
      double weightSum = CFloat.parseNoErr(view.getWeightSum());

      if (weightSum != null) {
        (component as IWeightSum).setWeightSum(weightSum);
      }
    }
  }

  void setLabel(ViewState.ViewState view, Component component) {
    if (component is ILabel) {
      (component as ILabel).setLabel(FormUtils.getLabel_(view, controller));
    }
  }

  void setHint(ViewState.ViewState view, Component component) {
    if (component is IHint) {
      UIRenderer renderer = controller.getRenderer();

      ViewState.ViewState vLbl = view.getLblView();
      final bool noLblView = vLbl == null;
      String hint = renderer.get(noLblView ? view.getHint() : vLbl.getHint());

      if (hint == null) {
        // Use 'null' because CMS form can be customized not to show the hint by putting blank
        String suffix = renderer.get(
            view.isMandatory(renderer) ? "view.mandatory" : "view.optional");
        String label = FormUtils.eval_(
            noLblView ? view.getLabel() : vLbl.getLabel(), renderer);
        hint = Utils.isBlank(label)
            ? suffix
            : renderer.get("view.default.hint", [label, suffix]);
      }

      (component as IHint).setHint(Utils.toBlank(hint));
    }
  }

  void setTitle(ViewState.ViewState view, Component component) {
    if (component is ITitle) {
      (component as ITitle)
          .setTitle(FormUtils.eval(view.getTitle(), controller));
    }
  }

  void setSubTitle(ViewState.ViewState view, Component component) {
    if (component is ISubTitle) {
      (component as ISubTitle)
          .setSubTitle(FormUtils.eval(view.getSubTitle(), controller));
    }
  }

  void setFont(View view, Component component) {
    if (component is IFont) {
      IFont iFont = component as IFont;
      UI ui = FormUtils.getUI_(controller.getRenderer());

      String color = FormUtils.resolveAsString(view.getFontColor(), controller);
      color = Utils.isBlank(color)
          ? FormUtils.resolveAsString(ui.getBaseFontColor(), controller)
          : color;

      if (!Utils.isBlank(color)) {
        iFont.setFontColor(color);
      }

      String size = FormUtils.resolveAsString(view.getFontSize(), controller);
      size = Utils.isBlank(size)
          ? FormUtils.resolveAsString(ui.getBaseFontSize(), controller)
          : size;

      if (!Utils.isBlank(size)) {
        iFont.setFontSize(size);
      }

      FontStyle style = FontStyle.toStyle(
          FormUtils.resolveAsString(view.getFontStyle(), controller), false);

      if (style != null) {
        iFont.setFontStyle(style);
      }
    }
  }

  void setStyle(String style, Component component) {
    style = FormUtils.resolveAsString(style, controller);

    if (!Utils.isBlank(style)) {
      component.setStyle(style);
    }
  }

  void setEnabled(ViewState.ViewState view, Component component) {
    if (component is IEnable) {
      (component as IEnable).setEnabled(view.isEnabled());
    }
  }

  void setPattern(View view, Component component) {
    if (component is IPattern) {
      String pattern = FormUtils.resolveAsString(view.getPattern(), controller);

      if (!Utils.isBlank(pattern)) {
        (component as IPattern).setPattern(pattern);
      }
    }
  }

  void setSelectionDate(View view, Component component) {
    if (component is ISelectDate) {
      // TODO Doesn't support default time selection.
      DateTime date = CDateTimePicker.toDate(
          FormUtils.resolve(view.getSelectionDate(), controller));

      if (date != null) {
        (component as ISelectDate).setSelectionDate(date);
      }
    }
  }

  void setImportable(View view, Component component) {
    if (component is IImport) {
      (component as IImport).setImportable(view.isImportable());
    }
  }

  void setExportable(View view, Component component) {
    if (component is IExport) {
      (component as IExport).setExportable(view.isExportable());
    }
  }

  void setKeyboard(View view, Component component) {
    if (component is IKeyboard) {
      String keyboard = view.getKeyboard();

      if (keyboard != null) {
        (component as IKeyboard).setKeyboard(keyboard);
      }
    }
  }

  void setMask(View view, Component component) {
    if (component is IMask) {
      String mask = view.getMask();

      if (!Utils.isBlank(mask)) {
        (component as IMask)
            .setMask(mask, view.getPlaceholder(), view.isUseMaskedValue());
      }
    }
  }

  void setLink(View view, Component component) {
    component = LabelledComponent.getComponent(component);

    if (component is ILink) {
      (component as ILink).setLink(view.getLink(controller.getRenderer()));
    }
  }

  void setIcon(View view, Component component) {
    component = LabelledComponent.getComponent(component);

    if (component is IIcon) {
      String icon = view.getIcon();

      if (!Utils.isBlank(icon)) {
        (component as IIcon).setIcon(icon);
      }
    }
  }

  void setLengthAndNumeric(Field field, Component component) {
    if (component is ILength) {
      UIRenderer renderer = controller.getRenderer();
      int maxLength = field.getMaxLength(renderer);

      // call the method despite maxLength is null because of the 'enabled' condition of length constraint
      (component as ILength).setLength(
          Numbers.nullOrBelowOne(maxLength) ? Defaults.MISSING : maxLength);

      if (component is InputText) {
        // TODO We don't want this to happen for all fields, make it configurable.
        component.setNumberType(field.isNumberType(renderer),
            FormUtils.getLocale_(renderer).getDigits());
      }
    }
  }

  void setImageAttrs(Field field, Component component) {
    if (component is ICapture) {
      UIRenderer renderer = controller.getRenderer();
      Constraint cImage = field.get(CImage.CImage.TYPE, renderer);
      ICapture capturer = component as ICapture;
      bool cameraSwitch, flash;

      if (cImage != null) {
        StringObjectMap args = cImage.getArgs_(renderer);
        IntRange width =
            IntRange.get(CImage.Args.MIN_WIDTH, CImage.Args.MAX_WIDTH, args);
        IntRange height =
            IntRange.get(CImage.Args.MIN_HEIGHT, CImage.Args.MAX_HEIGHT, args);
        int quality = CNumber.convertNoErr(args[CImage.Args.QUALITY]);

        if (width != null || height != null) {
          bool autoScale = CBoolean.convertNoErr(args[CImage.Args.AUTO_SCALE]);
          capturer.setDimension(width, height, Utils.toFalse(autoScale));
        }

        if (quality != null) {
          capturer.setQuality(quality);
        }

        cameraSwitch = CBoolean.convertNoErr(args[CImage.Args.CAMERA_SWITCH]);
        flash = CBoolean.convertNoErr(args[CImage.Args.FLASH]);
      }

      capturer.setFeatures(
          Utils.toDefault(cameraSwitch, Defaults.CAMERA_SWITCH),
          Utils.toDefault(flash, Defaults.FLASH));
      capturer.onConstraintChanged();
    }
  }

  void setFileAttrs(Field field, Component component) {
    if (component is ISelectFile) {
      UIRenderer renderer = controller.getRenderer();
      Constraint cFile = field.get(CFile.CFile.TYPE, renderer);

      if (cFile != null) {
        StringObjectMap args = cFile.getArgs_(renderer);
        String exts = CString.convert(args[CFile.Args.EXT]);
        IntRange size = IntRange.get(CFile.Args.MIN, CFile.Args.MAX, args);

        ISelectFile selector = component as ISelectFile;

        if (!Utils.isBlank(exts)) {
          selector.setExts(exts);
        }

        if (size != null) {
          bool autoCompress =
              CBoolean.convertNoErr(args[CFile.Args.AUTO_COMPRESS]);
          selector.setFileSize(size, Utils.toFalse(autoCompress));
        }
      }
    }
  }

  void setArray(ViewState.ViewState view, Component component) {
    if (view.canArray()) {
      Field field = view.getField();

      if (field != null) {
        String arrayId = view
            .getArray(); // a field that dependents on another, we should use selectedArray.

        if (!Utils.isBlank(arrayId)) {
          Features.getArrayValueHandler(controller.getRenderer())
              .setArray(arrayId, component as IArray);
        }
      }
    }
  }

  void changeDir(bool ltr, ViewState.ViewState container) {
    // nothing to do here
  }

  void changeLocale() {
    UIRenderer renderer = controller.getRenderer();

    for (ViewState.ViewState view in renderer.getAllViewStates()) {
      Component component = view.getComponent();
      InputComponent cInput;
      final Object value = view.getValue();

      if (view.isLabelledComponent()) {
        LabelledComponent labelledComponent = view.getLabelledComponent();
        cInput = labelledComponent.getInputComponent();

        // reset this value as it isn't retained on postback in JSF
        labelledComponent.setMandatory(view.isMandatory(renderer));
      }

      setLabel(view, component);
      setHint(view, component);
      setTitle(view, component);
      setSubTitle(view, component);
      setError(view, component);
      setArray(view, component);

      if (cInput is ILocale) {
        (cInput as ILocale).localeChanged();
      }

      // setLabel() call will already change the label value, we don't want to sustain stale label.
      if (value != null && !view.isLabel()) {
        view.setValue(value);
      }
    }
  }

  void setTitle_(ITitle root, List<View> layout) {
    ViewState.ViewState vTop = FocusUtils.getTopVisibleView(layout, controller);

    if (vTop != null) {
      Component cFirst = vTop.getComponent();

      if (cFirst is ITitle) {
        root.setTitle((cFirst as ITitle).getTitle());
        cFirst.setRendered(
            true); // sets the page title by re-rendering the component
      }
    }
  }

  void setError(ViewState.ViewState view, Component component) {
    if (component is IErr) {
      ViewState.ErrMsg errMsg = view.getErrMsg();
      String msg =
          controller.getRenderer().get(errMsg.getMsgId(), [errMsg.getArgs()]);
      (component as IErr).setErrMsg(Utils.toBlank(msg));
    }
  }

  bool isAllowCapture(View view) {
    return CBoolean.convertNoErr(
        FormUtils.resolve(view.getAllowCapture(), controller));
  }

  bool isAllowUpload(View view) {
    return CBoolean.convertNoErr(
        FormUtils.resolve(view.getAllowUpload(), controller));
  }

  bool isAllowRotate(View view) {
    return CBoolean.convertNoErr(
        FormUtils.resolve(view.getAllowRotate(), controller));
  }

  Alert getAlert();

  LinearLayout linearLayout(Orientation orientation);

  GridLayout gridLayout();

  Popup popup(String title, bool closeable, String width, String height);

  TabPanel tabPanel();

  Tab tab();

  Table table(TableAction action);

  Column column();

  IdColumn idColumn();

  TickColumn tickColumn();

  RadioColumn radioColumn();

  ImageColumn imageColumn();

  FpColumn fpColumn();

  ActionColumn actionColumn();

  Spacer spacer();

  Label label();

  Message message();

  LabelledComponent labelledComponent();

  InputText inputText(bool autoComplete);

  Barcode barcode(bool manualInput);

  Card card(bool manualInput);

  InputArea inputArea();

  Password password();

  Virtual virtual();

  Image image();

  DocViewer docViewer(String path);

  SelectOne selectOne(bool manualInput);

  SelectMany selectMany(bool manualInput);

  Captcha captcha();

  Camera camera(bool allowCapture, bool allowUpload, bool allowRotate);

  TickBox tickBox();

  RadioButton radioButton();

  SingleChoice singleChoice();

  MultipleChoice multipleChoice();

  Link link(String target);

  Button button();

  Submit submit();

  DatePicker datePicker(bool manualInput);

  TimePicker timePicker(bool manualInput);

  DateTimePicker dateTimePicker(bool manualInput);

  FilePicker filePicker();

  Fingerprint fingerprint(bool allowCapture, bool allowUpload);

  Signature signature(
      bool allowCapture, bool allowUpload, bool allowRotate, String device);

  Ocr ocr(
      bool twoLinesMrz, bool allowCapture, bool allowUpload, bool allowRotate);

  Passport passport(bool performOcr, bool allowUpload, bool allowRotate);

  Scanner scanner();
}
