/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:collection' show LinkedHashSet;

import '../../../../form/structure/ordered_map.dart';
import '../../../../form/structure/t_map.dart';
import '../../../enums/action.dart';
import '../../../enums/code.dart';
import '../../../structure/i18n/message_array.dart';
import '../../../structure/model/constraint.dart';
import '../../../structure/model/field.dart';
import '../../../structure/model/model.dart';
import '../../../structure/model/script.dart';
import '../../../util/form_utils.dart';
import '../../../util/stop_watch.dart';
import '../../../util/structure/string_set.dart';
import '../../../util/utils.dart';
import '../../../validation/converter/c_boolean.dart';
import '../action/event_handler.dart';
import '../action/handler/action_handler.dart';
import '../action/relative_render_action.dart';
import '../action/toggle_action.dart';
import '../view_state.dart';
import 'features.dart';
import 'ui_renderer.dart';

class EventProcessor {
  final UIRenderer renderer;
  final Set<ViewState> viewsForOnChangeListener;

  EventProcessor(this.renderer)
      : this.viewsForOnChangeListener = LinkedHashSet();

  void handle(
      String rootId, bool ignoreListeners, OrderedMap<ViewState> allViews) {
    StopWatch stopWatch = StopWatch.start();

    handleContainer(allViews[rootId].getChildren(), allViews, ignoreListeners);

    if (!ignoreListeners) {
      EventHandler eventHandler = Features.getEventHandler(renderer);

      try {
        for (ViewState view in viewsForOnChangeListener) {
          eventHandler.addOnChangeValue(view);
        }
      } finally {
        Utils.clear([viewsForOnChangeListener]);
      }
    }

    StopWatch.stop(stopWatch);
  }

  void handleContainer(List<ViewState> views, OrderedMap<ViewState> allViews,
      bool ignoreListeners) {
    if (Utils.isBlank(views)) {
      return;
    }

    TMap<Script> scripts = FormUtils.getScripts(renderer);

    for (ViewState view in views) {
      addOnChangeListener(view, ignoreListeners, false, scripts, allViews);
      addOnChangeListener(view, ignoreListeners, true, scripts, allViews);

      if (!ignoreListeners) {
        addOnChangeListenerForScript(view.getLabel(), view, scripts, allViews);
        addOnChangeListenerForScript(view.getTitle(), view, scripts, allViews);
        addOnChangeListenerForScript(
            view.getSubTitle(), view, scripts, allViews);
      }

      addOnChangeListener_(view, scripts, allViews);

      refreshRendering(view);

      if (view.isContainer()) {
        handleContainer(view.getChildren(), allViews, ignoreListeners);
      } else if (!ignoreListeners) {
        // Add onClick event if it can
        Features.getEventHandler(renderer).addOnClick(view);
      }
    }
  }

  void addOnChangeListener(ViewState view, bool ignoreListeners, bool rendered,
      TMap<Script> scripts, OrderedMap<ViewState> allViews) {
    final String scriptId =
        rendered ? view.getRenderedIf() : view.getEnabledIf();
    Script script = Model.getScript(scriptId, scripts);

    if (script != null) {
      RelativeRenderAction.refresh(view, scriptId, rendered, renderer);

      if (!ignoreListeners) {
        addOnChangeListener__(view, script, allViews);
      }
    }

    if (!ignoreListeners) {
      // Add listener to view itself
      viewsForOnChangeListener.add(view);
    }
  }

  void addOnChangeListenerForScript(String text, ViewState view,
      TMap<Script> scripts, OrderedMap<ViewState> allViews) {
    Script script = FormUtils.hasScript(text)
        ? Model.getScript(Utils.skipFirst(text), scripts)
        : null;

    if (script != null) {
      addOnChangeListener__(view, script, allViews);
      viewsForOnChangeListener.add(view);
    }
  }

  void addOnChangeListener_(
      ViewState view, TMap<Script> scripts, OrderedMap<ViewState> allViews) {
    MessageArray msgArray =
        FormUtils.getI18n(renderer).getMsgArray(view.getArray());

    if (msgArray != null) {
      for (Script script in msgArray.getConditions(scripts)) {
        addOnChangeListener__(view, script, allViews);
      }
    }

    Field field = view.getField();

    if (field != null) {
      for (Constraint constraint in field.getConstraints().values) {
        if (constraint.canImmediate_(renderer)) {
          viewsForOnChangeListener.add(view);
        }

        Script script =
            Model.getScript(constraint.getEnabledIf(renderer), scripts);

        if (script != null) {
          addOnChangeListener__(view, script, allViews);
        }

        for (String arg in constraint.getArgs().values) {
          if (Code.SCRIPT.starts(arg)) {
            script = Model.getScript(Utils.skipFirst(arg), scripts);

            if (script != null) {
              addOnChangeListener__(view, script, allViews);
            }
          }
        }
      }
    }
  }

  void addOnChangeListener__(
      ViewState view, Script script, OrderedMap<ViewState> allViews) {
    StringSet views =
        script.getViews(view, renderer); // views involved in the script

    if (!Utils.isBlank(views)) {
      for (String viewId in views) {
        ViewState dependsOn = allViews[viewId];

        if (dependsOn != null) {
          dependsOn.addDependent(view);
          viewsForOnChangeListener.add(dependsOn);
        }
      }
    }
  }

  void refreshRendering(ViewState view) {
    // if view supports dependency, re-render the listed components
    // TODO It can blow up if View is IArray
    if (view.getType().canDepend()) {
      StringSet render = view.getRender_(renderer);

      if (!Utils.isBlank(render)) {
        for (String viewId in render) {
          ViewState vRender = FormUtils.getView_(viewId, renderer);

          if (vRender != null) {
            view.addDependent(vRender);
          }
        }

        final bool flag = CBoolean.convertNoErr(view.getValue());

        if (flag != null) {
          ActionHandler handler =
              ActionHandler.getInstance(Action.RENDER.name(), false, renderer);
          ToggleAction().toggleRender(view, flag, handler);
        }
      }
    }
  }
}
