/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/structure/t_map.dart';
import '../../../bundle/base_bundle.dart';
import '../../../bundle/custom_bundle.dart';
import '../../../config/config.dart';
import '../../../structure/i18n/i18n.dart';
import '../../../structure/i18n/locale.dart';
import '../../../util/form_utils.dart';
import '../../../util/stop_watch.dart';
import '../../../util/utils.dart';
import 'aggregate_bundle.dart';
import 'params/renderer_params.dart';
import 'ui_renderer.dart';

class Messages {
  static final TMap<AggregateBundle> BUNDLE_CACHE = TMap();

  final UIRenderer renderer;
  final I18n i18n;

  Locale locale;

  // keep it as class members for better performance
  String language;

  Messages(UIRenderer renderer) : this.params(renderer, renderer.getParams());

  Messages.params(UIRenderer renderer, RendererParams params)
      : this.full(renderer, params.getI18n(), params.getCurrentLocale());

  Messages.full(this.renderer, this.i18n, Locale locale) {
    init(locale);
  }

  void init(Locale locale) {
    final bool noLocale = locale == null;

    this.locale = locale;
    this.language = noLocale ? null : locale.getId();

    // Only in development mode, StopWatch will log some time taking translation
    if (!Config.PRODUCTION_MODE) {
      // In development mode, we don't do caching because it may change on the fly.
      Utils.clear([BUNDLE_CACHE]);
    }
  }

  String get(String i18nMsgId, String bundleMsgId, List<Object> args) {
    StopWatch stopWatch = StopWatch.start();

    bool i18nMsgMissing = Utils.isBlank(i18nMsgId);
    bool bundleMsgMissing = Utils.isBlank(bundleMsgId);

    if (i18nMsgMissing && bundleMsgMissing) {
      return null; // at least one is expected
    }

    String msg;

    if (!i18nMsgMissing) {
      msg = inI18nOrBundle(i18nMsgId);
    }

    if (FormUtils.hasScript(msg)) {
      // check if i18n message got script?
      msg = FormUtils.evalScript(msg, renderer);
    } else {
      if (!bundleMsgMissing && Utils.isBlank(msg)) {
        msg = inI18nOrBundle(bundleMsgId);
      }

      if (!Utils.isBlank(msg) && !Utils.isBlank(args)) {
        msg = Utils.format(msg, args);
      }
    }

    if (stopWatch != null) {
      // just to save toDefault() step
      StopWatch.stop_(Utils.toDefault(i18nMsgId, bundleMsgId), stopWatch);
    }

    return msg;
  }

  String inI18nOrBundle(String msgId) {
    if (Utils.hasSpace(msgId)) {
      return msgId; // don't translate
    }

    String msg = i18n.getText_(
        msgId, locale, this, renderer); // Look for message in i18n first
    return msg == null
        ? getCachedBundle().getMsg(msgId)
        : msg; // message can be configured as blank
  }

  AggregateBundle getCachedBundle() {
    AggregateBundle bundle = BUNDLE_CACHE[language];

    // bundle for required language isn't cached
    if (bundle == null) {
      bundle = AggregateBundle(language, CustomBundle.map, BaseBundle.map);
      BUNDLE_CACHE.put(language, bundle);
    }

    return bundle;
  }
}
