/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../form/structure/ordered_map.dart';
import '../../../../form/structure/string_object_map.dart';
import '../../../../form/structure/t_map.dart';
import '../../../config/defaults.dart';
import '../../../contract/contract_utils.dart';
import '../../../enums/code.dart';
import '../../../enums/field_type.dart';
import '../../../enums/mode.dart';
import '../../../enums/orientation.dart';
import '../../../enums/scope.dart';
import '../../../enums/view_type.dart';
import '../../../error/concurrent_view_exception.dart';
import '../../../generator/form_cache.dart';
import '../../../structure/model/field.dart';
import '../../../structure/ui/method/method.dart';
import '../../../structure/ui/popup.dart' as UIPopup;
import '../../../structure/ui/ui.dart';
import '../../../structure/ui/view.dart';
import '../../../util/date_utils.dart';
import '../../../util/enum_utils.dart';
import '../../../util/form_utils.dart';
import '../../../util/log/logger.dart';
import '../../../util/script/condition_evaluator.dart';
import '../../../util/stop_watch.dart';
import '../../../util/structure/string_set.dart';
import '../../../util/utils.dart';
import '../../../util/view_utils.dart';
import '../component/base/component.dart';
import '../component/behavior/i_contain.dart';
import '../component/button/button.dart';
import '../component/layout/container.dart';
import '../component/layout/popup.dart';
import '../field_state.dart';
import '../form_controller.dart';
import '../method_state.dart';
import '../page/error_page.dart';
import '../page/raw_page.dart';
import '../util/focus_utils.dart';
import '../view_state.dart';
import 'component_factory.dart';
import 'features.dart';
import 'messages.dart';
import "params/renderer_params.dart";

// TODO Use underscore prefix for all the hard-coded component ids.
class UIRenderer {
  FormController controller;
  RendererParams params;
  DateTime initTime; // to know at what time form rendering began
  Messages msgs;
  Features features;

  OrderedMap<ViewState> allViews;
  StringSet inputViews;
  OrderedMap<MethodState> methodStates;
  ViewState vPopup; // currently displayed popup

  StringObjectMap data; // TODO Might possibly cause memory issues in Android.
  StringSet rejectedFields;

  String title, subTitle;

  bool rendering = false; // if renderer is busy constructing the form.
  bool localeChanging =
      false; // if renderer is in locale changing mode, listeners should deactivate.
  bool changedOnServer =
      false; // if loaded data has changed and pushed back to server (e.g. Save, Accept etc.)?

  UIRenderer(this.controller, this.params) {
    this.msgs = Messages(this);
    this.features = Features(this);
  }

  Future<Container> render<T>(ComponentFactory factory) async {
    StopWatch stopWatch = StopWatch.start();

    this.rendering = true;
    this.initTime = DateUtils.now();
    final bool showRaw = params.isShowRaw();
    bool supportsMode = false; // supports current mode?
    UI ui = params.getBaseUI();

    try {
      if (showRaw) {
        return createRawPage(factory);
      }

      supportsMode = ui.supportsMode(params
          .getCurrentMode()); // if unsupported, an exception will be thrown
      ContractUtils.onRender(ui.getPreRender(), this);
      Container root = await create(ui, factory);
      FormCache.set_(params.getFormId(), params.getForm());
      return root;
    } finally {
      this.rendering = false;

      if (!showRaw && supportsMode) {
        ContractUtils.onRender(ui.getPostRender(), this);
      }

      StopWatch.stop(stopWatch);
    }
  }

  Container createRawPage<T extends FormController>(
      ComponentFactory<T> factory) {
    this.allViews = OrderedMap(); // refresh the map

    ViewState vRoot =
        factory.createRoot(Defaults.ID_ROOT, Orientation.VERTICAL, null);
    allViews.put(vRoot.getId(), vRoot);

    construct_(
        RawPage.create(params), vRoot, TMap<FieldState>(), factory, false);
    updateTitle(null, null); // will use title/subTitle from 'params' object

    return vRoot.getComponent() as Container;
  }

  Future<Container> create<T extends FormController>(
      UI ui, ComponentFactory<T> factory) async {
    this.allViews = OrderedMap();
    this.inputViews = StringSet();
    this.methodStates = OrderedMap();

    // 1. Construct root view
    ViewState vRoot = factory.createRoot(
        Defaults.ID_ROOT, ui.getOrientation(), ui.getFocus());
    allViews.put(vRoot.getId(), vRoot);

    Container cRoot = vRoot.getComponent() as Container;
    Mode mode = params.getCurrentMode();

    // 2. Check for permission to initiate the form in the given mode.
    if (!params.getPermissions().permitted(mode)) {
      cRoot.add(
          ErrorPage.custom(get("form.permission.missing", [mode]), factory));
      return cRoot;
    }

    // 3. Fetch form data if we are not in CREATE mode.
    if (!checkMode([Mode.CREATE])) {
      try {
        setData(await fetchData());
      } on ConcurrentViewException {
        cRoot.add(ErrorPage.custom(get("form.concurrent.view.error"), factory));
        return cRoot; // we are not allowed to concurrently display data to multiple users
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
      }

      if (Utils.isBlank(data)) {
        cRoot.add(ErrorPage.custom(get("form.data.fetch.failure"), factory));
        return cRoot; // can't continue as we have no data to show.
      }

      final String msgId = ContractUtils.checkData(ui.getDataCheck(), this);

      if (!Utils.isBlank(msgId)) {
        cRoot.add(ErrorPage.custom(get(msgId), factory));
        return cRoot; // can't continue as this data is rejected for some reason.
      }
    }

    updateTitle(null, null); // will use title/subTitle from 'params' object

    List<View> layout = ui.getLayout();

    // 4. Construct virtual/hidden views undeclared for fields in the model
    TMap<FieldState> fieldStates = TMap();
    construct(fieldStates, vRoot, factory);

    // 5. Construct views in UI bean
    construct_(layout, vRoot, fieldStates, factory, false);
    Utils.clear([fieldStates]);

    // 6th and 7th steps are consolidated within
    handleData(data);

    // 8. Reflect all view attributes impacted by their constraints.
    reflectConstraints(layout, factory);

    // 9. Set the active views of all input components
    vRoot.setActiveViews();

    factory.setTitle_(cRoot, layout);

    return cRoot;
  }

  Future<StringObjectMap> fetchData() async {
    final Object formPK = controller.getFormPK();

    if (formPK is StringObjectMap) {
      return formPK;
    }

    Logger.info("Provided form primary key to fetch the data is $formPK");
    return formPK == null
        ? null
        : await Features.getDataProvider(this).fetch(formPK);
  }

  void handleData(StringObjectMap data) {
    handleData_(data, allViews, Defaults.ID_ROOT);
  }

  // TODO Step (2) should be prior to (1)?
  void handleData_(
      StringObjectMap data, OrderedMap<ViewState> views, String rootId) {
    // 1. Handle array/value attributes of all input components
    Features.getArrayValueHandler(this).handle(data, views);

    // 2. Handle interdependencies of all the views
    Features.getEventProcessor(this).handle(rootId, false, views);
  }

  void construct<T extends FormController>(
      TMap<FieldState> fields, ViewState root, ComponentFactory<T> factory) {
    final bool show = params.getBaseUI().isShowHiddenFields();
    ViewState container;

    for (Field field in params.getModel().getFieldMap().values) {
      if (field.isHidden()) {
        // view isn't declared for this field in the UI
        if (container == null) {
          // wrap all views inside a panel
          View view = View.modes(
              "pnlHidden",
              show ? ViewType.GRIDLAYOUT : ViewType.LINEARLAYOUT,
              show,
              show ? null : Scope.NONE.id());
          container = construct__(view, root, fields, factory, false);
        }

        final String fieldId = field.getId();
        View view = View.onClick(
            fieldId,
            FieldType.getDefaultView(show, field.getType()),
            show ? fieldId : "",
            "");
        construct__(view, container, fields, factory, false);
      }
    }
  }

  void construct_(List<View> layout, ViewState container,
      TMap<FieldState> fields, ComponentFactory factory, bool forPopup) {
    if (!Utils.isBlank(layout)) {
      for (View view in layout) {
        construct__(view, container, fields, factory, forPopup);
      }
    }
  }

  ViewState construct__(View view, ViewState vContainer,
      TMap<FieldState> fields, ComponentFactory factory, bool forPopup) {
    view.resolveType(this);
    final String viewId =
        ViewUtils.validate(view, vContainer, allViews, params);
    FieldState field = FieldState.get(view, fields, this);
    ViewState viewState = ViewState.view(view, vContainer, field,
        isRejected(viewId), params.permitted(view), this);
    allViews.put(viewId, viewState);

    if (inputViews != null && viewState.canInput()) {
      inputViews.add(viewId);
    }

    Component component = factory.create(viewState, view, forPopup);
    viewState.setComponent(component);

    if (viewState.isContainer()) {
      // recursive call to create nested layout
      construct_(view.getLayout(), viewState, fields, factory, forPopup);
    }

    IContain container = vContainer.getComponent() as IContain;

    if (viewState.isLabelledComponent()) {
      // add the created component(s) to its container
      viewState.getLabelledComponent().addToContainer(container);
    } else {
      container.add(component);
    }

    vContainer.addChild(viewState);

    return viewState;
  }

  void reflectConstraints(List<View> layout, ComponentFactory factory) {
    if (Utils.isBlank(layout)) {
      return;
    }

    // this iteration connects all input views for label views set via forId
    for (View view in layout) {
      // this loop can not be merged with next loop
      ViewState viewState = allViews[view.getId()];

      if (viewState != null) {
        ViewState vFor = allViews[viewState.getForId()];

        if (vFor != null && vFor.isLabel()) {
          vFor.setLblView(viewState);
        }
      }
    }

    for (View view in layout) {
      ViewState viewState = allViews[view.getId()];

      if (viewState != null) {
        factory.reflectConstraints(viewState);

        // recursive call to apply attributes for nested layout
        if (viewState.isContainer()) {
          reflectConstraints(view.getLayout(), factory);
        }

        applyMode(viewState, false);
      }
    }
  }

  bool isRejected(String viewId) {
    if (checkMode([Mode.REVISE])) {
      if (rejectedFields == null) {
        String rejectedFieldIds = Utils.isBlank(data)
            ? null
            : data.getAsString(Code.SYS.prefix("REJECTED_FIELDS"));
        rejectedFields = Utils.splitByComma(rejectedFieldIds);
        rejectedFields = rejectedFields == null ? StringSet() : rejectedFields;
      }

      return rejectedFields.contains(viewId);
    }

    return null;
  }

  void applyMode(ViewState view, bool modeChanging) {
    if (view.isRoot()) {
      return;
    }

    if (!view.supportsMode(this)) {
      view.setRendered(
          false); // hide the view if it doesn't support the current form mode.
      return;
    }

    // On initial creation of the form, view rendering depends on the view inclusion and configured 'rendered'
    // attribute. If view supports the mode, setIncluded() will change rendered attribute to be true which will be
    // wrong because it could be configured as false. This is only valid for CREATE mode.
    if (modeChanging || checkMode([Mode.PREVIEW])) {
      final bool rendered =
          checkMode([Mode.CREATE]) ? view.isInitialRendered() : true;
      view.setRendered(
          ConditionEvaluator.eval_(view.getRenderedIf(), rendered, this));
    }

    // if mode is changing we can toggle the 'enabled' property of a view but if the form is initially created, we
    // should take in consideration the 'enabled' attribute for the given view.
    bool preEnabled = modeChanging || view.isEnabled();
    bool enabled = true; // by default and for CREATE mode it should be true.

    // We don't have to check for other conditions if preEnabled is false. This is because of the AND condition.
    if (preEnabled) {
      if (modeChanging && checkMode([Mode.CREATE])) {
        enabled = ConditionEvaluator.eval_(
            view.getEnabledIf(), view.isInitialEnabled(), this);
      } else if (checkMode([Mode.UPDATE])) {
        enabled = view.canUpdate(this);
      } else if (checkMode([
        Mode.PRE_SAVE_VIEW,
        Mode.PREVIEW,
        Mode.POST_SAVE_VIEW,
        Mode.REJECT,
        Mode.REVISE
      ])) {
        // all popup views should be enabled. Outside popup all the views should be disabled except for buttons.
        enabled = vPopup != null || view.getComponent() is Button;

        if (!enabled) {
          // if 'enabled' is still false
          if (checkMode([Mode.PREVIEW])) {
            // some views are explicitly enabled in preview mode
            enabled = view.isPreviewEdit(this);
          } else if (checkMode([Mode.REVISE])) {
            // check for rejected fields for the REVISE mode
            enabled = view.isRejected(
                this); // enable if view was rejected so that the value can be changed
          }
        }
      }
    }

    view.setEnabled(preEnabled && enabled);
  }

  bool changeMode(Mode mode) {
    StopWatch stopWatch = StopWatch.start();

    try {
      params
          .getBaseUI()
          .supportsMode(mode); // if unsupported, an exception will be thrown

      if (!params.changeMode(mode)) {
        return false;
      }

      for (ViewState viewState in getAllViewStates()) {
        applyMode(viewState, true);
      }

      // Re-evaluate conditions/inclusions/enablement
      Features.getEventProcessor(this).handle(Defaults.ID_ROOT, true, allViews);
    } finally {
      StopWatch.stop(stopWatch);
    }

    return true;
  }

  bool changeLocale(String language) {
    StopWatch stopWatch = StopWatch.start();

    if (!params.changeLocale(language)) {
      return false;
    }

    localeChanging = true;

    try {
      msgs.init(params.getCurrentLocale());

      ComponentFactory factory = Features.getFactory_(this);
      factory.changeLocale();
      factory.changeDir(FormUtils.isLtr(this), allViews[Defaults.ID_ROOT]);

      updateTitle(title, subTitle);
    } finally {
      localeChanging = false;
      StopWatch.stop(stopWatch);
    }

    return true;
  }

  void createPopup(UIPopup.Popup popup) {
    StopWatch stopWatch = StopWatch.start();
    // TODO We don't need this, just prefix 'root_' to given popup's id?
    const String rootId = Defaults.ID_POPUP_ROOT;

    if (Utils.got(rootId, allViews)) {
      Logger.warn(
          "Can't create another popup until the previous one is closed.");
      return;
    }

    ComponentFactory factory = Features.getFactory_(this);

    // 1. Create root layout for the popup.
    ViewState vRoot =
        factory.createRoot(rootId, popup.getOrientation(), popup.getFocus());
    allViews.put(rootId, vRoot);

    // 2. Create the popup and add root layout in it.
    const String popupId = Defaults.ID_POPUP;
    final String width = popup.getWidth();
    final String height = popup.getHeight();

    Popup dlgPopup = factory.popup(
        get(popup.getTitle()), popup.isCloseable(), width, height);
    dlgPopup.setIdentity(popupId);
    factory.setSize(width, height, dlgPopup);
    factory.setStyle(popup.getStyle(), dlgPopup);

    // 3. Transform Popup to ViewState.
    vPopup = ViewState.popup(popupId, true, popup.getFocus());
    vPopup.setComponent(dlgPopup);
    vPopup.addChild(vRoot);
    allViews.put(popupId, vPopup);

    List<View> layout = popup.getLayout();

    // TODO Should 'fields' map be shared between construct() and here?
    // 4. Create the inner layout.
    construct_(layout, vRoot, TMap<FieldState>(), factory, true);

    // 5. Reflect view attributes in popup layout impacted by constraints.
    reflectConstraints(layout, factory);

    // 6. Get all the created views in the popups.
    OrderedMap<ViewState> popupViews = vRoot.getSubViews();
    popupViews.put(rootId, vRoot);

    if (!Utils.isBlank(popupViews)) {
      // 7th and 8th step consolidated within
      handleData_(null, popupViews, rootId);
    }

    // 9. Show the popup.
    dlgPopup.add(vRoot.getComponent());
    vPopup.setRendered(true);

    StopWatch.stop(stopWatch);
  }

  void closePopup() {
    if (vPopup != null) {
      ViewUtils.removeReferences(vPopup, allViews, inputViews);
      (vPopup.getComponent() as Popup).clear();
      vPopup.setRendered(false);
      vPopup = null;
    }
  }

  String getFocus() {
    return FocusUtils.getFocus(
        vPopup == null ? Defaults.ID_ROOT : Defaults.ID_POPUP_ROOT, allViews);
  }

  void updateTitle(String title, String subTitle) {
    this.title = Utils.toDefault(title, params.getTitle());
    this.subTitle = Utils.toDefault(subTitle, params.getSubTitle());

    controller.setTitle_(FormUtils.eval_(this.title, this),
        FormUtils.eval_(this.subTitle, this));
  }

  void reset() {
    initTime = null;

    Utils.clear([allViews, methodStates, data, inputViews, rejectedFields]);

    vPopup = null;
    title = subTitle = null;
  }

  void cleanUp() {
    reset();

    if (features != null) {
      features.cleanUp();
      features = null;
    }

    controller = null;
    params = null;
    msgs = null;
  }

  FormController getController() {
    return this.controller;
  }

  String get(String msgId, [Object args]) {
    return getMsg(null, msgId, args);
  }

  String getMsg(String i18nMsgId, String bundleMsgId, List<Object> args) {
    return msgs.get(i18nMsgId, bundleMsgId, args);
  }

  Messages getMsgs() {
    return this.msgs;
  }

  Features getFeatures() {
    return this.features;
  }

  OrderedMap<ViewState> getAllViews() {
    return this.allViews;
  }

  Iterable<ViewState> getAllViewStates() {
    return getAllViews().values;
  }

  StringSet getInputViews() {
    return this.inputViews;
  }

  MethodState getMethodState(Method method) {
    final String methodId = method.getId();
    MethodState methodState = methodStates[methodId];

    if (methodState == null) {
      methodState = MethodState(method);
      methodStates.put(methodId, methodState);
    }

    return methodState;
  }

  RendererParams getParams() {
    return this.params;
  }

  bool checkMode(List<Mode> modes) {
    return EnumUtils.match_(params.getCurrentMode(), modes);
  }

  ViewState getPopupView() {
    return this.vPopup;
  }

  bool isLocaleChanging() {
    return this.localeChanging;
  }

  bool isChangedOnServer() {
    return this.changedOnServer;
  }

  void setChangedOnServer(bool changedOnServer) {
    this.changedOnServer = changedOnServer;
  }

  bool isRendering() {
    return this.rendering;
  }

  DateTime getInitTime() {
    return this.initTime;
  }

  StringObjectMap getData() {
    return this.data;
  }

  void setData(StringObjectMap data) {
    this.data = Utils.isBlank(data) ? null : data;
  }
}
