/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../form/structure/ordered_map.dart';
import '../../config/defaults.dart';
import '../../enums/action.dart';
import '../../enums/field_type.dart';
import '../../enums/scope.dart';
import '../../enums/view_type.dart';
import '../../error/unsupported_operation_exception.dart';
import '../../structure/model/field.dart';
import '../../structure/ui/view.dart';
import '../../util/enum_utils.dart';
import '../../util/form_utils.dart';
import '../../util/script/condition_evaluator.dart';
import '../../util/stop_watch.dart';
import '../../util/structure/string_set.dart';
import '../../util/structure/unique_string_map.dart';
import '../../util/utils.dart';
import '../../validation/converter.dart';
import '../../validation/converter/c_boolean.dart';
import 'component/base/component.dart';
import 'component/base/input_component.dart';
import 'component/behavior/i_array.dart';
import 'component/behavior/i_enable.dart';
import 'component/behavior/i_err.dart';
import 'component/behavior/i_focus.dart';
import 'component/behavior/i_input.dart';
import 'component/composition/labelled_component.dart';
import 'component/label.dart';
import 'field_state.dart';
import 'renderer/ui_renderer.dart';

// TODO Put all attributes in a map?
class ViewState {
  final String
      id; // unique id of the view in the whole layout (should be as per the JSF standards)
  final ViewType
      type; // Type of the UI component that belongs to this ViewState
  final bool permitted; // View is permitted to be used by the user?
  bool excludedOnFetch =
      false; // When data is fetched for this view, it can be excluded from the form.

  bool initialRendered =
      false; // View was configured to be initially rendered (via XML/JSON)?
  bool rendered = false; // View is rendered?
  String renderedIf; // Script to evaluate if view is rendered?

  bool initialEnabled =
      false; // View was configured to be initially enabled (via XML/JSON)?
  bool enabled = false; // View is enabled?
  String enabledIf; // Script to evaluate if view is enabled?

  final StringSet modes; // modes supported by this view
  final String
      initialLabel; // Initially configured i18n label key (via XML/JSON)
  String label; // Current label of this view (defaults to initialLabel)
  final String hint; // Prompt text/hint for components that have input-text
  final String
      initialTitle; // Initially configured i18n header/title key of the page/activity (via XML/JSON)
  String title; // Current title of this view (defaults to initialTitle)
  final String subTitle; // Sub header/title of the page/activity
  String
      onClick; // Action enum or any custom method. (onClick event for IAction)
  final String params; // view parameters
  final String popupId; // popup to show on button click.
  StringSet
      render; // View ids to render/hide (+/- prefix) when action is executed successfully
  final StringSet
      enable; // View ids to enable/disable (+/- prefix) when action is executed successfully
  final StringSet validate; // View ids to validate
  final bool immediate; // if true, no field will be validated on button click
  final String
      focus; // View to focus when a layout is rendered or after a button's action is successful.
  String
      parentOf; // This view is a parent of which view? This is defined in field and view tags.
  final FieldState field; // Corresponding field state for this view
  ErrMsg errMsg; // Error message if its an input component
  Component component; // Component belonging to the configured facet
  List<ViewState>
      children; // if view is a layout/container, this holds child views
  Set<ViewState> dependents; // views which are dependent on this view
  final StringSet
      clear; // Views to nullify the values (onClick or onValueChange)
  final UniqueStringMap
      set; // Views to set the values (onClick or onValueChange)
  final StringSet reset; // Views to reset the values (onClick or onValueChange)
  final String forId; // use for separated label and input-component
  ViewState vLbl; // Label view that is configured for this view (using forId)
  final String
      onValueChange; // Interface to invoke when the input view detects value change
  final String onSuccess; // What to do when action is successful?
  final bool toastError; // To toast the error or show it in the adjacent label?
  final bool rejected; // View's value is rejected and should be revised?
  final String previewEdit; // View is editable explicity in PREVIEW mode?
  final ViewState container; // parent container view

  ViewState(
      this.id,
      this.type,
      bool permitted,
      String rendered,
      String enabled,
      String modes,
      String label,
      this.hint,
      String title,
      this.subTitle,
      this.onClick,
      this.params,
      this.popupId,
      String render,
      String enable,
      String validate,
      bool immediate,
      this.focus,
      String clear,
      String set,
      String reset,
      this.forId,
      this.onValueChange,
      this.onSuccess,
      bool toastError,
      bool rejected,
      this.previewEdit,
      String parentOf,
      this.field,
      this.container)
      : this.permitted = Utils.toTrue(permitted),
        this.modes = Utils.splitByComma(modes),
        this.initialLabel = label,
        this.initialTitle = title,
        this.render = Utils.splitByComma(render),
        this.enable = Utils.splitByComma(enable),
        this.validate = Utils.splitByComma(validate),
        this.immediate = Utils.toFalse(immediate),
        this.clear = Utils.splitByComma(clear),
        this.set = FormUtils.pairsToMap(set),
        this.reset = Utils.splitByComma(reset),
        this.toastError = Utils.toFalse(toastError),
        this.rejected = Utils.toFalse(rejected) {
    FormUtils.checkTypeCompatibility(id, type, field);

    setRendered_(rendered);
    this.initialRendered = this.rendered;
    setEnabled_(enabled);
    this.initialEnabled = this.enabled;
    this.parentOf = Utils.isBlank(parentOf)
        ? (field == null ? null : field.getParentOf())
        : parentOf;
    this.errMsg = canInput() ? ErrMsg() : null;
  }

  ViewState.popup(String id, bool isPopup, String focus)
      : this(
            id,
            isPopup ? ViewType.POPUP : ViewType.LINEARLAYOUT,
            null,
            (!isPopup).toString(),
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            focus,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null,
            null);

  static ViewState view(View view, ViewState vContainer, FieldState field,
      bool rejected, bool permitted, UIRenderer renderer) {
    return ViewState(
        view.getId(),
        view.getType(),
        permitted,
        view.getRendered(),
        view.getEnabled(),
        view.getModes(),
        view.getLabel(renderer),
        view.getHint(renderer),
        view.getTitle(renderer),
        view.getSubTitle(renderer),
        view.getOnClick(),
        view.getParams(),
        view.getPopup(),
        view.getRender(),
        view.getEnable(),
        view.getValidate(),
        view.isImmediate(),
        view.getFocus(),
        view.getClear(),
        view.getSet(),
        view.getReset(),
        view.getForId(),
        view.getOnValueChange(),
        view.getOnSuccess(),
        view.isToastError(),
        rejected,
        view.getPreviewEdit(),
        view.getParentOf(),
        field,
        vContainer);
  }

  void addChild(ViewState view) {
    if (!isContainer()) {
      throw UnsupportedOperationException.msg(
          "Can only add child view to a container!");
    }

    if (children == null) {
      children = [];
    }

    children.add(view);
  }

  String getId() {
    return this.id;
  }

  String getFieldId() {
    return canInput() ? Utils.toDefault(field.getId(), id) : null;
  }

  bool isRoot() {
    return Utils.equalsIgnoreCase(Defaults.ID_ROOT, id);
  }

  bool isLabel() {
    return component is Label;
  }

  bool isLabelledComponent() {
    return component is LabelledComponent;
  }

  ViewType getType() {
    return this.type;
  }

  FieldType getFieldType() {
    return FormUtils.getFieldType(type, field);
  }

  bool isContainer() {
    return getType().isContainer();
  }

  void refresh(String scriptId, bool rendered, UIRenderer renderer) {
    if (permitted) {
      bool flag = ConditionEvaluator.eval(scriptId, renderer);

      if (flag != null) {
        if (rendered) {
          setRendered(flag && supportsMode(renderer));
        } else {
          setEnabled(flag);
        }
      }
    }
  }

  bool isIncluded(UIRenderer renderer) {
    return !excludedOnFetch &&
        ConditionEvaluator.eval_(getInclusion(), Defaults.INCLUDED, renderer);
  }

  void setExcludedOnFetch(bool excludedOnFetch) {
    this.excludedOnFetch = excludedOnFetch;
  }

  bool isInitialRendered() {
    return this.initialRendered;
  }

  bool isRendered() {
    return this.rendered;
  }

  String getRenderedIf() {
    return this.renderedIf;
  }

  // rendered attribute may be true but its container may not be, in such case view is not considered as rendered.
  bool isReallyRendered() {
    bool isRendered = this.isRendered();

    if (container != null) {
      isRendered = container.isReallyRendered() && isRendered;
    }

    return isRendered;
  }

  void setRendered(bool rendered) {
    // 1. View should have the needed permission.
    // 2. Rendered attribute can't be altered for root.
    // 3. New 'rendered' value is different to current value.
    // 4. Component shouldn't be null.
    if (this.permitted &&
        !isRoot() &&
        this.rendered != rendered &&
        this.component != null) {
      this.rendered = rendered;
      this.component.setRendered(this.rendered);

      setActiveViews();
    }
  }

  bool isActive() {
    StringSet activeViews = getActiveViews();
    return activeViews != null && activeViews.contains(id);
  }

  StringSet getActiveViews() {
    return field == null ? null : field.getActiveViews();
  }

  void setActiveViews() {
    if (field != null && canInput()) {
      StringSet activeViews = getActiveViews();

      if (isRendered()) {
        activeViews.add(id);
      } else {
        activeViews.remove(id);
      }
    } else if (isContainer()) {
      if (!Utils.isBlank(children)) {
        for (ViewState child in children) {
          child.setActiveViews();
        }
      }
    }
  }

  void setRendered_(String rendered) {
    final bool isRendered =
        this.permitted ? CBoolean.parseNoErr(rendered) : CBoolean.NO;
    final bool cantParse = isRendered == null;
    this.rendered = cantParse ? true : isRendered;
    this.renderedIf = cantParse ? rendered : null;
  }

  void forceRender() {
    if (isRoot()) {
      // hide all top panels before rendering one
      List<ViewState> vChildren = getChildren();

      if (!Utils.isBlank(vChildren)) {
        for (ViewState vChild in vChildren) {
          if (vChild.isContainer()) {
            vChild.setRendered(false);
          }
        }
      }
    } else {
      container.forceRender();
    }

    setRendered(true);
  }

  bool isInitialEnabled() {
    return this.initialEnabled;
  }

  bool isEnabled() {
    return this.enabled;
  }

  String getEnabledIf() {
    return this.enabledIf;
  }

  void setEnabled(bool enabled) {
    // 1. View should have the needed permission.
    // 2. Enabled attribute can't be altered for root.
    // 3. New 'enabled' value is different to current value.
    // 4. Component should support enable feature and it shouldn't be null.
    if (this.permitted &&
        !isRoot() &&
        this.enabled != enabled &&
        this.component is IEnable) {
      this.enabled = enabled;
      (this.component as IEnable).setEnabled(enabled);
    }
  }

  void setEnabled_(String enabled) {
    final bool isEnabled = CBoolean.parseNoErr(enabled);
    final bool cantParse = isEnabled == null;
    this.enabled = cantParse ? Defaults.ENABLED : isEnabled;
    this.enabledIf = cantParse ? enabled : null;
  }

  /// Check if this view supports the current form mode.
  bool supportsMode(UIRenderer renderer) {
    if (Utils.isBlank(modes)) {
      // if nothing is configured then we refer to it's container
      return isRoot() || container == null || container.supportsMode(renderer);
    }

    return !modes.contains(Scope.NONE.id()) &&
        modes.contains(FormUtils.getMode(renderer).toString());
  }

  bool isFocusable() {
    return unwrapLabelledComponent() is IFocus && isEnabled();
  }

  /// View has to be persisted or it's just for view?
  bool canPersist(UIRenderer renderer) {
    Field field = getField();
    bool hidden = field != null && field.isHidden();
    return canPersist_(!hidden && !isLabel(), renderer);
  }

  bool canPersist_(bool defaultValue, UIRenderer renderer) {
    Field field = getField();
    return Utils.toDefault(
        field == null ? null : field.canPersist(defaultValue, renderer),
        defaultValue);
  }

  /// View is updatable in 'update' form mode?
  bool canUpdate(UIRenderer renderer) {
    Field field = getField();
    return canPersist(renderer) &&
        Utils.toTrue(field == null ? null : field.canUpdate(renderer));
  }

  /// View is rejectable in 'reject' form mode?
  ///
  /// If view can't be updated, it can neither be rejected.
  bool isRejectable(UIRenderer renderer) {
    Field field = getField();
    return canUpdate(renderer) &&
        Utils.toTrue(field == null ? null : field.isRejectable());
  }

  bool isPreviewEdit(UIRenderer renderer) {
    return Utils.toFalse(
        ConditionEvaluator.eval_(previewEdit, CBoolean.NO, renderer));
  }

  /// View was rejected in the reject mode? If so, the view's value could be revised by the user.
  bool isRejected(UIRenderer renderer) {
    return rejected && isRejectable(renderer);
  }

  bool isExtra() {
    Field field = getField();
    return field == null ? Defaults.EXTRA : field.isExtra();
  }

  String getInitialLabel() {
    return this.initialLabel;
  }

  String getLabel() {
    return Utils.toDefault(label, initialLabel);
  }

  void setLabel(String label) {
    this.label = label;
  }

  String getHint() {
    return this.hint;
  }

  String getInitialTitle() {
    return this.initialTitle;
  }

  String getTitle() {
    return Utils.toDefault(title, initialTitle);
  }

  void setTitle(String title) {
    this.title = title;
  }

  String getSubTitle() {
    return this.subTitle;
  }

  String getOnClick() {
    return this.onClick;
  }

  void setOnClick(String onClick) {
    this.onClick = onClick;
  }

  String getParams() {
    return this.params;
  }

  String getPopupId() {
    return this.popupId;
  }

  StringSet getRender() {
    return this.render;
  }

  StringSet getRender_(UIRenderer renderer) {
    return FormUtils.expand(render, id, false, false, renderer);
  }

  void setRender(StringSet render) {
    this.render = render;
  }

  StringSet getEnable() {
    return this.enable;
  }

  StringSet getEnable_(UIRenderer renderer) {
    return FormUtils.expand(enable, id, false, true, renderer);
  }

  StringSet getValidate() {
    return this.validate;
  }

  StringSet getValidate_(UIRenderer renderer) {
    return FormUtils.expand(validate, id, false, true, renderer);
  }

  bool isImmediate() {
    return this.immediate;
  }

  String getFocus() {
    return this.focus;
  }

  String getForId() {
    return this.forId;
  }

  ViewState getLblView() {
    return this.vLbl;
  }

  void setLblView(ViewState vLbl) {
    this.vLbl = vLbl;
  }

  String getOnValueChange() {
    return this.onValueChange;
  }

  String getOnSuccess() {
    return Utils.isBlank(this.onSuccess)
        ? (EnumUtils.match(onClick, [Action.SAVE, Action.UPDATE])
                ? Action.PREVIEW
                : Action.NOTHING)
            .name()
        : this.onSuccess;
  }

  String getParentOf() {
    return this.parentOf;
  }

  Field getField() {
    return field == null ? null : field.getField();
  }

  bool isMandatory(UIRenderer renderer) {
    return field != null && field.isMandatory(renderer);
  }

  bool hasError() {
    if (errMsg != null && !Utils.isBlank(errMsg.getMsgId())) {
      return true;
    }

    if (!Utils.isBlank(children)) {
      // if view is a container with children
      for (ViewState view in children) {
        if (view.hasError()) {
          return true;
        }
      }
    }

    return false;
  }

  String getConversionError() {
    Field field = getField();
    return field == null ? null : field.getConversionError();
  }

  ErrMsg getErrMsg() {
    return this.errMsg;
  }

  void clearErrMsg() {
    if (errMsg != null && canInputAndErr()) {
      errMsg.clear();

      if (canErr()) {
        // clear the error that was set before
        (component as IErr).setErrMsg(null);
      }
    }
  }

  Component unwrapLabelledComponent() {
    Component component = getInputComponent();
    return component == null ? this.component : component;
  }

  InputComponent getInputComponent() => component is InputComponent
      ? LabelledComponent.getComponent(component) as InputComponent
      : null;

  /// Mostly label and input-component are vertically aligned but there can be layout in which label is on top of the
  /// input-component which means when we need to scroll to this view, we should scroll to the component on top.
  Component getComponentOnTop() {
    if (isLabelledComponent()) {
      LabelledComponent labelledComponent = getLabelledComponent();
      Label label = labelledComponent.getLblText();
      return label == null ? labelledComponent.getInputComponent() : label;
    }

    return getComponent();
  }

  Component getComponent() {
    return this.component;
  }

  void setComponent(Component component) {
    this.component = component;
  }

  LabelledComponent getLabelledComponent() {
    return getComponent() as LabelledComponent;
  }

  OrderedMap<ViewState> getSubViews() {
    OrderedMap<ViewState> views;

    if (isContainer()) {
      List<ViewState> children = getChildren();

      if (!Utils.isBlank(children)) {
        views = OrderedMap();

        for (ViewState viewState in children) {
          OrderedMap<ViewState> subViews = viewState.getSubViews();

          if (!Utils.isBlank(subViews)) {
            views.addAll(subViews);
          }

          views.put(viewState.getId(), viewState);
        }
      }
    }

    return views;
  }

  List<ViewState> getChildren() {
    return this.children;
  }

  void setChildren(List<ViewState> children) {
    this.children = children;
  }

  Set<ViewState> getDependents() {
    return this.dependents;
  }

  void addDependent(ViewState dependent) {
    if (dependent != null) {
      if (dependents == null) {
        dependents = {};
      }

      dependents.add(dependent);
    }
  }

  String getContainerId() {
    return container == null ? null : container.getId();
  }

  ViewState getContainer() {
    return this.container;
  }

  bool canInput() {
    // using 'component instanceof IInput' can be misleading (e.g. 'AColumn extends ALabel')
    return type.canInput();
  }

  bool canErr() {
    return component is IErr;
  }

  bool canArray() {
    return getInputComponent() is IArray;
  }

  bool canInputAndErr() {
    return canInput() && (toastError || canErr());
  }

  ViewState getParent(UIRenderer renderer) {
    for (ViewState view in renderer.getAllViewStates()) {
      Field field = view.getField();

      if (field != null && Utils.equalsIgnoreCase(id, getParentOf())) {
        return view;
      }
    }

    return null;
  }

  Set<String> getParentOf_(OrderedMap<ViewState> allViews) {
    Set<String> parentOfIds;

    if (field != null) {
      ViewState child = allViews[getParentOf()];

      if (child != null) {
        parentOfIds = {};
        parentOfIds.add(child.getId());

        Utils.addAll(parentOfIds, [child.getParentOf_(allViews)]);
      }
    }

    return parentOfIds;
  }

  StringSet getClear() {
    return this.clear;
  }

  StringSet getClear_(UIRenderer renderer) {
    return FormUtils.expand(clear, id, true, true, renderer);
  }

  UniqueStringMap getSet() {
    return this.set;
  }

  StringSet getReset() {
    return this.reset;
  }

  StringSet getReset_(UIRenderer renderer) {
    return FormUtils.expand(reset, id, true, true, renderer);
  }

  String getInclusion() {
    return field == null ? null : field.getInclusion();
  }

  String getPreviewValue() {
    return field == null ? null : field.getPreviewValue();
  }

  Object getValue() {
    if (!canInput()) {
      return null;
    }

    StopWatch stopWatch = StopWatch.start();

    Object value = (component as IInput).getVal();
    final String previewValue = getPreviewValue();
    value = Utils.isBlank(previewValue) || previewValue != value ? value : null;

    StopWatch.stop(stopWatch);

    return value;
  }

  Object getConvertedValue() {
    return Converter.convertNoErr(getFieldType(), getValue());
  }

  bool setValue(Object value) {
    if (!canInput()) {
      return false;
    }

    StopWatch stopWatch = StopWatch.start();
    (component as IInput).setVal(value);
    StopWatch.stop_(value, stopWatch);

    return true;
  }

  String getDefinedArray() {
    return field == null ? null : field.getDefinedArray();
  }

  String getArray() {
    return field == null ? null : field.getArray();
  }

  void setArray(String array) {
    if (field != null) {
      field.setArray(array);
    }
  }

  String getInvalidItemMsg() {
    Field field = getField();
    String invalidItemMsg = field == null ? null : field.getInvalidItemMsg();
    return Utils.toDefault(invalidItemMsg, "array.invalid.item");
  }

  bool renderAndFocus(bool forceRender) {
    bool isReallyRendered = this.isReallyRendered();

    if (!isReallyRendered && forceRender) {
      this.forceRender();
      isReallyRendered = true;
    }

    if (isReallyRendered && isFocusable()) {
      (component as IFocus).focus();
    }

    return isReallyRendered;
  }

  @override
  int get hashCode {
    return 31 + (id == null ? 0 : id.hashCode);
  }

  @override
  bool operator ==(Object obj) {
    if (identical(this, obj)) {
      return true;
    }

    if (obj == null) {
      return false;
    }

    if (runtimeType != obj.runtimeType) {
      return false;
    }

    ViewState other = obj;

    if (id == null) {
      if (other.id != null) {
        return false;
      }
    } else if (!Utils.equalsIgnoreCase(id, other.id)) {
      return false;
    }

    return true;
  }

  @override
  String toString() => getId();
}

class ErrMsg {
  String
      msgId; // if view can input, this is the error message id in i18n or bundle
  List<Object> args; // arguments for templated message

  void set(String msgId, List<Object> args) {
    this.msgId = Utils.trim(msgId);
    this.args = args;
  }

  void clear() {
    set(null, null);
  }

  String getMsgId() {
    return this.msgId;
  }

  List<Object> getArgs() {
    return this.args;
  }
}
