/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../error/illegal_argument_exception.dart';
import '../../numbers.dart';
import '../../utils.dart';

class Range<T extends num> {
  // Either will be non-null
  final T min;
  final T max;

  Range(this.min, this.max) {
    if (Utils.noneNull([min, max]) && min.compareTo(max) > 0) {
      throw IllegalArgumentException("Min shouldn't be greater than max.");
    }
  }

  static Range<T> get<T extends num>(T min, T max) {
    return Utils.allNull([min, max]) ? null : Range(min, max);
  }

  static T getMax<T extends num>(Range<T> r1, Range<T> r2) {
    return Numbers.getMax(
        r1 == null ? null : r1.max, r2 == null ? null : r2.max);
  }

  bool isMinNatural() {
    return min != null && min.toInt() > 0;
  }

  bool isMaxNatural() {
    return max != null && max.toInt() > 0;
  }

  bool belowMin(T number) {
    return Utils.noneNull([min, number]) && number.compareTo(min) < 0;
  }

  bool beyondMax(T number) {
    return Utils.noneNull([max, number]) && number.compareTo(max) > 0;
  }

  bool isInRange(T number) {
    if (number == null) {
      return false;
    }

    if (min == null) {
      return max.compareTo(number) >= 0;
    }

    if (max == null) {
      return min.compareTo(number) <= 0;
    }

    return min.compareTo(number) <= 0 && max.compareTo(number) >= 0;
  }

  bool minEqualsMax() {
    return Utils.noneNull([min, max]) && min.compareTo(max) == 0;
  }

  bool overlaps(Range<T> that) {
    try {
      return (min.compareTo(that.max) < 0 || min.compareTo(that.max) == 0) &&
          (that.min.compareTo(max) < 0 || that.min.compareTo(max) == 0);
    } catch (e) {
      return false;
    }
  }

  static bool isNonOverlapping<T extends num>(List<Range<T>> ranges) {
    final int size = Utils.getSize(ranges);

    for (int i = 0; i < size; i++) {
      for (int j = i + 1; j < size; j++) {
        if (ranges[i].overlaps(ranges[j])) {
          return false;
        }
      }
    }

    return true;
  }

  @override
  int get hashCode =>
      31 * (31 + (max == null ? 0 : max.hashCode)) +
      (min == null ? 0 : min.hashCode);

  @override
  bool operator ==(Object obj) {
    if (identical(this, obj)) {
      return true;
    }

    if (obj == null || runtimeType != obj.runtimeType) {
      return false;
    }

    Range other = obj;

    if (max == null) {
      if (other.max != null) {
        return false;
      }
    } else if (max != other.max) {
      return false;
    }

    if (min == null) {
      if (other.min != null) {
        return false;
      }
    } else if (min != other.min) {
      return false;
    }

    return true;
  }

  @override
  String toString() => "Range [min=$min, max=$max]";
}
