/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:io' show File;
import 'dart:typed_data' show Uint8List;

import '../../error/file_not_found_exception.dart';
import '../../error/illegal_argument_exception.dart';
import '../io_utils.dart';
import '../log/logger.dart';
import '../stop_watch.dart';
import '../utils.dart';
import 'algorithm.dart';
import 'des.dart';

class Cypher {
  static const int MODE_ENCRYPT = 1;
  static const int MODE_DECRYPT = 2;

  final Algorithm algorithm;
  final String iv;
  final String key;

  Cypher(Algorithm algorithm) : this.full(algorithm, null, null);

  Cypher.algo(String algorithm, String iv, String key)
      : this.full(Algorithm.toAlgorithm(algorithm, false), iv, key);

  Cypher.full(this.algorithm, String iv, String key)
      : this.iv = Utils.toDefault(iv, algorithm == null ? null : algorithm.iv),
        this.key = algorithm == null ? null : algorithm.truncate(key) {
    if (algorithm == null) {
      throw IllegalArgumentException("Please provide algorithm.");
    }
  }

  Uint8List encrypt(String text) {
    return encrypt_(Utils.toBytes__(text));
  }

  Uint8List encrypt_(Uint8List content) {
    return process(content, MODE_ENCRYPT);
  }

  Uint8List encryptNoErr(String text) {
    return encryptNoErr_(Utils.toBytes__(text));
  }

  Uint8List encryptNoErr_(Uint8List content) {
    try {
      return encrypt_(content);
    } catch (e) {
      return content;
    }
  }

  String decrypt(String text) {
    return Utils.toString_(decrypt_(Utils.toBytes__(text)));
  }

  String decryptNoErr(String text) {
    try {
      return decrypt(text);
    } catch (e) {
      return text;
    }
  }

  Uint8List decrypt_(Uint8List content) {
    return process(content, MODE_DECRYPT);
  }

  Uint8List decryptNoErr_(Uint8List content) {
    try {
      return decrypt_(content);
    } catch (e) {
      return content;
    }
  }

  bool encryptFile(String file) {
    Logger.info("Encrypting file '$file'");
    return processFile(file, MODE_ENCRYPT);
  }

  bool encryptFileNoErr(String file) {
    try {
      return encryptFile(file);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return false;
    }
  }

  bool decryptFile(String file) {
    Logger.info("Decrypting file '$file'");
    return processFile(file, MODE_DECRYPT);
  }

  bool decryptFileNoErr(String file) {
    try {
      return decryptFile(file);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return false;
    }
  }

  Uint8List process(Uint8List content, int mode) {
    return doFinal(content, MODE_ENCRYPT == mode, prepare());
  }

  bool processFile(String filePath, int mode) {
    StopWatch stopWatch = StopWatch.start();
    bool success = false;

    try {
      File file = IOUtils.getFileIfExists(filePath);

      if (file == null) {
        throw FileNotFoundException();
      }

      // TODO How to process the file without loading it's content? (like in Java)
      IOUtils.writeFile_(process(IOUtils.readFile_(file), mode), file);

      success = true;
    } finally {
      StopWatch.stop(stopWatch);
    }

    return success;
  }

  // TODO Implement other algorithms?
  dynamic prepare() {
    if (Algorithm.DES128 == algorithm) {
      // NOTE: 'dart_des' package was too slow, hence removed.
      // If paddingType isn't PKCS7, then content encrypted by CMS Dart can't
      // be decrypted by CMS Java and vice versa.
      // NOTE: If original package is used, following should be:
      //       return DES(
      //           key: Utils.toBytes(key),
      //           iv: iv == null ? DES.IV_ZEROS : Utils.toBytes(iv),
      //           paddingType: DESPaddingType.PKCS7);
      return DES(Utils.toBytes__(key));
    } else if (Algorithm.AES256 == algorithm) {
      // TODO Delete it or keep it for future reference?
      // NOTE: 'aes_crypt' package had problems, hence removed.
      // AesCrypt crypt = AesCrypt();
      // crypt.setOverwriteMode(AesCryptOwMode.on);
      // crypt.setUserData(createdBy: '');
      // crypt.aesSetMode(AesMode.ecb);
      //
      // Uint8List keySpec = Utils.toBytes__(key);
      //
      // if (Utils.isBlank(iv)) {
      //   crypt.aesSetKeys(keySpec);
      // } else {
      //   crypt.aesSetKeys(keySpec, Utils.toBytes__(iv));
      // }
      //
      // return crypt;
    }

    throw UnimplementedError("Algorithm isn't implemented: $algorithm");
  }

  // TODO Implement other algorithms?
  Uint8List doFinal(Uint8List content, bool encrypt, dynamic cipher) {
    if (Algorithm.DES128 == algorithm) {
      return encrypt ? cipher.encrypt(content) : cipher.decrypt(content);
    } else if (Algorithm.AES256 == algorithm) {
      return encrypt ? cipher.aesEncrypt(content) : cipher.aesDecrypt(content);
    }

    throw UnimplementedError("Algorithm isn't implemented: $algorithm");
  }
}
