/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'package:intl/intl.dart' show DateFormat;

import '../config/defaults.dart';
import '../validation/converter/c_date.dart';
import 'numbers.dart';
import 'utils.dart';

class DateUtils {
  static const String ENGLISH = "en_US";

  const DateUtils._();

  static DateFormat getFormatter(String pattern) {
    return DateFormat(Utils.trim(pattern), ENGLISH);
  }

  static String formatDefault(DateTime date) {
    return format_(Defaults.DATE_TIME_CONVERSION_FORMAT, date);
  }

  static String format(int milliseconds, String pattern) {
    return format_(pattern, millisToDate(milliseconds));
  }

  static String format_(String pattern, DateTime date) {
    return date == null ? null : getFormatter(pattern).format(date);
  }

  static int epoch() {
    return dateToMillis_(now());
  }

  static DateTime now() {
    return getCalendar();
  }

  static DateTime today() {
    return ignoreTime();
  }

  static DateTime getTodayCalendar() {
    return now();
  }

  static DateTime toDate(int hour, int minute) {
    return getCalendar_(hour, minute);
  }

  static DateTime toDate_(int day, int month, int year) {
    return getCalendar__(day, month, year);
  }

  static DateTime toDate__(int day, int month, int year, int hour, int minute) {
    return getCalendar___(day, month, year, hour, minute);
  }

  static DateTime ignoreTime() {
    return ignoreHourAndMinute(0, 0, 0, 0);
  }

  static DateTime endDay() {
    return ignoreHourAndMinute(23, 59, 59, 999);
  }

  static int dateToMillis(String sDate) {
    return dateToMillis_(
        CDate.parseNoErr(sDate, Defaults.DATE_TIME_CONVERSION_FORMAT));
  }

  static int dateToMillis_(DateTime date) {
    return date == null ? Defaults.MISSING_L : date.millisecondsSinceEpoch;
  }

  static DateTime millisToDate(int milliseconds) {
    return milliseconds == null
        ? null
        : DateTime.fromMillisecondsSinceEpoch(milliseconds);
  }

  static DateTime alterToday(bool inPast, int years) {
    return alterToday_(inPast, years, null, null);
  }

  static DateTime alterToday_(bool inPast, int years, int months, int days) {
    return alter(now(), inPast, years, months, days); // initially today
  }

  static DateTime alter(
      DateTime calendar, bool inPast, int years, int months, int days) {
    int year = calendar.year;

    if (Numbers.notZero(years)) {
      year = getYear(calendar);
      year = inPast ? year - years : year + years;
    }

    int month = calendar.month;

    if (Numbers.notZero(months)) {
      month = getMonth(calendar);
      month = inPast ? month - months : month + months;
    }

    int day = calendar.day;

    if (Numbers.notZero(days)) {
      day = getDate(calendar);
      day = inPast ? day - days : day + days;
    }

    return DateTime(year, month, day, calendar.hour, calendar.minute,
        calendar.second, calendar.millisecond, calendar.microsecond);
  }

  static DateTime ignoreHourAndMinute(
      int hour, int min, int second, int millisecond) {
    DateTime today = getCalendar_(hour, min);
    return DateTime(today.year, today.month, today.day, today.hour,
        today.minute, second, millisecond, today.microsecond);
  }

  static DateTime getCalendar() {
    return DateTime.now();
  }

  static DateTime getCalendar_(int hour, int minute) {
    return setTime(hour, minute, getCalendar());
  }

  static DateTime getCalendar__(int day, int month, int year) {
    DateTime calendar = ignoreHourAndMinute(0, 0, 0, 0);
    return DateTime(year, month, day, calendar.hour, calendar.minute,
        calendar.second, calendar.millisecond, calendar.microsecond);
  }

  static DateTime getCalendar___(
      int day, int month, int year, int hour, int minute) {
    return setTime(hour, minute, getCalendar__(day, month, year));
  }

  static DateTime setTime(int hour, int minute, DateTime calendar) {
    return DateTime(calendar.year, calendar.month, calendar.day, hour, minute,
        calendar.second, calendar.millisecond, calendar.microsecond);
  }

  static int getDate(DateTime calendar) {
    return calendar == null ? Defaults.MISSING : calendar.day;
  }

  static int getMonth(DateTime calendar) {
    return calendar == null ? Defaults.MISSING : calendar.month;
  }

  static int getYear(DateTime calendar) {
    return calendar == null ? Defaults.MISSING : calendar.year;
  }

  static bool isStale(int time, int threshold) {
    return epoch() - time >= threshold;
  }
}
