/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../form/structure/ordered_map.dart';
import '../../form/structure/string_object_map.dart';
import '../../form/structure/t_map.dart';
import '../config/defaults.dart';
import '../enums/code.dart';
import '../enums/constraint_type.dart';
import '../enums/dir.dart';
import '../enums/field_type.dart';
import '../enums/mode.dart';
import '../enums/platform.dart';
import '../enums/scope.dart';
import '../enums/sort_method.dart';
import '../enums/view_type.dart';
import '../error/illegal_argument_exception.dart';
import '../generator/reader/form_reader.dart';
import '../structure/i18n/i18n.dart';
import '../structure/i18n/locale.dart';
import '../structure/i18n/option.dart';
import '../structure/model/constraint.dart';
import '../structure/model/field.dart';
import '../structure/model/global_constraint.dart';
import '../structure/model/model.dart';
import '../structure/model/script.dart';
import '../structure/ui/ui.dart';
import '../ui/discrete/action/data/value_resolver.dart';
import '../ui/discrete/common/on_value_change_listener.dart';
import '../ui/discrete/component/base/input_component.dart';
import '../ui/discrete/field_state.dart';
import '../ui/discrete/form_controller.dart';
import '../ui/discrete/renderer/features.dart';
import '../ui/discrete/renderer/params/renderer_params.dart';
import '../ui/discrete/renderer/ui_renderer.dart';
import '../ui/discrete/view_state.dart';
import '../validation/constraint/c_range.dart' as CRange;
import '../validation/converter/c_number.dart';
import '../validation/converter/c_string.dart';
import 'script/script_evaluator.dart';
import 'structure/string_set.dart';
import 'structure/unique_string_map.dart';
import 'utils.dart';

class FormUtils {
  static const String ESC_EQ = "@eq";
  static const String ESC_CM = "@cm";

  const FormUtils._();

  static UIRenderer getRenderer(FormController controller) {
    return controller == null ? null : controller.getRenderer();
  }

  static Field getField(ViewState view) {
    return view == null ? null : view.getField();
  }

  static Field getField_(String fieldId, FormController controller) {
    return getField__(fieldId, getRenderer(controller));
  }

  static Field getField__(String fieldId, UIRenderer renderer) {
    return getModel(renderer).getField(fieldId);
  }

  static TMap<GlobalConstraint> getConstraints(UIRenderer renderer) {
    Model model = getModel(renderer);
    return model == null ? TMap<GlobalConstraint>() : model.getConstraintMap();
  }

  static Constraint getConstraint(
      String fieldId, ConstraintType type, FormController controller) {
    UIRenderer renderer = getRenderer(controller);
    return getConstraint_(getField__(fieldId, renderer), type, renderer);
  }

  static Constraint getConstraint_(
      Field field, ConstraintType type, UIRenderer renderer) {
    return field == null ? null : field.get(type, renderer);
  }

  static StringObjectMap getConstraintArgs(
      String fieldId, ConstraintType type, FormController controller) {
    Constraint constraint = getConstraint(fieldId, type, controller);
    return constraint == null
        ? StringObjectMap()
        : constraint.getArgs_(getRenderer(controller));
  }

  static int getMaxAllowed(Field field, UIRenderer renderer) {
    Constraint cRange = getConstraint_(field, CRange.CRange.TYPE, renderer);
    return cRange == null
        ? null
        : CNumber.convertNoErr(cRange.getArg___(CRange.Args.MAX, renderer));
  }

  static RendererParams getParams(UIRenderer renderer) {
    return renderer == null ? null : renderer.getParams();
  }

  static Model getModel(UIRenderer renderer) {
    RendererParams params = getParams(renderer);
    return params == null ? null : params.getModel();
  }

  static UI getUI(FormController controller) {
    return getUI_(getRenderer(controller));
  }

  static UI getUI_(UIRenderer renderer) {
    RendererParams params = getParams(renderer);
    return params == null ? null : params.getBaseUI();
  }

  static I18n getI18n(UIRenderer renderer) {
    RendererParams params = getParams(renderer);
    return params == null ? null : params.getI18n();
  }

  static Locale getLocale(FormController controller) {
    return getLocale_(getRenderer(controller));
  }

  static Locale getLocale_(UIRenderer renderer) {
    RendererParams params = getParams(renderer);
    return params == null ? null : params.getCurrentLocale();
  }

  static Dir getDir(UIRenderer renderer) {
    Locale locale = getLocale_(renderer);
    return locale == null ? Defaults.DIR : locale.getDir();
  }

  static bool isLtr(UIRenderer renderer) {
    return isLtr_(getDir(renderer));
  }

  static bool isLtr_(Dir dir) {
    return Dir.LTR == dir;
  }

  static Mode getMode(UIRenderer renderer) {
    RendererParams params = getParams(renderer);
    return params == null ? null : params.getCurrentMode();
  }

  static bool isShowRaw(FormController controller) {
    RendererParams params = getParams(getRenderer(controller));
    return params != null && params.isShowRaw();
  }

  static String getLabel(String viewId, FormController controller) {
    return getLabel_(getView(viewId, controller), controller);
  }

  static String getLabel_(ViewState view, FormController controller) {
    return view == null ? null : eval(view.getLabel(), controller);
  }

  static String getLabel__(ViewState view, UIRenderer renderer) {
    return view == null ? null : eval_(view.getLabel(), renderer);
  }

  static String getFontSize(FormController controller) {
    UIRenderer renderer = getRenderer(controller);
    UI ui = getUI_(renderer);
    return ui == null
        ? null
        : CString.convert(resolve_(ui.getBaseFontSize(), renderer));
  }

  static Object getFieldValue(String fieldId, UIRenderer renderer) {
    Field field = getField__(fieldId, renderer);
    return field == null ? null : field.getConvertedValue(renderer);
  }

  static Object getValue(String viewId, FormController controller) {
    return getValue_(viewId, getRenderer(controller));
  }

  static Object getValue_(String viewId, UIRenderer renderer) {
    ViewState view = getView_(viewId, renderer);
    return view == null ? null : view.getValue();
  }

  static OrderedMap<ViewState> getAllViews(FormController controller) {
    return getAllViews_(getRenderer(controller));
  }

  static OrderedMap<ViewState> getAllViews_(UIRenderer renderer) {
    return renderer == null ? null : renderer.getAllViews();
  }

  static ViewState getView(String viewId, FormController controller) {
    return getView_(viewId, getRenderer(controller));
  }

  static ViewState getView_(String viewId, UIRenderer renderer) {
    OrderedMap<ViewState> allViews = getAllViews_(renderer);
    return allViews == null ? null : allViews[viewId];
  }

  static ViewState getView__(
      String viewId, bool checkNull, bool shouldInput, UIRenderer renderer) {
    ViewState view = getView_(viewId, renderer);
    final bool vMissing = view == null;

    if (checkNull && vMissing) {
      String msg = renderer.isRendering()
          ? "Renderer is not done constructing the view: "
          : "Missing view: ";
      throw IllegalArgumentException(msg + viewId);
    }

    if (shouldInput && (vMissing || !view.canInput())) {
      throw IllegalArgumentException("View can't input: $viewId");
    }

    return view;
  }

  static void closePopup(FormController controller) {
    UIRenderer renderer = getRenderer(controller);

    if (renderer != null) {
      renderer.closePopup();
    }
  }

  static Object getExtra(String key, FormController controller) {
    StringObjectMap extras = controller.getExtras();
    return extras == null ? null : extras[key];
  }

  static bool hasScript(String text) {
    return Code.SCRIPT.starts(text);
  }

  static Script getScript(String scriptId, UIRenderer renderer) {
    return Model.getScript(scriptId, getScripts(renderer));
  }

  static TMap<Script> getScripts(UIRenderer renderer) {
    Model model = getModel(renderer);
    return model == null ? TMap<Script>() : model.getScriptMap();
  }

  static String eval(String text, FormController controller) {
    return eval_(text, getRenderer(controller));
  }

  static String eval_(String text, UIRenderer renderer) {
    if (hasScript(text)) {
      text = evalScript(text, renderer);
    }

    return Utils.toBlank(renderer.get(text));
  }

  static String evalScript(String text, UIRenderer renderer) {
    return renderer == null
        ? ""
        : CString.convert(
            ScriptEvaluator.eval(Utils.skipFirst(text), null, renderer));
  }

  static Object resolve(String value, FormController controller) {
    return Features.getResolver_(controller).resolve(value);
  }

  static Object resolve_(String value, UIRenderer renderer) {
    return Features.getResolver(renderer).resolve(value);
  }

  static String resolveAsString(String value, FormController controller) {
    return CString.convert(resolve(value, controller));
  }

  static String resolveAsString_(String value, UIRenderer renderer) {
    return CString.convert(resolve_(value, renderer));
  }

  static FieldType getFieldType(ViewType type, FieldState field) {
    return type.canInput()
        ? Utils.toDefault(
            field == null ? null : field.getType(), type.fieldType())
        : null;
  }

  static void checkTypeCompatibility(
      String id, ViewType type, FieldState field) {
    FieldType fieldType = getFieldType(type, field);

    if (fieldType != null && !fieldType.supports(type)) {
      throw IllegalArgumentException(
          "$type [$id] can't produce $fieldType value!");
    }
  }

  static Platform getPlatform(UIRenderer renderer) {
    ValueResolver resolver = Features.getResolver(renderer);
    return resolver == null ? null : resolver.getPlatform();
  }

  static bool isMobile(UIRenderer renderer) {
    return isMobile_(Features.getResolver(renderer));
  }

  static bool isMobile_(ValueResolver resolver) {
    final Platform platform = resolver.getPlatform();
    return Platform.ANDROID == platform || Platform.IOS == platform;
  }

  static void fireOnValueChange(OnValueChangeListener onValueChange) {
    if (onValueChange != null) {
      onValueChange.perform();
    }
  }

  static void fireOnValueChange_(
      bool ensureChange, InputComponent inputComponent) {
    OnValueChangeListener onValueChange = inputComponent.onValueChange();

    if (onValueChange != null) {
      // don't call other signature of this method
      onValueChange.perform_(inputComponent.getVal(), ensureChange);
    }
  }

  static void fireOnValueChange__(
      Object val, bool ensureChange, OnValueChangeListener onValueChange) {
    if (onValueChange != null) {
      onValueChange.perform_(val, ensureChange);
    }
  }

  static StringSet expand(StringSet ids, String srcId, bool deduce,
      bool forInputViews, UIRenderer renderer) {
    if (Utils.isBlank(ids)) {
      return ids;
    }

    StringSet expandedIds = StringSet();

    for (String id in ids) {
      if (Scope.ALL.check(id)) {
        if (forInputViews) {
          // in this case, '@all' is same as '@allinputViews'
          Utils.addAll(expandedIds, [renderer.getInputViews()]);
        } else {
          for (ViewState view in renderer.getAllViewStates()) {
            expandedIds.add(view.getId());
          }
        }
      } else if (Scope.ALLINPUTVIEWS.check(id)) {
        Utils.addAll(expandedIds, [renderer.getInputViews()]);
      } else if (hasScript(id)) {
        final Object result =
            ScriptEvaluator.eval(Utils.skipFirst(id), null, renderer);

        if (result is String) {
          Utils.addAll(expandedIds, [
            expand(Utils.splitByComma(result), srcId, deduce, forInputViews,
                renderer)
          ]);
        }
      } else if (deduce && Code.EXCLUDE.starts(id)) {
        expandedIds.remove(Utils.skipFirst(id));
      } else {
        if (Scope.THIS.check(id)) {
          expandedIds.add(srcId);
        } else if (Scope.THIS.check(Utils.skipFirst(id))) {
          expandedIds.add(id[0] + srcId);
        } else {
          StringSet viewIds =
              getUI_(renderer).getGroupViews(id); // try it as group id

          if (Utils.isBlank(viewIds)) {
            // wasn't a view group id, just a view id though
            expandedIds.add(id);
          } else {
            Utils.addAll(expandedIds, [viewIds]);
          }
        }
      }
    }

    return expandedIds;
  }

  static List<String> getArrayKeys(String arrayId, UIRenderer renderer) {
    I18n i18n = getI18n(renderer);
    return i18n == null ? null : i18n.getArrayKeys(arrayId);
  }

  static UniqueStringMap pairsToMap(String inlinePairs) {
    UniqueStringMap map;
    List<String> pairs =
        Utils.splitByCommaAsList(inlinePairs); // value should be case-sensitive

    if (!Utils.isBlank(pairs)) {
      map = UniqueStringMap();

      for (String pair in pairs) {
        List<String> keyAndVal = Utils.split(pair, "=");

        if (!Utils.isBlank(keyAndVal)) {
          map.put(
              keyAndVal[0], keyAndVal.length > 1 ? escape(keyAndVal[1]) : null);
        }
      }
    }

    return map;
  }

  static String escape(String text) {
    final String trimmed = Utils.trim(text);
    return trimmed == null
        ? null
        : trimmed.replaceAll(ESC_EQ, "=").replaceAll(ESC_CM, ",");
  }

  static List<Option> sortOptions(List<Option> options, SortMethod sortBy) {
    if (sortBy != null && Utils.getSize(options) > 1) {
      if (SortMethod.ID_ASC == sortBy) {
        options.sort((opt1, opt2) =>
            Utils.compareToIgnoreCase(opt1.getId(), opt2.getId()));
      } else if (SortMethod.ID_DESC == sortBy) {
        options.sort((opt1, opt2) =>
            Utils.compareToIgnoreCase(opt2.getId(), opt1.getId()));
      } else if (SortMethod.MSG_ASC == sortBy) {
        options.sort((opt1, opt2) =>
            Utils.compareToIgnoreCase(opt1.getMsgId(), opt2.getMsgId()));
      } else if (SortMethod.MSG_DESC == sortBy) {
        options.sort((opt1, opt2) =>
            Utils.compareToIgnoreCase(opt2.getMsgId(), opt1.getMsgId()));
      } else if (SortMethod.SHUFFLE == sortBy) {
        options.shuffle();
      } else if (SortMethod.REVERSE == sortBy) {
        options = List.from(options.reversed);
      }
    }

    return options;
  }

  static String translate(String key, String arrayId, String lang,
      String i18nRaw, UIRenderer renderer) {
    return translate_(
        key, arrayId, lang, FormReader().toI18n(i18nRaw), renderer);
  }

  /// Translates the specific key of a message-array in the i18n.
  static String translate_(
      String key, String arrayId, String lang, I18n i18n, UIRenderer renderer) {
    Option option = i18n.getArrayOption(key, arrayId);
    String msgId = option == null ? key : option.getMsgId();
    return Utils.toDefault(i18n.getText(msgId, lang, renderer), msgId);
  }
}
