/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:io' show File, FileSystemEntity, Platform;
import 'dart:typed_data' show Uint8List;

import '../config/defaults.dart';
import '../error/illegal_argument_exception.dart';
import 'log/logger.dart';
import 'utils.dart';

class IOUtils {
  static const int EOF = Defaults.MISSING;
  static const int BYTES_IN_KB = 1024;
  static const int BUFFER_SIZE = BYTES_IN_KB * 64; // 65536 KBs (recommended)

  const IOUtils._();

  static Uint8List readFileNoErr(String path) {
    return readFileNoErr_(getFile(path));
  }

  static Uint8List readFileNoErr_(File file) {
    try {
      return readFile_(file);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return null;
    }
  }

  static Uint8List readFile(String path) {
    return readFile_(getFile(path));
  }

  static Uint8List readFile_(File file) {
    return fileExists(file) && FileSystemEntity.isFileSync(file.path)
        ? file.readAsBytesSync()
        : null;
  }

  static String readFileAsString(String path) {
    return Utils.toUtf8NoErr(readFile(path));
  }

  static String writeFileNoErr(List<int> content, String path) {
    try {
      return writeFile(content, path);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return null;
    }
  }

  static String writeFile(List<int> content, String path) {
    if (Utils.isBlank(content) || Utils.isBlank(path)) {
      throw IllegalArgumentException("Content or path can't be blank!");
    }

    return writeFile_(content, File(path)).path;
  }

  static File writeFile_(List<int> content, File file) {
    if (file != null) {
      file.writeAsBytesSync(content, flush: true);
    }

    return file;
  }

  static Uint8List readResource(String name) {
    return null; // TODO Implement
  }

  static bool rename(String oldName, String newName) {
    return rename_(getFileIfExists(oldName), newName);
  }

  static bool rename_(File file, String newName) {
    if (file != null) {
      final String path = file.path;
      final int lastSeparator = path.lastIndexOf(Platform.pathSeparator);
      file.renameSync(path.substring(0, lastSeparator + 1) + newName);
      return true;
    }

    return false;
  }

  static void delete(List<String> paths) {
    if (!Utils.isBlank(paths)) {
      for (String path in paths) {
        delete_([getFileIfExists(path)]);
      }
    }
  }

  static void delete_(List<File> files) {
    if (!Utils.isBlank(files)) {
      for (File file in files) {
        if (file != null) {
          file.deleteSync(recursive: true);
        }
      }
    }
  }

  static String getFileName(File file) {
    return file == null ? null : file.uri.pathSegments.last;
  }

  static File getFile(String path) {
    return Utils.isBlank(path) ? null : File(path);
  }

  static File getFileIfExists(String path) {
    File file = getFile(path);
    return fileExists(file) ? file : null;
  }

  static bool fileExists_(String path) {
    return getFileIfExists(path) != null;
  }

  static bool fileExists(File file) {
    return file != null && file.existsSync();
  }

  static bool copy(String src, String destination) {
    return copy_(getFile(src), getFile(destination));
  }

  static bool copy_(File src, File destination) {
    if (!fileExists(src)) {
      Logger.error("Source file doesn't exist!");
      return false;
    }

    if (destination == null) {
      Logger.error("Destination file is null!");
      return false;
    }

    if (fileExists(destination)) {
      Logger.warn("Destination file already exists!");
      return false;
    }

    try {
      src.copySync(destination.path);
      return true;
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return false;
    }
  }

  static String stripExt(File file) {
    return stripExt_(getFileName(file));
  }

  static String stripExt_(String fileName) {
    final int index =
        Utils.isBlank(fileName) ? Defaults.MISSING : fileName.lastIndexOf('.');
    return index > 0 ? fileName.substring(0, index) : fileName;
  }

  static String getExt(File file) {
    return getExt_(getFileName(file));
  }

  static String getExt_(String fileName) {
    final int index =
        Utils.isBlank(fileName) ? Defaults.MISSING : fileName.lastIndexOf('.');
    return index > 0 ? fileName.substring(index + 1) : "";
  }

  static int getSize_(String path) {
    // in KBs
    return getSize(getFileIfExists(path));
  }

  static int getSize(File file) {
    // in KBs
    return file == null ? 0 : toKBs(file.lengthSync());
  }

  static int getSize__(List<int> content) {
    // in KBs
    return Utils.isBlank(content) ? 0 : toKBs(content.length);
  }

  static int toKBs(int number) {
    return (number / BYTES_IN_KB).round();
  }

  static int toBytes(int number) {
    return number * BYTES_IN_KB;
  }
}
