/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:io' show Platform;
import 'dart:typed_data' show Uint8List;

import 'package:flutter_js/javascript_runtime.dart' show JavascriptRuntime;

import '../../../form/bean/server_response.dart';
import '../../enums/code.dart';
import '../../enums/constraint_type.dart';
import '../../enums/field_type.dart';
import '../../enums/mode.dart';
import '../../enums/value_type.dart';
import '../../error/illegal_argument_exception.dart';
import '../../structure/i18n/option.dart';
import '../../structure/model/constraint.dart';
import '../../structure/model/field.dart';
import '../../structure/model/model.dart';
import '../../ui/discrete/action/data/data_comparator.dart';
import '../../ui/discrete/component/base/component.dart';
import '../../ui/discrete/component/base/input_component.dart';
import '../../ui/discrete/component/button/button.dart';
import '../../ui/discrete/form_controller.dart';
import '../../ui/discrete/renderer/features.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../ui/discrete/view_state.dart';
import '../../validation/constraint/c_mandatory.dart';
import '../../validation/converter.dart';
import '../../validation/converter/c_string.dart';
import '../../validation/validator.dart';
import '../enum.dart';
import '../form_utils.dart';
import '../io_utils.dart';
import '../log/logger.dart';
import '../utils.dart';
import 'condition_evaluator.dart';
import 'el_cypher.dart';
import 'script_evaluator.dart';

// NOTE: For consistency, don't overload methods as DART don't allow it
class ELFunctions extends ELCypher {
  static const List<String> VIEWS_FUNCS = [
    "run",
    "runArgs",
    "runC",
    "field",
    "view",
    "inView",
    "noVal",
    "val",
    "val_",
    "fVal",
    "setVal",
    "rendered",
    "hidden",
    "enabled",
    "disabled",
    "included",
    "excluded"
  ];

  final UIRenderer renderer;

  ELFunctions(FormController controller)
      : this.renderer = controller == null ? null : controller.getRenderer() {
    if (controller == null) {
      throw IllegalArgumentException("Controller can't be null");
    }
  }

  @override
  void register(JavascriptRuntime engine) {
    super.register(engine);

    // TODO We can't return Object neither can we pass Object as parameter

    reg('run', (args) => run(args[0]), engine);
    reg('runArgs', (args) => runArgs(args[0], args[1]), engine);
    reg('runC', (args) => runC(args[0]), engine);
    reg('field', (args) => field(args[0]), engine);
    reg('view', (args) => view(args[0]), engine);
    reg('inView', (args) => inView(args[0]), engine);
    reg('con', (args) => con(args[0], args[1]), engine);
    reg('conArg', (args) => conArg(args[0], args[1], args[2]), engine);
    reg('com', (args) => com(args[0]), engine);
    reg('inCom', (args) => inCom(args[0]), engine);
    reg('noVal', (args) => noVal(args[0]), engine);
    reg('val', (args) => val(args[0]), engine);
    reg('val_', (args) => val_(args[0]), engine);
    reg('fVal', (args) => fVal(args[0]), engine);
    reg('resolve', (args) {
      final Object value = resolve(args[0]);
      // FIXME This is just a workaround... flutter_js can't understand if value is Object or Enum
      return value is Enum ? value.toString() : value;
    }, engine);
    reg('resolveAsStr', (args) => resolveAsStr(args[0]), engine);
    reg('setVal', (args) => setVal(args[0], args[1]), engine);
    reg('clrVal', (args) => clrVal(args), engine);
    reg('perm', (args) => perm(args[0]), engine);
    reg('check', (args) => check(args[0]), engine);
    reg('check_', (args) => check_(args[0], args[1]), engine);
    reg('rmi', (args) async => await rmi(args[0]), engine);
    reg('rmi_', (args) async => await rmi_(args[0], args[1]), engine);
    reg('rmi__', (args) async => await rmi__(args[0], args[1]), engine);
    reg('rmiList', (args) async => await rmiList(args[0]), engine);
    reg('fetchList', (args) => fetchList(args[0], args[1]), engine);
    reg('rendered', (args) => rendered(args[0]), engine);
    reg('hidden', (args) => hidden(args[0]), engine);
    reg('render', (args) => render(args), engine, ret: false);
    reg('hide', (args) => hide(args), engine, ret: false);
    reg('enabled', (args) => enabled(args[0]), engine);
    reg('disabled', (args) => disabled(args[0]), engine);
    reg('enable', (args) => enable(args), engine, ret: false);
    reg('disable', (args) => disable(args), engine, ret: false);
    reg('click', (args) => click(args[0]), engine);
    reg('included', (args) => included(args[0]), engine);
    reg('excluded', (args) => excluded(args[0]), engine);
    reg('msg', (args) => msg(args[0], args.sublist(1)), engine);
    reg('log', (args) => log(args[0]), engine, ret: false);
    reg('info', (args) => info(args[0]), engine, ret: false);
    reg('err', (args) => err(args[0]), engine, ret: false);
    reg('popup', (args) => popup(args[0]), engine, ret: false);
    reg('reset', (args) async => await reset(), engine, ret: false);
    reg('reset_', (args) async => await reset_(args[0]), engine, ret: false);
    reg('veql', (args) => veql(args[0], args[1]), engine);
    reg('eql', (args) => eql(args[0], args[1]), engine);
    reg('neql', (args) => neql(args[0], args[1]), engine);
    reg('any', (args) => any(args[0], args.sublist(1)), engine);
    reg('mode', (args) => mode(args), engine);
    reg('platform', (args) => platform(args), engine);
    reg('ltr', (args) => ltr(), engine);
    reg('rtl', (args) => rtl(), engine);
    reg('blank', (args) => blank(args[0]), engine);
    reg('convert', (args) => convert(args[0], args[1]), engine);
    reg('changeLocale', (args) => changeLocale(args[0]), engine);
    reg('sysEnv', (args) => sysEnv(args[0]), engine);
    reg('sysProp', (args) => sysProp(args[0]), engine);
    reg('readFile', (args) => readFile(args[0]), engine);
    reg('saveForm', (args) async => await saveForm(args[0], args[1]), engine);
    reg('isRendering', (args) => isRendering(), engine);
  }

  Object run(String scriptId) => runArgs(scriptId, null);

  // FIXME How to pass Map<String, Object> from JS?
  Object runArgs(String scriptId, Map<String, Object> args) =>
      ScriptEvaluator.runScript(scriptId, null, args, renderer);

  bool runC(String scriptId) => ConditionEvaluator.eval(scriptId, renderer);

  Field field(String fieldId) => FormUtils.getField__(fieldId, renderer);

  ViewState view(String viewId) =>
      FormUtils.getView__(viewId, true, false, renderer);

  ViewState inView(String viewId) =>
      FormUtils.getView__(viewId, true, true, renderer);

  Constraint con(String fieldId, String type) => FormUtils.getConstraint(
      fieldId, ConstraintType.toType(type, false), getController());

  String conArg(String fieldId, String id, String arg) {
    Field field = this.field(fieldId);
    Constraint constraint = field == null ? null : field.getConstraints()[id];

    if (constraint != null && constraint.isGlobal()) {
      constraint =
          Model.getConstraint(constraint, FormUtils.getConstraints(renderer));
    }

    return constraint == null
        ? ""
        : CString.convert(constraint.getArg___(arg, renderer));
  }

  Component com(String viewId) => view(viewId).getComponent();

  InputComponent inCom(String viewId) => inView(viewId).getInputComponent();

  bool noVal(String viewId) => !CMandatory.check(val(viewId));

  Object val(String viewId) => val_(inView(viewId));

  Object val_(ViewState view) => view == null ? null : view.getConvertedValue();

  Object fVal(String fieldId) => FormUtils.getFieldValue(fieldId, renderer);

  Object resolve(String value) => FormUtils.resolve_(value, renderer);

  String resolveAsStr(String value) =>
      FormUtils.resolveAsString_(value, renderer);

  bool setVal(String viewId, Object value) => inView(viewId).setValue(value);

  bool clrVal(List<dynamic> viewIds) {
    bool success = true;

    for (String viewId in viewIds) {
      try {
        success = setVal(viewId, null) && success;
      } on IllegalArgumentException catch (e, stackTrace) {
        success = false;
        Logger.error(e, stackTrace: stackTrace);
      }
    }

    return success;
  }

  bool perm(String permissionId) =>
      renderer.getParams().getPermissions().check_(permissionId);

  bool check(String viewId) => Features.getValidator(renderer).run([viewId]);

  // TODO viewId can be any value?
  bool check_(String constraintId, String viewId) {
    ViewState view = inView(viewId);
    Constraint constraint =
        Model.getConstraint_(constraintId, FormUtils.getConstraints(renderer));
    Validator validator = Features.getValidator(renderer);
    return constraint.isCustom()
        ? validator.runCustomConstraint(constraint, view, false)
        : validator.runConstraint(constraint, view, val_(view), false);
  }

  Future<bool> rmi(String methodId) async => await rmi_(methodId, null);

  Future<bool> rmi_(String methodId, String viewId) async =>
      await rmi__(methodId, view(viewId));

  // FIXME ViewState as param wont' work
  Future<bool> rmi__(String methodId, ViewState vSrc) async {
    final Object result =
        await Features.getRMIAction(renderer).invoke(methodId, vSrc);
    return result is ServerResponse && ServerResponse.isSuccessful(result);
  }

  Future<List<Option>> rmiList(String methodId) async =>
      await Features.getRMIAction(renderer).invokeFetchList(methodId);

  bool fetchList(bool fetch, String viewIds) =>
      Features.getDataActions(renderer)
          .fetchOrClearLists(fetch, Utils.splitByComma(viewIds));

  bool rendered(String viewId) {
    try {
      return view(viewId).isRendered();
    } on IllegalArgumentException {
      return false;
    }
  }

  bool hidden(String viewId) => !rendered(viewId);

  void render(List<dynamic> viewIds) {
    for (String viewId in viewIds) {
      view(viewId).setRendered(true);
    }
  }

  void hide(List<dynamic> viewIds) {
    for (String viewId in viewIds) {
      view(viewId).setRendered(false);
    }
  }

  bool enabled(String viewId) {
    try {
      return view(viewId).isEnabled();
    } on IllegalArgumentException {
      return false;
    }
  }

  bool disabled(String viewId) => !enabled(viewId);

  void enable(List<dynamic> viewIds) {
    for (String viewId in viewIds) {
      view(viewId).setEnabled(true);
    }
  }

  void disable(List<dynamic> viewIds) {
    for (String viewId in viewIds) {
      view(viewId).setEnabled(false);
    }
  }

  bool click(String viewId) => (com(viewId) as Button).click();

  bool included(String viewId) => inView(viewId).isIncluded(renderer);

  bool excluded(String viewId) => !included(viewId);

  String msg(String msgId, List<Object> args) => renderer.get(msgId, args);

  void log(Object object) {
    String msg =
        object is List ? Utils.toString___(object) : Utils.toString_O(object);

    if (!Utils.isBlank(msg)) {
      Logger.debug(msg);
    }
  }

  void info(Object object) {
    String msg =
        object is List ? Utils.toString___(object) : Utils.toString_O(object);

    if (!Utils.isBlank(msg)) {
      Features.getAlert_(renderer).info(msg);
    }
  }

  void err(Object object) {
    String msg =
        object is List ? Utils.toString___(object) : Utils.toString_O(object);

    if (!Utils.isBlank(msg)) {
      Features.getAlert_(renderer).error(msg);
    }
  }

  void popup(String popupId) =>
      Features.getFormActions(renderer).showPopup_(popupId);

  Future<void> reset() async =>
      await Features.getFormActions(renderer).doReset(null, null);

  Future<void> reset_(String viewId) async =>
      await Features.getFormActions(renderer).doReset(null, view(viewId));

  bool veql(String view1, String view2) => eql(val(view1), val(view2));

  bool eql(Object obj1, Object obj2) => !neql(obj1, obj2);

  bool neql(Object obj1, Object obj2) => DataComparator.isChanged(obj1, obj2);

  bool any(Object value, List<Object> elements) {
    if (Utils.isBlank(elements)) {
      IllegalArgumentException("Elements can't be blank for any() function!");
    }

    for (Object element in elements) {
      if (element is String) {
        element = resolve(element);
      }

      if (eql(value, element)) {
        return true;
      }
    }

    return false;
  }

  bool mode(List<dynamic> modes) {
    final Mode mode =
        resolve(Code.RESOLVE.prefix(ValueType.MODE.toString())) as Mode;
    return Utils.match(mode.toString(), modes.cast<String>());
  }

  bool platform(List<dynamic> platforms) => Utils.match(
      Features.getResolver(renderer).getPlatform().toString(),
      platforms.cast<String>());

  bool ltr() => FormUtils.isLtr(renderer);

  bool rtl() => !ltr();

  bool blank(Object value) => Utils.isBlank(value);

  Object convert(String fieldType, Object value) =>
      Converter.convertNoErr(FieldType.toType(fieldType, false), value);

  bool changeLocale(String language) => renderer.changeLocale(language);

  String sysEnv(String name) {
    try {
      return Platform.environment[name];
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return null;
    }
  }

  String sysProp(String name) => null;

  Uint8List readFile(String path) => IOUtils.readFileNoErr(path);

  Future<bool> saveForm(String viewId, bool update) async =>
      await Features.getDataActions(renderer).doPersist(view(viewId), update);

  bool isRendering() => renderer.isRendering();

  FormController getController() => renderer.getController();
}
