/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:ffi' show Pointer;

import 'package:flutter_js/javascript_runtime.dart' show JavascriptRuntime;
import 'package:flutter_js/javascriptcore/jscore_runtime.dart';
import 'package:flutter_js/js_eval_result.dart';
import 'package:flutter_js/quickjs/quickjs_runtime2.dart' show QuickJsRuntime2;

import '../../config/config.dart';
import '../../enums/platform.dart';
import '../../error/illegal_argument_exception.dart';
import '../../structure/model/script.dart';
import '../../ui/discrete/form_controller.dart';
import '../../ui/discrete/renderer/features.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../ui/discrete/view_state.dart';
import '../form_utils.dart';
import '../log/logger.dart';
import '../stop_watch.dart';
import '../structure/string_set.dart';
import '../utils.dart';
import 'el_functions.dart';

/// @see "https://pub.dev/packages/flutter_js"
class ScriptEvaluator {
  static const String SRC = "src";
  static const String RENDERER = "r";
  static const String CONTROLLER = "c";

  const ScriptEvaluator._();

  static Object runScript(String scriptId, ViewState vSrc,
      Map<String, Object> args, UIRenderer renderer) {
    return runScript_(
        FormUtils.getScript(scriptId, renderer), vSrc, args, renderer);
  }

  static Object runScript_(Script script, ViewState vSrc,
      Map<String, Object> args, UIRenderer renderer) {
    return script.onClient_()
        ? runOnClient(script, vSrc, args, renderer)
        : eval__(script, vSrc, args, renderer);
  }

  static bool runOnClient(Script script, ViewState vSrc,
      Map<String, Object> args, UIRenderer renderer) {
    return renderer.getController().runScript(script.getText(), vSrc, args);
  }

  static Object eval(
      String scriptId, Map<String, Object> args, UIRenderer renderer) {
    return eval_(FormUtils.getScript(scriptId, renderer), args, renderer);
  }

  static Object eval_(
      Script script, Map<String, Object> args, UIRenderer renderer) {
    return eval__(script, null, args, renderer);
  }

  static Object eval__(Script script, ViewState vSrc, Map<String, Object> args,
      UIRenderer renderer) {
    if (script.cached_(vSrc, renderer)) {
      return script.getResult();
    }

    StopWatch stopWatch = StopWatch.start();
    final String text = script.getText();

    if (Config.DEBUG) {
      Logger.debug("Executing script: $script");
    }

    try {
      dynamic value = execute(
          text, getContext(vSrc, args, renderer), Features.getEngine(renderer));
      // cache the result if script is cacheable
      script.setResult(value, vSrc, renderer);

      if (Config.DEBUG) {
        if (Script.notInline(script)) {
          Logger.debug("Script: '${script.getId()}' => $value");
        } else {
          Logger.debug("Expression: '$text' => $value");
        }
      }

      return value;
    } catch (e, stackTrace) {
      Logger.error("Problematic Script: $script", stackTrace: stackTrace);
      rethrow;
    } finally {
      StopWatch.stop_(script, stopWatch);
    }
  }

  static dynamic execute(
      String text, Map<String, Object> context, JavascriptRuntime engine,
      {bool prepare = false}) {
    // 1. Prepare function arguments
    // FIXME We can't send objects to JS (no reflection)
    // final String args = toArgs(context);

    // 2. Prepare final script
    final String finalText = prepare ? prepareText(text) : text;
    // TODO Use 'args' for both cases?
    final String script =
        text.contains("return ") ? "(function() { $finalText })();" : finalText;

    // 3. Execute script
    JsEvalResult result = engine.evaluate(script);

    // 4. Get script value and return
    final dynamic rawResult = result.rawResult;
    return rawResult is Pointer ? engine.convertValue(result) : rawResult;
  }

  static String toArgs(Map<String, Object> context) {
    StringBuffer args = StringBuffer();

    if (context != null) {
      context.forEach((key, value) {
        args.write("var $key, ");
      });
    }

    return Utils.toBlank(Utils.skipLast(args.toString()));
  }

  static bool isScript(String text) => Utils.contains_(text, false, ["{", ";"]);

  // add implicitly available variables (should be used case-sensitively)
  static Map<String, Object> getContext(
      ViewState vSrc, Map<String, Object> args, UIRenderer renderer) {
    Map<String, Object> context = {
      SRC: vSrc, // source view, can be null
      RENDERER: renderer,
      CONTROLLER: renderer.getController()
    };

    if (!Utils.isBlank(args)) {
      for (String key in args.keys) {
        if (Utils.match(key, [SRC, RENDERER, CONTROLLER])) {
          throw IllegalArgumentException("Reserved variable used: $key");
        }

        context[key] = args[key];
      }
    }

    return context;
  }

  static JavascriptRuntime getEngine(UIRenderer renderer) {
    JavascriptRuntime.debugEnabled = Config.DEBUG;

    // TODO Simply use JavascriptCoreRuntime for consistency? But QuickJsRuntime is better for Android.
    JavascriptRuntime engine =
        Platform.ANDROID == FormUtils.getPlatform(renderer)
            ? QuickJsRuntime2()
            : JavascriptCoreRuntime();
    engine.evaluate(_commonJs);

    // platform dependent functions are added via controller
    FormController controller = renderer?.getController();

    if (controller != null) {
      ELFunctions functions = controller.getELFunctions();
      functions ??= ELFunctions(controller);
      functions.register(engine);
    }

    return engine;
  }

  static StringSet findViews(String text, ViewState vSrc, UIRenderer renderer) {
    // TODO Implement this just like we have it in ScriptEvaluator.java
    return StringSet();
  }

  // To have consistent script between JS and JEXL in CMS Java, we do some tricks here.
  static String prepareText(String text) {
    // TODO Not the best way to replace operators
    text = text.replaceAll(' and ', ' && '); // 'x and y' -> 'x && y'
    text = text.replaceAll(' or ', ' || '); // 'x or y' -> 'x || y'
    text = text.replaceAll('not ', '!'); // 'not cond' -> '!cond'
    text = text.replaceAll(' eq ', ' == '); // 'x eq y' -> 'x == y'
    text = text.replaceAll(' ne ', ' != '); // 'x ne y' -> 'x != y'

    text = text.replaceAll(' lt ', ' < '); // 'x lt y' -> 'x < y'
    text = text.replaceAll(' le ', ' <= '); // 'x le y' -> 'x <= y'
    text = text.replaceAll(' gt ', ' > '); // 'x gt y' -> 'x > y'
    text = text.replaceAll(' ge ', ' >= '); // 'x ge y' -> 'x >= y'

    // TODO Support 'Starts With' (=^) operator?
    // TODO Support 'Not Starts With' (!^) operator?
    // TODO Support 'Ends With' (=$) operator?
    // TODO Support 'Not Ends With' (!$) operator?

    text = text.replaceAll(' div ', ' / '); // 'x div y' -> 'x / y'
    text = text.replaceAll(' mod ', ' % '); // 'x mod y' -> 'x % y'

    // '"2" =~ ["1", "2"]' -> '"2"?.isInList("1", "2") ?? false' (left-side can be null)
    text = text.replaceAllMapped(RegExp('[ ]*=~[ ]*([^\\]]+)\\]'), (match) {
      final String group = match.group(0);
      return '?.$_prototype_isInList(${group.substring(group.indexOf("[") + 1, group.length - 1)}) ?? false';
    });

    // TODO Support 'Not-In' (!~) operator?

    return text;
  }
}

const String _prototype_isInList = 'isInList';

// TODO Object.defineProperty() doesn't work for JavascriptCoreRuntime()
// as may not be allowed on iOS / Safari (not sure)
const String _commonJs = """
  var argsToJson = function() {
    var json = '[';
    var lastIndex = arguments.length - 1;
  
    for (var i = 0; i <= lastIndex; i++) {
      var val = arguments[i];
      json += typeof val === 'string' ? '"' + val + '"' : val;
      json += i == lastIndex ? '' : ', ';
    }
  
    return json + ']';
  };

  if (!String.prototype.$_prototype_isInList) {
    Object.defineProperty(String.prototype, '$_prototype_isInList', {
      get: () => function(...args) {
        let value = this.valueOf();
        for (let i = 0, l = args.length; i < l; i += 1) {
          if (arguments[i] === value) return true;
        }
        return false;
      }
    });
  }
""";
