/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:convert' show utf8, LineSplitter;
import 'dart:io' show Platform;
import 'dart:typed_data' show Uint8List;

import '../config/defaults.dart';
import '../ui/discrete/form_controller.dart';
import '../ui/discrete/renderer/ui_renderer.dart';
import '../validation/converter/c_string.dart';
import 'form_utils.dart';
import 'log/logger.dart';
import 'structure/string_set.dart';
import 'uuid.dart';

class Utils {
  static const String UTF_8 = "UTF-8";
  static const String SQ = "'";
  static const String DQ = '"';

  static final RegExp REGEX_NUMERIC = RegExp(r'^-?[0-9]+$');
  static final RegExp REGEX_ALPHA = RegExp(r'^[a-zA-Z]+$');
  static final RegExp REGEX_IPV4MAYBE =
      RegExp(r'^(\d?\d?\d)\.(\d?\d?\d)\.(\d?\d?\d)\.(\d?\d?\d)$');
  static final RegExp REGEX_IPV6 =
      RegExp(r'^::|^::1|^([a-fA-F0-9]{1,4}::?){1,7}([a-fA-F0-9]{1,4})$');

  const Utils._();

  static bool isBlank(dynamic obj) {
    if (obj == null) {
      return true;
    }

    if (obj is String) {
      obj = obj.trim();
    }

    try {
      // 'String, Iterable, Map' all have isEmpty function
      return obj.isEmpty;
    } catch (e) {
      return false;
    }
  }

  static bool isBlankDeep(Iterable array) {
    if (!isBlank(array)) {
      for (Object item in array) {
        if (!isBlank(item)) {
          return false;
        }
      }
    }

    return true;
  }

  static bool isUnavailable(String str) {
    return isBlank(str) || equalsIgnoreCase(Defaults.NA, str);
  }

  static bool anyNull([Object objects]) {
    if (!isBlank(objects)) {
      for (Object object in objects) {
        if (object == null) {
          return true;
        }
      }
    }

    return false;
  }

  static bool allNull([Object objects]) {
    if (!isBlank(objects)) {
      for (Object object in objects) {
        if (object != null) {
          return false;
        }
      }
    }

    return true;
  }

  static bool noneNull([Object objects]) {
    if (!isBlank(objects)) {
      for (Object object in objects) {
        if (object == null) {
          return false;
        }
      }
    }

    return true;
  }

  static bool anyBlank(List<String> strings) {
    if (!isBlank(strings)) {
      for (String string in strings) {
        if (isBlank(string)) {
          return true;
        }
      }
    }

    return false;
  }

  static bool allBlank(List<String> strings) {
    if (!isBlank(strings)) {
      for (String string in strings) {
        if (!isBlank(string)) {
          return false;
        }
      }
    }

    return true;
  }

  static int count(String str, String searchStr) {
    return isBlank(str) ? 0 : str.length - str.replaceAll(searchStr, "").length;
  }

  static List<String> blankArray(int size) {
    return List.filled(size, "");
  }

  static List<String> toArray(Iterable collection) {
    return isBlank(collection) ? null : collection.toList();
  }

  static List<String> split(String str, String delimiters) {
    return isBlank(str) ? null : str.trim().split(RegExp("[$delimiters]"));
  }

  // Can't use generic type here as it warns all the usages of it
  static void addAll(dynamic source, [Iterable collections]) {
    if (source != null && collections != null) {
      for (Iterable collection in collections) {
        if (!isBlank(collection)) {
          try {
            source.addAll(collection);
          } catch (e) {
            // nothing to do here
          }
        }
      }
    }
  }

  static void clear(List<dynamic> lists) {
    if (lists != null) {
      for (dynamic list in lists) {
        if (list != null) {
          try {
            list.clear();
          } catch (e) {
            // nothing to do here
          }
        }
      }
    }
  }

  static void removeAll(dynamic source, [List collections]) {
    if (noneNull([source, collections])) {
      for (List collection in collections) {
        if (collection != null) {
          for (Object item in collection) {
            if (item != null) {
              source.remove(item);
            }
          }
        }
      }
    }
  }

  static int getSize(dynamic collection) {
    return isBlank(collection) ? 0 : collection.length;
  }

  static List<String> splitLines(String str) {
    return isBlank(str) ? null : LineSplitter().convert(str.trim());
  }

  static List<String> splitByCommaAsList(String str) {
    return split(str, ",");
  }

  static StringSet splitByComma(String str) {
    List<String> split = splitByCommaAsList(str);
    return split == null ? null : StringSet.list(split);
  }

  static List<String> splitByComma_(String str, bool toLower) {
    return splitByDelimiters(str, ",", toLower);
  }

  static List<String> splitByDelimiters(
      String str, String delimiters, bool toLower) {
    if (isBlank(str)) {
      return null;
    }

    List<String> values = split(str, delimiters);
    int length = values.length;
    List<String> tValues = blankArray(length);

    for (int i = 0; i < length; i++) {
      tValues[i] = values[i].trim();

      if (toLower) {
        tValues[i] = tValues[i].toLowerCase();
      }
    }

    return tValues;
  }

  static String trimAndToLower(String str) {
    return isBlank(str) ? null : str.trim().toLowerCase();
  }

  static String trimAndToUpper(String str) {
    return isBlank(str) ? null : str.trim().toUpperCase();
  }

  static String trim(String str) {
    return isBlank(str) ? null : str.trim();
  }

  static String trimOrToBlank(String str, bool trimIt) {
    return trimIt ? toBlank(trim(str)) : toBlank(str);
  }

  static String escapeAndToBlank(String str) {
    return toBlank(escape(str));
  }

  static String escapeEnquote(String str) {
    return enquote(escape(str));
  }

  static String escape(String str) {
    try {
      return isBlank(str)
          ? str
          : Uri.encodeComponent(str).replaceAll("+", "%20");
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return str;
    }
  }

  static String unescape(String str) {
    try {
      return isBlank(str) ? str : Uri.decodeComponent(str);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      return str;
    }
  }

  static T get<T>(int index, List<T> list) {
    try {
      return list[index];
    } catch (e) {
      return null;
    }
  }

  static V get_<K, V>(Object key, Map<K, V> map) {
    try {
      return map[key];
    } catch (e) {
      return null;
    }
  }

  static String getAsString<T>(String key, Map<String, T> map) {
    return CString.convert(get_(key, map));
  }

  static bool got(Object key, Map map) {
    return map != null && map.containsKey(key);
  }

  static StringSet asSet(List<String> elements) {
    return isBlank(elements) ? null : StringSet.list(elements);
  }

  static List<T> nonNullList<T>(List<T> elements) {
    if (elements == null) {
      return null;
    }

    List<T> list = <T>[];

    for (T element in elements) {
      if (element != null) {
        list.add(element);
      }
    }

    return list.isEmpty ? null : list;
  }

  static String toString_O(Object obj) {
    if (obj == null) {
      return "";
    }

    return obj is List ? "byte[${obj.length}]" : obj.toString().trim();
  }

  static String castToString(Object obj) {
    return obj is String ? obj : null;
  }

  static String toString_(List<int> bytes) {
    String string;

    try {
      string = toString__(bytes, false);
    } catch (e) {
      // this can never happen since UTF_8 won't be used.
    }

    return string;
  }

  static String toString__(List<int> bytes, bool toUtf8) {
    return isBlank(bytes)
        ? null
        : toNull(toUtf8 ? utf8.decode(bytes) : String.fromCharCodes(bytes));
  }

  static String toString___(List array) {
    return toString____(array, ",");
  }

  static String toString____(List array, String delimiter) {
    if (isBlank(array)) {
      return null;
    }

    StringBuffer sb = StringBuffer();

    for (Object object in array) {
      if (object != null) {
        String item = object.toString();

        if (!isBlank(item)) {
          sb.write(item.trim());
          sb.write(delimiter);
        }
      }
    }

    return skipLast(sb.toString());
  }

  static String toCsv(Iterable set) {
    return isBlank(set) ? null : set.join(",");
  }

  static List<int> toBytes(String str) {
    List<int> bytes;

    try {
      bytes = toBytes_(str, false);
    } catch (e) {
      // this can never happen since UTF_8 won't be used.
    }

    return bytes;
  }

  static List<int> toBytes_(String str, bool _utf8) {
    final String value = isBlank(str) ? null : str.trim();
    return isBlank(value)
        ? null
        : _utf8
            ? utf8.encode(value)
            : value.codeUnits;
  }

  static Uint8List toBytes__(String str) {
    final List<int> bytes = toBytes(str);
    return bytes == null ? null : Uint8List.fromList(bytes);
  }

  static String toUtf8(List<int> bytes) {
    return toString__(bytes, true);
  }

  static String toUtf8NoErr(List<int> bytes) {
    try {
      return toUtf8(bytes);
    } catch (e) {
      return toString_(bytes);
    }
  }

  static String toUtf8_(String str) {
    try {
      return toUtf8(toBytes_(str, true));
    } catch (e) {
      return str;
    }
  }

  static String toProperCase(String str) {
    return toProperCase_(str, " '-/");
  }

  // @param actionables cause the character following to be capitalized
  static String toProperCase_(String str, String actionables) {
    StringBuffer formatted = StringBuffer();

    if (!isBlank(str)) {
      bool capNext = true;

      str.trim().runes.forEach((rune) {
        String c = String.fromCharCode(rune);
        c = capNext ? c.toUpperCase() : c.toLowerCase();
        formatted.write(c);
        capNext = actionables.indexOf(c) >= 0; // explicit cast not needed
      });
    }

    return toNull(formatted.toString());
  }

  static bool match(String match, List<String> candidates) {
    if (!isBlank(match) && !isBlank(candidates)) {
      final String tMatch = trim(match);

      if (tMatch != null) {
        // don't use isBlank() here
        for (String candidate in candidates) {
          if (equalsIgnoreCase(trimOrToBlank(candidate, true), tMatch)) {
            return true;
          }
        }
      }
    }

    return false;
  }

  static bool match_(int number, List<int> numbers) {
    if (!isBlank(numbers)) {
      for (int candidate in numbers) {
        if (number == candidate) {
          return true;
        }
      }
    }

    return false;
  }

  static bool matches(String text, String regex) {
    return RegExp(regex).hasMatch(text);
  }

  static bool contains(String text, List<String> candidates) {
    return contains_(text, true, candidates);
  }

  static bool contains_(
      String text, bool insensitive, List<String> candidates) {
    final bool noText = text == null; // don't use isBlank() here

    if (noText && isBlank(candidates)) {
      return true;
    }

    if (!noText) {
      text = insensitive ? text.toLowerCase() : text;

      for (String candidate in candidates) {
        // don't use isBlank() here
        if (candidate != null &&
            text.contains(insensitive ? candidate.toLowerCase() : candidate)) {
          return true;
        }
      }
    }

    return false;
  }

  static bool contains__(Exception e, List<String> candidates) {
    return contains(getMsg(e), candidates);
  }

  static bool hasLineBreak(String str) {
    return contains_(str, false, [Defaults.NL, Defaults.EOL]);
  }

  static bool hasSpace(String str) {
    return contains_(str, false, [" "]);
  }

  static String skipFirst(String str) {
    final String trimmed = trim(str);

    try {
      return trim(trimmed.substring(1));
    } catch (e) {
      return trimmed;
    }
  }

  static String skipFirstAndLast(String str) {
    final String trimmed = trim(str);

    try {
      return trimmed.substring(
          1, trimmed.length - 1); // don't trim the final string
    } catch (e) {
      return trimmed;
    }
  }

  static String skipLast(String str) {
    final String trimmed = trim(str);

    try {
      // don't trim the final string
      return trimmed.substring(0, trimmed.length - 1);
    } catch (e) {
      return trimmed;
    }
  }

  static bool startsWith(List<int> bytes, List<String> prefixes) {
    if (!isBlank(bytes)) {
      final String firstChar = String.fromCharCode(bytes.first).toLowerCase();

      for (String prefix in prefixes) {
        if (firstChar == prefix.toLowerCase()) {
          // compare it insensitively
          return true;
        }
      }
    }

    return false;
  }

  static bool startsWith_(String str, List<String> prefixes) {
    if (!isBlank(str) && !isBlank(prefixes)) {
      final String tStr = trimAndToLower(str);

      for (String prefix in prefixes) {
        // compare it insensitively
        if (!isBlank(prefix) && tStr.startsWith(prefix.toLowerCase())) {
          return true;
        }
      }
    }

    return false;
  }

  static bool endsWith(String str, List<String> suffixes) {
    if (!isBlank(str) && !isBlank(suffixes)) {
      final String tStr = trimAndToLower(str);

      for (String suffix in suffixes) {
        // compare it insensitively
        if (!isBlank(suffix) && tStr.endsWith(suffix.toLowerCase())) {
          return true;
        }
      }
    }

    return false;
  }

  static bool startsEndsWithSame(String str, String boundary) {
    return startsEndsWith(str, boundary, boundary);
  }

  static bool startsEndsWith(String str, String prefix, String suffix) {
    if (!isBlank(str)) {
      final String tStr = trimAndToLower(str);

      if (tStr.length > 1) {
        // compare it insensitively
        return tStr.startsWith(prefix.toLowerCase()) &&
            str.endsWith(suffix.toLowerCase());
      }
    }

    return false;
  }

  // Avoids NPE in the string evaluation
  static String toBlank(String str) {
    return toDefault(str, "");
  }

  static String toNull(String str) {
    return toDefault(str, null);
  }

  static bool toFalse(bool flag) {
    return toDefault(flag, false);
  }

  static bool toTrue(bool flag) {
    return toDefault(flag, true);
  }

  static T toDefault<T>(T value, T defaultValue) {
    if (value is String) {
      return isBlank(value) ? defaultValue : value;
    } else if (value is List) {
      return isBlank(value) ? defaultValue : value;
    }

    return value == null ? defaultValue : value;
  }

  static String prefix(String text, String prefix) {
    return startsWith_(text, [prefix]) ? text : prefix + text;
  }

  static String prefixSlash(String path) {
    return startsWith_(path, ["/", "\\"]) ? path : "/" + path;
  }

  static String suffix(String text, String suffix) {
    return endsWith(text, [suffix]) ? text : text + suffix;
  }

  static String suffixSlash(String path) {
    return endsWith(path, ["/", "\\"]) ? path : "$path/";
  }

  // E.g. expand("a-zA-Z") -> abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
  static String expand(String regex) {
    StringBuffer characters = StringBuffer();
    String start;
    bool rangeDetected = false;

    regex.runes.forEach((rune) {
      String c = String.fromCharCode(rune);

      if (start == null) {
        if (c != '-') {
          start = c;
        }
      } else {
        if (rangeDetected) {
          characters.write(c == '-' ? start : expand_(start, c));
          start = null;
          rangeDetected = false;
        } else {
          rangeDetected = c == '-';

          if (!rangeDetected) {
            characters.write(start);
            start = c;
          }
        }
      }
    });

    if (start != null) {
      characters.write(start);
    }

    return characters.toString();
  }

  // E.g. expand('0', '9') -> 0123456789
  static String expand_(String start, String end) {
    StringBuffer characters = StringBuffer();

    for (int i = start.codeUnitAt(0); i <= end.codeUnitAt(0); i++) {
      characters.writeCharCode(i);
    }

    return characters.toString();
  }

  /// Format the provided text with given arguments. In case of any errors, we just return the original text.
  /// For support normal string format and standard message bundle format.
  static String format(String text, List<Object> args) {
    String result = text;

    // TODO Do we need if-else like in Java's Utils.format()?
    try {
      for (int i = 0; i < args.length; i++) {
        Object value = get(i, args);
        value = isBlank(value) ? "" : value.toString();
        result = result.replaceAll('{$i}', value);
      }
    } catch (e) {
      // ignore the exception.
    }

    return result;
  }

  static String getLanguage() {
    return Platform.localeName; // TODO Is this correct?
  }

  static String getMsg(Exception e) {
    return e == null ? null : e.toString();
  }

  static String getMsg_(
      FormController controller, String i18nMsgId, String bundleMsgId,
      [Object args]) {
    return getMsg__(
        FormUtils.getRenderer(controller), i18nMsgId, bundleMsgId, args);
  }

  static String getMsg__(
      UIRenderer renderer, String i18nMsgId, String bundleMsgId,
      [Object args]) {
    if (renderer == null) {
      Logger.warn("Can't get message because controller or renderer is null.");
      return null;
    }

    return renderer.getMsg(i18nMsgId, bundleMsgId, args);
  }

  static bool hasMsg(String i18nMsgId, UIRenderer renderer) {
    final String msg =
        renderer == null ? null : renderer.getMsg(i18nMsgId, null, null);
    return !isBlank(msg) && !equalsIgnoreCase(i18nMsgId, msg);
  }

  static String getStackTrace(StackTrace t) {
    return t == null ? null : t.toString();
  }

  static String getUUID() => Uuid().v4().replaceAll("-", "").toUpperCase();

  static String getShortUUID() => getUUID().substring(20);

  static String enquote(Object obj) {
    return obj == null
        ? null
        : SQ + (obj is String ? obj : obj.toString()) + SQ;
  }

  static String doubleQuote(Object obj) {
    return obj == null
        ? null
        : DQ + (obj is String ? obj : obj.toString()) + DQ;
  }

  // NOTE: Below functions are NOT in CMS Java

  static bool equalsIgnoreCase(String string1, String string2) {
    return string1?.toLowerCase() == string2?.toLowerCase();
  }

  static int compareToIgnoreCase(String string1, String string2) {
    return string1.toLowerCase().compareTo(string2?.toLowerCase());
  }

  // TODO Following functions should only work with char or int (String is wrong)
  // TODO isDigit() and isLetter() only works for English
  static bool isDigit(String s) {
    return !isBlank(s) && REGEX_NUMERIC.hasMatch(s[0]);
  }

  static bool isLetter(String s) {
    return !isBlank(s) && REGEX_ALPHA.hasMatch(s[0]);
  }

  static bool isLowerCase(String s) {
    return !isBlank(s) && s[0] == s[0].toLowerCase();
  }

  static bool isUpperCase(String s) {
    return !isBlank(s) && s[0] == s[0].toUpperCase();
  }

  static String firstToLower(String name) {
    return name[0].toLowerCase() + name.substring(1);
  }

  // NOTE: Following functions are copy-pasted from 'validators: ^3.0.0' package

  /// check if the string [str] is a URL
  ///
  /// * [protocols] sets the list of allowed protocols
  /// * [requireTld] sets if TLD is required
  /// * [requireProtocol] is a `bool` that sets if protocol is required for validation
  /// * [allowUnderscore] sets if underscores are allowed
  /// * [hostWhitelist] sets the list of allowed hosts
  /// * [hostBlacklist] sets the list of disallowed hosts
  static bool isURL(String str,
      {List<String> protocols = const ['http', 'https', 'ftp'],
      bool requireTld = true,
      bool requireProtocol = false,
      bool allowUnderscore = false,
      List<String> hostWhitelist = const [],
      List<String> hostBlacklist = const []}) {
    if (str == null ||
        str.length == 0 ||
        str.length > 2083 ||
        str.startsWith('mailto:')) {
      return false;
    }

    var protocol,
        user,
        auth,
        host,
        hostname,
        port,
        port_str,
        path,
        query,
        hash,
        split;

    // check protocol
    split = str.split('://');
    if (split.length > 1) {
      protocol = shift(split);
      if (protocols.indexOf(protocol) == -1) {
        return false;
      }
    } else if (requireProtocol == true) {
      return false;
    }
    str = split.join('://');

    // check hash
    split = str.split('#');
    str = shift(split);
    hash = split.join('#');
    if (hash != null && hash != "" && RegExp(r'\s').hasMatch(hash)) {
      return false;
    }

    // check query params
    split = str.split('?');
    str = shift(split);
    query = split.join('?');
    if (query != null && query != "" && RegExp(r'\s').hasMatch(query)) {
      return false;
    }

    // check path
    split = str.split('/');
    str = shift(split);
    path = split.join('/');
    if (path != null && path != "" && RegExp(r'\s').hasMatch(path)) {
      return false;
    }

    // check auth type urls
    split = str.split('@');
    if (split.length > 1) {
      auth = shift(split);
      if (auth.indexOf(':') >= 0) {
        auth = auth.split(':');
        user = shift(auth);
        if (!RegExp(r'^\S+$').hasMatch(user)) {
          return false;
        }
        if (!RegExp(r'^\S*$').hasMatch(user)) {
          return false;
        }
      }
    }

    // check hostname
    hostname = split.join('@');
    split = hostname.split(':');
    host = shift(split);
    if (split.length > 0) {
      port_str = split.join(':');
      try {
        port = int.parse(port_str, radix: 10);
      } catch (e) {
        return false;
      }
      if (!RegExp(r'^[0-9]+$').hasMatch(port_str) ||
          port <= 0 ||
          port > 65535) {
        return false;
      }
    }

    if (!isIP(host) &&
        !isFQDN(host,
            requireTld: requireTld, allowUnderscores: allowUnderscore) &&
        host != 'localhost') {
      return false;
    }

    if (hostWhitelist.isNotEmpty && !hostWhitelist.contains(host)) {
      return false;
    }

    if (hostBlacklist.isNotEmpty && hostBlacklist.contains(host)) {
      return false;
    }

    return true;
  }

  static shift(List l) {
    if (l.length >= 1) {
      var first = l.first;
      l.removeAt(0);
      return first;
    }
    return null;
  }

  /// check if the string [str] is IP [version] 4 or 6
  ///
  /// * [version] is a String or an `int`.
  static bool isIP(String str, [/*<String | int>*/ version]) {
    version = version.toString();
    if (version == 'null') {
      return isIP(str, 4) || isIP(str, 6);
    } else if (version == '4') {
      if (!REGEX_IPV4MAYBE.hasMatch(str)) {
        return false;
      }
      var parts = str.split('.');
      parts.sort((a, b) => int.parse(a) - int.parse(b));
      return int.parse(parts[3]) <= 255;
    }
    return version == '6' && REGEX_IPV6.hasMatch(str);
  }

  /// check if the string [str] is a fully qualified domain name (e.g. domain.com).
  ///
  /// * [requireTld] sets if TLD is required
  /// * [allowUnderscore] sets if underscores are allowed
  static bool isFQDN(String str,
      {bool requireTld = true, bool allowUnderscores = false}) {
    var parts = str.split('.');
    if (requireTld) {
      var tld = parts.removeLast();
      if (parts.length == 0 || !RegExp(r'^[a-z]{2,}$').hasMatch(tld)) {
        return false;
      }
    }

    for (var part in parts) {
      if (allowUnderscores) {
        if (part.contains('__')) {
          return false;
        }
      }
      if (!RegExp(r'^[a-z\\u00a1-\\uffff0-9-]+$').hasMatch(part)) {
        return false;
      }
      if (part[0] == '-' ||
          part[part.length - 1] == '-' ||
          part.indexOf('---') >= 0) {
        return false;
      }
    }
    return true;
  }
}
