/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../enums/code.dart';
import '../../error/class_cast_exception.dart';
import '../../structure/model/field.dart';
import '../../ui/discrete/component/base/component.dart';
import '../../ui/discrete/component/behavior/i_array.dart';
import '../../ui/discrete/component/behavior/i_label.dart';
import '../../ui/discrete/component/behavior/i_title.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../ui/discrete/view_state.dart';
import '../../validation/converter/c_binary.dart';
import '../../validation/converter/c_boolean.dart';
import '../../validation/converter/c_byte.dart';
import '../../validation/converter/c_character.dart';
import '../../validation/converter/c_decimal.dart';
import '../../validation/converter/c_float.dart';
import '../../validation/converter/c_long.dart';
import '../../validation/converter/c_number.dart';
import '../../validation/converter/c_short.dart';
import '../../validation/converter/c_string.dart';
import '../../validation/converter/conversion_error.dart';
import '../../validation/error_marker.dart';
import '../form_utils.dart';
import '../utils.dart';
import 'bounds_utils.dart';

class ValueUtils {
  const ValueUtils._();

  static List<String> getViews(
      String mainView, List<String> boundedViews, UIRenderer renderer) {
    Set<String> predecessors = {};
    predecessors.add(mainView);
    return getViews_(predecessors, boundedViews, renderer);
  }

  static List<String> getViews_(Set<String> predecessors,
      List<String> boundedViews, UIRenderer renderer) {
    if (Utils.isBlank(boundedViews)) {
      return null;
    }

    List<String> views = [];

    for (String boundedView in boundedViews) {
      String viewId = BoundsUtils.getViewId(boundedView);

      if (!predecessors.contains(viewId)) {
        predecessors.add(viewId);

        // to ensure that the logics works case-insensitive
        getViews__(viewId, predecessors, views, renderer);
      }
    }

    return views;
  }

  static void getViews__(String viewId, Set<String> predecessors,
      List<String> views, UIRenderer renderer) {
    ViewState view = FormUtils.getView_(viewId, renderer);

    if (view != null) {
      Field field = view.getField();

      if (field != null) {
        Utils.addAll(views,
            [getViews_(predecessors, field.getReplicate(renderer), renderer)]);
      }

      views.add(view.getId());
    }
  }

  static void setValues(
      Map<String, Object> data, bool unescape, UIRenderer renderer) {
    if (!Utils.isBlank(data)) {
      for (String viewId in data.keys) {
        setValue(viewId, data[viewId], unescape, renderer);
      }
    }
  }

  static void setValue(
      String viewId, Object value, bool unescape, UIRenderer renderer) {
    final bool setLabel = Code.EFFECT.starts(viewId);
    ViewState view = FormUtils.getView_(
        setLabel ? Utils.skipFirst(viewId) : viewId, renderer);

    if (view != null) {
      if (setLabel || view.isContainer()) {
        Component component = view.getComponent();
        final String text = value as String;

        if (component is ITitle) {
          view.setTitle(text);
          (component as ITitle)
              .setTitle(renderer.get(view.getTitle())); // get the title again
        } else if (component is ILabel) {
          view.setLabel(text);
          (component as ILabel)
              .setLabel(renderer.get(view.getLabel())); // get the label again
        }
      } else if (view.canInput()) {
        if (view.canArray() &&
            value != null &&
            !(view.getInputComponent() as IArray).isValidItem(value)) {
          ErrorMarker.mark(
              view, renderer, null, view.getInvalidItemMsg(), [value]);
          value = null;
        }

        if (unescape && value is String) {
          value = Utils.unescape(value as String);
        }

        view.setValue(value);
      }
    }
  }

  /// Casts the value to the target class. Only primitive data types (including String) are supported. If provided
  /// value is of String data type then it attempts to parse it to targeted class. But if value's class doesn't match
  /// any supported data types then the value is as it is returned.
  ///
  /// @throws ClassCastException
  ///         When target class belongs to one of the supported data types but given value isn't of target or String.
  static T cast<T>(Object value, Type target) {
    if (value == null) {
      return null;
    }

    final Type srcClass = value.runtimeType;

    // If value's class is the same as required.
    if (srcClass == target) {
      return value as T;
    }

    try {
      if (CByte.checkType(value)) {
        return CByte.convert(value) as T;
      } else if (CShort.checkType(value)) {
        return CShort.convert(value) as T;
      } else if (CNumber.checkType(value)) {
        return CNumber.convert(value) as T;
      } else if (CLong.checkType(value)) {
        return CLong.convert(value) as T;
      } else if (CFloat.checkType(value)) {
        return CFloat.convert(value) as T;
      } else if (CDecimal.checkType(value)) {
        return CDecimal.convert(value) as T;
      } else if (CCharacter.checkType(value)) {
        return CCharacter.convert(value) as T;
      } else if (CString.checkType(value)) {
        return CString.convert(value) as T;
      } else if (CBoolean.checkType(value)) {
        return CBoolean.convert(value) as T;
      } else if (CBinary.checkType(value)) {
        return CBinary.convert(value) as T;
      }
    } on ConversionError {
      // nothing to do here
    }

    throw ClassCastException("Can't cast $srcClass to $target.");
  }
}
