/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'package:collection/collection.dart' show ListEquality;

import '../../config/defaults.dart';
import '../../enums/constraint_type.dart';
import '../../enums/field_type.dart';
import '../../error/illegal_argument_exception.dart';
import '../../ui/discrete/renderer/ui_renderer.dart';
import '../../util/date_utils.dart';
import '../../util/enum.dart';
import '../../util/enum_utils.dart';
import '../../util/log/logger.dart';
import '../../util/utils.dart';
import 'c_abstract.dart';

class CCompare extends CAbstract<Object> {
  static final ConstraintType TYPE = ConstraintType.COMPARE;

  final Object crossValue;
  final FieldType type;
  final Operator op;
  final String opMsg;
  final bool sensitive;
  final bool isWsq;
  final double threshold;

  CCompare(this.crossValue, this.type, this.op, this.opMsg, this.sensitive,
      this.isWsq, this.threshold)
      : super(TYPE);

  static CCompare get(Object crossValue, FieldType type, String op,
      bool sensitive, UIRenderer renderer) {
    return get_(crossValue, type, op, sensitive, false, null, renderer);
  }

  static CCompare get_(Object crossValue, FieldType type, String op,
      bool sensitive, bool isWsq, double threshold, UIRenderer renderer) {
    Operator operator =
        Utils.isBlank(op) ? Operator.EQ : Operator.toOperator(op, false);
    String operatorMsg = renderer.get(operator.msgId());

    return CCompare(toNullAndTrim(crossValue), type, operator, operatorMsg,
        Utils.toTrue(sensitive), isWsq, Utils.toDefault(threshold, 50));
  }

  @override
  String isValid(Object value) {
    return compare(toNullAndTrim(value), crossValue) ? null : getMsgId();
  }

  bool compare(Object valueA, Object valueB) {
    // handle null values first
    final bool valueAIsNull = valueA == null;
    final bool valueBIsNull = valueB == null;

    // it's a valid case if both are null but only if EQ operator is in use
    if (valueAIsNull && valueBIsNull) {
      return Operator.EQ == op;
    }

    if (valueAIsNull || valueBIsNull) {
      return Operator.NE == op;
    }

    bool valid = true;

    if (FieldType.NUMBER == type) {
      valid = compareNumbers(valueA as int, valueB as int);
    } else if (FieldType.LONG == type) {
      valid = compareLong(valueA as int, valueB as int);
    } else if (FieldType.DECIMAL == type) {
      valid = compareDecimals(valueA as double, valueB as double);
    } else if (EnumUtils.match_(
        type, [FieldType.CHARACTER, FieldType.STRING])) {
      throwIfNotEqOrNe();

      final String first = Utils.toBlank(valueA.toString());
      final String second = Utils.toBlank(valueB.toString());
      valid =
          sensitive ? first == second : Utils.equalsIgnoreCase(first, second);
    } else if (EnumUtils.match_(
        type, [FieldType.DATE, FieldType.TIME, FieldType.DATETIME])) {
      valid = compareDates(valueA as DateTime, valueB as DateTime);
    } else if (FieldType.BINARY == type) {
      throwIfNotEqOrNe();
      valid = compareBinary(toBinary(valueA), toBinary(valueB));
    } else {
      // for any other field type, we just do EQ
      valid = valueA == valueB;
    }

    return Operator.NE == op ? !valid : valid;
  }

  bool compareNumbers(int valueA, int valueB) {
    if (Operator.LT == op) {
      return valueA < valueB;
    }

    if (Operator.LE == op) {
      return valueA <= valueB;
    }

    if (Operator.GT == op) {
      return valueA > valueB;
    }

    if (Operator.GE == op) {
      return valueA >= valueB;
    }

    return valueA == valueB; // default operator is EQ
  }

  bool compareLong(int valueA, int valueB) {
    if (Operator.LT == op) {
      return valueA < valueB;
    }

    if (Operator.LE == op) {
      return valueA <= valueB;
    }

    if (Operator.GT == op) {
      return valueA > valueB;
    }

    if (Operator.GE == op) {
      return valueA >= valueB;
    }

    return valueA == valueB; // default operator is EQ
  }

  bool compareDecimals(double valueA, double valueB) {
    if (Operator.LT == op) {
      return valueA < valueB;
    }

    if (Operator.LE == op) {
      return valueA <= valueB;
    }

    if (Operator.GT == op) {
      return valueA > valueB;
    }

    if (Operator.GE == op) {
      return valueA >= valueB;
    }

    return valueA == valueB; // default operator is EQ
  }

  bool compareDates(DateTime valueA, DateTime valueB) {
    return compareLong(
        DateUtils.dateToMillis_(valueA), DateUtils.dateToMillis_(valueB));
  }

  bool compareBinary(List<int> valueA, List<int> valueB) {
    final bool same = ListEquality().equals(valueA, valueB);

    if (!same && isWsq) {
      // if they are completely the same, no need to check further.
      try {
        // TODO Support it's native support (for Android only).
        double score =
            0.0; // FIXME WsqFingerprintMatcher.compareWsq(valueA, valueB)
        return score > threshold;
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
      }
    }

    return same;
  }

  static Object toNullAndTrim(Object value) {
    if (value is String) {
      final String sValue = value;
      return Utils.isBlank(sValue) ? null : sValue.trim();
    }

    return value;
  }

  void throwIfNotEqOrNe() {
    if (!EnumUtils.match_(op, [Operator.EQ, Operator.NE])) {
      throw IllegalArgumentException(
          "Only EQ or NE operator is allowed for '$type' comparision.");
    }
  }

  List<int> toBinary(Object value) {
    return value as List<int>;
  }

  @override
  List<Object> getMsgArgs() {
    return [opMsg, crossValue];
  }
}

class Args {
  static const String SENSITIVE = Defaults.SENSITIVE;
  static const String OPERATOR = Defaults.OPERATOR;
  static const String VALUE = Defaults.VALUE;
  static const String THRESHOLD = "threshold"; // only used for WSQ comparison

}

class Operator extends Enum<Operator> {
  // (EQ, =), (NE !=), (LT, <), (LE, <=), (GT, >), (GE, >=)
  static const EQ = Operator("EQ");
  static const NE = Operator("NE");
  static const LT = Operator("LT");
  static const LE = Operator("LE");
  static const GT = Operator("GT");
  static const GE = Operator("GR");

  const Operator(String name) : super(name);

// TODO Should be different messages for all operators with only 'value' argument
  String msgId() {
    return "validation.compare.${name().toLowerCase()}";
  }

  static Operator toOperator(String operator, bool throwOnMismatch) {
    return EnumUtils.toEnum_(operator, _ENUMS, throwOnMismatch);
  }

  static const _ENUMS = [EQ, NE, LT, LE, GT, GE];
}
