/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:typed_data' show Uint8List;

import '../../../form/structure/string_object_map.dart';
import '../../enums/constraint_type.dart';
import '../../ui/discrete/form_controller.dart';
import '../../util/bean/range/float_range.dart';
import '../../util/bean/range/int_range.dart';
import '../../util/enum.dart';
import '../../util/enum_utils.dart';
import '../../util/form_utils.dart';
import '../../util/log/logger.dart';
import '../../util/utils.dart';
import '../converter/c_boolean.dart';
import '../converter/c_string.dart';
import 'c_abstract.dart';

abstract class CFace<T> extends CAbstract<Uint8List> {
  static final ConstraintType TYPE = ConstraintType.FACE;

  static const Type DEFAULT_TYPE = Type.DETECTOR;
  static const int DEFAULT_MIN_FACES = 1;

  static const double MAX_CONFIDENCE = 100.0; // TODO What's the best value?

  Params params;

  CFace(Params params) : super(TYPE) {
    this.params = params == null ? Params() : params;
  }

  @override
  String isValid(Uint8List value) {
    List<T> faces = Utils.isBlank(value) ? null : findFaces(value);

    if (faces == null) {
      return null; // couldn't locate the image or some error occurred
    }

    bool valid = countFaces(faces);

    for (T face in faces) {
      if (!valid) {
        break;
      }

      if (face != null) {
        if (params.confidence != null) {
          valid = calculateConfidence(face) >= params.confidence;
        }

        if (valid && params.eyesDistance != null) {
          valid = checkEyesDistance(face);
        }

        if (valid && params.yAngle != null) {
          valid = checkYAngle(face);
        }

        if (valid && params.zAngle != null) {
          valid = checkZAngle(face);
        }
      }
    }

    return valid ? null : getMsgId();
  }

  @override
  List<Object> getMsgArgs() {
    return [params.faces.min, params.faces.max];
  }

  List<T> findFaces(Uint8List value);

  bool countFaces(List<T> faces) {
    int facesCount = 0;

    for (T face in faces) {
      if (face != null) {
        facesCount++;
      }
    }

    Logger.info("Faces count = $facesCount");
    return params.faces.isInRange(facesCount);
  }

  @override
  String getMsgId() {
    if (params.faces.max == null) {
      return "validation.face.min.invalid";
    }

    if (params.faces.min == null) {
      return "validation.face.max.invalid";
    }

    return params.faces.minEqualsMax()
        ? "validation.face.equal.invalid"
        : super.getMsgId();
  }

  int calculateConfidence(T face);

  bool checkEyesDistance(T face);

  bool checkYAngle(T face);

  bool checkZAngle(T face);
}

class Args {
  static const String TYPE = "type";
  static const String MIN_FACES = "min_faces";
  static const String MAX_FACES = "max_faces";
  static const String CONFIDENCE = "confidence"; // in percentage
  static const String MIN_EYES_DISTANCE = "min_eyes_distance"; // in pixels
  static const String MAX_EYES_DISTANCE = "max_eyes_distance"; // in pixels
  static const String MIN_Y_ANGLE = "min_y_angle";
  static const String MAX_Y_ANGLE = "max_y_angle";
  static const String MIN_Z_ANGLE = "min_z_angle";
  static const String MAX_Z_ANGLE = "max_z_angle";

  // effects only the component not the constraint
  static const String OVERLAY_MASK = "overlay_mask"; // true or false
  static const String CROP_IMAGE = "crop_image"; // true or false
  static const String REMOVE_BACKGROUND = "remove_background"; // true or false

  // related to Params.java in mt-client-desk-aliveness for 3DiVi impl
  static const String DRAW_FACE = "drawFace";

  static const String MIN_PITCH = "minPitch";
  static const String MAX_PITCH = "maxPitch";

  static const String MIN_ROLL = "minRoll";
  static const String MAX_ROLL = "maxRoll";

  static const String MIN_YAW = "minYaw";
  static const String MAX_YAW = "maxYaw";

  static const String LEFT_EYE_MIN_X = "leftEyeMinX";
  static const String LEFT_EYE_MAX_X = "leftEyeMaxX";

  static const String LEFT_EYE_MIN_Y = "leftEyeMinY";
  static const String LEFT_EYE_MAX_Y = "leftEyeMaxY";

  static const String RIGHT_EYE_MIN_X = "rightEyeMinX";
  static const String RIGHT_EYE_MAX_X = "rightEyeMaxX";

  static const String RIGHT_EYE_MIN_Y = "rightEyeMinY";
  static const String RIGHT_EYE_MAX_Y = "rightEyeMaxY";

  static const String AGE_GROUP = "ageGroup";

  static const String AGE_YEARS_MIN = "ageYearsMin";
  static const String AGE_YEARS_MAX = "ageYearsMax";

  static const String GENDER = "gender";
  static const String EMOTION = "emotion";

  static const String FACE_QUALITY_MIN = "faceQualityMin";
  static const String FACE_QUALITY_MAX = "faceQualityMax";

  static const String TOTAL = "total";

  static const String MIN_LIGHTING = "minLighting";
  static const String MAX_LIGHTING = "maxLighting";

  static const String MIN_NOISE = "minNoise";
  static const String MAX_NOISE = "maxNoise";

  static const String MIN_SHARPNESS = "minSharpness";
  static const String MAX_SHARPNESS = "maxSharpness";

  static const String MIN_FLARE = "minFlare";
  static const String MAX_FLARE = "maxFlare";

  static const String FACE_CUT_TYPE = "faceCutType";
}

class Type extends Enum<Type> {
  static const DETECTOR = Type("DETECTOR");
  static const LIKELIHOOD = Type("LIKELIHOOD");
  static const ALIVENESS = Type("ALIVENESS"); // Checks for eye blinks
  static const ALIVENESS_PASSIVE = Type(
      "ALIVENESS_PASSIVE"); // Checks passively for facial aliveness using 3DiVi SDK

  const Type(String name) : super(name);

  static Type toType(String type, bool throwOnMismatch) {
    return EnumUtils.toEnum_(type, _ENUMS, throwOnMismatch);
  }

  static Type toType_(Object oType) {
    return Utils.toDefault(
        toType(CString.convert(oType), false), CFace.DEFAULT_TYPE);
  }

  static const List<Type> _ENUMS = [
    DETECTOR,
    LIKELIHOOD,
    ALIVENESS,
    ALIVENESS_PASSIVE
  ];
}

class AgeGroup extends Enum<AgeGroup> {
  static const KID = AgeGroup("KID");
  static const YOUNG = AgeGroup("YOUNG");
  static const ADULT = AgeGroup("ADULT");
  static const SENIOR = AgeGroup("SENIOR");

  const AgeGroup(String name) : super(name);

  static AgeGroup toAgeGroup(String ageGroup, bool throwOnMismatch) {
    return EnumUtils.toEnum_(ageGroup, _ENUMS, throwOnMismatch);
  }

  static const List<AgeGroup> _ENUMS = [KID, YOUNG, ADULT, SENIOR];
}

class Gender extends Enum<Gender> {
  static const FEMALE = Gender("FEMALE");
  static const MALE = Gender("MALE");

  const Gender(String name) : super(name);

  static Gender toGender(String gender, bool throwOnMismatch) {
    return EnumUtils.toEnum_(gender, _ENUMS, throwOnMismatch);
  }

  static const List<Gender> _ENUMS = [FEMALE, MALE];
}

class Emotion extends Enum<Emotion> {
  static const NEUTRAL = Emotion("NEUTRAL");
  static const HAPPY = Emotion("HAPPY");
  static const ANGRY = Emotion("ANGRY");
  static const SURPRISE = Emotion("SURPRISE");

  const Emotion(String name) : super(name);

  static Emotion toEmotion(String emotion, bool throwOnMismatch) {
    return EnumUtils.toEnum_(emotion, _ENUMS, throwOnMismatch);
  }

  static const List<Emotion> _ENUMS = [NEUTRAL, HAPPY, ANGRY, SURPRISE];
}

class FaceCutType extends Enum<FaceCutType> {
  static const BASE = FaceCutType("BASE");
  static const FULL_FRONTAL = FaceCutType("FULL_FRONTAL");
  static const TOKEN_FRONTAL = FaceCutType("TOKEN_FRONTAL");

  const FaceCutType(String name) : super(name);

  static FaceCutType toFaceCutType(String faceCutType, bool throwOnMismatch) {
    return EnumUtils.toEnum_(faceCutType, _ENUMS, throwOnMismatch);
  }

  static const List<FaceCutType> _ENUMS = [BASE, FULL_FRONTAL, TOKEN_FRONTAL];
}

class Params {
  Type type;
  IntRange faces;
  double confidence; // in percentage
  FloatRange eyesDistance;
  FloatRange yAngle;
  FloatRange zAngle;

  Params() : this.type(null);

  Params.type(Type type) : this.full(type, null, null, null, null, null);

  Params.full(Type type, IntRange faces, double confidence,
      FloatRange eyesDistance, FloatRange yAngle, FloatRange zAngle) {
    this.type = Utils.toDefault(type, CFace.DEFAULT_TYPE);
    this.faces = faces == null
        ? IntRange.get_(CFace.DEFAULT_MIN_FACES, CFace.DEFAULT_MIN_FACES)
        : faces;
    this.confidence = confidence;
    this.eyesDistance = eyesDistance;
    this.yAngle = yAngle;
    this.zAngle = zAngle;
  }
}

class UIParams {
  final Type type;
  final bool overlayMask;
  final bool cropImage;
  final bool removeBackground;

  const UIParams._(
      this.type, this.overlayMask, this.cropImage, this.removeBackground);

  static UIParams get(String viewId, FormController controller) {
    StringObjectMap args =
        FormUtils.getConstraintArgs(viewId, CFace.TYPE, controller);
    Type type = Type.toType(args[Args.TYPE], false);
    bool overlayMask =
        Type.DETECTOR == type && getFlag(Args.OVERLAY_MASK, false, args);
    bool cropImage = getFlag(Args.CROP_IMAGE, false, args);
    bool removeBackground = overlayMask &&
        !cropImage &&
        getFlag(Args.REMOVE_BACKGROUND, false, args);

    return UIParams._(type, overlayMask, cropImage, removeBackground);
  }

  static bool getFlag(String arg, bool defaultValue, StringObjectMap args) {
    return Utils.toDefault(
        CBoolean.convertNoErr(Utils.get_(arg, args)), defaultValue);
  }
}
