/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../config/defaults.dart';
import '../../enums/constraint_type.dart';
import '../../error/illegal_argument_exception.dart';
import '../../util/date_utils.dart';
import '../../util/utils.dart';
import '../converter/c_date.dart';
import '../converter/c_decimal.dart';
import '../converter/c_number.dart';
import 'c_abstract.dart';

class CRange extends CAbstract<Object> {
  static final ConstraintType TYPE = ConstraintType.RANGE;

  static const String MSG_MIN = "validation.range.min.invalid";
  static const String MSG_MAX = "validation.range.max.invalid";

  final Object min;
  final Object max;

  /// Args min/max are normally numbers except for Date/Time field types.
  CRange(this.min, this.max) : super(TYPE) {
    if (Utils.allNull([min, max])) {
      throw IllegalArgumentException(
          "Either ${Args.MIN} or ${Args.MAX} is required.");
    }
  }

  @override
  String isValid(Object value) {
    if (value == null) {
      return null; // Considered valid if nothing is provided. This should be handled by mandatory constraint.
    }

    if (value is DateTime) {
      return dateChecker(value);
    }

    if (value is double) {
      return doubleChecker(CDecimal.convertNoErr(value),
          CDecimal.convertNoErr(min), CDecimal.convertNoErr(max));
    }

    final int minimum = CNumber.convertNoErr(min);
    final int maximum = CNumber.convertNoErr(max);

    if (value is num) {
      return numberChecker(value, minimum, maximum);
    }

    if (value is List) {
      return sizeChecker((value).length, minimum, maximum);
    }

    return null;
  }

  String dateChecker(DateTime date) {
    // initially try to treat them as numbers (range in years)
    final int minYear = CNumber.convertNoErr(min);
    final int maxYear = CNumber.convertNoErr(max);

    // if minYear and maxYear are not numbers, treat them as dates (inclusive)
    final DateTime start = minYear == null
        ? CDate.convertNoErr(min)
        : DateUtils.alterToday(minYear < 0, minYear.abs());

    if (min != null && start == null) {
      throw IllegalArgumentException("Couldn't parse ${Args.MIN} date.");
    }

    final DateTime end = maxYear == null
        ? CDate.convertNoErr(max)
        : DateUtils.alterToday(maxYear < 0, maxYear.abs());

    if (max != null && end == null) {
      throw IllegalArgumentException("Couldn't parse ${Args.MAX} date.");
    }

    if (Utils.noneNull([start, end]) && start.isAfter(end)) {
      throw IllegalArgumentException("Start date should be before end date.");
    }

    return checkDateRange(date, start, end);
  }

  String checkDateRange(DateTime date, DateTime min, DateTime max) {
    if (min == null) {
      return date.isAfter(max) ? MSG_MAX : null;
    }

    if (max == null) {
      return date.isBefore(min) ? MSG_MIN : null;
    }

    return date.isBefore(min) || date.isAfter(max) ? getMsgId() : null;
  }

  String doubleChecker(double value, double min, double max) {
    if (min == null) {
      return value <= max ? null : MSG_MAX;
    }

    if (max == null) {
      return value >= min ? null : MSG_MIN;
    }

    return value >= min && value <= max ? null : getMsgId();
  }

  String numberChecker(num value, int min, int max) {
    final int longValue = value.toInt();

    if (min == null) {
      return longValue <= max ? null : MSG_MAX;
    }

    if (max == null) {
      return longValue >= min ? null : MSG_MIN;
    }

    return longValue >= min && longValue <= max ? null : getMsgId();
  }

  String sizeChecker(int size, int min, int max) {
    if (min == null) {
      return size <= max ? null : MSG_MAX;
    }

    if (max == null) {
      return size >= min ? null : MSG_MIN;
    }

    return size >= min && size <= max ? null : getMsgId();
  }

  @override
  String getMsgId() {
    return min == max ? "validation.range.equal.invalid" : super.getMsgId();
  }

  @override
  List<Object> getMsgArgs() {
    return [min, max];
  }
}

class Args {
  static const String MIN = Defaults.MIN;
  static const String MAX = Defaults.MAX;
}
