/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../config/defaults.dart';
import '../../enums/constraint_type.dart';
import '../../error/illegal_argument_exception.dart';
import '../../util/enum.dart';
import '../../util/enum_utils.dart';
import '../../util/utils.dart';
import 'c_abstract.dart';

class CText extends CAbstract<String> {
  static final ConstraintType TYPE = ConstraintType.TEXT;

  Operator op;
  List<String> refValues; // CSV (should be regex in case of MATCHES op)
  bool sensitive;

  CText(Operator op, bool sensitive, String refValues) : super(TYPE) {
    if (op == null) {
      throw IllegalArgumentException("${Args.OPERATOR} can't be blank!");
    }

    this.op = op;
    this.sensitive = Utils.toFalse(sensitive);
    this.refValues = Utils.splitByCommaAsList(refValues);

    if (Utils.isBlank(this.refValues)) {
      throw IllegalArgumentException("${Args.VALUE} can't be blank!");
    }
  }

  @override
  String isValid(String value) {
    if (Utils.isBlank(value)) {
      return null;
    }

    final String srcValue = getValue(value);

    for (String refValue in refValues) {
      if (isValid_(srcValue, getValue(refValue))) {
        return null;
      }
    }

    return getMsgId();
  }

  String getValue(String value) {
    return (sensitive ? value : value.toLowerCase()).trim();
  }

  bool isValid_(String srcValue, String refValue) {
    if (Operator.CONTAINS == op) {
      return srcValue.contains(refValue);
    } else if (Operator.NOTCONTAINS == op) {
      return !srcValue.contains(refValue);
    } else if (Operator.EQUALS == op) {
      return srcValue == refValue;
    } else if (Operator.NOTEQUALS == op) {
      return srcValue != refValue;
    } else if (Operator.MATCHES == op) {
      return Utils.matches(srcValue, refValue);
    } else if (Operator.NOTMATCHES == op) {
      return !Utils.matches(srcValue, refValue);
    } else if (Operator.STARTSWITH == op) {
      return srcValue.startsWith(refValue);
    } else if (Operator.NOTSTARTSWITH == op) {
      return !srcValue.startsWith(refValue);
    } else if (Operator.ENDSWITH == op) {
      return srcValue.endsWith(refValue);
    } else if (Operator.NOTENDSWITH == op) {
      return !srcValue.endsWith(refValue);
    }

    throw IllegalArgumentException("Unknown operator: $op");
  }

  @override
  String getMsgId() {
    return op.msgId();
  }

  @override
  List<Object> getMsgArgs() {
    return [refValues.join("' or '")];
  }
}

class Args {
  static const String SENSITIVE = Defaults.SENSITIVE;
  static const String OPERATOR = Defaults.OPERATOR;
  static const String VALUE = Defaults.VALUE;
}

class Operator extends Enum<Operator> {
  static const CONTAINS = Operator("CONTAINS");
  static const NOTCONTAINS = Operator("NOTCONTAINS");
  static const EQUALS = Operator("EQUALS");
  static const NOTEQUALS = Operator("NOTEQUALS");
  static const MATCHES = Operator("MATCHES");
  static const NOTMATCHES = Operator("NOTMATCHES");
  static const STARTSWITH = Operator("STARTSWITH");
  static const NOTSTARTSWITH = Operator("NOTSTARTSWITH");
  static const ENDSWITH = Operator("ENDSWITH");
  static const NOTENDSWITH = Operator("NOTENDSWITH");

  const Operator(String name) : super(name);

  String msgId() {
    return "validation.text.${name().toLowerCase()}.invalid";
  }

  static Operator toOperator(String operator, bool throwOnMismatch) {
    return EnumUtils.toEnum_(operator, _ENUMS, throwOnMismatch);
  }

  static const List<Operator> _ENUMS = [
    CONTAINS,
    NOTCONTAINS,
    EQUALS,
    NOTEQUALS,
    MATCHES,
    NOTMATCHES,
    STARTSWITH,
    NOTSTARTSWITH,
    ENDSWITH,
    NOTENDSWITH
  ];
}
