/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:typed_data' show Uint8List;

import '../../../config/defaults.dart';
import '../../../enums/constraint_type.dart';
import '../../../error/illegal_argument_exception.dart';
import '../../../util/log/logger.dart';
import '../../../util/numbers.dart';
import '../../../util/utils.dart';
import '../c_abstract.dart';

class CFingerprint extends CAbstract<Uint8List> {
  static final ConstraintType TYPE = ConstraintType.FINGERPRINT;

  static const String ERR_NFIQ = "validation.fingerprint.nfiq";
  static const String ERR_MINUTIAE = "validation.fingerprint.minutiae";
  static const String ERR_PARTIAL = "validation.fingerprint.partial";

  static const int UNKNOWN = 0;

  final int nfiqThreshold;
  final int minutiaeCount;
  final int grayThreshold;
  final String pattern;
  final bool natively;

  CFingerprint() : this.full(null, null, null, null, null);

  CFingerprint.full(this.nfiqThreshold, this.minutiaeCount, int grayThreshold,
      this.pattern, bool natively)
      : this.grayThreshold = Utils.toDefault(grayThreshold, 240),
        this.natively = Utils.toFalse(natively),
        super(TYPE) {
    if (Numbers.belowOne(nfiqThreshold) && nfiqThreshold > 5) {
      throw IllegalArgumentException(
          "${Args.NFIQ} can be 1 - 5 (excellent - poor) only.");
    }

    if (Numbers.belowOne(minutiaeCount)) {
      throw IllegalArgumentException(
          "${Args.MINUTIAE} should be a natural number.");
    }
  }

  @override
  String isValid(Uint8List value) {
    if (Utils.isBlank(value)) {
      return null;
    }

    if (shouldValidate(true)) {
      try {
        final int score = computeNFIQScore(value); // 1-5 (excellent - poor)

        if (score < UNKNOWN || score > nfiqThreshold) {
          return ERR_NFIQ;
        }
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
        return ERR_NFIQ;
      }
    }

    if (shouldValidate(false)) {
      try {
        final int score = computeMinutiae(value);

        if (score < UNKNOWN || score < minutiaeCount) {
          return ERR_MINUTIAE;
        }
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
        return ERR_MINUTIAE;
      }
    }

    if (!Utils.isBlank(pattern)) {
      try {
        if (!isFingerprintComplete(value)) {
          return ERR_PARTIAL;
        }
      } catch (e, stackTrace) {
        Logger.error(e, stackTrace: stackTrace);
        return ERR_PARTIAL;
      }
    }

    return null; // this constraint is dynamically run when user scans/uploads fingerprint.
  }

  bool shouldValidate(bool nfiq) {
    final int threshold = nfiq ? nfiqThreshold : minutiaeCount;
    return threshold != null && threshold > UNKNOWN;
  }

  int computeNFIQScore(Uint8List value) {
    // FIXME
    // NFIQBean score = NFIQ.computeNFIQScore(value);
    dynamic score;
    return score == null || score.getReturnCode() != UNKNOWN
        ? Defaults.MISSING
        : score.getNfiq();
  }

  int computeMinutiae(Uint8List value) {
    // FIXME
    // MinutiaeBean minutiae = NFIQ.detectMinutiae(value);
    dynamic minutiae;
    return minutiae == null || minutiae.getReturnCode() != UNKNOWN
        ? Defaults.MISSING
        : Utils.getSize(minutiae.getMinutiae().getMinutias());
  }

  bool isFingerprintComplete(Uint8List value) {
    return true; // just pass this check since FP manager would already have done it.
  }
}

class Args {
  static const String NFIQ =
      "nfiq"; // stands for 'NIST Fingerprint Image Quality'
  static const String MINUTIAE = "minutiae"; // Precise details of fingerprint
  static const String GRAY_THRESHOLD =
      "gray.threshold"; // Pixel threshold for partial check
  static const String PATTERN = "pattern"; // Pattern for partial check
  static const String BIT_RATE = "bit.rate"; // range (0.75 - 2.25)
  static const String NATIVELY =
      "natively"; // Calculate NFIQ and MINUTIAE natively (Android only)

}
