/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../../util/log/logger.dart';

/// Class to Check for blurriness of an image. The algorithm is as follows: <br>
/// <br>
/// 1. Grey scale and resize image<br>
/// 2. Convolve around a Laplace kernel defined below<br>
/// 3. Calculate variance of result and compare with user defined threshold
///
/// References:
///
/// https://www.pyimagesearch.com/2015/09/07/blur-detection-with-opencv/
/// https://www.pyimagesearch.com/2020/06/15/opencv-fast-fourier-transform-fft-for-blur-detection-in-images-and-video-streams/
/// https://github.com/indyka/blur-detection
/// https://medium.com/snapaddy-tech-blog/mobile-image-blur-detection-with-machine-learning-c0b703eab7de
/// https://medium.com/@sagardhungel/laplacian-and-its-use-in-blur-detection-fbac689f0f88
///
/// FIXME We can't possibly have 1 universal threshold to support all resolutions / cameras.
/// We need new algorithm which is free of such parameters:
///
/// https://github.com/jainullas/ImageBlurDetection
/// https://stackoverflow.com/questions/21107560/simple-way-to-check-if-an-image-bitmap-is-blur
/// https://pastebin.com/8hc1sURE
/// https://medium.com/snapaddy-tech-blog/mobile-image-blur-detection-with-machine-learning-c0b703eab7de
abstract class BlurCheck<T> {
  const BlurCheck();

  int calculate(T original) {
    Logger.debug("Image size ${getWidth(original)}x${getHeight(original)}");
    T blackAndWhite, convolve;

    try {
      blackAndWhite = toGrayscale(original);
      recycle(original);
      convolve = this.convolve(blackAndWhite);
      recycle(blackAndWhite);
      return getVariance(convolve);
    } finally {
      recycle(blackAndWhite);
      recycle(convolve);
    }
  }

  /// Threshold is the desired variance i.e. the higher the sharper (1500 is a good start). Threshold is concrete-class
  /// specific e.g. android needs threshold of 100 due to difference in Bitmap encodings.
  static bool isBlurry(int variance, int threshold) {
    return variance > 0 &&
        variance <
            threshold; // negative variance indicates calculation wasn't done
  }

  int getWidth(T img);

  int getHeight(T img);

  /// Preprocess image by resizing and greyscaling it.
  T toGrayscale(T image);

  /// Apply 3x3 kernel on image
  T convolve(T blackAndWhite);

  /// Calculate variance of all pixels. Assumes greyscale i.e. R=G=B
  int getVariance(T img);

  int calculateVariance(int mean, List<int> pixels) {
    int total = 0, difference = 0;

    for (int pixel in pixels) {
      difference = mean - pixel;
      total += difference * difference;
    }

    return total ~/ pixels.length;
  }

  void recycle(T img);
}
