/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'dart:typed_data' show Uint8List;

import '../../../config/defaults.dart';
import '../../../enums/constraint_type.dart';
import '../../../util/bean/range/int_range.dart';
import '../../../util/log/logger.dart';
import '../../../util/utils.dart';
import '../c_abstract.dart';
import 'blur/blur_check.dart';

abstract class CImage<T> extends CAbstract<Uint8List> {
  static final ConstraintType TYPE = ConstraintType.IMAGE;

  final Params params;

  int blurVariance;

  CImage(Params params)
      : this.params = params == null ? Params() : params,
        super(TYPE);

  @override
  String isValid(Uint8List value) {
    T image;
    blurVariance = null;

    try {
      image = params.verifyNothing() || Utils.isBlank(value)
          ? null
          : toImage(value);

      if (image != null) {
        if ((params.verifyWidth() &&
                !params.width.isInRange(getWidth(image))) ||
            (params.verifyHeight() &&
                !params.height.isInRange(getHeight(image)))) {
          return "validation.image.dimension";
        }

        if (params.verifyBlur()) {
          blurVariance = getBlurVariance(image);

          if (BlurCheck.isBlurry(blurVariance, params.blurThreshold)) {
            return "validation.image.blurry";
          }
        }
      }
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
    } finally {
      if (image != null) {
        recycle(image);
      }
    }

    return null;
  }

  @override
  List<Object> getMsgArgs() {
    final bool verifyWidth = params.verifyWidth();
    final bool verifyHeight = params.verifyHeight();

    return [
      Utils.toDefault(verifyWidth ? params.width.min : null, 0),
      Utils.toDefault(verifyWidth ? params.width.max : null, Defaults.NA),
      Utils.toDefault(verifyHeight ? params.height.min : null, 0),
      Utils.toDefault(verifyHeight ? params.height.max : null, Defaults.NA),
      Utils.toDefault(blurVariance, Defaults.NA)
    ];
  }

  T toImage(Uint8List value);

  int getWidth(T image);

  int getHeight(T image);

  int getBlurVariance(T image);

  void recycle(T image);
}

class Args {
  static const String MIN_WIDTH = "minWidth";
  static const String MAX_WIDTH = "maxWidth";
  static const String MIN_HEIGHT = "minHeight";
  static const String MAX_HEIGHT = "maxHeight";
  static const String AUTO_SCALE = "autoScale";

  static const String BLUR_THRESHOLD = "blurThreshold";
  static const String QUALITY = "quality";

  // used in CameraActivty for Android (nothing to do with constraint)
  static const String CAMERA_SWITCH = "cameraSwitch";
  static const String FLASH = "flash";
}

class Params {
  final IntRange width;
  final IntRange height;
  final int blurThreshold;

  Params() : this.full(null, null, null); // checks for nothing

  Params.full(this.width, this.height, this.blurThreshold);

  bool verifyWidth() {
    return width != null;
  }

  bool verifyHeight() {
    return height != null;
  }

  bool verifyBlur() {
    return blurThreshold != null && blurThreshold > 0;
  }

  bool verifyNothing() {
    return !verifyWidth() && !verifyHeight() && !verifyBlur();
  }
}
