/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../../config/defaults.dart';
import '../../../enums/constraint_type.dart';
import '../../../ui/discrete/util/language/languages.dart' as Languages;
import '../../../util/enum.dart';
import '../../../util/enum_utils.dart';
import '../../../util/structure/string_set.dart';
import '../../../util/utils.dart';
import '../../converter/c_boolean.dart';
import 'c_pattern.dart';

/// Validates that the field has a valid alpha-numeric value. But if configured to check for:
///
/// 1) Alphabets: it verifies that the value has alphabets of the language only.
/// 2) Numbers: it verifies that the value has numbers of the language only.
///
/// Language depends on the configured language. If unprovided, English is used.
class CAlphaNumeric extends CPattern {
  static final ConstraintType TYPE = ConstraintType.ALPHANUMERIC;

  static const String SPECIAL_CHARACTERS = ",\\.\\-_~@#%&*+ /<>():;\\r\\\n";

  Mode mode;
  StringSet languageSet;
  bool allowLanguageMix;

  /// The user can pass custom regex but if its null default regex is used. If language is blank, EN is used as
  /// default. If mode is null, ALPHANUMERIC is used as default. We can allow/forbid special characters when the mode
  /// is ALPHA or ALPHANUMERIC.
  CAlphaNumeric(String regex, String languages, Mode mode,
      bool allowSpecialChars, String specialChars, bool allowLanguageMix)
      : super.regex(
            TYPE,
            regex,
            getRegex_(
                regex, languages, mode, allowSpecialChars, specialChars)) {
    this.mode = Utils.toDefault(mode, Mode.ALPHANUMERIC);
    this.languageSet =
        Utils.isBlank(regex) ? Utils.splitByComma(languages) : null;
    // if user has provided custom regex, we can't differentiate between distinct language characters.
    this.allowLanguageMix =
        Utils.getSize(languageSet) <= 1 || Utils.toTrue(allowLanguageMix);
  }

  static String getRegex_(String userRegex, String languages, Mode mode,
      bool allowSpecialChars, String specialChars) {
    if (!Utils.isBlank(userRegex)) {
      return null;
    }

    StringBuffer letters = StringBuffer();
    StringBuffer numbers = StringBuffer();
    StringSet languageSet =
        Utils.splitByComma(languages); // TODO Duplicated in constructor.

    if (Utils.isBlank(languageSet)) {
      letters.write(Languages.Languages.DEFAULT.letters);
      numbers.write(Languages.Languages.DEFAULT.digits);
    } else {
      for (String langId in languageSet) {
        Languages.Language language = Languages.Languages.get(langId);

        if (language != null) {
          letters.write(language.letters);
          numbers.write(language.digits);
        }
      }
    }

    String defaultRegex =
        getLanguageCharacters(letters.toString(), numbers.toString(), mode);

    if (Mode.NUMERIC != mode && CBoolean.YES == allowSpecialChars) {
      defaultRegex +=
          RegExp.escape(Utils.toDefault(specialChars, SPECIAL_CHARACTERS));
    }

    // on Java side, we have asterisk at the end but Dart doesn't work with that
    return "[$defaultRegex\\s]+\$";
  }

  @override
  String isValid(String value) {
    String msgId = super.isValid(value) == null ? null : mode.msgId();

    if (Utils.isBlank(msgId) && !allowLanguageMix) {
      msgId = countDistinctLangChars(value) <= 1
          ? null
          : "validation.alphanumeric.mix.invalid";
    }

    return msgId;
  }

  int countDistinctLangChars(String value) {
    int languagesCount = 0;

    for (String langId in languageSet) {
      Languages.Language language = Languages.Languages.get(langId);

      if (language != null) {
        String languageChars =
            getLanguageCharacters(language.letters, language.digits, mode);

        if (Utils.matches(value, ".*[$languageChars]+.*")) {
          languagesCount++;
        }
      }
    }

    return languagesCount;
  }

  static String getLanguageCharacters(
      String letters, String numbers, Mode mode) {
    if (Mode.ALPHA == mode) {
      return letters;
    }

    if (Mode.NUMERIC == mode) {
      return numbers;
    }

    return letters + numbers;
  }
}

class Mode extends Enum<Mode> {
  static const ALPHA = Mode("ALPHA");
  static const NUMERIC = Mode("NUMERIC");
  static const ALPHANUMERIC = Mode("ALPHANUMERIC");

  const Mode(String name) : super(name);

  String msgId() {
    return "validation.${name().toLowerCase()}.invalid";
  }

  static Mode toMode(String mode, bool throwOnMismatch) {
    return EnumUtils.toEnum_(mode, _ENUMS, throwOnMismatch);
  }

  static const List<Mode> _ENUMS = [ALPHA, NUMERIC, ALPHANUMERIC];
}

class Args {
  static const String REGEX = Defaults.REGEX;
  static const String LANGUAGE = Defaults.LANGUAGE;
  static const String MODE = Defaults.MODE;
  static const String ALLOW_SPECIAL_CHARS = "allowSpecialChars";
  static const String SPECIAL_CHARS = "specialChars";
  static const String ALLOW_LANGUAGE_MIX = "allowLanguageMix";
}
