/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import '../../form/structure/string_object_map.dart';
import '../enums/constraint_type.dart';
import '../enums/field_type.dart';
import '../enums/view_type.dart';
import '../error/illegal_argument_exception.dart';
import '../structure/model/constraint.dart';
import '../ui/discrete/renderer/ui_renderer.dart';
import '../ui/discrete/view_state.dart';
import '../util/bean/range/float_range.dart';
import '../util/bean/range/int_range.dart';
import '../util/form_utils.dart';
import '../util/log/logger.dart';
import '../util/numbers.dart';
import '../util/stop_watch.dart';
import '../util/structure/string_set.dart';
import '../util/utils.dart';
import 'constraint/c_abstract.dart';
import 'constraint/c_check_case.dart' as CCheckCase;
import 'constraint/c_compare.dart' as CCompare;
import 'constraint/c_count.dart' as CCount;
import 'constraint/c_date.dart' as CDate;
import 'constraint/c_distinct.dart' as CDistinct;
import 'constraint/c_duplicate.dart' as CDuplicate;
import 'constraint/c_exclude.dart' as CExclude;
import 'constraint/c_face.dart' as CFace;
import 'constraint/c_file.dart' as CFile;
import 'constraint/c_group_mandatory.dart' as CGroupMandatory;
import 'constraint/c_length.dart' as CLength;
import 'constraint/c_link.dart';
import 'constraint/c_mandatory.dart';
import 'constraint/c_no_space.dart';
import 'constraint/c_range.dart' as CRange;
import 'constraint/c_repetition.dart' as CRepetition;
import 'constraint/c_script.dart' as CScript;
import 'constraint/c_table_duplicate.dart' as CTableDuplicate;
import 'constraint/c_text.dart' as CText;
import 'constraint/c_word.dart' as CWord;
import 'constraint/fingerprint/c_fingerprint.dart' as CFingerprint;
import 'constraint/image/c_image.dart' as CImage;
import 'constraint/image/c_image_matcher.dart';
import 'constraint/image/c_smart_scan.dart';
import 'constraint/pattern/c_alpha_numeric.dart' as CAlphaNumeric;
import 'constraint/pattern/c_email.dart';
import 'constraint/pattern/c_hexadecimal.dart';
import 'constraint/pattern/c_ip_address.dart';
import 'constraint/pattern/c_password.dart';
import 'constraint/pattern/c_pattern.dart' as CPattern;
import 'constraint/pattern/c_phone_number.dart';
import 'constraint/pattern/c_user_name.dart';
import 'converter/c_boolean.dart';
import 'converter/c_float.dart';
import 'converter/c_number.dart';
import 'converter/c_string.dart';
import 'converter/conversion_error.dart';
import 'converter.dart';

abstract class ConstraintFactory {
  final UIRenderer renderer;

  const ConstraintFactory(this.renderer);

  CAbstract get(ViewState view, Constraint constraint) {
    StopWatch stopWatch = StopWatch.start();
    ConstraintType type = constraint.getType();
    StringObjectMap args = constraint.getArgs_(renderer);
    CAbstract checker;

    if (CMandatory.TYPE == type) {
      checker = mandatory();
    } else if (CGroupMandatory.CGroupMandatory.TYPE == type) {
      String fields = CString.convert(args[CGroupMandatory.Args.FIELDS]);
      int count = CNumber.convertNoErr(args[CGroupMandatory.Args.COUNT]);

      checker = groupMandatory(fields, count);
    } else if (CLength.CLength.TYPE == type) {
      IntRange range = IntRange.get(CLength.Args.MIN, CLength.Args.MAX, args);
      String language = CString.convert(args[CLength.Args.LANGUAGE]);

      checker = length(range, language);
    } else if (CRange.CRange.TYPE == type) {
      Object min = args[CRange.Args.MIN];
      Object max = args[CRange.Args.MAX];

      checker = range(min, max);
    } else if (CCheckCase.CCheckCase.TYPE == type) {
      CCheckCase.Case _case = CCheckCase.Case.toCase(
          CString.convert(args[CCheckCase.Args.CASE]), false);

      checker = checkCase(_case);
    } else if (CExclude.CExclude.TYPE == type) {
      bool sensitive = CBoolean.convertNoErr(args[CExclude.Args.SENSITIVE]);
      String chars = CString.convert(args[CExclude.Args.CHARS]);
      int position = CNumber.convertNoErr(args[CExclude.Args.POSITION]);

      checker = exclude(sensitive, chars, position);
    } else if (CNoSpace.TYPE == type) {
      checker = noSpace();
    } else if (CLink.TYPE == type) {
      checker = link();
    } else if (CWord.CWord.TYPE == type) {
      IntRange range = IntRange.get(CWord.Args.MIN, CWord.Args.MAX, args);

      checker = word(range);
    } else if (CDate.CDate.TYPE == type) {
      CDate.Mode mode =
          CDate.Mode.toMode(CString.convert(args[CDate.Args.MODE]), false);
      bool allowToday = CBoolean.convertNoErr(args[CDate.Args.ALLOW_TODAY]);
      int years = CNumber.convertNoErr(args[CDate.Args.YEARS]);
      int months = CNumber.convertNoErr(args[CDate.Args.MONTHS]);
      int days = CNumber.convertNoErr(args[CDate.Args.DAYS]);

      checker = date(mode, allowToday, years, months, days);
    } else if (CCompare.CCompare.TYPE == type) {
      checker = getCompare(view, args);
    } else if (CCount.CCount.TYPE == type) {
      IntRange range = IntRange.get(CCount.Args.MIN, CCount.Args.MAX, args);
      bool sensitive = CBoolean.convertNoErr(args[CCount.Args.SENSITIVE]);
      String chars = CString.convert(args[CCount.Args.CHARS]);

      checker = count(range, sensitive, chars);
    } else if (CRepetition.CRepetition.TYPE == type) {
      int max = CNumber.convertNoErr(args[CRepetition.Args.MAX]);
      bool sensitive = CBoolean.convertNoErr(args[CRepetition.Args.SENSITIVE]);
      String chars = CString.convert(args[CRepetition.Args.CHARS]);

      checker = repetition(max, sensitive, chars);
    } else if (CDistinct.CDistinct.TYPE == type) {
      int min = CNumber.convertNoErr(args[CDistinct.Args.MIN]);
      bool sensitive = CBoolean.convertNoErr(args[CDistinct.Args.SENSITIVE]);
      String ignore = CString.convert(args[CDistinct.Args.IGNORE]);

      checker = distinct(min, sensitive, ignore);
    } else if (CText.CText.TYPE == type) {
      bool sensitive = CBoolean.convertNoErr(args[CText.Args.SENSITIVE]);
      CText.Operator op = CText.Operator.toOperator(
          CString.convert(args[CText.Args.OPERATOR]), false);
      String refValues = CString.convert(args[CText.Args.VALUE]);

      checker = text(op, sensitive, refValues);
    } else if (CFace.CFace.TYPE == type) {
      CFace.Type faceType = CFace.Type.toType(args[CFace.Args.TYPE], false);
      IntRange faces =
          IntRange.get(CFace.Args.MIN_FACES, CFace.Args.MAX_FACES, args);
      double confidence = CFloat.convertNoErr(args[CFace.Args.CONFIDENCE]);
      FloatRange eyesDistance = FloatRange.get(
          CFace.Args.MIN_EYES_DISTANCE, CFace.Args.MAX_EYES_DISTANCE, args);
      FloatRange yAngle =
          FloatRange.get(CFace.Args.MIN_Y_ANGLE, CFace.Args.MAX_Y_ANGLE, args);
      FloatRange zAngle =
          FloatRange.get(CFace.Args.MIN_Z_ANGLE, CFace.Args.MAX_Z_ANGLE, args);

      checker = face(CFace.Params.full(
          faceType, faces, confidence, eyesDistance, yAngle, zAngle));
    } else if (CImage.CImage.TYPE == type) {
      IntRange width =
          IntRange.get(CImage.Args.MIN_WIDTH, CImage.Args.MAX_WIDTH, args);
      IntRange height =
          IntRange.get(CImage.Args.MIN_HEIGHT, CImage.Args.MAX_HEIGHT, args);
      int blurThreshold =
          CNumber.convertNoErr(args[CImage.Args.BLUR_THRESHOLD]);

      checker = image(CImage.Params.full(width, height, blurThreshold));
    } else if (CSmartScan.TYPE == type) {
      checker =
          smartScan(); // no need for arguments as this constraint do nothing
    } else if (CImageMatcher.TYPE == type) {
      checker =
          imageMatcher(); // no need for arguments as this constraint do nothing
    } else if (CFingerprint.CFingerprint.TYPE == type) {
      int nfiqThreshold = CNumber.convertNoErr(args[CFingerprint.Args.NFIQ]);
      int minutiaeCount =
          CNumber.convertNoErr(args[CFingerprint.Args.MINUTIAE]);
      int grayThreshold =
          CNumber.convertNoErr(args[CFingerprint.Args.GRAY_THRESHOLD]);
      String pattern = CString.convert(args[CFingerprint.Args.PATTERN]);
      bool natively = CBoolean.convertNoErr(args[CFingerprint.Args.NATIVELY]);

      checker = fingerprint(
          nfiqThreshold, minutiaeCount, grayThreshold, pattern, natively);
    } else if (CFile.CFile.TYPE == type) {
      IntRange size = IntRange.get(CFile.Args.MIN, CFile.Args.MAX, args);

      checker = file(size);
    } else if (CDuplicate.CDuplicate.TYPE == type) {
      checker = getDuplicate(view, args);
    } else if (CTableDuplicate.CTableDuplicate.TYPE == type) {
      String tableId = CString.convert(args[CTableDuplicate.Args.TABLE]);
      int columnIndex = CNumber.convertNoErr(args[CTableDuplicate.Args.COLUMN]);
      int max = CNumber.convertNoErr(args[CTableDuplicate.Args.MAX]);

      checker = tableDuplicate(tableId, columnIndex, max);
    } else if (CScript.CScript.TYPE == type) {
      String expression = CString.convert(args[CScript.Args.EXPR]);
      bool ignoreIfNull =
          CBoolean.convertNoErr(args[CScript.Args.IGNORE_IF_NULL]);

      checker = script(expression, ignoreIfNull);
    } else if (CAlphaNumeric.CAlphaNumeric.TYPE == type) {
      String regex = CString.convert(args[CAlphaNumeric.Args.REGEX]);
      String language = CString.convert(args[CAlphaNumeric.Args.LANGUAGE]);
      CAlphaNumeric.Mode mode = CAlphaNumeric.Mode.toMode(
          CString.convert(args[CAlphaNumeric.Args.MODE]), false);
      bool allowSpecialChars =
          CBoolean.convertNoErr(args[CAlphaNumeric.Args.ALLOW_SPECIAL_CHARS]);
      String specialChars =
          CString.convert(args[CAlphaNumeric.Args.SPECIAL_CHARS]);
      bool allowLanguageMix =
          CBoolean.convertNoErr(args[CAlphaNumeric.Args.ALLOW_LANGUAGE_MIX]);

      return alphaNumeric(regex, language, mode, allowSpecialChars,
          specialChars, allowLanguageMix);
    } else {
      checker = getPattern(type, args);
    }

    StopWatch.stop_(type, stopWatch);

    if (checker == null) {
      throw IllegalArgumentException(
          "Constraint '$type' isn't supported or wrongly configured!");
    }

    return checker;
  }

  CAbstract getPattern(ConstraintType type, StringObjectMap args) {
    final String regex = CString.convert(args[CPattern.Args.REGEX]);

    if (CPattern.CPattern.TYPE == type) {
      bool enDigits = CBoolean.convertNoErr(args[CPattern.Args.EN_DIGITS]);
      bool find = CBoolean.convertNoErr(args[CPattern.Args.FIND]);
      bool sensitive = CBoolean.convertNoErr(args[CPattern.Args.SENSITIVE]);

      return pattern(regex, enDigits, find, sensitive);
    }

    if (!Utils.isBlank(regex)) {
      throw IllegalArgumentException(
          "${CPattern.Args.REGEX} can not de defined for $type");
    }

    if (CEmail.TYPE == type) {
      return email();
    } else if (CHexadecimal.TYPE == type) {
      return hexadecimal();
    } else if (CIpAddress.TYPE == type) {
      return ipAddress();
    } else if (CPhoneNumber.TYPE == type) {
      return phoneNumber();
    } else if (CUserName.TYPE == type) {
      return username();
    } else if (CPassword.TYPE == type) {
      return password();
    }

    return null;
  }

  CCompare.CCompare getCompare(ViewState view, StringObjectMap args) {
    FieldType type = view.getFieldType();

    try {
      Object refValue = Converter.convert(type, args[CCompare.Args.VALUE]);
      bool sensitive = CBoolean.convertNoErr(args[CCompare.Args.SENSITIVE]);
      String op = CString.convert(args[CCompare.Args.OPERATOR]);
      bool isWsq = ViewType.FINGERPRINT == view.getType();
      double threshold =
          isWsq ? CFloat.convertNoErr(args[CCompare.Args.THRESHOLD]) : null;

      return compare(sensitive, op, type, refValue, isWsq, threshold);
    } on ConversionError catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      throw IllegalArgumentException(
          "CCompare: Reference value should be of $type type");
    }
  }

  CDuplicate.CDuplicate getDuplicate(ViewState view, StringObjectMap args) {
    StringSet viewIds =
        Utils.splitByComma(CString.convert(args[CDuplicate.Args.FIELDS]));

    if (Utils.isBlank(viewIds)) {
      throw IllegalArgumentException(
          "CDuplicate: ${CDuplicate.Args.FIELDS} arg is mandatory.");
    }

    int max = CNumber.convertNoErr(CString.convert(args[CDuplicate.Args.MAX]));

    if (Numbers.nullOrBelowOne(max)) {
      throw IllegalArgumentException(
          "CDuplicate: ${CDuplicate.Args.MAX} is required.");
    }

    List<Object> values = [];
    FieldType fSrcType = view.getFieldType();
    final String vSrcId = view.getId();

    for (String viewId in viewIds) {
      // ignore checking duplicate with itself so it is more convenient to copy-paste a list across fields
      if (Utils.equalsIgnoreCase(vSrcId, viewId)) {
        continue;
      }

      ViewState vRef = FormUtils.getView_(viewId, renderer);

      if (vRef == null) {
        throw IllegalArgumentException(
            "CDuplicate: viewId '$viewId' does not exist");
      }

      FieldType refFieldType = vRef.getFieldType();

      if (fSrcType == refFieldType && vRef.canInput()) {
        try {
          values.add(Converter.convert(fSrcType, vRef.getValue()));
        } on ConversionError catch (e, stackTrace) {
          Logger.error(e, stackTrace: stackTrace);
          throw IllegalArgumentException(
              "CDuplicate: $viewId should be $fSrcType type");
        }
      }
    }

    return duplicate(values, max);
  }

  CMandatory mandatory() {
    return CMandatory();
  }

  CGroupMandatory.CGroupMandatory groupMandatory(String fields, int count) {
    return CGroupMandatory.CGroupMandatory(
        fields, count, renderer.getAllViews());
  }

  CLength.CLength length(IntRange range, String language) {
    return CLength.CLength(range, language);
  }

  CRange.CRange range(Object min, Object max) {
    return CRange.CRange(min, max);
  }

  CCheckCase.CCheckCase checkCase(CCheckCase.Case _case) {
    return CCheckCase.CCheckCase.full(_case);
  }

  CExclude.CExclude exclude(bool sensitive, String chars, int position) {
    return CExclude.CExclude(sensitive, chars, position);
  }

  CNoSpace noSpace() {
    return CNoSpace();
  }

  CLink link() {
    return CLink();
  }

  CWord.CWord word(IntRange range) {
    return CWord.CWord(range);
  }

  CDate.CDate date(
      CDate.Mode mode, bool allowToday, int years, int months, int days) {
    return CDate.CDate(mode, allowToday, years, months, days);
  }

  CCompare.CCompare compare(bool sensitive, String op, FieldType type,
      Object refValue, bool isWsq, double threshold) {
    return CCompare.CCompare.get_(
        refValue, type, op, sensitive, isWsq, threshold, renderer);
  }

  CCount.CCount count(IntRange range, bool sensitive, String chars) {
    return CCount.CCount(range, sensitive, chars);
  }

  CRepetition.CRepetition repetition(int max, bool sensitive, String chars) {
    return CRepetition.CRepetition(max, sensitive, chars);
  }

  CDistinct.CDistinct distinct(int min, bool sensitive, String ignore) {
    return CDistinct.CDistinct(min, sensitive, ignore);
  }

  CText.CText text(CText.Operator op, bool sensitive, String refValues) {
    return CText.CText(op, sensitive, refValues);
  }

  CFace.CFace face(CFace.Params params);

  CImage.CImage image(CImage.Params params);

  CSmartScan smartScan() {
    return CSmartScan();
  }

  CImageMatcher imageMatcher() {
    return CImageMatcher();
  }

  CFingerprint.CFingerprint fingerprint(int nfiqThreshold, int minutiaeCount,
      int grayThreshold, String pattern, bool natively) {
    return CFingerprint.CFingerprint.full(
        nfiqThreshold, minutiaeCount, grayThreshold, pattern, natively);
  }

  CFile.CFile file(IntRange size) {
    return CFile.CFile(size);
  }

  CDuplicate.CDuplicate duplicate(List<Object> values, int max) {
    return CDuplicate.CDuplicate(values, max);
  }

  CTableDuplicate.CTableDuplicate tableDuplicate(
      String tableId, int columnIndex, int max) {
    return CTableDuplicate.CTableDuplicate(
        tableId, columnIndex, max, renderer.getAllViews());
  }

  CScript.CScript script(String expression, bool ignoreIfNull) {
    return CScript.CScript(expression, ignoreIfNull, renderer);
  }

  CAlphaNumeric.CAlphaNumeric alphaNumeric(
      String regex,
      String language,
      CAlphaNumeric.Mode mode,
      bool allowSpecialChars,
      String specialChars,
      bool allowLanguageMix) {
    return CAlphaNumeric.CAlphaNumeric(regex, language, mode, allowSpecialChars,
        specialChars, allowLanguageMix);
  }

  CPattern.CPattern pattern(
      String regex, bool enDigits, bool find, bool sensitive) {
    return CPattern.CPattern.full(null, regex, null, enDigits, find, sensitive);
  }

  CEmail email() {
    return CEmail();
  }

  CHexadecimal hexadecimal() {
    return CHexadecimal();
  }

  CIpAddress ipAddress() {
    return CIpAddress();
  }

  CPhoneNumber phoneNumber() {
    return CPhoneNumber();
  }

  CUserName username() {
    return CUserName();
  }

  CPassword password() {
    return CPassword();
  }
}
