/*
 * ----------------------------------------------------------------------------
 * Copyright © 2009 by Mobile-Technologies Limited. All rights reserved. All intellectual property rights in and/or in
 * the computer program and its related documentation and technology are the sole Mobile-Technologies Limited property.
 * This computer program is under Mobile-Technologies Limited copyright and cannot be in whole or in part reproduced,
 * sublicensed, leased, sold or used in any form or by any means, including without limitation graphic, electronic,
 * mechanical, photocopying, recording, taping or information storage and retrieval systems without Mobile-Technologies
 * Limited prior written consent. The downloading, exporting or reexporting of this computer program or any related
 * documentation or technology is subject to any export rules, including US regulations.
 * ----------------------------------------------------------------------------
 */

import 'package:collection/collection.dart' show ListEquality;

import '../../form/structure/ordered_map.dart';
import '../../form/structure/string_object_map.dart';
import '../../form/structure/t_map.dart';
import '../contract/contract_utils.dart';
import '../enums/constraint_type.dart';
import '../enums/mode.dart';
import '../enums/scope.dart';
import '../structure/i18n/message_array.dart';
import '../structure/model/constraint.dart';
import '../structure/model/field.dart';
import '../structure/model/global_constraint.dart';
import '../structure/model/model.dart';
import '../ui/discrete/component/base/input_component.dart';
import '../ui/discrete/component/select_one.dart';
import '../ui/discrete/renderer/features.dart';
import '../ui/discrete/renderer/ui_renderer.dart';
import '../ui/discrete/view_state.dart';
import '../util/base64.dart';
import '../util/form_utils.dart';
import '../util/log/logger.dart';
import '../util/stop_watch.dart';
import '../util/utils.dart';
import 'constraint/c_abstract.dart';
import 'converter/c_binary.dart';
import 'converter/conversion_error.dart';
import 'converter.dart';
import 'error_marker.dart';

// TODO Conversion should be done prior to running the constraints.

// Three cases to handle:
//
// 1) Validate the displayed region. (Next Button)
// 2) Validate all the included and updatable views in the form. (Save/Update Button)
// 3) Validate given list of fields with configured constraints. (Validate Button)
class Validator {
  final UIRenderer renderer;

  const Validator(this.renderer);

  bool run(List<String> viewIds) {
    return run_(viewIds, true);
  }

  bool run_(Iterable<String> viewIds, bool filter) {
    return run__(viewIds, filter, false, true);
  }

  bool runAll(bool markError) {
    return run__([Scope.ALL.id()], true, false, markError);
  }

  bool run__(
      Iterable<String> viewIds, bool filter, bool forTable, bool markError) {
    List<ViewState> views = filterAndSort(viewIds, filter, forTable);
    bool valid = true;

    if (!Utils.isBlank(views)) {
      final bool forceRenderInvalidView =
          FormUtils.getUI_(renderer).isForceRenderInvalidView(renderer);
      bool invalidViewShown =
          false; // only top invalid view as in the model should be rendered/focused

      for (ViewState view in views) {
        try {
          final bool isValidView =
              run___(view, view.getValue(), false, markError);

          if (!isValidView && !invalidViewShown) {
            invalidViewShown = view.renderAndFocus(forceRenderInvalidView);

            if (invalidViewShown) {
              renderer.getController().scrollToView(view);
            }
          }

          valid = isValidView && valid;
        } catch (e, stackTrace) {
          Logger.error(e, stackTrace: stackTrace);
        }
      }
    }

    return valid;
  }

  bool runWithoutMarking(ViewState view, Object value) {
    return run___(view, value, false, false);
  }

  /// Validates specific view with its given value.
  bool run___(
      ViewState view, Object value, bool immediateOnly, bool markError) {
    if (!view.canInputAndErr()) {
      return true;
    }

    StopWatch stopWatch = StopWatch.start();
    bool valid = verifyArrayOptions(view, markError) &&
        verifyArrayText(view, value, markError);

    if (valid) {
      try {
        value = Converter.convert(view.getFieldType(), value);
        valid = runConstraints(view, value, immediateOnly, markError);
      } on ConversionError catch (e) {
        valid = false;

        if (markError) {
          ErrorMarker.mark(
              view, renderer, view.getConversionError(), e.getMsgId());
        }
      }
    }

    StopWatch.stop_(view, stopWatch);

    if (!valid) {
      Logger.debug("Validation failed for viewId '${view.getId()}'");
    }

    return valid;
  }

  /// For a view that can have array to choose the value from, this method verifies that the currently loaded array is
  /// what it should be. So for example, if 'state' drop-down has 'Thailand' as its selected value then we expect the
  /// connected drop-down 'city' to have relevant 'Thailand' cities only. If by any chance its not the case then we
  /// show error on the form and ultimately validation fails.
  bool verifyArrayOptions(ViewState view, bool markError) {
    if (view.canArray() && Utils.isBlank(view.getDefinedArray())) {
      // Should be 'IArray' view without any array
      ViewState vParent = view.getParent(
          renderer); // finds the parent view of given view in parameter.

      if (vParent != null) {
        final String parentSelectionId = vParent.getValue()
            as String; // selected option for the parent's view
        final String childArrayId = view.getArray();

        // If parent's selected value is null, there shouldn't be any array loaded for the child's drop-down.
        if (Utils.allBlank([parentSelectionId, childArrayId])) {
          return true;
        }

        final String parentArrayId =
            vParent.getArray(); // currently loaded parent's array id
        // expected message array that should be loaded for the view
        MessageArray expectedArray = FormUtils.getI18n(renderer)
            .getMsgArray_(parentArrayId, parentSelectionId);
        final String expectedArrayId =
            expectedArray == null ? "" : expectedArray.getId(); // expect null

        // verify the currently loaded message array with the expected message array.
        if (!Utils.equalsIgnoreCase(expectedArrayId, childArrayId)) {
          if (markError) {
            ErrorMarker.mark(view, renderer, null, "array.invalid.list");
          }

          return false; // TODO Clear and reload the correct array?
        }
      }
    }

    return true;
  }

  /// When SelectOne can be filtered manually by the user, the text in the component can be invalid which means the
  /// original value will be null. To avoid this inconsistency, we show the error to the user that the text doesn't
  /// represent any valid item.
  bool verifyArrayText(ViewState view, Object value, bool markError) {
    if (value == null) {
      InputComponent component = view.getInputComponent();

      if (component is SelectOne && !Utils.isBlank(component.getInput())) {
        if (markError) {
          ErrorMarker.mark(view, renderer, null, "array.invalid.text");
        }

        return false;
      }
    }

    return true;
  }

  /// Filters out all the given views that shouldn't be validated and sorts the collection as per their order in UI.
  List<ViewState> filterAndSort(
      Iterable<String> viewIds, bool filter, bool forTable) {
    List<ViewState> views = [];

    if (!Utils.isBlank(viewIds)) {
      final bool allViews = viewIds.contains(Scope.ALL.id());
      final bool region = !allViews && viewIds.contains(Scope.REGION.id());

      for (ViewState view in renderer.getAllViewStates()) {
        if (allViews || region || viewIds.contains(view.getId())) {
          if (!filter || shouldValidate(view, forTable, region)) {
            views.add(view);
          }
        }
      }
    }

    return views;
  }

  /// Should Validate only if:<br>
  /// <br>
  /// 1. View is an input component that can err.<br>
  /// 2. Field can be persisted or it is to be added in a table.<br>
  /// 3. If region is true, check if view is really rendered, otherwise check if view is active.<br>
  /// 4. View supports the current form mode.
  bool shouldValidate(ViewState view, bool forTable, bool region) {
    if (!view.canInputAndErr()) {
      return false;
    }

    final bool isActive = region ? view.isReallyRendered() : view.isActive();

    if (isActive) {
      view.clearErrMsg();
    }

    return (forTable || view.canPersist(renderer)) &&
        isActive &&
        view.supportsMode(renderer);
  }

  bool runConstraints(
      ViewState view, Object value, bool immediateOnly, bool markError) {
    Field field = view.getField();
    OrderedMap<Constraint> constraints =
        field == null ? null : field.getConstraints();

    if (!Utils.isBlank(constraints)) {
      TMap<GlobalConstraint> globalConstraints =
          FormUtils.getConstraints(renderer);
      bool
          bytesChanged; // calculate it only once as we can have multiple binary constraints

      for (Constraint constraint in constraints.values) {
        constraint = constraint.isGlobal()
            ? Model.getConstraint(constraint, globalConstraints)
            : constraint;

        // We can not run the constraint if:
        // 1. We only need to run 'immediate' constraints.
        // 2. Constraint is not configured to run on a disabled component.
        // 3. Constraint is disabled.
        if (immediateOnly && !constraint.isImmediate(renderer) ||
            constraint.ignoreIfDisabled_() && !view.isEnabled() ||
            constraint.isDisabled(renderer)) {
          continue;
        }

        bool valid = true;

        try {
          if (constraint.isCustom()) {
            valid = runCustomConstraint(constraint, view, markError);
          } else {
            // cache the result
            if (bytesChanged == null) {
              bytesChanged =
                  this.bytesChanged(constraint.getType(), view, value);
            }

            if (Utils.toTrue(bytesChanged)) {
              valid = runConstraint(constraint, view, value, markError);
            }
          }
        } catch (e) {
          // catch and continue with the next constraint
          Logger.error(
              "Constraint '${constraint.getId()}' error for view '${view.getId()}': $e");
        }

        if (!valid) {
          return false; // apply only one error
        }
      }
    }

    return true;
  }

  /// If binary value of a field isn't changed in UPDATE mode, skip the binary constraints on it's view.
  bool bytesChanged(ConstraintType type, ViewState view, Object value) {
    if (renderer.checkMode([Mode.UPDATE])) {
      StringObjectMap data = renderer.getData();

      if (!Utils.isBlank(data) &&
          CBinary.isBinary(value) &&
          type.binary() &&
          view != null) {
        Object oldValue = data[view.getFieldId()];

        if (oldValue is String) {
          oldValue = Base64.decode(oldValue as String);
        }

        return !ListEquality().equals(oldValue, value);
      }
    }

    return null; // for all other modes, we don't check
  }

  bool runCustomConstraint(
      Constraint constraint, ViewState view, bool markError) {
    final String impl = constraint.getIdUnprefixed(); // IConstraint impl

    try {
      String bundleMsgId =
          ContractUtils.doCheck(impl, view, constraint.getArgs_(renderer));

      if (markError) {
        ErrorMarker.mark(view, renderer, constraint.getMsg(), bundleMsgId);
      }

      return Utils.isBlank(bundleMsgId);
    } catch (e, stackTrace) {
      Logger.error(e, stackTrace: stackTrace);
      Features.getAlert_(renderer)
          .error("Error invoking custom constraint '$impl'.");
      return true; // if something goes wrong during method invocation, we should consider this as a bug to fix.
    }
  }

  // if the given value of the field is valid for the given constraint
  bool runConstraint(
      Constraint constraint, ViewState view, Object value, bool markError) {
    CAbstract checker =
        Features.getConstraintFactory(renderer).get(view, constraint);
    String bundleMsgId;

    try {
      bundleMsgId = checker.isValid(value);

      if (markError) {
        return ErrorMarker.mark(view, renderer, constraint.getMsg(),
            bundleMsgId, checker.getMsgArgs());
      }
    } catch (e) {
      // ignore if the given constraint doesn't support the value data type.
    }

    return Utils.isBlank(bundleMsgId);
  }
}
